/*
 *
 *  AM --- AmigaMail
 *  (C) 1991, 1992 by Christian Riede
 *
 *  AM is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY.  No author or distributor accepts responsibility to anyone
 *  for the consequences of using it or for whether it serves any
 *  particular purpose or works at all, unless he says so in writing.
 *  Refer to the GNU General Public License, Version 1, for full details.
 *  
 *  Everyone is granted permission to copy, modify and redistribute AM,
 *  but only under the conditions described in the GNU General Public
 *  License, Version 1.  A copy of this license is supposed to have been 
 *  given to you along with AM so you can know your rights and responsi-
 *  bilities.  It should be in a file named COPYING.  Among other things,
 *  the copyright notice and this notice must be preserved on all copies.
 *
 *  
 *
 */

#include "am.h"
#include "server.h"

/* get one header line, handle LWSP correctly */
/* returns TRUE if more headerlines are following, else FALSE */
static int GetHeaderLine(FILE *file,char *s,int len)
{
	int copy = TRUE,header = TRUE;
	int count=0,c,d;

	do 
	{
		switch(c=getc(file))
		{
			case EOF: 
				header = FALSE;
				copy = FALSE;
				break;

			case '\n':
				copy=FALSE;
				
				/* one character lookahead */
				d=getc(file);
				ungetc(d,file);

				if (d==' ' || d=='\t') 
					copy=TRUE;

				if (d=='\n') /* end of header */
					header = FALSE;

				break;

			default: 
				if (count++<len-1)
					*s++=c;
				break;
		}
	}
	while (copy);
	
	*s=0;

	return(header);
}

/* return pointer to field body, i.e. */
/* first character after first white space */
static char *FieldBody(char *s)
{
	while (*s && *s!=' ' && *s!='\t')
		s++;

	if (*s==' ' || *s=='\t') s++;

	return(s);
}

/* compare with case ignored, t must be uppercase */
static int HdrCmp(char *s, char *t)
{
	while (*t)
	{
		if (islower(*s))
		{
			if (toupper((int)*s++) != *t++) 
				return(FALSE);
		}
		else if (*s++ != *t++)
			return(FALSE);
	}

	return(TRUE);
}

/* parse header, add to list */
struct Mail *AddMail(int i,ULONG num)
{
	struct Mail *New,*Scan;
	char s[MAIL_FIELDLENGTH];
	FILE *Mailfile;
	int j,isheader;
	struct tm Actual_Date;

	/* build filename of mail */
	sprintf(s,"UUMAIL:%s.mail/%ld",User[i],num);

	/* open file */
	if (!(Mailfile = fopen(s,"r")))
		return(0);

	/* allocate space for new list element */
	if (!(New = (struct Mail *)
		AllocMem(sizeof(struct Mail),MEMF_CLEAR)))
	{
		writelog("Out of Memory","addmail");
		fclose(Mailfile);
		return(0);
	}

	New->Number = num;
	New->read |= MAIL_NEW;
	New->CTE = CTE_8BIT;

	/* clear Actual_Date */
	Actual_Date.tm_hour = 0;
	Actual_Date.tm_min = 0;
	Actual_Date.tm_mon = 0;
	Actual_Date.tm_mday = 0;
	Actual_Date.tm_yday = 0;
	Actual_Date.tm_year = 0;

	New->To = GetString(GetNumber("",TRUE,&Strings[i]),&Strings[i]);
	New->Subject = GetString(GetNumber("",TRUE,&Strings[i]),&Strings[i]);
	New->InReplyTo = GetString(GetNumber("",TRUE,&Strings[i]),&Strings[i]);
	New->ReplyTo = GetString(GetNumber("",TRUE,&Strings[i]),&Strings[i]);
	New->MsgId = GetString(GetNumber("",TRUE,&Strings[i]),&Strings[i]);

	/* parse mail-header */
	j=0;
	do /* read all headerlines */
	{
		/* copy next headerline to s */
		isheader = GetHeaderLine(Mailfile,s,MAIL_FIELDLENGTH);

		if (!j++)
			if (strstr(s,"(ARCHIVE)")) /* "(ARCHIVE)" in first headerline? */
				New->read |= MAIL_ARCHIVE|MAIL_READ; /* mail is from me */

		/* parse fields */

		if (HdrCmp(s,"FROM:"))
			New->From = GetString(GetNumber(FieldBody(s),TRUE,&Strings[i]),&Strings[i]);

		if (HdrCmp(s,"TO: "))
			New->To = GetString(GetNumber(FieldBody(s),TRUE,&Strings[i]),&Strings[i]);

		if (HdrCmp(s,"SUBJECT: "))
			New->Subject = GetString(GetNumber(FieldBody(s),TRUE,&Strings[i]),&Strings[i]);

		if (HdrCmp(s,"IN-REPLY-TO: "))
			New->InReplyTo = GetString(GetNumber(FieldBody(s),TRUE,&Strings[i]),&Strings[i]);

		if (HdrCmp(s,"REPLY-TO: "))
			New->ReplyTo = GetString(GetNumber(FieldBody(s),TRUE,&Strings[i]),&Strings[i]);

		if (HdrCmp(s,"MESSAGE-ID: "))
			New->MsgId = GetString(GetNumber(FieldBody(s),TRUE,&Strings[i]),&Strings[i]);
		
		if (HdrCmp(s,"DATE: "))
		{
			/* clear date */
			memset(&Actual_Date,0,sizeof(struct tm));

			/* scan date */
			/* only RFC-Conform dates with slight extentions allowed */
			if (ParseDate(FieldBody(s),&Actual_Date))
			{
				Actual_Date.tm_yday = 0;
				New->Date = mktime(&Actual_Date);
			}
			else 
				New->Date = 0;
		}

	}
	while (isheader);


	/* close file */
	fclose(Mailfile);
	

	/* init displaytext */
	New->m_SelectNode.ln_Name = New->DisplayText;

	/* put into list */
	/* sorted by number */
	for (Scan = (struct Mail *) Mailboxes[i].lh_Head;
		Scan->m_Node.ln_Succ;
		Scan = (struct Mail *)Scan->m_Node.ln_Succ)
		if (Scan->Number > num)
			break;

	New->m_Node.ln_Type = AMNT_MAILBOX;
	New->m_SelectNode.ln_Type = AMNT_SELECT_LIST;

	Insert(&Mailboxes[i],&(New->m_Node),Scan->m_Node.ln_Pred);

	return(New);
}
