/*
 *
 *  AM --- AmigaMail
 *  (C) 1991, 1992 by Christian Riede
 *
 *  AM is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY.  No author or distributor accepts responsibility to anyone
 *  for the consequences of using it or for whether it serves any
 *  particular purpose or works at all, unless he says so in writing.
 *  Refer to the GNU General Public License, Version 1, for full details.
 *  
 *  Everyone is granted permission to copy, modify and redistribute AM,
 *  but only under the conditions described in the GNU General Public
 *  License, Version 1.  A copy of this license is supposed to have been 
 *  given to you along with AM so you can know your rights and responsi-
 *  bilities.  It should be in a file named COPYING.  Among other things,
 *  the copyright notice and this notice must be preserved on all copies.
 *
 *  
 *
 */


#include "am.h"


void AddAddress(struct List *Addressbook,char *Realname,char *Username,int rrt,char *ReplyTo,int CTE)
{
	struct Address *Address;

	if (!(Address = AllocMem(sizeof(struct Address),MEMF_CLEAR)))
		return;

	if (!(Address->Node.ln_Name = AllocMem(160,MEMF_CLEAR)))
		return;
	
	if (!(Address->Realname = AllocMem(160,MEMF_CLEAR)))
		return;
	
	if (!(Address->Username = AllocMem(160,MEMF_CLEAR)))
		return;
	
	if (!(Address->ReplyTo = AllocMem(160,MEMF_CLEAR)))
		return;
	
	strcpy(Address->Realname,Realname);
	strcpy(Address->Username,Username);
	strcpy(Address->ReplyTo,ReplyTo);
	Address->ReturnReceiptTo = rrt;
	Address->CTE = CTE;

	sprintf(Address->Node.ln_Name,"%-30.30s <%s>",Realname,Username);
	Address->Node.ln_Type = AMNT_ADDRESSBOOK;

	AddTail(Addressbook,(struct Node *)Address);

}

void RemAddress(struct Address *Address)
{
	Remove(&(Address->Node));

	FreeMem(Address->Node.ln_Name,160);
	FreeMem(Address->Realname,160);
	FreeMem(Address->Username,160);
	FreeMem(Address->ReplyTo,160);
	FreeMem(Address,sizeof(struct Address));
}

#if 0
void PrintAddressbook(struct List *Addressbook)
{
	struct Address *Address;

	for (Address = (struct Address *)Addressbook->lh_Head;
		Address->Node.ln_Succ;
		Address = (struct Address *)Address->Node.ln_Succ)
	{
		printf("Username:%s\n",Address->Username);
		printf("Realname:%s\n",Address->Realname);
		printf("Reply-To:%s\n",Address->ReplyTo);
		printf("RRT:     %ld\n",Address->ReturnReceiptTo);
		puts("--------------------------------------------");
	}
		
}
#endif

void ReadAddressBook(struct List *Addressbook)
{
	FILE *File;
	char InputBuffer[MAIL_FIELDLENGTH];
	char Header[MAIL_FIELDLENGTH];
	char Body[MAIL_FIELDLENGTH];
	char Username[160];
	char Realname[160];
	char ReplyTo[160];
	int rrt,cte;
	char *s,*t;

	/* init the Addressbook-list */
	Addressbook->lh_Head = (struct Node *)&(Addressbook->lh_Tail);
	Addressbook->lh_TailPred = (struct Node *)&(Addressbook->lh_Head);
	Addressbook->lh_Tail = 0;
	Addressbook->lh_Type = AMNT_ADDRESSBOOK;

	if (!(File = fopen("UULIB:Addressbook","r")))
		return;

	*Realname = *Username = *ReplyTo = 0;
	rrt = -1;

	while (fgets(InputBuffer,MAIL_FIELDLENGTH,File))
	{
		if (*InputBuffer=='\n') /* blank line, save address */
		{
			if (*Realname && *Username) /* valid address read? */
				AddAddress(Addressbook,Realname,Username,rrt,ReplyTo,cte);
			*Realname = *Username = *ReplyTo = 0;
			rrt = -1;
			cte = 2;
		}
		else /* read line */
		{
			if (sscanf(InputBuffer,"%s ",Header)==1)
			{
				/* body may contain blanks */
				s=index(InputBuffer,' ')+1;
				t=Body;
				while (*s && *s!='\n') *t++=*s++;
				*t=0;
				/* delete trailing blanks */
				while (t>Body) {--t; if (*t==' ' || *t=='\t') *t=0; else break;}

				if (!strcmp(Header,"Realname"))
					strcpy(Realname,Body);
				else if (!strcmp(Header,"Username"))
					strcpy(Username,Body);
				else if (!strcmp(Header,"Reply-To"))
					strcpy(ReplyTo,Body);
				else if (!strcmp(Header,"RRT"))
					sscanf(Body,"%d",&rrt);
				else if (!strcmp(Header,"CTE"))
					sscanf(Body,"%d",&cte);
			}
		}
	}

	fclose(File);

#if 0
	PrintAddressbook(Addressbook);
#endif
}

void WriteAddressBook(struct List *Addressbook)
{
	FILE *File;
	struct Address *Address;

	if (!(File = fopen("UULIB:Addressbook","w")))
		return;

	while (Address = (struct Address *)RemHead(Addressbook))
	{
		if (Address->Username)
			fprintf(File,"Username %s\n",Address->Username);

		if (Address->Realname)
			fprintf(File,"Realname %s\n",Address->Realname);

		if (Address->ReplyTo)
			fprintf(File,"Reply-To %s\n",Address->ReplyTo);

		fprintf(File,"RRT %d\nCTE %d\n\n",
			Address->ReturnReceiptTo,
			Address->CTE); /* add extra blank line! */

		FreeMem(Address->Node.ln_Name,160);
		FreeMem(Address->Realname,160);
		FreeMem(Address->Username,160);
		FreeMem(Address->ReplyTo,160);
		FreeMem(Address,sizeof(struct Address));
	}

	fclose(File);
}


void FreeAddressBook(struct List *Addressbook)
{
	struct Address *Address;

	while (Address = (struct Address *)RemHead(Addressbook))
	{
		FreeMem(Address->Node.ln_Name,160);
		FreeMem(Address->Realname,160);
		FreeMem(Address->Username,160);
		FreeMem(Address->ReplyTo,160);
		FreeMem(Address,sizeof(struct Address));
	}
}

