/*
 *
 *  AM --- AmigaMail
 *  (C) 1991, 1992 by Christian Riede
 *
 *  AM is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY.  No author or distributor accepts responsibility to anyone
 *  for the consequences of using it or for whether it serves any
 *  particular purpose or works at all, unless he says so in writing.
 *  Refer to the GNU General Public License, Version 1, for full details.
 *  
 *  Everyone is granted permission to copy, modify and redistribute AM,
 *  but only under the conditions described in the GNU General Public
 *  License, Version 1.  A copy of this license is supposed to have been 
 *  given to you along with AM so you can know your rights and responsi-
 *  bilities.  It should be in a file named COPYING.  Among other things,
 *  the copyright notice and this notice must be preserved on all copies.
 *
 *  
 *
 */

#include "am.h"
#include "server.h"

UBYTE *vers = (UBYTE *)("\0$VER: contentsserver "AMIGAMAIL_VERSION);

struct MsgPort *ServerPort;

char User[MAXUSERS][USERLEN+1];
int  LockCount[MAXUSERS];
struct List Mailboxes[MAXUSERS];
int Changes[MAXUSERS];
struct List NotifyPortList[MAXUSERS];
struct Strings Strings[MAXUSERS];
int NumUsers=0;

char Command[80];
char InputBuffer[MAIL_FIELDLENGTH];
char seqname[160];
char filename[160];

static int getuser(char *s)
{
	int i;

	for (i=0;i<NumUsers;++i)
		if (!strcmp(s,User[i]))
			return(i);

	return(-1);
}

int main(int argc, char *argv[])
{
	int run;
	struct AMMessage *Msg;
	struct Mail *Mail,*NewMail;
	struct NotifyPort *NewPort;
	FILE *kusers;
	int CurrentUser;
	int i;
	ULONG sequential;
	FILE *seq;
	char *s;
	BPTR l;

	if (FindPort(SERVERNAME))
	{
		puts("Server already running!");
		exit(0);
	}

	ReadConfig(0);

	/* create NIL: filehandles for System() */
	if (!(AMSystemTags[0].ti_Data = (ULONG)
		Open((UBYTE *)"NIL:",MODE_OLDFILE)))
		exit(23);
	if (!(AMSystemTags[1].ti_Data = (ULONG)
		Open((UBYTE *)"NIL:",MODE_NEWFILE)))
		exit(24);

	if (!(ServerPort = CreatePort(SERVERNAME,0)))
	{
		Close((BPTR)AMSystemTags[0].ti_Data);
		Close((BPTR)AMSystemTags[1].ti_Data);
		exit(21);
	}

#if DEBUG
	kprintf("ServerPort = 0x%lx\n",(ULONG)ServerPort);
#endif

	/* read in known users */
	if (!(kusers = fopen("GETTY:passwd","r")))
	{
		Close((BPTR)AMSystemTags[0].ti_Data);
		Close((BPTR)AMSystemTags[1].ti_Data);
		DeletePort(ServerPort);
		exit(22);
	}

	CurrentUser = 0;
	while (fgets(InputBuffer,MAIL_FIELDLENGTH,kusers) && CurrentUser<MAXUSERS)
	{
		if (InputBuffer[0]=='#') continue;
		if (InputBuffer[0]==' ') continue;
		if (InputBuffer[0]=='\t') continue;
		if (InputBuffer[0]=='\n') continue;

		i=0;
		s=User[CurrentUser];

		while (i<USERLEN && InputBuffer[i]!=',')
			*s++=InputBuffer[i++];

		InputBuffer[i]=0;

		CurrentUser++;
	}

	NumUsers = CurrentUser;

	fclose(kusers);

	/* initialize the NotifyPortList for each user */
	for (i=0;i<NumUsers;++i)
	{
		NotifyPortList[i].lh_Head = (struct Node *)&(NotifyPortList[i].lh_Tail);
		NotifyPortList[i].lh_TailPred = (struct Node *)&(NotifyPortList[i].lh_Head);
		NotifyPortList[i].lh_Tail = 0;
		NotifyPortList[i].lh_Type = AMNT_NOTIFY_PORT_LIST;
	}

#if DEBUG
	kputs("Server running!\n");
#endif

	run = TRUE;
	while (run)
	{
		WaitPort(ServerPort);

		while (Msg = (struct AMMessage *)GetMsg(ServerPort))
		{
			switch(Msg->Action)
			{
				case ACT_DIE:
#if DEBUG
					kputs("contentsserver: ACTION_DIE\n");
#endif
					run = FALSE;
					for (i=0;i<NumUsers;++i)
						if (LockCount[i]) 
						{
							run = TRUE; 
							break;
						}
					Msg->Number = !run;
					break;

				case ACT_LOCKCONTENTS: /* Parms: Msg->Username */
#if DEBUG
					kputs("contentsserver: ACT_LOCKCONTENTS\n");
#endif
					if ((i=getuser(Msg->Username))>=0)
					{
						if (!LockCount[i])
						{
							Mailboxes[i].lh_Head = (struct Node *)&(Mailboxes[i].lh_Tail);
							Mailboxes[i].lh_TailPred = (struct Node *)&(Mailboxes[i].lh_Head);
							Mailboxes[i].lh_Tail = 0;
							Mailboxes[i].lh_Type = AMNT_MAILBOX;
							
							if (!ReadContents(Msg->Username,&Mailboxes[i],&Strings[i]))
							{
								sprintf(Command,"Contentsfile of user %s corrupt!",User[i]);
								writelog(Command,"contentsserver");
#if 0
								/* contents file corrupt, creating a new one */
								sprintf(Command,"%s %s",MAKECONTENTS,Msg->Username);
								System((UBYTE *)Command,AMSystemTags);
								ReadContents(Msg->Username,&Mailboxes[i],&Strings[i]);
#endif

							}
							Changes[i]=0;
						}
						
						(LockCount[i])++;
						Msg->Number = TRUE;
					}
					else
						Msg->Number = FALSE;
					break;
				
				case ACT_UNLOCKCONTENTS: /* Parms: Msg->Username */
#if DEBUG
					kputs("contentsserver: ACT_UNLOCKCONTENTS\n");
#endif
					if ((i=getuser(Msg->Username))>=0)
					{
						/* reply immediately */
						Msg->Number = TRUE;
						ReplyMsg((struct Message *)Msg);
						Msg = 0;

						if (LockCount[i]>0)
						{
							LockCount[i]--;
							if (!LockCount[i])
								if (Changes[i])
									WriteContents(User[i],&Mailboxes[i],&Strings[i]);
								else
								{
									FreeMailbox(&Mailboxes[i]);
									FreeStrings(&Strings[i]);
								}
						}
					}
					else
						Msg->Number = FALSE;
					break;


				case ACT_GETCONTENTS: /* Parms: Msg->Username, Msg->Mailbox */
#if DEBUG
					kputs("contentsserver: ACT_GETCONTENTS\n");
#endif
					if ((i=getuser(Msg->Username))>=0)
					{
						/* delete all entries in client's list */
						while (Mail = (struct Mail *)RemHead(Msg->Mailbox))
							FreeMail(Mail);

						/* give him the new list */
						for (Mail = (struct Mail *)Mailboxes[i].lh_Head;
							Mail->m_Node.ln_Succ;
							Mail = (struct Mail *)Mail->m_Node.ln_Succ)
						{
							/* allocate new struct */
							if (!(NewMail = AllocMem(sizeof(struct Mail),MEMF_CLEAR)))
								break;

							/* copy from our list */
							*NewMail = *Mail;

							/* add to the client's list */
							AddHead(Msg->Mailbox,(struct Node *)NewMail);
						}
						Msg->Number = TRUE;
					}
					else
						Msg->Number = FALSE;
					break;

				case ACT_MARK_READ: /* Parms: Msg->Username, Msg->Number */
#if DEBUG
					kputs("contentsserver: ACT_MARK_READ\n");
#endif
					if ((i=getuser(Msg->Username))>=0)
					{
						if (Mail = GetNum(&Mailboxes[i],Msg->Number))
						{
							if (!(Mail->read & MAIL_READ))
								(Changes[i])++;
							Mail->read |= MAIL_READ;

							/* notify clients */
							Notify(i,NOTIFY_ITEM_CHANGED,Msg->Number,0);

							Msg->Number = TRUE;
						}
						else
							Msg->Number = FALSE;
					}
					else
						Msg->Number = FALSE;
					break;

				case ACT_MARK_UNREAD: /* Parms: Msg->Username, Msg->Number */
#if DEBUG
					kputs("contentsserver: ACT_MARK_UNREAD\n");
#endif
					if ((i=getuser(Msg->Username))>=0)
					{
						if (Mail = GetNum(&Mailboxes[i],Msg->Number))
						{

							if (Mail->read & MAIL_READ)
								(Changes[i])++;
							Mail->read &= ~MAIL_READ;

							/* notify clients */
							Notify(i,NOTIFY_ITEM_CHANGED,Msg->Number,0);
	
							Msg->Number = TRUE;
						}
						else
							Msg->Number = FALSE;
					}
					else
						Msg->Number = FALSE;
					break;

				case ACT_MARK_OLD: /* Parms: Msg->Username, Msg->Number */
#if DEBUG
					kputs("contentsserver: ACT_MARK_OLD\n");
#endif
					if ((i=getuser(Msg->Username))>=0)
					{
						if (Mail = GetNum(&Mailboxes[i],Msg->Number))
						{
							if (Mail->read & MAIL_NEW)
								(Changes[i])++;
							Mail->read &= ~MAIL_NEW;

							Msg->Number = TRUE;
						}
						else
							Msg->Number = FALSE;
					}
					else
						Msg->Number = FALSE;
					break;

				case ACT_ADD_ITEM: /* Parms: Msg->Username, Msg->Number */
#if DEBUG
					kputs("contentsserver: ACT_ADD_ITEM\n");
#endif
					if ((i=getuser(Msg->Username))>=0)
					{
						if (Mail = AddMail(i,Msg->Number))
						{
							/* return a copy of the new item */
							NewMail = AllocMem(sizeof(struct Mail),MEMF_CLEAR);
							if (NewMail)
								*NewMail = *Mail;

							Msg->NewMail = NewMail;

							/* give a copy of the new item to all who requested */
							Notify(i,NOTIFY_NEWITEM,Msg->Number,NewMail);

							(Changes[i])++;
						}
						else
							Msg->NewMail = 0;
					}
					else
						Msg->NewMail = 0;
					break;
				
				case ACT_DELETE_ITEM: /* Parms: Msg->Username, Msg->Number */
#if DEBUG
					kputs("contentsserver: ACT_DELETE_ITEM\n");
#endif
					if ((i=getuser(Msg->Username))>=0)
					{
						if (Mail = GetNum(&Mailboxes[i],Msg->Number))
						{
							Remove(&(Mail->m_Node));
							Notify(i,NOTIFY_ITEM_DELETED,Msg->Number,0);
							FreeMail(Mail);
							(Changes[i])++;
							Msg->Number = TRUE;
						}
						else
							Msg->Number = FALSE;
					}
					else
						Msg->Number = FALSE;
					break;
				
				case ACT_NEW_SEQ: /* Parms: Msg->Username */
#if DEBUG
					kputs("contentsserver: ACT_NEQ_SEQ\n");
#endif
					Msg->Number = 0; /* 0 indicate failure */
					if ((i=getuser(Msg->Username))>=0)
					{
						/* get seq */
						sprintf(seqname,"uumail:%s.mail/seq",User[i]);
						if (seq = fopen(seqname,"r"))
						{
							fscanf(seq,"%ld",&sequential);
							fclose(seq);
						}
						else
							sequential = 1;


						sprintf(filename,"uumail:%s.mail/%ld",User[i],sequential);
						while (l = Lock((UBYTE *)filename,ACCESS_READ)) /* file exists ? */
						{
							UnLock(l);
							sprintf(filename,"uumail:%s.mail/%d",User[i],++sequential);
						}

						/* increment sequential */
						if (seq = fopen(seqname,"w"))
						{
							fprintf(seq,"%d",sequential+1);
							fclose(seq);
						}

						Msg->Number = sequential;
					}
					break;

				case ACT_NEW_MSGID: /* Parms: Msg->Username */
#if DEBUG
					kputs("contentsserver: ACT_NEW_MSGID\n");
#endif
					GetNewMsgId(Msg->MsgId);
					break;

				case ACT_NOTIFY_ON: /* Parms: Msg->Username, Msg->NotifyPort */
#if DEBUG
					kputs("contentsserver: ACT_NOTIFY_ON\n");
#endif
					if ((i=getuser(Msg->Username))>=0)
					{
						if (NewPort = AllocMem(sizeof(struct NotifyPort),MEMF_CLEAR))
						{
							NewPort->np_Node.ln_Type = AMNT_NOTIFY_PORT_LIST;
							NewPort->np_Port = Msg->NotifyPort;
							AddHead(&(NotifyPortList[i]),&(NewPort->np_Node));
							Msg->Number = TRUE;
						}
						else
							Msg->Number = FALSE;
					}
					else
						Msg->Number = FALSE;
					break;
				
				case ACT_NOTIFY_OFF: /* Parms: Msg->Username, Msg->NotifyPort */
#if DEBUG
					kputs("contentsserver: ACT_NOTIFY_OFF\n");
#endif
					if ((i=getuser(Msg->Username))>=0)
					{
						for (NewPort = (struct NotifyPort *)NotifyPortList[i].lh_Head;
							NewPort->np_Node.ln_Succ;
							NewPort = (struct NotifyPort *)NewPort->np_Node.ln_Succ)
						{
							if (NewPort->np_Port == Msg->NotifyPort)
							{
								Remove(&(NewPort->np_Node));
								FreeMem(NewPort,sizeof(struct NotifyPort));
								Msg->Number = TRUE;
								break;
							}
						}
						Msg->Number = FALSE;
					}
					else
						Msg->Number = FALSE;
					break;

				case ACT_VERSION:
#if DEBUG
					kputs("contentsserver: ACT_VERSION\n");
#endif
					Msg->MsgId = AMIGAMAIL_VERSION;
					break;

				default:
#if DEBUG
					printf("unknown action: %ld",Msg->Action);
#endif
					break;
			}
			if (Msg)  /* Msg set to 0 if already replied */
				ReplyMsg((struct Message *)Msg);
		}
	}

	DeletePort(ServerPort);

	exit(0);
}
