/************************************************************************
 * arndaemon - The daemon process for creating/updating databases       *
 *----------------------------------------------------------------------*
 * (C)opyright 1992 by Roland Bless, Amiga-Version, SAS/C 6.3           *
 * Written by Roland Bless, ALL RIGHTS RESERVED                         *
 * Created this file 24-Apr-1992                                        *
 * Another Blessed Software Product                                     *
 ************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>
#include <dos.h>

#include <exec/exec.h>
#include <devices/timer.h>
#include <libraries/dos.h>
#include <libraries/dosextens.h>
#include <dos/dosextens.h>
#include <proto/timer.h>
#include <proto/dos.h>
#include <proto/exec.h>

                           /* Defines & macros */
#define CHECK_PERIOD 8
#define INACTIVITY_PERIOD 60  /* This period must elapse */
#define INACTIVITY_CHECK 2    /* Every two seconds is inactivity checked */
#define MAXLINE 256
#define SUPERLINE 256
#define MEMTYPE NULL
#define ERROUT stdout
#define RETURN_ERROR 10
#define false 0
#define true  1
#define SEPARATOR ','
#define DEFAULT_ACTIVATE "relaynews,rnews,expire,trimnews"
#define DEFAULT_CALL     "arnmaster"
#define SIGS (SIGBREAKF_CTRL_C|SIGBREAKF_CTRL_D|SIGBREAKF_CTRL_E|SIGBREAKF_CTRL_F)
#define FOREVER for(;;)
#define CLI(proc) ((struct CommandLineInterface *)(BADDR(proc->pr_CLI)))

                                /* typedefs */
typedef struct timerequest TIMEREQ;
typedef struct MsgPort     PORT;
typedef struct Process     PROC;
typedef struct Library     LIB;
typedef struct IORequest   IOREQ;
typedef struct CliProcList CPL;

                       /* ErrorMsg Codes */
enum { OK, NOMEM, NOTIMERDEV, NONULLDEV, STRANGE, INFO, USAGE };
                       /* switches */
enum { SW_INFO=0x01, SW_PRIORITY=0x02 };

char *InitMessage= "ArnDaemon %s (%s/%s)\nCopyright  1992 by Roland Bless,\
 Byteable Software Products\n\
Written 1992 by Roland Bless, all rights reserved.\n\n";

char *ErrMsg[]= {
  "OK.\n",
  "FATAL ERROR: Can't get enough memory!",
  "FATAL ERROR: Can't open timer.device",
  "WARNING: NO Null-device found",
  "STRANGE ERROR: %s"
};  /* Messages for ErrorHandle/Numbers */

static int DaysofMonth[]=
  { 31, 28, 31, 30,
    31, 30, 31, 31,
    30, 31, 30, 31 };

char *tmp= NULL,       /* temporary multi-purpose string */
     *activate=NULL,   /* on which programs do we start program, see DEFAULT_ACTIVATE */
     *program=NULL,    /* program that is executed, see DEFAULT_CALL */
     *prgname=NULL,    /* This programs callname */
     *logfilename=NULL;

long OS_VER, OldPriority;  /* OS Version-Code and Old_Priority */
int GlobalERROR,           /* Error Code */
    switches=0;            /* command line options */

TIMEREQ TReq;                 /* Timer-request */
PORT    *TPort= NULL;         /* Timer-msgport */
struct FileHandle *NullFH;    /* NULL:-device */
CPL   *RootCliProc= NULL, *Tail= NULL;  /* RootCliProcess */
APTR  ProcTable,              /* Dos-Process-table */
      oldConsoleTask;
BOOL  treq_open= false;       /* timer Device open? */
ULONG CheckPeriod= CHECK_PERIOD;
PROC  *this_proc;             /* process struct of this task */


  /************************ E x t e r n a l s *****************************/

/* from arndaemon_ver.c */
extern char *Version, *Last_UpdateD, *Last_UpdateT;

                       /***********************/


    /* time functions */
ULONG GetTime(void)
{
  struct DateStamp MyDSt;

  DateStamp(&MyDSt);
  return (ULONG) MyDSt.ds_Days*86400+MyDSt.ds_Minute*60+MyDSt.ds_Tick/TICKS_PER_SECOND;
}


int leapyear (int year)
{
  return (year % 4 == 0 && year % 100 != 0 || year % 400 == 0) ? 366 : 365;
}

struct tm *
mytime(struct tm *ptm, register ULONG clock)
{
  register int year= 1978, month= 0;

  ptm->tm_sec= clock % 60;
  clock/= 60;
  ptm->tm_min= clock % 60;
  clock/= 60;
  ptm->tm_hour= clock % 24;
  clock/= 24;
  ptm->tm_wday= clock % 7;

  while (clock >= 366)
  {
    clock-= leapyear(year);
    year++;
  }
  if (clock == 365L && leapyear(year)!= 366)
  {
    clock-=365;
    year++;
  }

  DaysofMonth[1]= leapyear(year)==366? 29 : 28;

  ptm->tm_yday= clock;

  while (clock >= DaysofMonth[month])
  {
    clock-=DaysofMonth[month];
    month++;
  }

  clock++;
  month++;

  ptm->tm_mday= (short) clock;
  ptm->tm_mon=  month;
  ptm->tm_year= year;

  return ptm;
}


char*        /* returns valid ARPANET-Format string of current time */
TimeStr(void)
{
  struct tm Time;
  static char timestr[40];

  static char *Month[]=
  {
    "Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"
  };

  if (mytime(&Time, GetTime()))
  {
    sprintf(timestr, "%02d %s %02d %02d:%02d:%02d",
            Time.tm_mday,
            Month[Time.tm_mon-1],
            Time.tm_year-1900,
            Time.tm_hour,
            Time.tm_min,
            Time.tm_sec);

    return timestr;
  }
  else
    return NULL;
}

                 /***/

void logmessage(char *text)
{
  FILE *logfile;

  if (logfilename==NULL) return; /* No logfile */

  logfile= fopen(logfilename,"a");
  if (logfile)
  {
    fprintf(logfile, "%s%s -(%s)- %s\n", prgname, Version, TimeStr(), text);
    fclose(logfile);
  }
}


void ErrorHandle(int ErrMsgNr, char *addinfo)
{
  char *errtmp, *errtmp2;

  if (ErrMsgNr == INFO)
  {
     fprintf(ERROUT, InitMessage, Version, Last_UpdateD, Last_UpdateT);
     ErrMsgNr= USAGE;
  }
  if (ErrMsgNr == USAGE)
  {
     fprintf(ERROUT, "ArnDaemon usage: %s [-i?] [-L logfile][-X program][-A programlist][-P(n)T(n)]\n", prgname);
     ErrMsgNr= OK;
  }
  else
  {
    errtmp= AllocMem((ULONG) 2*MAXLINE,MEMTYPE);
    errtmp2=errtmp+MAXLINE;

    if (errtmp)
    {
      sprintf(errtmp2,ErrMsg[ErrMsgNr],addinfo);
      sprintf(errtmp, "\n%s: %s\n", prgname, errtmp2);
      fprintf(ERROUT, errtmp);
    }
    else
      fprintf(ERROUT, "\n%s: %s", prgname, ErrMsg[ErrMsgNr]);

    logmessage(errtmp2); /* log error-msg */

    if (errtmp) FreeMem(errtmp, (ULONG) 2*MAXLINE);
    GlobalERROR= ErrMsgNr;
  }
}


char *
CopyString(char *source)
/* Allocates String, copies it and returns pointer to it */
{
  char *dest;

  dest=AllocMem((ULONG) strlen(source)+1UL,NULL);
  if (dest)
     strcpy(dest,source);
  else
    ErrorHandle(NOMEM,NULL);

  return dest;
}


void FreeString(char *string)
{
  if (string) FreeMem(string, (ULONG) strlen(string)+1UL);
}


int
CloseDaemon(void)
{
  logmessage("Terminated.");

  if (tmp) FreeMem(tmp, (ULONG) SUPERLINE+1UL);
  FreeString(activate);
  FreeString(program);

  if (TPort) DeletePort(TPort);
  if (treq_open) CloseDevice((IOREQ *) &TReq);

  if (switches & SW_PRIORITY)
     SetTaskPri(FindTask(NULL),OldPriority);

  if (NullFH) Close((BPTR) NullFH);

  this_proc->pr_ConsoleTask = oldConsoleTask;
  return (0);
}


void InitDaemon(void)
{
  char *nulldev;
  struct ExecBase *ExecBase;
  struct DosLibrary *DosBase;

  this_proc= (PROC *) FindTask(NULL);
  oldConsoleTask= this_proc->pr_ConsoleTask;

  tmp= (char *) AllocMem((ULONG) SUPERLINE+1L, MEMTYPE);

  if (!(tmp))
  {
    ErrorHandle(NOMEM,NULL);
    CloseDaemon();
    exit(RETURN_ERROR);
  }


  if (!(ExecBase = (struct ExecBase *) OpenLibrary("exec.library", 0L)))
  {
    ErrorHandle(STRANGE,NULL);
    CloseDaemon();
    exit(RETURN_ERROR);
  }
  else
  {
    OS_VER=ExecBase->LibNode.lib_Version;      /* Get the OS-Version */
    CloseLibrary((struct Library *) ExecBase);
  }

  if (!(DosBase = (struct DosLibrary *) OpenLibrary("dos.library", 0L)))
  {
    ErrorHandle(STRANGE,NULL);
    CloseDaemon();
    exit(RETURN_ERROR);
  }
  else
  {
    ProcTable= BADDR(DosBase->dl_Root->rn_TaskArray);

    if (OS_VER>=36L)
    {
      RootCliProc= (CPL *) DosBase->dl_Root->rn_CliList.mlh_Head;
      Tail= (CPL *) &DosBase->dl_Root->rn_CliList.mlh_Tail;
#ifdef DEBUG
    printf("List:%lx Head: %lx Tail:%lx TailPred:%lx\n", &DosBase->dl_Root->rn_CliList,
            RootCliProc, DosBase->dl_Root->rn_CliList.mlh_Tail,
            DosBase->dl_Root->rn_CliList.mlh_TailPred);
#endif
    }
    CloseLibrary((LIB *) DosBase);
  }

  /*    Execute()
        Under V37, if an input filehandle is passed, and it's either
        interactive or a NIL: filehandle, the pr_ConsoleTask of the new
        process will be set to that filehandle's process (the same applies
        to SystemTagList()).
  */

  if (OS_VER<37)
   nulldev="NULL:Arn";
  else
   nulldev="NIL:";

  if (!(NullFH= (struct FileHandle *) Open(nulldev, 1006L)))
  {
    ErrorHandle(NONULLDEV,NULL);
  }
  else
  if (OS_VER<37)
     this_proc->pr_ConsoleTask = DeviceProc("NULL:");

  if (OpenDevice("timer.device", 0, (IOREQ *) &TReq, UNIT_VBLANK))
  {
    ErrorHandle(NOTIMERDEV,NULL);
    CloseDaemon();
    exit(RETURN_ERROR);
  }
  else
   treq_open= true;
}


BOOL
SpecialPartialStringMatch(char *stringlist, char *string_start, char *string_end)
{
  short int act_pos= 0;
  char *string= string_start;

  while(*stringlist!='\0')
  {
    while((string+act_pos)<=string_end && stringlist[act_pos] &&
          (toupper(stringlist[act_pos])==toupper(string[act_pos])))
       act_pos++;
    if (stringlist[act_pos]=='\0' || stringlist[act_pos]==SEPARATOR)
    {
      return true;
    }
    else /* no match */
    {
      if (string+act_pos>string_end) /* try next substring in stringlist */
      {
        stringlist+=act_pos;
        while(*stringlist && *stringlist!=SEPARATOR) stringlist++;
        if (*stringlist==SEPARATOR) stringlist++;
        string= string_start; /* reset string to beginning */
      }
      else /* try next partial match in string (taskname) */
       string++;

      act_pos=0;
    }
  }
  return false; /* not found */
}


/* #define DEBUG */
BOOL SearchArray(char *activate, PORT **cpl_Array)
/* Seaches an array containing PTRs to Process IDs (msgports) */
{
  char *tname_end, *taskname;
  ULONG max_taskname_index;
  BOOL found= false;
  LONG index, maxindex;
  PORT *TaskPort;
  struct CommandLineInterface *CliP;

  maxindex= (LONG) cpl_Array[0];
#ifdef DEBUG
    printf("maxindex:%ld\n",maxindex);
#endif

  for(index= 1; index<=maxindex && !found; index++)
  {
    TaskPort=cpl_Array[index]; /* The tasks msgport */
    if (TaskPort) /* CLI-Process is existent */
    {
      /* get the baby's name */
#ifdef DEBUG
      printf("pr_TaskNum\n");
#endif
      if (((PROC *)(TaskPort->mp_SigTask))->pr_TaskNum) /* Has a pr_CLI */
      {
        /* cli_CommandName is a f*cking BCPL String! */
#ifdef DEBUG
      printf("cli_CommandName\n");
#endif
        if (CliP= CLI(((PROC *)(TaskPort->mp_SigTask))))
        {
          if (CliP->cli_CommandName)
            taskname= (char *) BADDR(CliP->cli_CommandName);
          else
            continue;
        }
        else
            continue;

        if (taskname==NULL) continue;
        max_taskname_index=(ULONG) taskname[0];
        tname_end= taskname+max_taskname_index+1;
#ifdef DEBUG
        strncpy(tmp, &taskname[1], (int) max_taskname_index);
        tmp[max_taskname_index]='\0';
        printf("checking:%2ld %s\n",index, tmp);
#endif
      }
      else
      {
        continue;
      }
      if (max_taskname_index==0) continue; /* skip:"No command loaded" */

      /* check for taskname in string activate, which contains
         several names separated by commata                    */

      found= SpecialPartialStringMatch(activate,taskname+1,tname_end);
    } /* endif TaskPort */
  } /* end for all indices */

  return found;
}
#undef DEBUG

BOOL
FindProgram(char *activate)
/* searches in the CLI-Process-list for tasks with names
   like in activate. If one matches 'true' is returned
   The tasknames must be separated by a comma */
{
  CPL *pCliList;
  PORT **cpl_Array;
  BOOL found= false;

#ifdef DEBUG
  printf("Called FindProgram\n");
#endif

  Forbid(); /* No Task switching while traversing the Process list */

  if (OS_VER<36) /* limit of 20 CLIs */
  {
    found= SearchArray(activate, (PORT **) ProcTable);
  }
  else
  {
    for (pCliList= RootCliProc; pCliList!=NULL && pCliList!=Tail && !found;
                                pCliList=(CPL*) (pCliList->cpl_Node.mln_Succ))
    {
#ifdef DEBUG
      printf("outer loop! pcli: %lx succ:%lx prev:%lx first:%ld\n",
           pCliList,pCliList->cpl_Node.mln_Succ,pCliList->cpl_Node.mln_Pred,
             pCliList->cpl_First);
#endif
      cpl_Array= pCliList->cpl_Array;
      if (cpl_Array==NULL) break;
      found= SearchArray(activate, cpl_Array);
    }
  }

#ifdef DEBUG
  printf("leaving FindProgram\n");
  if (found) printf("FOUND!\n");
#endif
  Permit(); /* Enable taskswitching */

  return found;
}
#undef DEBUG


int
GetNumberfromSwitch(char *str, long *val)
/* Gets number from a switch, result is stored in val,
   returns number of converted chars                   */
{
  char buf, *t;

  *val=0L;

  for (t=str; *t!='\0' && (isdigit(*t) || *t=='-'); t++);

  buf=*t;
  *t='\0';
  if (*str=='\0')
  {
    *t=buf;
    return 0;
  }
  else
   *val= atol(str);

  *t=buf;

  return (int) (t-str);
}


int
SetPriority(char *prio)
/* Sets this tasks priority and sets SW_PRIORITY the original priority
   is stored in OldPriority */
{
  long priority;
  int  len;

  len= GetNumberfromSwitch(prio,&priority);

  if (priority>=-127 || priority<=128)
  {
     OldPriority= SetTaskPri(FindTask(NULL),priority);
     switches|=SW_PRIORITY;
  }
  return len;
}


int
checkswitches(int *sw, int argc, char **argvp)
/* handels all command-line parameters */
{
  int i, j;
  BOOL skipnextoption= false;

  logfilename= activate= NULL;

  for (*sw= 0,j=1; j<argc;j++)
  {
    if (*argvp[j]=='-')
    {
      for (i=1; argvp[j][i]!='\0'; i++)
      {
        switch(argvp[j][i])
        {
          case 'i':
          case '?': *sw|= SW_INFO;
                    break;
          case 'L': if (j+1<=argc)
                    {
                      logfilename=argvp[j+1];
                      skipnextoption=true;
                    }
                    break;
          case 'P': i+=SetPriority(&(argvp[j][i+1]));
                    /* SW_PRIORITY is set if prio is changed */
                    break;
          case 'A': if (j+1<=argc)
                    {
                      activate= CopyString(argvp[j+1]);
                      skipnextoption=true;
                    }
                    break;
          case 'X': if (j+1<=argc)
                    {
                      program= CopyString(argvp[j+1]);
                      skipnextoption=true;
                    }
                    break;
          case 'T': i+=GetNumberfromSwitch(&(argvp[j][i+1]),&CheckPeriod);
                    if (CheckPeriod==0) CheckPeriod=1UL;
                    break;

           default: ErrorHandle(USAGE,NULL);
                    return 1;
                    break;
        }

        if (skipnextoption) /* Parameter value given */
        {
          j++; /* skip groupname */
          skipnextoption= false;
          break;
        }
      }
    }
    else
    {
      ErrorHandle(USAGE,NULL);
      return 1;
      break;
    }

  }

  if (activate==NULL)
  {
    activate= CopyString(DEFAULT_ACTIVATE);
  }
  if (program==NULL)
  {
    program= CopyString(DEFAULT_CALL);
  }
#ifdef DEBUG
  printf("program: %s\n",program);
  printf("activate: %s\n",activate);
  if (logfilename) printf("logfile: %s\n",logfilename);
#endif
  return 0;
}


#ifdef __SASC
int nop_func(void)
{
  return 0;
}
#endif


main(int argc, char **argv)
{
  long mask, InactivityCounter=0L;
  unsigned int detected;

  detected=0;
  prgname=argv[0];

#ifdef __SASC
  onbreak(&nop_func);
#endif

  fclose(stderr);

  InitDaemon();
  checkswitches(&switches, argc, argv);

  if (switches & SW_INFO)
  {
    ErrorHandle(INFO,NULL);
    CloseDaemon();
    exit(0);
  }

  sprintf(tmp, InitMessage, Version, Last_UpdateD, Last_UpdateT);
  printf(tmp);

  logmessage("Startup");

  /* Initiate Timerequest */
  TPort= CreatePort(NULL,0);
  TReq.tr_time.tv_secs = CheckPeriod;
  TReq.tr_time.tv_micro= 0;
  TReq.tr_node.io_Message.mn_ReplyPort = TPort;
  TReq.tr_node.io_Command = TR_ADDREQUEST;
  SendIO((IOREQ*) &TReq);                                  /*  timeout */
  SetSignal(0UL,(ULONG) (SIGS|(1UL << TPort->mp_SigBit)));

  FOREVER
  {
    mask= Wait(SIGS | (1UL << TPort->mp_SigBit));
    /* Sleep until a break signal comes or timer device wakes you up */

#ifdef DEBUG
    printf("got signals!\n");
#endif
    if (mask & SIGS)
    {
      logmessage("Break");
      break;
    }

    if (CheckIO((IOREQ *) &TReq))
    { /* request ready? (not necessary, since all other signals break) */
      WaitIO((IOREQ *)&TReq);
      SetSignal(0UL,(ULONG) (SIGS|(1UL << TPort->mp_SigBit)));

      if (detected) /* we have found the program and waited for PAUSE_PERIOD */
      {
        if (!FindProgram(activate)) /* program has terminated */
        {
          if (InactivityCounter)
          {
            InactivityCounter--;
            TReq.tr_time.tv_secs= INACTIVITY_CHECK;
          }
          else
          {
            sprintf(tmp,"Calling %s",program);
            logmessage(tmp);
            Execute(program, (BPTR) NullFH, (BPTR) NullFH); /* call program */
            sprintf(tmp,"%s returned.",program);
            logmessage(tmp);
            detected= 0;
            TReq.tr_time.tv_secs= CheckPeriod; /* check again for normal intervals */
          }
        }
        else /* program was started again during INACTIVITY_PERIOD */
        {
          TReq.tr_time.tv_secs= INACTIVITY_CHECK; /* wait again for inactivity */
          InactivityCounter= INACTIVITY_PERIOD/INACTIVITY_CHECK+1L;
        }
      }
      else
      {
        if (FindProgram(activate)) /* I found the program! */
        {
          detected= 1;
          TReq.tr_time.tv_secs= INACTIVITY_CHECK; /* CHECK every X secs for inactivity */
          InactivityCounter= INACTIVITY_PERIOD/INACTIVITY_CHECK+1L; /* reset counter */
        }
        else /* nothing detected, yet -> wait again */
        {
          TReq.tr_time.tv_secs= CheckPeriod;
        }
      }
      TReq.tr_time.tv_micro= 0;
#ifdef DEBUG
      printf("wait %lu\n",TReq.tr_time.tv_secs);
#endif
      SendIO((IOREQ *) &TReq); /* wait another period */
      SetSignal(0UL,(ULONG) (SIGS|(1UL << TPort->mp_SigBit)));
    }
  }
  /* only break exits this loop */
  AbortIO((IOREQ *) &TReq);
  WaitIO((IOREQ *) &TReq);

  CloseDaemon();
  exit(0);
}
