/***********************************************************************
 * file: ArnIO.c -- Routines for In-/Output-functions                  *
 *                                                                     *
 * Arn - (C)opyright by Roland Bless, All rights reserved,             *
 * -------- YOU ARE NOT ALLOWED TO DISTRIBUTE A CHANGED SOURCE ------- *
 ***********************************************************************/

#ifndef GLOBALVARS_H
 #include "globalvars.h"
#endif
#ifndef ENUMS_H
 #include "enums.h"
#endif
#ifndef EXEC_MEMORY_H
 #include <exec/memory.h>
#endif

#include "amiga_ansi.h"
#include "proto/termcap_proto.h"
#include "proto/tgoto_proto.h"
#ifdef LATTICE_50
#include "proto/raw_proto.h"
#endif

/* Prototypes for functions defined in ArnIO.c */
/* proto$start */
void ConDevWriteLen(char *text,
              int length);
void ConDevWrite(char *text);
void ConDevWriteLn(char *text);
void ConDevWriteChar(char c);
void ConDevRead(struct IOStdReq *StdReq,
                char *buffer,
                ULONG length);
void DispMsg(char *text);
void AuxWrite(char *text);
void AuxWriteLn(char *text);
void AuxWriteChar(char c);
void GotoXY(int x,
            int y);
void ConClear(void);
void ConCR(void);
void ClrLine(void);
void ClrScr(void);
void StatusLine(void);
void InvON(void);
void InvOFF(void);
char GetKeyConDev(void);
char GetUpKey(void);
char GetKeyAux(void);
char WaitForSpace(char *prompt);
char More(int Percent,
          ULONG ArtNr,
          char *Status,
          long togo,
          ULONG unread);
short int CursorKeys(char *key);
void Getline(char *s,
             int maxline);
void initmyf(void);
char *myfgets(register  char *s,
              int len,
              struct FileHandle *pfh);
char *tgetstrreq(char *id, char **area);
void InitTerminalCodes(void);
/* proto$end */

struct MsgPort *replyport;
struct StandardPacket *packet;
extern char directcommand;
extern char *TermName;
char *commonkeys= NULL;

/*******************/

char *p_fbuf, *p_fbufEOF, *p_fbufEND, *fbuffer;
long ReadBytes, fbuflen;
char *tc_entry= NULL,*tc_vector= NULL, *nullstr="\0";

  /*------------------------------------------------------------------------*/

void ConDevWriteLen(char *text, int length)
{
  conWriteReq->io_Command= CMD_WRITE;
  conWriteReq->io_Data= (APTR) text;
  conWriteReq->io_Length= length;
  DoIO((struct IORequest *) conWriteReq); /* DoIO removes the message from the msgport
                                             and clears all signals */
}


void ConDevWrite(char *text)
{
  conWriteReq->io_Command= CMD_WRITE;
  conWriteReq->io_Data= (APTR) text;
  conWriteReq->io_Length= -1;
  DoIO((struct IORequest *) conWriteReq);
}


void ConDevWriteLn(char *text)
{
  conWriteReq->io_Command= CMD_WRITE;
  conWriteReq->io_Data= (APTR) text;
  conWriteReq->io_Length= -1;
  DoIO((struct IORequest *) conWriteReq);
  conWriteReq->io_Command= CMD_WRITE;
  conWriteReq->io_Data= (APTR) "\n";
  conWriteReq->io_Length= 1L;
  DoIO((struct IORequest *) conWriteReq);
}



void
AuxWrite(char *text)
{
  fputs(text,stdout);
  fflush(stdout);
}


void
AuxWriteLn(char *text)
{
  fputs(text,stdout);
  puts("\r");
  fflush(stdout);
}


void ConDevWriteChar(char c)
{
  ConDevWriteLen(&c,1);
}


void
AuxWriteChar(char c)
{
  fputc(c,stdout);
  fflush(stdout);
}

void ConDevRead(struct IOStdReq *StdReq, char *buffer, ULONG length)
{
  StdReq->io_Command= CMD_READ;
  StdReq->io_Data= (APTR) buffer;
  StdReq->io_Length= length;
  SendIO((struct IORequest *) StdReq);
}


void DispMsg(char *text)
{
  (*ConWrite)(EmphON);
  (*ConWrite)(text);
  (*ConWrite)(EmphOFF);
}


void
GotoXY(int x, int y) /* position cursor to this coords */
{
  char buffer[20];

  if (auxmode)
  {
    (*ConWrite)(tgoto(tc_GOTO_XY,x-1,y-1));
  }
  else /* faster for AMIGA mode */
  {
    sprintf(buffer,tc_GOTO_XY,y,x);
    (*ConWrite)(buffer);
  }
}


void ConClear(void) /* removes any pending input characters from the console */
{
  if (auxmode)
  {
    /* fflush(stdin); */
    return;
  }
  conReadReq->io_Command=CMD_CLEAR; /* causes a bug under OS2.04 without setpatch >=V37.28 */
  DoIO((struct IORequest *) conReadReq);
}


void ConCR(void)  /* Puts CRLF to Console */
{
  (*ConWrite)("\r\n");
}


void ClrScr(void)
{
  (*ConWrite)(tc_CLRSCR);
}


void ClrLine(void)
{
  (*ConWriteChar)(RETURN);
  (*ConWrite)(tc_CLR_TO_EOL);
}


void
StatusLine(void)
/* sets the cursor on the last Screenline (first column) */
{
  GotoXY(1,MaxScrLines);
  ClrLine();
}


void
InvON(void)
{
  (*ConWrite)(tc_INV_ON);
}


void InvOFF(void)
{
  (*ConWrite)(tc_INV_OFF);
}


char
GetKeyAux(void)
{
  return (char) getchar();
}

char
GetKeyConDev(void)
{
  char c='\0';
  BYTE error;

  conReadReq->io_Command= CMD_READ;
  conReadReq->io_Data= (APTR) &c;
  conReadReq->io_Length= 1;
  error= DoIO((struct IORequest *) conReadReq);
  if (error!=0) fprintf(ERROUT,"GetKey Error:%x\n",error);

  /* former
     ConRead(conReadReq, &c, 1L);
     Wait(1L<<conReadPort->mp_SigBit);
     GetMsg(conReadPort);
  */

  return c;
}


char
GetUpKey(void)
{
  char c;

  c= (*GetKey)();

  return (char) toupper(c);
}


char WaitForSpace(char *prompt)
{
  char c;
  char *defprompt= "[Type a key to continue]";

  if (!prompt) prompt=defprompt;

  InvON();
  (*ConWrite)(prompt);
  InvOFF();
  c= (*GetKey)();
  ClrLine();

  return c;
}



char More(int Percent, ULONG ArtNr, char *Status,long togo, ULONG unread)
{
  char mtmp[80];  /* **** Must be big enough!!! **** */

  if (directcommand!='\0')
     return directcommand;

  if (Percent>=0 && Percent <=100)
  {
    if (togo>=0)
      sprintf(mtmp,"--- MORE (%d%%) <%lu> [%ld togo] %s ---", Percent,ArtNr,togo,Status);
    else
      sprintf(mtmp,"--- MORE (%d%%) <%lu> (%lu unread) %s ---", Percent,ArtNr,unread,Status);
  }
  else
     sprintf(mtmp,"--- MORE <%lu> <%lu> %s ---",ArtNr,unread,Status);

  InvON();
  (*ConWrite)(mtmp);
  InvOFF();
  ConClear();
  mtmp[0]=(*GetKey)();
  ClrLine();  /* ClearLine */

  return mtmp[0];
}


short int
CursorKeys(char *key)
/* returns the CUNUM_X if cursor sequence was detected
 * WARNING: this procedure presumes that all 4 CURSOR strings
 * begin with the same chars (commonkeys) and that only one
 * distinct character follows!
 */
{
  if (auxmode)
  {
    char *c= commonkeys;
    int i;

    if (c)
    {
      while(*c && (*key==*c))
      {
        c++;
        if (*c)
           *key= (*GetKey)();
#ifdef DEBUG
        printf("1 got:%x\n",*key);
        Delay(150L);
#endif
      }
      if (*c) /* wrong sequence -> ungetc */
      {
        return 0;
      }
      else /* sequence okay so far */
      {
        i= c-commonkeys;

        *key=(*GetKey)();
#ifdef DEBUG
        printf("2 got:%x\n",*key);
        Delay(150L);
#endif
        if (*key==tc_KEY_CURSOR_UP[i])
          return CUNUM_UP;
        else
        if (*key==tc_KEY_CURSOR_DOWN[i])
          return CUNUM_DOWN;
        else
        if (*key==tc_KEY_CURSOR_LEFT[i])
          return CUNUM_LEFT;
        else
        if (*key==tc_KEY_CURSOR_RIGHT[i])
          return CUNUM_RIGHT;
        else
        {
          return 0;
        }
      }
    }
  }
  else /* AMIGA stuff */
  if (*key=='\x9b')
  {
     *key= (*GetKey)();
     switch(*key)
     {
       case 'A':return CUNUM_UP;
                 break;
       case 'B':return CUNUM_DOWN;
                break;
       case 'C':return CUNUM_RIGHT;
                break;
       case 'D':return CUNUM_LEFT;
                break;
     }
  }
  return 0;
}


void Getline(char *s, int maxline)
{
  int cp=0, strend=0, tmpcp, cursor_num=0;
  char c;

  if (s)
     strend= strlen(s);

  (*ConWrite)(s);

  do
  {
    for (cp=strend; (c= (*GetKey)())!=RETURN && cp<MAXLINE;)
    {
      if (cursor_num= CursorKeys(&c))
      {
        if (cursor_num==CUNUM_RIGHT)
        {
          if (cp<strend)
          {
            cp++;
            (*ConWrite)(tc_CURSOR_RIGHT);
          }
        }
        else
        if (cursor_num==CUNUM_LEFT)
        {
          if (cp>0)
          {
            cp--;
            (*ConWrite)(tc_CURSOR_LEFT);
          }
        }
      }
      else
      if (c==*tc_KEY_BACKSPACE) /* backspace */
      {
        if (cp>0)
        {
          for(tmpcp= cp-1; tmpcp<strend; tmpcp++)
              s[tmpcp]= s[tmpcp+1];
          s[tmpcp]='\0';
          cp--; strend--;
          (*ConWrite)(tc_CURSOR_LEFT);
          (*ConWrite)(tc_DELCHAR);
        }
      }
      else
      if (c==*tc_KEY_DELCHAR)
      {
        for(tmpcp= cp; tmpcp<strend; tmpcp++)
           s[tmpcp]= s[tmpcp+1];
        s[tmpcp]='\0';
        strend--;
        (*ConWrite)(tc_DELCHAR);
      }
      else
      if (strend<maxline)  /* insert a character */
      {
        (*ConWrite)(tc_INSCHAR); (*ConWriteChar)(c);
        for (tmpcp= strend; tmpcp>=cp; tmpcp--)
          s[tmpcp+1]=s[tmpcp];
        strend++;
        s[cp]=c;
        cp++;
#ifdef TEST
        printf("%s\n",s);
#endif
      }
#ifdef TEST
      else
       printf("No print: %c %d %x\n", c, c, c);
#endif
    }
  }
  while (c!=RETURN);
  s[strend]='\0';
#ifdef TEST
        printf(":%s:\n",s);
#endif
}


void initmyf(void)
{

  fbuffer= mBuffer;             /* Buffer-address */
  fbuflen= M_BUFFERSIZE;        /* Buffer-length */
  p_fbufEND= fbuffer+fbuflen;   /* Pointer to end of buffer */
  p_fbufEOF= p_fbufEND+1;       /* Pointer to EOF-byte in buffer */
  p_fbuf= p_fbufEOF+1;          /* fill buffer! */
}


char *myfgets(register char *s, int len, struct FileHandle *pfh)
{
  static char *smax= NULL;
  char *t=s;

  smax= s+len-1;

#ifdef DEBUG
      printf("read: %ld p_fbuf:%lx p_fbufEND:%lx p_fbufEOF:%lx\n",
             ReadBytes,p_fbuf, p_fbufEND, p_fbufEOF);
#endif

  do
  {
    if (p_fbuf>=p_fbufEND && p_fbuf!=p_fbufEOF)
    {
      ReadBytes= Read((BPTR) pfh, fbuffer, fbuflen);
      if (ReadBytes<fbuflen)
         p_fbufEOF=fbuffer+ReadBytes; /* points right behind last byte */

      p_fbuf= fbuffer;
#ifdef DEBUG
      printf("Filled buffer!:\n");
      printf("read: %ld p_fbuf:%lx p_fbufEND:%lx p_fbufEOF:%lx\n",
             ReadBytes,p_fbuf, p_fbufEND, p_fbufEOF);
#endif
    }
    while (p_fbuf<p_fbufEND && p_fbuf!=p_fbufEOF && s<smax && *p_fbuf!='\n')
    {
      *s= *p_fbuf;
      s++; p_fbuf++;
    }
    if (*p_fbuf=='\n')
    {
      *s= *p_fbuf;
      s++; p_fbuf++;
      *s='\0';
      break;
    }
  }
  while (s<smax && p_fbuf!=p_fbufEOF);

  if (p_fbuf==p_fbufEOF)
#ifdef DEBUG
  {
     printf("EOF reached!\n\n");
#endif
     return NULL;
#ifdef DEBUG
  }
#endif
  else
     return t;
}


char *
tgetstrreq(char *id, char **area)
/* This routine checks for missing required entries */
/* calling the original tgetstr() */
{
  char *strp;

  strp= tgetstr(id,area);
  if (strp==NULL)
     ErrorHandle(INCOMPLETE_TERMCAP,id);

  return strp;
}


void
InitTerminalCodes(void)
{
  MaxChars=80;
  MaxScrLines= 24;
  EmphON= EmphOFF= nullstr; /* needed for DispMsg */

  if (auxmode)
  {
    char **c,*tmp;

    tc_entry =AllocMem(TERMCAP_MAXSIZE, MEMTYPE);
    tc_vector=AllocMem(TERMCAP_MAXSIZE, MEMTYPE);
    tmp=tc_vector;
    c= &tmp;
    if (tc_entry==NULL || tc_vector==NULL)
    {
      ErrorHandle(NOMEM,NULL);
      return;
    }
    if (tgetent(tc_entry,TermName))
    {
      tc_BOLDFACE_ON=   tgetstr("md",c);
      tc_BOLDFACE_OFF=  tgetstr("me",c);
      tc_UNDERLINE_ON=  tgetstr("us",c);
      tc_UNDERLINE_OFF= tgetstr("ue",c);
      tc_SLANTED_ON=    tgetstr("as",c);
      tc_SLANTED_OFF=   tgetstr("ae",c);
      tc_INV_ON=        tgetstr("mr",c);
      if (tc_INV_ON==NULL) /* no "mr" available */
      {
        tc_INV_ON=        tgetstr("so",c); /* use standout mode instead */
        tc_INV_OFF=       tgetstr("se",c);
      }
      else
        tc_INV_OFF=     tgetstr("me",c);
      tc_KEY_DELCHAR=   tgetstr("kD",c);
      tc_CURSOR_UP=     tgetstrreq("up",c);
      tc_CURSOR_DOWN=   tgetstrreq("do",c);
      tc_CURSOR_LEFT=   tgetstrreq("le",c);
      tc_CURSOR_RIGHT=  tgetstrreq("nd",c);
      tc_DELCHAR=       tgetstrreq("dc",c);
      if (tc_DELCHAR==NULL)
         tc_DELCHAR= tgetstrreq("ec",c);
      tc_INSCHAR=       tgetstrreq("ic",c);
      tc_GOTO_XY=       tgetstrreq("cm",c);
      tc_CLRSCR=        tgetstrreq("cl",c);
      tc_CLR_TO_EOL=    tgetstrreq("ce",c);
      tc_KEY_BACKSPACE= tgetstrreq("kb",c);
      tc_KEY_CURSOR_UP=     tgetstrreq("ku",c);
      tc_KEY_CURSOR_DOWN=   tgetstrreq("kd",c);
      tc_KEY_CURSOR_LEFT=   tgetstrreq("kl",c);
      tc_KEY_CURSOR_RIGHT=  tgetstrreq("kr",c);
      MaxChars= tgetnum("co");
      MaxScrLines= tgetnum("li");
      if (tc_BOLDFACE_ON==NULL) tc_BOLDFACE_ON= tc_INV_ON;
      if (tc_BOLDFACE_OFF==NULL) tc_BOLDFACE_OFF= tc_INV_OFF;
      if (tc_UNDERLINE_ON==NULL) tc_UNDERLINE_ON= tc_INV_ON;
      if (tc_UNDERLINE_OFF==NULL) tc_UNDERLINE_OFF= tc_INV_OFF;
      if (tc_SLANTED_ON==NULL) tc_SLANTED_ON= tc_INV_ON;
      if (tc_SLANTED_OFF==NULL) tc_SLANTED_OFF= tc_INV_OFF;

      EmphON= tc_BOLDFACE_ON;
      EmphOFF= tc_BOLDFACE_OFF;

      if (!tc_INV_ON || !tc_INV_OFF)     /* in the worst case */
         tc_INV_ON= tc_INV_OFF= nullstr; /* ignore display attributes */

      /* check for required entries */
      if (!tc_KEY_CURSOR_UP || !tc_KEY_CURSOR_DOWN ||
          !tc_KEY_CURSOR_LEFT || !tc_KEY_CURSOR_RIGHT ||
          !tc_CURSOR_RIGHT || !tc_CURSOR_LEFT ||
          !tc_CURSOR_UP || !tc_CURSOR_RIGHT ||
          !tc_DELCHAR || !tc_INSCHAR || !tc_GOTO_XY || !tc_CLRSCR ||
          !tc_CLR_TO_EOL || !tc_KEY_BACKSPACE)
      {
         if (tc_entry) FreeMem(tc_entry,TERMCAP_MAXSIZE);
         return;
      }

      commonkeys= AllocMem((ULONG)strlen(tc_KEY_CURSOR_UP)+1UL,MEMTYPE);
      if (commonkeys)
      {
        int i= 0;

        while(tc_KEY_CURSOR_UP[i] &&
              (tc_KEY_CURSOR_UP[i]==tc_KEY_CURSOR_DOWN[i]) &&
              (tc_KEY_CURSOR_DOWN[i]==tc_KEY_CURSOR_LEFT[i]) &&
              (tc_KEY_CURSOR_LEFT[i]==tc_KEY_CURSOR_RIGHT[i]))
        {
          commonkeys[i]=tc_KEY_CURSOR_UP[i];
          i++;
        }
        commonkeys[i]='\0';
      }
    }
    else
      ErrorHandle(NOTERMCAP,TermName);

    if (tc_entry) FreeMem(tc_entry,TERMCAP_MAXSIZE);
    return;
  }

  tc_BOLDFACE_ON=  BOLDFACE_ON;
  tc_UNDERLINE_ON= UNDERLINE_ON;
  tc_SLANTED_ON=   SLANTED_ON;
  tc_INV_ON=       INV_ON;
  tc_CURSOR_UP=    CURSOR_UP;
  tc_CURSOR_DOWN=  CURSOR_DOWN;
  tc_CURSOR_LEFT=  CURSOR_LEFT;
  tc_CURSOR_RIGHT= CURSOR_RIGHT;
  tc_CURSOR_UP=    CURSOR_UP;
  tc_KEY_DELCHAR=  KEY_DELCHAR;
  tc_KEY_BACKSPACE= KEY_BACKSPACE;
  tc_KEY_CURSOR_DOWN=  CURSOR_DOWN;
  tc_KEY_CURSOR_LEFT=  CURSOR_LEFT;
  tc_KEY_CURSOR_RIGHT= CURSOR_RIGHT;
  tc_KEY_CURSOR_UP=    CURSOR_UP;
  tc_INSCHAR=          INSCHAR;
  tc_DELCHAR=          DELCHAR;
  tc_GOTO_XY=          GOTO_XY;
  tc_CLRSCR=           CLRSCR;
  tc_CLR_TO_EOL=       CLR_TO_EOL;


  tc_BOLDFACE_OFF= BOLDFACE_OFF;
  tc_UNDERLINE_OFF= UNDERLINE_OFF;
  tc_SLANTED_OFF= SLANTED_OFF;
  tc_INV_OFF= INV_OFF;

  EmphON=tc_BOLDFACE_ON;
  EmphOFF=tc_BOLDFACE_OFF;
}

