/************************************************************************
 * arnmaster - The master process for creating/updating databases       *
 *----------------------------------------------------------------------*
 * (C)opyright 1991-93 by Roland Bless, Amiga-Version, SAS/C 6.3        *
 * Written by Roland Bless, ALL RIGHTS RESERVED                         *
 * Created this file 25-Oct-1991                                        *
 * Another Blessed Software Product                                     *
 ************************************************************************/

/************************************************************************
 * Mode of operation:                                                   *
 * ArnMaster creates special databases for newsgroups, which contain    *
 * the following information:                                           *
 * Article Number, Number of Lines, Date, Article-Status-Flags,         *
 * Length of From:(max. 16), Shortened From:,                           *
 * and depending on the article status:                                 *
 * Length of Subject, Full Subject, Number of FollowUps                 *
 * ----                                                                 *
 * Note that this is a database with "variable" record size!            *
 * ArnMaster reads the active-file, then the old active-file and        *
 * compares which articles are expired and which are new. It reads in   *
 * the existing database (if any) and reads in new articles, scans their*
 * header for new information, then inserts them into the article list. *
 * After this the new database is written back.                         *
 ***********************************************************************/

/*
  Database Format:

  ULONG ArtNr | long Lines | ULONG date | UCHAR ArtStat | UCHAR FromLen |
               char FromString[FromLen]
  [ | UCHAR SubjLen | char SubjLen[SubjLen] | ULONG FollowUpCount ]

*/

/* #define DEBUG */

#include <proto/exec.h>
#include <proto/dos.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <ctype.h>
#include <time.h>
#include <dos.h>

#include "timezones.h"

#ifndef YEAR0
#define YEAR0 1978
#endif

                /**** external protos ****/
extern void CreateDir_IfNotExists(char *name, char *name_end);
               /*************************/


#ifdef AZTEC
 #undef BUFSIZ
 #define BUFSIZ 4096
#endif
#ifdef __SASC
  _buffsize= 4096;
 #undef BUFSIZ
 #define BUFSIZ 4096
#endif

#define MAXLINE 255            /* Maximal 256 Zeichen pro Zeile */
#define SUPERLINE 1024         /* Super-long line */
#define M_BUFFERSIZE 0x4000UL  /* Multi-purpose-buffer size: 16k */
#define SIZEOF_HEADERBUF 0x1000UL /* Header-Buffer */
#define MEMTYPE NULL          /* Fast, Chip or whatever */
#define true 1                /* Boolean as int */
#define false 0
#define FOPEN_RETRY_MAX 5
#define ERROUT stderr
#define RETURN_ERROR 10
#define RETURN_OK 0
#define MAXFROM 16
#define MAXGROUPLEN_DISPLAY 25
#define UCHAR unsigned char
#define MASTERINFO "T:masterinfo"
#define RED 1     /* red for red-black trees   */
#define BLACK 0   /* black for red-black trees */
#define right(x)  (x)->RightSubject
#define left(x)   (x)->LeftSubject
#define parent(x) (x)->ParentSubject
#define color(x)  (x)->color
#define getcolor(x) ((x)==NULL ? BLACK : (x)->color)   /* color must return BLACK for x==NULL */
#define LEFT       0
#define RIGHT      1

/* master-config */
#define DEFAULT_NEWSDIR "UUNEWS:"
#define MASTERCONFIG_1  "UULIB:.arnmasterrc"
#define MASTERCONFIG_2  "UULIB:.arnrc"
#define CONF_ACTIVEFILE   "ACTIVEFILE"
#define CONF_NEWSDIR      "NEWSDIR"
#define CONF_DBASEDIR     "DBASEDIR"
#define CONF_LOGFILE      "ARNMASTERLOG"
#define CONF_DOTDIRS      "DOTDIRS"
#define CONF_UPDATEACTIVE "UPDATEACTIVE"
#define CONF_COUNTLINES   "COUNTLINES"

                            /* Error Numbers */
enum { OK, NOMEM, NOACTIVE, NOCONFIG, NOFROM, NOSUBJ, DBASECORRUPT,
       NONEWSGROUP, WRITE_CHECK, STRANGE, ABORT, INFO, USAGE };

                          /* Flags for Articles */
enum { UNREAD=0x00, READ=0x01, UNUSED_FLAG=0x02, NONEXISTENT=0x04,
 PARENTART=0x08, FOLLOWUPART=0x10, PA_WITHF=0x20, FUP_WITHF=0x40,
 XREF= 0x80 };

/* PARENTART  PA_WITHF  FOLLOWUP  FUP_WITHF  NONEXISTENT XREF */
/* must all fit into the lower byte!!!                        */

                          /* Flags for Newgroups */
enum { SUBSCRIBED=0, UNSUBSCRIBED=1, UNUSED_GFLAG=2, MODERATED=4 };


enum { SW_PRIORITY=0x01, SW_INFO=0x02, SW_GOTOGROUP=0x04, SW_DEBUG=0x08,
       SW_CLEANUP=0x10, SW_UPDATEACTIVE=0x20, SW_SINGLEDIR=0x40,
       SW_COUNTLINES=0x80, SW_CONSISTENCY_CHECK=0x100 };


                    /******** structures *******/

struct art_struct
{
  short Status;                        /* Article Status Flags */
  struct subj_struct *Subject;         /* Pointer to subject */
  char *From;                          /* Pointer to author */
  struct art_struct *PrevArt,*NextArt; /* Article thread */
  long Lines;                          /* number of lines */
  ULONG NrofFollowUps;                 /* number of FollowUps */
  ULONG Date;                          /* Date: */
  void *ExtData;                       /* for future enhancements */
};


struct subj_struct /* binary (red-black) tree structure */
{
  char *Subject;                     /* pointer to the subject itself */
  struct art_struct *RootArt;        /* pointer to first article in thread */
  struct subj_struct *ParentSubject, /* pointer to parent node */
                     *LeftSubject,   /* pointer to left child */
                     *RightSubject;  /* pointer to right child */
  char   color;                      /* for red-black trees */
};


struct              /* standard structure to read in */
{
  ULONG ArtNr; /* because of longword alignment chars must be defined later */
  long Lines;
  ULONG Date;
  UCHAR ArtStat;
  UCHAR FromLen;
} mask;


struct Markers      /* single linked list for read marks in grouplist */
{
  char *Mark;               /* This is the string of ASCII-Ranges */
  int lastpos;              /* length of string Mark, Mark[lastpos]='\0' */
  struct Markers *NextMark; /* Next MarkString */
};


struct GroupInfoList    /* This is a structure for a double-linked list */
{                       /* of newsgroups */
  char *Name;                   /* Name of the Group */
  struct Markers *Marks;        /* The list of read articles */
  char Status;                  /* Status byte of this Newsgroup */
  char Active_Char;             /* Character from active-file */
  ULONG firstArtNr;              /* First Article Number */
  ULONG OldfirstArtNr;           /* First Article Number from the last time */
  ULONG lastArtNr;               /* Last  Article Number */
  ULONG OldlastArtNr;            /* Last Article Number from the last time */
  struct subj_struct *RootSubjP; /* RootP of subject list */
  struct GroupInfoList *PrevGI; /* Previous newsgroup */
  struct GroupInfoList *NextGI; /* Next newsgroup */
};

              /**************** G L O B A L S *******************/

extern int _OSERR;

                                   /***/
struct GroupInfoList *RootGI;
char *tmp,
     *fnametmp,
     *activefile,
     *logfile,
     *prgname,
     *mBuffer,
     *HeaderBuf,
     *pathtogroup,
     *pathtodbase,
     FromTmp[MAXFROM+1],
     formatstr[40];

static short int DaysofMonth[]=
  { 31, 28, 31, 30,
    31, 30, 31, 31,
    30, 31, 30, 31 };

static char *Weekday[]=
{
  "Sun",
  "Mon",
  "Tue",
  "Wed",
  "Thu",
  "Fri",
  "Sat"
};

static char *Month[]=
{
  "Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"
};

char *nobody  ="(nobody)",
     *none    ="(none)",
     *newsdir =NULL,
     *dbasedir=NULL;

#ifdef MYFORMAT
char groupname_tmp[MAXGROUPLEN_DISPLAY+1];
#endif

short switches;
ULONG sizeofHeaderBuf= SIZEOF_HEADERBUF, TotalDBbytes= 0UL;
ULONG OldPriority, FirstArtNr, TotalScanned, TotalGroups,
     ScannedGroups, recounted_lines= 0UL, expired=0UL;
int GlobalERROR, len;
struct subj_struct *RootSubjP;
struct art_struct *RootField;  /* dynamic array of array structs */
char *onlyonegroup;
FILE *logFP;
long TZ_Offset=0L;
ULONG valid_ds= 0UL, stopwatch;
BOOL Dots_to_Slashes= true;
                                   /***/

char *InitMessage= "ArnMaster %s (%s/%s)\nCopyright  1991-93 by Roland Bless,\
 Blessed Software Products\n\
Written 1991-93 by Roland Bless, all rights reserved.\n\n";

char *ErrMsg[]= {
  "OK.\n",
  "FATAL ERROR: Can't get enough memory!",
  "ERROR: Can't find the ACTIVEFILE!",
  "ERROR: Can't find the Config-File",
  "ERROR: Wrong DBase-Format - missing From:! %s",
  "ERROR: Wrong DBase-Format - missing Subject:! %s",
  "ERROR: DBase-format corrupt: %s",
  "WARNING: No directory for %s",
  "WARNING: Too few datasets written in %s",
  "STRANGE ERROR: %s",
  "Break signal detected! Aborting...\n",
};  /* Messages for ErrorHandle/Numbers */



/* prototype */
int CloseMaster(void);


  /************************ E x t e r n a l s *****************************/

/* from version.c */
extern char *Version, *Last_UpdateD, *Last_UpdateT;

                       /***********************/

ULONG max(ULONG x, ULONG y)
{
  return x<y ? y: x;
}


ULONG GetTime(void)
{
  struct DateStamp MyDSt;

  DateStamp(&MyDSt);
  return (ULONG) MyDSt.ds_Days*86400+MyDSt.ds_Minute*60+MyDSt.ds_Tick/TICKS_PER_SECOND;
}


int leapyear (int year)
{
  return (year % 4 == 0 && year % 100 != 0 || year % 400 == 0) ? 366 : 365;
}


struct tm *
mytime(struct tm *ptm, register ULONG clock)
{
  register int year, month=0;

  ptm->tm_sec= clock % 60;
  clock/= 60;
  ptm->tm_min= clock % 60;
  clock/= 60;
  ptm->tm_hour= clock % 24;
  clock/= 24;
  ptm->tm_wday= clock % 7;

  year= YEAR0;
  while (clock >= 366)
  {
    clock-= leapyear(year);
    year++;
  }
  if (clock == 365L && leapyear(year)!= 366)
  {
    clock-=365;
    year++;
  }

  DaysofMonth[1]= leapyear(year)==366? 29 : 28;

  ptm->tm_yday= clock;

  while (clock >= DaysofMonth[month])
  {
    clock-=DaysofMonth[month];
    month++;
  }

  clock++;
  month++;

  ptm->tm_mday= (short) clock;
  ptm->tm_mon=  month;
  ptm->tm_year= year;

  return ptm;
}

#ifdef DEBUG
char*        /* returns ddmmyyyy-Format string of time */
ShortTimeStr(ULONG time)
{
  struct tm Time;
  static char timestr[25];

  if (mytime(&Time,time))
  {
    sprintf(timestr,"%02d %3s %04d %02d:%02d:%02d",
           Time.tm_mday,Month[Time.tm_mon-1],Time.tm_year,
           Time.tm_hour,Time.tm_min,Time.tm_sec);

    return timestr;
  }
  else
    return NULL;
}
#endif


char*        /* returns valid Short-Format string of current time */
TimeStr(void)
{
  struct tm Time;
  static char timestr[40];


  if (mytime(&Time, GetTime()))
  {
#ifdef DEBUG
    printf("Time: %02d-%02d-%02d %02d:%02d:%02d\n",
           Time.tm_mday,Time.tm_mon,Time.tm_year,
           Time.tm_hour,Time.tm_min,Time.tm_sec);
#endif
      sprintf(timestr, "%02d %s %02d %02d:%02d:%02d",
              Time.tm_mday,
              Month[Time.tm_mon-1],
              Time.tm_year-1900,
              Time.tm_hour,
              Time.tm_min,
              Time.tm_sec);
    return timestr;
  }
  else
    return NULL;
}


int
GetTimeOffSet(char *s)
{
  ULONG hashval= 0UL;
/*
  ADT    -03     Atlantic Daylight Time
  AHST   -10     Alaska-Hawaii Standard Time
  AST    -04     Atlantic Standard Time
  AT     -02     Azores Time
  BST    +01     British Summer Time
* BST    -03     Brazil Standard Time
* BST    -11     Bering Standard Time
* BT     -11     "Bering Time"
  BT     +03     Baghdad Time, USSR Zone 2
  CADT   +10.5   Central Australian Daylight Time
  CAST   +09.5   Central Australian Standard Time
  CAT    -10     Central Alaska Time
  CCT    +08     China Coast Time, USSR Zone 7
  CDT    -05     Central Daylight Time
  CEDT   +02     "Central European Daylight Time"
  CES    +02     "Central European Summer time"
  CEST   +02     Central European Summer Time
  CET    +01     Central European Time
  CST    -06     Central Standard Time
  EADT   +11     East Australian Daylight Time
  EAST   +10     East Australian Standard Time
  ECT    +01     "European Central Time"? (Norway)
  EDT    -04     Eastern Daylight Time
  EET    +02     Eastern European Time, USSR Zone 1
  EST    -05     Eastern Standard Time
  FST    +02     French Summer Time
* FST    +02     "Finish Standard Time" ?
  FWT    +01     French Winter Time
  GMT    +00     Greenwich Mean Time
  GST    -03     Greenland Standard Time
* GST    +10     Guam Std Time, USSR Zone 9
  GVA    +01     ? (Switzerland)
  HFE    +02     Heure Francais d'Ete
  HFH    +01     Heure Francais d'Hiver
  IDLE   +12     International Date Line, East
  IDLW   -12     INternational Date Line, West
  IST    +02     Israelian Standard Time
* IST    +05.5   Indian Standard Time
  IT     +03.5   Iran Time
  JST    +09     Japan Std Time, USSR Zone 8
  JT     +07.5   Java Time
  MDT    -06     Mountain Daylight Time
  MEST   +02     Middle European Summer Time
  MESZ   +02     "Mittel Europaeische Sommer Zeit" (Germany)
  MET    +01     Middle European Time
  MEWT   +01     Middle European Winter Time
  MEZ    +01     "Mittel Europaeische Zeit" (Germany)
  MST    -07     Mountain Standard Time
  MT     +08.5   Moluccas Time
  NDT    -02.5   Newfoundland Daylight Time
  NST    -03.5   Newfoundland Standard Time
* NST    +06.5   North Sumatra Time
  NT     -11     Nome Time
  NZDT   +13     New Zealand Daylight Time
  NZST   +12     New Zealand Standard Time
  NZT    +12     "New Zealand Time"
  PDT    -07     Pacific Daylight Time
  PST    -08     Pacific Standard Time
  SADT   +10.5   South Australian Daylight Time
  SAST   +09.5   South Australian Standard Time
  SET    +01     "Standard European Time"
  SST    +02     "Swedish Summer Time"
* SST    +08     Singapore Standard Time
* SST    +07     South Sumatra Time, USSR Zone 6
  SWT    +01     "Swedish Winter Time"
  UT     +00     Universal Time
  UTC    +00     Universal Time, coordinated
  WADT   +08     West Australian Daylight Time
  WAST   +07     West Australian Standard Time
  WAT    -01     West Africa Time
  WET    +00     Western European Time
  WZ     +00     "Weltzeit"
  YST    -09     Yukon Standard Time
  Z      +00     Universal Time
  ZP4    +04     GMT+4 hours.
  ZP5    +05     GMT+5 hours.
  ZP6    +06     GMT+6 hours.
*/
 if (*s=='+' || *s=='-' || isdigit(*s)) /* local differential (hhmm) */
   return (int) atol(s);
 else
 {
   while(*s)
   {
     hashval|=(((toupper(*s)-'A')+1) & 0x001F);
     hashval<<=5;
     s++;
   }
   switch (hashval)
   {
     case TZ_ADT : /* -03 Atlantic Daylight Time */
     case TZ_GST : /* -03     Greenland Standard Time */
     case TZ_C   :
                   return -300;
                   break;
     case TZ_AHST: /* -10 Alaska-Hawaii Standard Time */
     case TZ_HST : /* -10     Hawaii Standard */
     case TZ_HAST: /* -10     Hawaii-Aleutian Standard */
     case TZ_CAT : /* -10 Central Alaska Time */
     case TZ_K   :
                   return -1000;
                   break;
     case TZ_AST : /* -04 Atlantic Standard Time */
     case TZ_EDT : /* -04 Eastern Daylight Time */
     case TZ_D   :
                   return -400;
                   break;
     case TZ_AT  : /* -02 Azores Time */
     case TZ_B   :
                   return -200;
                   break;
     case TZ_BST : /* +01 British Summer Time */
     case TZ_CET : /* +01 Central European Time */
     case TZ_ECT : /* +01 "European Central Time"? (Norway) */
     case TZ_FWT : /* +01 French Winter Time */
     case TZ_GVA : /* +01 ? (Switzerland) */
     case TZ_HFH : /* +01 Heure Francais d'Hiver */
     case TZ_MET : /* +01 Middle European Time */
     case TZ_MEWT: /* +01 Middle European Winter Time */
     case TZ_MEZ : /* +01 "Mittel Europaeische Zeit" (Germany) */
     case TZ_N   :
     case TZ_SET : /* +01 "Standard European Time" */
     case TZ_SWT : /* +01 "Swedish Winter Time" */
                   return 100;
                   break;
     case TZ_BT  : /* -11 "Bering Time" */
     case TZ_NT  : /* -11 Nome Time */
     case TZ_L   :
                   return -1100;
                   break;
     case TZ_CADT: /* +10.5 Central Australian Daylight Time */
     case TZ_SADT: /* +10.5   South Australian Daylight Time */
                   return 1030;
                   break;
     case TZ_CAST: /* +09.5 Central Australian Standard Time */
     case TZ_SAST: /* +09.5   South Australian Standard Time */
                   return 930;
                   break;
     case TZ_CCT : /* +08 China Coast Time, CIS Zone 7 */
     case TZ_WADT: /* +08 West Australian Daylight Time */
     case TZ_HKT : /* +08     Hong Kong */
     case TZ_U   :
                   return 800;
                   break;
     case TZ_CDT : /* -05 Central Daylight Time */
     case TZ_EST : /* -05 Eastern Standard Time */
     case TZ_E   :
                   return -500;
                   break;
     case TZ_CEDT: /* +02 "Central European Daylight Time" */
     case TZ_CES : /* +02 "Central European Summer time" */
     case TZ_CEST: /* +02 Central European Summer Time */
     case TZ_FST : /* +02 French Summer Time */
     case TZ_EET : /* +02 Eastern European Time, CIS Zone 1 */
     case TZ_HFE : /* +02 Heure Francais d'Ete */
     case TZ_IST : /* +02 Israelian Standard Time */
     case TZ_MEST: /* +02 Middle European Summer Time */
     case TZ_MESZ: /* +02 "Mittel Europaeische Sommer Zeit" (G */
     case TZ_O   :
     case TZ_SST : /* +02 "Swedish Summer Time" */
                   return 200;
                   break;
     case TZ_P   :
     case TZ_MSK : /* +03     Moskow Winter */
                   return 300;
                   break;
     case TZ_Q   :
     case TZ_MSD : /* +04     Moskow Summer */
                   return 400;
                   break;
     case TZ_R   :
                   return 500;
                   break;
     case TZ_S   :
                   return 600;
                   break;
     case TZ_CST : /* -06 Central Standard Time */
     case TZ_MDT : /* -06 Mountain Daylight Time */
     case TZ_F   :
                   return -600;
                   break;
     case TZ_EADT: /* +11 East Australian Daylight Time */
     case TZ_X   :
                   return 1100;
                   break;
     case TZ_EAST: /* +10 East Australian Standard Time */
     case TZ_KDT : /* +09     Korean Daylight */
     case TZ_W   :
                   return 1000;
                   break;
     case TZ_GMT : /* +00 Greenwich Mean Time */
     case TZ_UT  : /* +00 Universal Time */
     case TZ_UTC : /* +00 Universal Time, coordinated */
     case TZ_CUT : /* +00 Coordinated Universal */
     case TZ_WET : /* +00 Western European Time */
     case TZ_WZ  : /* +00 "Weltzeit" */
     case TZ_Z   : /* +00     Universal Time */
                   return 0;
                   break;
     case TZ_IDLE: /* +12 International Date Line, East */
     case TZ_NZST: /* +12 New Zealand Standard Time */
     case TZ_NZT : /* +12 "New Zealand Time" */
     case TZ_Y   :
                   return +1200;
                   break;
     case TZ_IDLW: /* -12     INternational Date Line, West */
     case TZ_M   :
                   return -1200;
                   break;
     case TZ_IT  : /* +03.5   Iran Time */
                   return 330;
                   break;
     case TZ_JST : /* +09     Japan Std Time, CIS Zone 8 */
     case TZ_KST : /* +09     Korean Standard */
     case TZ_V   :
                   return 900;
                   break;
     case TZ_JT  : /* +07.5   Java Time */
                   return 730;
                   break;
     case TZ_MST : /* -07 Mountain Standard Time */
     case TZ_PDT : /* -07 Pacific Daylight Time */
     case TZ_G   :
                   return -700;
                   break;
     case TZ_MT  : /* +08.5   Moluccas Time */
                   return 830;
                   break;
     case TZ_NDT : /* -02.5   Newfoundland Daylight Time */
                   return -230;
                   break;
     case TZ_NST : /* -03.5   Newfoundland Standard Time */
                   return -330;
                   break;
     case TZ_NZDT: /* +13     New Zealand Daylight Time */
                   return +1300;
                   break;
     case TZ_PST : /* -08 Pacific Standard Time */
     case TZ_H   :
     case TZ_AKDT: /* -08     Alaska Daylight */
     case TZ_YDT : /* -08     Yukon Daylight */
                   return -800;
                   break;
     case TZ_WAST: /* +07     West Australian Standard Time */
     case TZ_T   :
                   return 700;
                   break;
     case TZ_A   : /* -01 Military */
     case TZ_WAT : /* -01 West Africa Time */
                   return -100;
                   break;
     case TZ_YST : /* -09     Yukon Standard Time */
     case TZ_I   :
     case TZ_AKST: /* -09     Alaska Standard */
     case TZ_HADT: /* -09     Hawaii-Aleutian Daylight */
                   return -900;
                   break;
     default: return 0;
              break;

   }
 }
}


int
GetMonth(char *s)
/* returns number of month from MMM string */
{
  switch(toupper(*s))
  {
    case 'J':s++;
             if (toupper(*s)=='U')
             {
               s++;
               if (toupper(*s)=='N') /* Jun */
                 return 6;
               else /* Jul */
                 return 7;
             }
             else /* Jan */
               return 1;
             break;
    case 'F': return 2;
              break;
    case 'M': s+=2;
              if (toupper(*s)=='R') /* Mar */
                return 3;
              else
                return 5; /* May */
              break;
    case 'A': s++;
              if (toupper(*s)=='P') /* Apr */
                return 4;
              else
                return 8;
              break;
    case 'S': return 9;
              break;
    case 'O': return 10;
              break;
    case 'N': return 11;
              break;
    case 'D': return 12;
              break;
     default: return 12;
              break;
  }
}



ULONG
ParseDate(char *datestring)
/* Returns system ULONG time from ctime()/ARPANET-Date-String */
{
  char month[4],timezone[10];
  struct tm tmstruct;
  ULONG dateval=0UL;
  short int i, offset;

  if (isalpha(*datestring))
     while(*datestring && (!isspace(*datestring))) datestring++; /* skip leading weekday */

  while(isspace(*datestring)) datestring++;

  if(isalpha(*datestring)) /* ctime format Wdy Mon DD HH:MM:SS YYYY */
  {
    sscanf(datestring,"%3s %d %d:%d:%d %d %9s",
           &month,
           &tmstruct.tm_mday,
           &tmstruct.tm_hour,
           &tmstruct.tm_min,
           &tmstruct.tm_sec,
           &tmstruct.tm_year,
           &timezone
           );

  }
  else          /* Wdy, DD Mon YY HH:MM:SS TIMEZONE */
  {
    sscanf(datestring,"%d %3s %d %d:%d:%d %9s",
           &tmstruct.tm_mday,
           &month,
           &tmstruct.tm_year,
           &tmstruct.tm_hour,
           &tmstruct.tm_min,
           &tmstruct.tm_sec,
           &timezone
           );
  }
  tmstruct.tm_mon= GetMonth(month);
  offset= GetTimeOffSet(timezone);
#ifdef DEBUG
   printf("%s -> %02d-%02d-%02d %02d:%02d:%02d %d", datestring,
           tmstruct.tm_mday,tmstruct.tm_mon,tmstruct.tm_year,
           tmstruct.tm_hour,tmstruct.tm_min,tmstruct.tm_sec,offset);
#endif
  if (tmstruct.tm_year<(YEAR0-1900)) /* Year 2000? :-) */
     tmstruct.tm_year+=2000;
  else
  if (tmstruct.tm_year<=100) /* between 78..99 -> mapped to 1978-1999 */
  {
    tmstruct.tm_year+=1900;
  }
  if (--tmstruct.tm_mday>0) dateval+=tmstruct.tm_mday*86400;
  DaysofMonth[1]= leapyear(tmstruct.tm_year)==366? 29 : 28;
  for (i= tmstruct.tm_mon-1; i>0; i--)
      dateval+=DaysofMonth[i-1]*86400;
  for (i= tmstruct.tm_year-1; i>=YEAR0; i--)
      dateval+= leapyear(i)*86400;

  dateval+= (LONG) ((tmstruct.tm_hour-offset/100)*60
                    +(tmstruct.tm_min-offset%100)   )*60
            +tmstruct.tm_sec;

#ifdef DEBUG
printf("=> %s\n",ShortTimeStr(dateval));
#endif
  return dateval;
}


FILE *
fretryopen(char *Name, char *mode)
/* This function tries an fopen(Name,mode) and if it fails
   because the file is in use, it tries again after waiting a
   second. After FOPEN_RETRY_MAX it gives up and returns NULL */
{
  int count= 0;
  FILE *fp;

  while ((fp= fopen(Name, mode))==NULL &&
         _OSERR==ERROR_OBJECT_IN_USE &&
         count<FOPEN_RETRY_MAX)
  {
    Delay(50L);
    count++;
  }

  return fp;
}


void
Display(char *text)
{
  printf("%s",text);
}



void ErrorHandle(int ErrMsgNr, char *addinfo)
{
  char *errtmp, *errtmp2;

  if (ErrMsgNr == USAGE)
  {
     fprintf(ERROUT, "ArnMaster usage: %s [-?idALCP(n)X] [-g|G groupname]\n", prgname);
     ErrMsgNr= INFO;
  }
  if (ErrMsgNr == INFO)
  {
     fprintf(ERROUT, "ArnMaster %s (%s/%s)  -- (C)opyright by Roland Bless\n", Version,
             Last_UpdateD, Last_UpdateT);
  }
  else
  {
    errtmp= AllocMem((ULONG) 2*MAXLINE,MEMTYPE);
    errtmp2=errtmp+MAXLINE;

    if (errtmp)
    {
      sprintf(errtmp2,ErrMsg[ErrMsgNr],addinfo);
      sprintf(errtmp, "\n%s: %s\n", prgname, errtmp2);
      Display(errtmp);
    }
    else
      fprintf(ERROUT, "\n%s: %s", prgname, ErrMsg[ErrMsgNr]);

    if (logFP) /* log error-msg */
       fprintf(logFP, "%s%s -(%s)- ! %s\n",
                   prgname, Version, TimeStr(), errtmp2);

    if (errtmp) FreeMem(errtmp, (ULONG) 2*MAXLINE);
    GlobalERROR= ErrMsgNr;
  }
}


char *
CopyString(char *source)
/* Allocates String, copies it and returns pointer to it */
{
  char *dest;

  dest= AllocMem((ULONG) strlen(source)+1UL,NULL);
  if (dest)
     strcpy(dest,source);
  else
    ErrorHandle(NOMEM,NULL);

  return dest;
}


void
FreeString(char *string)
/* Frees string which was allocated with AllocMem() */
{
  if (string)
     FreeMem(string, (ULONG) strlen(string)+1UL);
}


void
makeslashes(char *s)
{
  if (s)
  while (*s!='\0')
  {
    if (*s=='.') *s='/';
    s++;
  }
}


void
MakeDirName(char *dest, char* dirname, char *newsgroup)
/* Produces path to newsgroup, result goes to dest */
/* sideeffect!!: dest is overwritten, newsgroup: DOTS_TO_SLASHES */
{
  if (Dots_to_Slashes)
     makeslashes(newsgroup);

  if (dirname[strlen(dirname)-1]!=':')
     sprintf(dest, "%s/%s",dirname, newsgroup);
  else
     sprintf(dest, "%s%s",dirname, newsgroup);
}


BOOL
ArticleExists(ULONG ArtNr)
{
  struct FileLock *FL= NULL;

  sprintf(tmp, "%s/%lu", pathtogroup, ArtNr); /* pathtogroup must be set correctly */
  if (!(FL= (struct FileLock *) Lock(tmp, (ULONG) ACCESS_READ)))
  {
    if (IoErr()==ERROR_OBJECT_NOT_FOUND)
       return false;
    else
       return true; /* Lock could not be obtained for other reasons */
  }
  else
  {
    UnLock((BPTR) FL);
    return true;
  }
}


short int
islong(char *s)    /* returns true if s doesn't contain any non-numbers */
{
  char *ps= s;

  while (*s!='\0' && (isdigit(*s))) s++;
  if (*s=='\0' && s!=ps)
     return true;
  else
     return false;
}


void cuteoln(char *s)   /* Cuts off the '\n' from s */
{
  if (s)
  {
    while(*s!='\0' && *s!='\n')
         s++;
    *s='\0';
  }
}


char *skpwspc(char *s) /* returns pointer to first non white-space character */
{                      /* NOTE: s must be !=NULL !! */
  while (*s!='\0' && (*s==' ' || *s=='\t' || *s=='\n')) s++;

  return s;
  /* skips spaces and tabs & LFs, returns first occurance of non white space */
}


char *firststr(register char *s, register char *t)
    /* returns pointer to first byte following t in s */
{
  if (s==NULL || *s=='\0' || toupper(*s)!=toupper(*t))
     return NULL;
  else { s++;t++; }

  while (*t!='\0' && toupper(*s)==toupper(*t))
  {
    t++; s++;
  }
  if (*t != '\0')
     return NULL;
  else
     return s;
}


char *
NextWhiteSpc(register char *s)
{
  if (s)
    while (*s!='\0' && *s!=' ' && *s!='\t' && *s!='\n') s++;

  return s;
}



ULONG
strlen_filter(char *s)
/* returns strlen() and filters non-printable characters */
{
  ULONG i= 0;

  while (*s)
  {
    if (iscntrl(*s)) *s='?';
    s++; i++;
  }
  return i;
}


char *
StripSubject(char *Subject, short *Status)
/* Sets article Status accordingly! */
{
  char *c;

  *Status|= PARENTART;

  while (c= firststr(Subject, "Re:"))
  {
    Subject= skpwspc(c);
    *Status|= FOLLOWUPART;
    *Status&= ~PARENTART;
  }
  while (c= firststr(Subject, "Re^"))
  {
    if (c= strchr(c,':'))
       Subject= skpwspc(c+1);
    else
    {
      c= firststr(Subject, "Re^");
      Subject= skpwspc(NextWhiteSpc(c));
    }
    *Status|= FOLLOWUPART;
    *Status&= ~PARENTART;
  }

  /* Now cut trailing blanks and filter non-printable chars! */
  for (c= Subject; *c!='\0'; c++)
      if (iscntrl(*c)) *c='?';
  if (c>Subject)
  {
    c--;
    while (c>Subject && isspace(*c)) c--;
    *++c= '\0';
  }

  return Subject;
}


char *
findmatchingchar(char *s, char openchr, char closechr)
/* finds outmost pair of <openchr>,<closechr> and returns pointer to
   next char behind last <closechr>. if there is an unnequal # of
   open/closechr then NULL is returned.
*/
{
  int opencount= 0;

  do
  {
    if (*s==openchr)
       opencount++;
    else
    if (*s==closechr)
       opencount--;
    s++;
  }
  while (*s!='\0' && opencount>0);

  if (opencount!=0)
     return NULL;      /* no matching chars found */
  else
     return s;         /* found, s is now one char behind matched char */
}


void
SkipNickNames(char *realname)
/* deletes nicknames from realnames like "Fred (HULABALOO) Foobar" */
{
  char *startnick, *origreal= realname+1; /* +1 to save addition */


  while (*realname!='\0')
  {
    if (*realname=='(') /* (nick) */
    {
      startnick= realname;
      if (realname= findmatchingchar(realname,'(',')'))
      {
        realname= skpwspc(realname);
        if (*realname || startnick>origreal) /* There follows/preceeds text */
        {
          while (*realname!='\0')
            *startnick++= *realname++;
          *startnick= '\0';
        }
        /* copy the text behind the closing char to position of the open char */
      }
      else
       return; /* there was no matching pair of ()s */
    }
    else
    if (*realname=='`')  /* `nick' */
    {
      startnick= realname;
      if (realname= findmatchingchar(realname,'`','\''))
      {
        realname= skpwspc(realname);
        if (*realname || startnick>origreal) /* There follows/preceeds text */
        {
          while (*realname!='\0')
            *startnick++= *realname++;
          *startnick= '\0';
        }
        /* copy the text behind the closing char to position of the open char */
      }
      else
       return; /* there was no matching pair of `'s */
    }
    else
    if (*realname=='\'')  /* 'nick' */
    {
      startnick= realname++;
      while (*realname!='\0' && *realname!='\'' ) realname++;
      if (*realname=='\'') realname= skpwspc(++realname);
      if (*realname || startnick>origreal) /* There follows/preceeds text */
      {
        while (*realname!='\0')
          *startnick++= *realname++;
        *startnick='\0';
      }
    }
    else
    if (*realname=='\"')  /* "nick" */
    {
      startnick= realname++;
      while (*realname!='\0' && *realname!='\"' ) realname++;
      if (*realname=='\"') realname= skpwspc(++realname);
      if (*realname || startnick>origreal) /* There follows/preceeds text */
      {
        while (*realname!='\0')
          *startnick++= *realname++;
        *startnick= '\0';
      }
    }
    else
      realname++;
  }
}


void
CutRealName(char *dest, char *realname)
/* If realname too long this routine makes initials of prenames */
/* NOTE: <dest> need not exceed MAXFROM chars! Should not start with wspace */
/* filters non-printable characters */
{
  char *c,*lastnonwspc;
  int i= 0;

  if (strlen_filter(realname)>MAXFROM) /* Realname has to be shortened */
  {
    SkipNickNames(realname); /* deletes (Gaga) from <Fred (Gaga) Foobar> etc. */

    if (strlen(realname)>MAXFROM) /* Realname has to be shortened */
    {
      lastnonwspc= skpwspc(realname);  /* Get first non-white space */
      c= NextWhiteSpc(lastnonwspc);    /* Next Space? */
      while (*c!='\0' && i<MAXFROM)    /* while not reached end of realname  */
      {                                /* or exceeded MAXFROM in destination */
        dest[i++]= *lastnonwspc;       /* Copy first letter of word into dest */
        if (i<MAXFROM) dest[i++]= ' '; /* separate with a blank */
        lastnonwspc= skpwspc(c);       /* next non white-spc */
        c= NextWhiteSpc(lastnonwspc);  /* Next white-spc */
      }
      for (;i<MAXFROM;i++)   /* No further spaces, copy the rest into dest */
        dest[i]= *lastnonwspc++;
      dest[i]= '\0';          /* i==MAXFROM, so terminate dest */
    }
    else /* realname fits now */
    {
      strcpy(dest, realname);
    }
  }
  else /* realname fits */
  {
    strcpy(dest, realname);
  }
}


int
killrightbracket(char *s)
/* Overwrites the matching closing bracket with '\0' */
{
  char *t=s;

  if(s= findmatchingchar(s, '(',')'))
  {
    if (s>t) *--s= '\0'; /* Terminate string at last bracket */
    return 0;
  }
  else
    return 1; /* no bracket matched */
}


void
GetAuthor(char *fromstr,char *authorstr)
/* WARNING: authorstr must be at least MAXFROM+1 wide! */
/* from is SOURCE, authorstr is dest, fromstr is not empty */
{
  int i=0;
  char *c;

  fromstr= skpwspc(fromstr);
  if ((c= NextWhiteSpc(fromstr))=='\0')    /* No further space: foo@bar */
  {
    strncpy(authorstr,fromstr,MAXFROM);  /* return foo@bar */
    authorstr[MAXFROM]='\0'; /* Set the last byte to \0 */
    /* if fromstr exceeds MAXFROM, authorstr will not be null terminated otherwise! */
    return;
  }
  else
  {
    c= skpwspc(c);
    if (c && *c=='(')  /* That's: foo@bar (fred foobar) */
    {
      /* where's the closing brace? */

      if (killrightbracket(c)!=0) /* No closing bracket found! */
        cuteoln(c);

      c++;
      if (*c=='\0')      /* empty braces! */
      {
        strncpy(authorstr,fromstr,MAXFROM);  /* return foo@bar */
        authorstr[MAXFROM]= '\0'; /* Set the last byte to \0 */
        /* if fromstr exceeds MAXFROM, authorstr will not be null terminated otherwise! */
        return;
      }
      else
      {
        c= skpwspc(c); /* foo@bar (  fred foobar ) */
        CutRealName(authorstr,c);   /* return "fred foobar" / "f foobar" */
      }
    }
    else /* Should be: Fred Foobar <foo@bar>, contains at least a word */
    {
      while (c[i]!='\0' && c[i]!='<') i++;
      if (c[i--]=='<')
        c[i]= '\0';

      CutRealName(authorstr,fromstr);   /* return: Fred Foobar */
    }
  }
}


struct GroupInfoList *
FindGroup(char *s)
{
  struct GroupInfoList *pGI;

  for (pGI= RootGI; pGI!=NULL && stricmp(pGI->Name,s); pGI= pGI->NextGI);

  return pGI;
}


void
Entry(char *c,char **var)   /* Allocates Mem for variable (var) and copies c to var */
{
  int len;

  cuteoln(c);
  c= skpwspc(c);
  if (len= strlen(c))
  {
    *var= AllocMem((ULONG) len+1UL, MEMTYPE);
    if (*var)
      strcpy(*var,c);
    else
      ErrorHandle(NOMEM,NULL);
  }
}


#ifdef DEBUG
void ListGroups(void)
{
  struct GroupInfoList *pGI;
  char subscribed, moderated;
  int i;

  for(pGI= RootGI, i=1; pGI!=NULL; pGI=pGI->NextGI, i++)
  {
    if (pGI->Status & UNSUBSCRIBED)
       subscribed='U';
    else
       subscribed=' ';

    if (pGI->Status & MODERATED)
       moderated='M';
    else
       moderated=' ';

    sprintf(tmp,"%-58s: %c %c %6lu-%-6lu\n", pGI->Name,subscribed,moderated,
                                             pGI->firstArtNr,pGI->lastArtNr);

    Display(tmp);
  }
}
#endif


void
FreeGIL(void) /* MARKS ARE NOT FREED */
{
  struct GroupInfoList *pGI, *plGI;

  pGI= RootGI;

  while (pGI!=NULL)
  {
    plGI= pGI;
    pGI= pGI->NextGI;
#ifdef DEBUG
    sprintf(tmp,"Freeing newsgroup: %s\n", plGI->Name);
    Display(tmp);
#endif
    FreeString(plGI->Name);

    FreeMem(plGI,(ULONG) sizeof(struct GroupInfoList));
  }

  RootGI= NULL;
}


void
LeftRotate(struct subj_struct **RootNode, struct subj_struct *node_to_rotate)
/* assumes that node_to_rotate and right(node_to_rotate) is non-NULL */
/* Situation:     |                                  |
            node_to_rotate                       rightchild
            /           \             ==>       /          \
           a            rightchild       node_to_rotate     c
                        /        \       /            \
                       b          c     a              b
*/
{
  struct subj_struct *rightchild;

  rightchild= right(node_to_rotate);
  if ((right(node_to_rotate)= left(rightchild))!=NULL)
     parent(left(rightchild))= node_to_rotate;

  if ((parent(rightchild)= parent(node_to_rotate))==NULL) /* note_to_rotate was root */
     *RootNode= rightchild;                               /* now it's right child is root */
  else
  {
    if (node_to_rotate == left(parent(node_to_rotate))) /* was node_to_rotate left-child? */
       left(parent(node_to_rotate))= rightchild;
    else
       right(parent(node_to_rotate))= rightchild;
  }
  left(rightchild)= node_to_rotate;
  parent(node_to_rotate)= rightchild;
}


void
RightRotate(struct subj_struct **RootNode, struct subj_struct *node_to_rotate)
/* node_to_rotate and left(node_to_rotate) must be non-NULL! */
{
  struct subj_struct *leftchild;

  leftchild= left(node_to_rotate);
  if ((left(node_to_rotate)= right(leftchild))!=NULL)
    parent(right(leftchild))= node_to_rotate;

  if ((parent(leftchild)= parent(node_to_rotate))==NULL)
    *RootNode= leftchild;
  else
  {
    if (node_to_rotate == left(parent(node_to_rotate)))
       left(parent(node_to_rotate))= leftchild;
    else
       right(parent(node_to_rotate))= leftchild;
  }
  right(leftchild)= node_to_rotate;
  parent(node_to_rotate)= leftchild;
}


void
InsertIntoSubjectTree(struct subj_struct **RootNode, struct subj_struct *node_to_insert)
/* This procedure inserts a node into the red-black binary tree and returns nothing */
{
  struct subj_struct *x,*y= NULL;

  /* normal iterative binary tree insertion */
  for (x= *RootNode; x!=NULL; )
  {
    y= x;   /* store parent node */
    if (stricmp(node_to_insert->Subject, x->Subject)<0)
       x= left(x);
    else
       x= right(x);
  }
  if ((parent(node_to_insert)= y)==NULL)
     *RootNode= node_to_insert;
  else
  {
    if (stricmp(node_to_insert->Subject, y->Subject)<0)
       left(y)= node_to_insert;
    else
       right(y)= node_to_insert;
  }

  /* Red-Black tree special processing */
  color(node_to_insert)= RED;
  while (node_to_insert!=(*RootNode) && parent(node_to_insert) && color(parent(node_to_insert))==RED)
  {
    if ( parent(node_to_insert)==left(parent(parent(node_to_insert))) )
    { /* if parent is left child */
      y= right(parent(parent(node_to_insert)));
      if (y && color(y)==RED)
      {
        color(parent(node_to_insert))= color(y)= BLACK;
        color(parent(parent(node_to_insert)))= RED;
        node_to_insert= parent(parent(node_to_insert));
      }
      else
      {
        if (node_to_insert==right(parent(node_to_insert)))
        {
          node_to_insert= parent(node_to_insert);
          LeftRotate(RootNode, node_to_insert);
        }
        color(parent(node_to_insert))= BLACK;
        color(parent(parent(node_to_insert)))= RED;
        RightRotate(RootNode, parent(parent(node_to_insert)));
      }
    }
    else /* parent is right child */
    {
      y= left(parent(parent(node_to_insert)));
      if (y && color(y)==RED)
      {
        color(parent(node_to_insert))= color(y)= BLACK;
        color(parent(parent(node_to_insert)))= RED;
        node_to_insert= parent(parent(node_to_insert));
      }
      else
      {
        if (node_to_insert==left(parent(node_to_insert)))
        {
          node_to_insert= parent(node_to_insert);
          RightRotate(RootNode, node_to_insert);
        }
        color(parent(node_to_insert))= BLACK;
        color(parent(parent(node_to_insert)))= RED;
        LeftRotate(RootNode, parent(parent(node_to_insert)));
      }
    }
  }
  color(*RootNode)= BLACK;
}

#ifdef RECURSE
void
PrintSubjectTree2(struct subj_struct *RootNode)
{
  if (RootNode)
  {
    PrintSubjectTree2(left(RootNode));
    printf("%s\n",RootNode->Subject);
    PrintSubjectTree2(right(RootNode));
  }
}
#endif

void
TraverseSubjectTree(struct subj_struct *RootNode, void (*action)(struct subj_struct *subjP))
/* traverses binary-tree and executes action on its nodes.
   this is an iterative procedure! */
{
  struct subj_struct *x;

  x= RootNode;

  while (x)
  {
    if (left(x))
    {
      x= left(x);
    }
    else
    if (right(x))
    {
      if (left(x)==NULL) /* left subtree is empty, so all following nodes are greater */
         action(x);
      x= right(x);
    }
    else
    { /* we reached a leaf */
      action(x); /* print leaf */

      while (x) /* climbing up or entering right child */
      {
        if (parent(x) && x==left(parent(x))) /* x is left child */
        {  /* climb up until we can proceed with a right child */
           x= parent(x);
           action(x);
           if (right(x))
           {
              x= right(x);  /* enter right child */
              break;
           }
        }
        else /* x is right child or root */
          x= parent(x);
      }
    }
  }
}


void PrintSubject(struct subj_struct *x)
{
  printf("%s\n", x->Subject);
}


void
PrintSubjectTree(struct subj_struct *RootNode)
/* traverses binary-tree and prints its nodes.
   this is an iterative procedure! */
{
  TraverseSubjectTree(RootNode, PrintSubject);
}


void
FreeSubjectTree(struct subj_struct *subjp)
{
  if (subjp)
  {
    FreeString(subjp->Subject);
    FreeSubjectTree(subjp->LeftSubject);
    FreeSubjectTree(subjp->RightSubject);
    FreeMem(subjp, (ULONG) sizeof(struct subj_struct));
  }
}


void
FreeSubjectNode(struct subj_struct *subjP)
{
  /* FreeString() expanded, because of less overhead */
  if (subjP->Subject)
  {
    FreeMem(subjP->Subject, (ULONG) strlen(subjP->Subject)+1UL);
    subjP->Subject= NULL;
  }
  subjP->LeftSubject= subjP->RightSubject= subjP->ParentSubject= NULL;
  subjP->RootArt= NULL;
  FreeMem(subjP, (ULONG) sizeof(struct subj_struct));
}


#ifdef BUGGY
void
FreeSubjectTree(struct subj_struct *RootNode)
/* iterative deallocation of SubjectTree */
{
  TraverseSubjectTree(RootNode, FreeSubjectNode);
}
#endif


void
FreeSubjectTree2(struct subj_struct *RootNode)
/* traverses binary-tree and frees its nodes.
   this is an iterative procedure! */
{
  struct subj_struct *x, *parent, *right;

  x= RootNode;

  while (x)
  {
    if (left(x))
    {
      x= left(x);
    }
    else
    if (right(x))
    {
      x= right(x);
    }
    else
    { /* we reached a leaf */
      parent= parent(x);
      FreeSubjectNode(x);  /* free leaf */

      while (x) /* climbing up or entering right child */
      {
        if (parent && x==left(parent)) /* x was left child */
        {  /* climb up until we can proceed with a right child */
           x= parent;
           parent= parent(x);
           if (right= right(x))
           {
              x= right;  /* enter right child */
              break;
           }
           else /* left tree is already free, right node doesn't exist, so we have another leaf */
           {
             FreeSubjectNode(x);
           }
        }
        else /* x was right child or root */
        { /* if x was a right child, we can discard it's node */
          if (x= parent)
          {
            parent= parent(x);
            FreeSubjectNode(x);
          }
        }
      }
    }
  }
}





ULONG TreeHeight(struct subj_struct *subjp)
{
  if (subjp)
     return max(TreeHeight(subjp->LeftSubject), TreeHeight(subjp->RightSubject))+1;
  else
     return 0;
}


void
FreeGroup(struct GroupInfoList *pGI)
/* Frees all memory allocated by one group */
{
  struct art_struct *artp;
  ULONG NrOfArticles,i;

  if (!pGI || !RootField) return;

  NrOfArticles= pGI->lastArtNr - pGI->firstArtNr + 1L;

  if (NrOfArticles<=0) return;

  FreeSubjectTree2(pGI->RootSubjP);
  /* FreeSubjectTree(pGI->RootSubjP); */
  /* ALL Subject-strings AND STRUCTURES are now free */

  /* Now free all FROM'S */
  for (i=0,artp= RootField; i<NrOfArticles; ++i,artp= &RootField[i])
  {
    FreeString(artp->From);
  }

  FreeMem(RootField, (ULONG) sizeof(struct art_struct)*(ULONG) NrOfArticles);

  RootField= NULL;
  RootSubjP= NULL;
}



int
ReadActiveFile(void)
/* Reads the actual active-file */
{
  FILE *activefp=NULL;
  char *newsgroup, moderated;
  struct GroupInfoList **pGI, *prevGI;
  ULONG highest,lowest;

  pGI= &RootGI;
  prevGI= NULL;

  newsgroup=AllocMem(MAXLINE,MEMTYPE);

  if (!newsgroup)
  {
     ErrorHandle(NOMEM,NULL);
     return 1;
  }
  if (activefile) activefp= fretryopen(activefile,"r");
  if (activefp)
  {
    while (fgets(tmp, SUPERLINE, activefp))
    {
      highest= lowest= 0L;

      sscanf(tmp,"%s %lu %lu %c",newsgroup,&highest,&lowest,&moderated);

#ifdef DEBUG
      printf("%s",tmp);
#endif

      if (*newsgroup!='\0')
      {
        if (*pGI= AllocMem((ULONG) sizeof(struct GroupInfoList),MEMTYPE))
        {
          (*pGI)->Name=CopyString(newsgroup);
          if ((*pGI)->Name)
          {
            TotalGroups++;
            strcpy((*pGI)->Name, newsgroup);
            (*pGI)->Marks=NULL;
            (*pGI)->Status=0;
            if (toupper(moderated)=='M')
               (*pGI)->Status|=MODERATED;
            /* preserve this char for writing back the active-file */
            (*pGI)->Active_Char= moderated;

            /* 0 is a non existing or invalid article number! */
            /* case lowest==highest==0     -> empty group */
            /* case lowest==0 && highest>0 -> set lowest to 1L */
            /* case lowest>highest         -> empty group */

            (*pGI)->firstArtNr= lowest;
            (*pGI)->OldlastArtNr= (*pGI)->OldfirstArtNr= 0L;

            if (lowest==0L && highest) (*pGI)->firstArtNr= 1L;
            else
            if (lowest>highest) (*pGI)->firstArtNr= highest+1;

            (*pGI)->lastArtNr= highest;
            (*pGI)->PrevGI= prevGI;
            prevGI= *pGI;

            pGI= &(*pGI)->NextGI;
          }
          else
            ErrorHandle(NOMEM, NULL);
        }
        else
          ErrorHandle(NOMEM, NULL);
      }
    } /* end while */

    fclose(activefp);

    *pGI=NULL; /* No next group */
  }
  else
   ErrorHandle(NOACTIVE,NULL);

#ifdef DEBUG
  ListGroups();
#endif

  if (newsgroup) FreeMem(newsgroup, MAXLINE);
}



ReadOldActiveFile(void)
/* Reads the old (renamed) active-file that exists as active.bak */
{
  FILE *activefp;
  char moderated, *newsgroup;
  ULONG oldhighest,oldlowest;
  struct GroupInfoList *pGI;

  newsgroup=AllocMem(MAXLINE,MEMTYPE);

  if (!newsgroup)
  {
     ErrorHandle(NOMEM,NULL);
     return 1;
  }

  sprintf(fnametmp,"%s.bak",activefile);
  activefp= fretryopen(fnametmp,"r");
  if (activefp)
  {
    while (fgets(tmp, SUPERLINE, activefp))
    {
      oldhighest= oldlowest= 0UL;

      sscanf(tmp,"%s %lu %lu %c",newsgroup,&oldhighest,&oldlowest,&moderated);
      if (pGI=FindGroup(newsgroup))
      {
        if (oldlowest==0L && oldhighest) oldlowest= 1L;
        if (oldlowest>oldhighest) oldlowest= oldhighest+1; /* empty group */
        /* if there are too much news expired and the active file is reset */
        /* force a full scan */
        if (oldlowest>pGI->firstArtNr || oldhighest>pGI->lastArtNr)
        {
          oldhighest= oldlowest=0L;
        }
        pGI->OldlastArtNr= oldhighest;
        pGI->OldfirstArtNr= oldlowest;
      }
    }
    fclose(activefp);
  }

  if (newsgroup)
      FreeMem(newsgroup, MAXLINE);
  return (0); /*all ok*/
}


int
ScanArticles(struct GroupInfoList *pGI, struct art_struct *RootField)
/* Scans all articles in the group, so we can update lowest and highest
   article number for the active-file.
   A pointer to the group we want to scan must be given, an ERROR_CODE is
   returned:
   0 = all ok, 1 = break, 2 = NONEWSGROUP, 3 = NOMEM
   If RootField is non-NULL, each article is scanned for existence and
   marked as NONEXISTENT */
{
  struct FileLock *FL= NULL;
  struct FileInfoBlock *FIB;
  int Error= 0;
  register ULONG ArticleNr, lowest= LONG_MAX, highest= 0L;
  ULONG index, FirstArtNr, NrOfArticles;
  struct art_struct *artp;

  FIB= (struct FileInfoBlock *)
       AllocMem ((ULONG) sizeof(struct FileInfoBlock), MEMTYPE);

  if (!FIB)
  {
    ErrorHandle(NOMEM,NULL);
    Error= 3; /* ErrorCode */
  }
  else
  {
    strcpy(fnametmp,pGI->Name);

    if (!(switches & SW_SINGLEDIR))
    {
      MakeDirName(tmp, (dbasedir!=NULL ? dbasedir : newsdir), fnametmp);
#ifdef DEBUG
 printf("Opening: %s\n",tmp);
#endif
    }
    else
      strcpy(tmp,fnametmp);

    if (!(FL= (struct FileLock *) Lock(tmp, (ULONG) ACCESS_READ)))
    {
      ErrorHandle(NONEWSGROUP,pGI->Name);
      Error= 2;                   /* No Lock! */
    }
    else
    if (!Examine((BPTR) FL, FIB))
    {
      ErrorHandle(NONEWSGROUP,pGI->Name);
      Error= 2;                   /* No Dir! */
    }
    else
    {
#ifdef DEBUG
      printf("Examine Dir:\n");
#endif
      if (RootField)
      {
        FirstArtNr= pGI->firstArtNr;
        NrOfArticles= pGI->lastArtNr - pGI->firstArtNr + 1L;
        for (index= 0, artp= RootField; index<NrOfArticles; index++, artp= &RootField[index])
            artp->Status|= NONEXISTENT;
      }

      while (!Error && ExNext((BPTR) FL, FIB))
      {
        if (islong(FIB->fib_FileName))
        {
          ArticleNr= (ULONG) atol(FIB->fib_FileName);
          if (RootField)
             RootField[ArticleNr-FirstArtNr].Status &= ~NONEXISTENT;

#ifdef DEBUG
          printf("Got: %s=%lu\n", FIB->fib_FileName, ArticleNr);
#endif
          if (ArticleNr>highest) highest= ArticleNr;
          if (ArticleNr<lowest) lowest= ArticleNr;
        }
        if (SetSignal(0L,0L) & (SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D))
        {
          ErrorHandle(ABORT, NULL);
          Error= 1;
        }
      } /* end while */
    }
  }
  if (FIB) FreeMem (FIB,(ULONG) sizeof(struct FileInfoBlock));
  if (FL) UnLock ((BPTR) FL);

  if (!RootField) /* if checking for existance, don't update firstArtNr/lastArtNr! */
  {
    pGI->firstArtNr= (lowest==LONG_MAX) ? pGI->lastArtNr+1 : lowest;
                                         /* first case: empty group */
    if (highest) pGI->lastArtNr= highest;
#ifdef DEBUG
   printf("Got: %s lowest:%lu highest:%lu\n", pGI->Name, pGI->firstArtNr,pGI->lastArtNr);
   printf("ScanNews returns: %d\n", Error);
#endif
  }

  if (Error==2)
     GlobalERROR= OK;

  return Error;
}


void
UpdateActive(void)
/* Updates all active-file entries by scanning all groups */
{
  struct GroupInfoList *pGI;
  FILE *activefp;

  for (pGI=RootGI; pGI!=NULL; pGI=pGI->NextGI)
  {
    if (ScanArticles(pGI,NULL)==1) break; /* got break signal */
  }

  if (activefp= fretryopen(activefile,"w"))
  {
    for (pGI= RootGI; pGI!=NULL; pGI= pGI->NextGI)
    {
      fprintf(activefp,"%s %06lu %06lu %c\n",
              pGI->Name,pGI->lastArtNr,pGI->firstArtNr,pGI->Active_Char);
    }
    fclose(activefp);
  }
}


ULONG
GetRelArtNr(struct art_struct *ArtP)
/* Returns the relative article number from ArtP */
{
  if (ArtP)
     return ((ULONG)ArtP-(ULONG)RootField)/sizeof(struct art_struct);
  else
     return 0;
}


struct subj_struct *
ReadDataBase(struct GroupInfoList *pGI, struct art_struct *RootField)
/* Reads the existing database, ignoring all articles that are no longer */
/* active */
{
  FILE *dbasefp;                        /* file pointer */
  long readnr;                          /* nr of read datasets */
  struct art_struct *ActArtP,           /* actual article struct */
                    **PrevNextArtPP, *PrevArtP;
  ULONG FollowUpCount=0UL;              /* FollowUps following */
  struct subj_struct *SubjectStructP=NULL;
  char *Subject= NULL, *FromStr;
  BOOL VALID;                           /* article in active range? */
  UCHAR SubjLen=0;                      /* length of the subj */

  /* group is empty */
  if (pGI->lastArtNr<pGI->firstArtNr) return NULL;

  if ((switches & SW_CONSISTENCY_CHECK) && ScanArticles(pGI,RootField)==1) return NULL;

  ActArtP= RootField;      /* TODO */

  PrevNextArtPP= NULL;
  PrevArtP= NULL;

  RootSubjP= NULL;

  sprintf(fnametmp,"%s/.newsdb",pathtodbase); /* Name of the .newsdb */

  if (dbasefp=fretryopen(fnametmp,"r"))  /* Open the DataBase */
  {
    setvbuf(dbasefp, mBuffer, _IOFBF, M_BUFFERSIZE); /* for faster IO! */

    if (!feof(dbasefp))
    do
    {
      /* read in one standard mask */
      readnr= fread((char *) &mask,sizeof (mask),1 ,dbasefp);
      if (readnr)  /* we got it */
      {
        if (mask.ArtNr>pGI->lastArtNr || mask.ArtNr<pGI->firstArtNr)
        {
           VALID= false; /* article not in the active-range (expired) */
                         /* This must be ignored, but skipped! */
           expired++;
        }
        else
        {
          ActArtP= &RootField[mask.ArtNr-FirstArtNr]; /* Get struct in direct-access */
          if ((switches & SW_CONSISTENCY_CHECK) && (ActArtP->Status & NONEXISTENT))
          {
            VALID= false; expired++;
          }
          else
          {
            VALID= true;      /* Valid article-number */
            ActArtP->Status= mask.ArtStat; /* copy status */
            ActArtP->Lines= mask.Lines;   /* copy lines */
            ActArtP->Date= mask.Date;     /* copy date */
            valid_ds++;
#ifdef SLOW
            ActArtP->NrofFollowUps=0;      /* reset */
            ActArtP->NextArt= ActArtP->PrevArt= NULL; /* no links yet */
            ActArtP->ExtData=NULL;
#endif
          }
        }

#ifdef DEBUG
        printf("#%lu Stat:%x=%x %lx %ld FromLen:%d VALID:%d\n",
                mask.ArtNr,mask.ArtStat,ActArtP->Status,ActArtP->Date,
                mask.Lines,mask.FromLen,VALID);
#endif
        /* From is a required entry */
        if (mask.FromLen>0)  /* should always be >0 */
        {
#ifdef DEBUG
          if (mask.FromLen>16) printf("From: > 16 chars!\n");
#endif
          FromStr=AllocMem((ULONG) mask.FromLen+1UL, MEMTYPE);

          if (FromStr)
          {
            fread(FromStr,mask.FromLen,1,dbasefp); /* Read Author */
            FromStr[mask.FromLen]='\0';
#ifdef DEBUG
            printf("From: %s\n",FromStr);
#endif
          }
          else
          {
            ErrorHandle(NOMEM,NULL);
            break;
          }

          if (VALID)
             ActArtP->From=FromStr; /* link FROM into struct */
          else
             if (FromStr) FreeMem(FromStr, (ULONG) mask.FromLen+1UL);
        }
        else
        {
          sprintf(fnametmp,"%s/%lu",pGI->Name,mask.ArtNr);
          ErrorHandle(NOFROM,fnametmp); /* Database CORRUPT */
          break; /* leave while */
        }

        if (FollowUpCount==0) /*** Subject follows ***/
        {
          if (SubjectStructP && SubjectStructP->RootArt==NULL)
          { /* No valid article-nr followed yet, so free the reserved mem */
            if (SubjectStructP)
               FreeMem(SubjectStructP, (ULONG) sizeof(struct subj_struct));
            if (Subject)
               FreeMem(Subject, (ULONG) SubjLen+1);

            /* Subject= NULL; */
          }

          if (SubjectStructP= AllocMem((ULONG) sizeof(struct subj_struct),MEMTYPE))
          {
            if (VALID) /** this Subject is valid **/
            {
              ActArtP->Subject= SubjectStructP;    /* link SUBJECT-STRUCT */

              PrevNextArtPP= &ActArtP->NextArt;    /* For FollowUps */
              PrevArtP= ActArtP;                   /* For FollowUps */

              /* reference back to this first article containing subject */
              SubjectStructP->RootArt= ActArtP;
              SubjectStructP->LeftSubject=
              SubjectStructP->RightSubject=
              SubjectStructP->ParentSubject= NULL;
            }
            else /** NOT VALID, but preserve Subject if VALID FollowUp follows **/
            {
              SubjectStructP->RootArt= NULL;
              SubjectStructP->LeftSubject=
              SubjectStructP->RightSubject=
              SubjectStructP->ParentSubject= NULL;
              PrevArtP=NULL;
            }
            fread(&SubjLen, sizeof(UCHAR), 1, dbasefp);
#ifdef DEBUG
            printf("SubjLen: %d ",SubjLen);
#endif
            if (SubjLen>0) /* should always be >0 */
            {
              Subject= SubjectStructP->Subject= AllocMem((ULONG) SubjLen+1UL,MEMTYPE);
              if (Subject)
              {
                fread(Subject,SubjLen,1,dbasefp); /* get the SUBJECT */
                Subject[SubjLen]='\0';
#ifdef DEBUG
                printf("Subject: %s\n",Subject);
#endif
                /* Insert new Subject into SubjectTree */
                if (VALID) InsertIntoSubjectTree(&RootSubjP, SubjectStructP);
              }
              else
              {
                ErrorHandle(NOMEM,NULL);
                break;
              }
            }
            else
            {
              SubjectStructP->Subject= NULL;
              sprintf(fnametmp,"%s/%lu",pGI->Name,mask.ArtNr);
              ErrorHandle(NOSUBJ,fnametmp); /* DBase CORRUPT */
              break;
            }
          }
          else
          {
            ErrorHandle(NOMEM,NULL);
            break;
          }

          /** readin FollowUp count if there is one **/
          if (mask.ArtStat & (PA_WITHF | FUP_WITHF))
          {     /* readin FollowUp count */
            readnr= fread((char *) &FollowUpCount, sizeof(ULONG), 1, dbasefp);
#ifdef DEBUG
            printf("FUP_count: %lu\n",FollowUpCount);
#endif
            if (readnr==0) /* feof */
            {
              ErrorHandle(DBASECORRUPT,"expecting followups");
              break;
            }
          }
        } /* endif FollowUpcount==0 */
        else
        {           /**** FOLLOWUP-handling ****/
          if (VALID) /* link followup into art_struct list */
          {
            ActArtP->Subject= SubjectStructP; /* Fups have the same subject */
            ActArtP->PrevArt= PrevArtP;

            if (SubjectStructP->RootArt) /* Subject-article was valid */
            {
              SubjectStructP->RootArt->NrofFollowUps++;
#ifdef DEBUG
            printf("NrOfFupS: %lu\n",SubjectStructP->RootArt->NrofFollowUps);
#endif
              *PrevNextArtPP= ActArtP; /* Link last NextArt to this struct */
            }
            else /* The former subject: article w. subj was INVALID */
            {
              InsertIntoSubjectTree(&RootSubjP, SubjectStructP);
              /* reference back to this first article containing subject */
              SubjectStructP->RootArt= ActArtP;
            }

            /* preserve addresses of this struct for next links */
            PrevNextArtPP= &ActArtP->NextArt;
            PrevArtP=ActArtP;
          }
          FollowUpCount--;
        } /**** follow up handling ****/
      } /* if readnr */
    }
    while(!feof(dbasefp));

    if (PrevNextArtPP) *PrevNextArtPP=NULL;

    TotalDBbytes+= ftell(dbasefp);
    fclose (dbasefp);
  } /* opened database */

#ifdef DEBUG
  PrintSubjectTree(RootSubjP);
  printf("Height: %lu\n", TreeHeight(RootSubjP));
#endif
  if (GlobalERROR) /* Corrupt databases cause crashes */
  {
    pGI->RootSubjP= RootSubjP;
    FreeGroup(pGI);
    CloseMaster();
    exit(RETURN_ERROR);
  }
  return RootSubjP;
}



ULONG
WriteThread(FILE *dbasefp, struct subj_struct *subjP)
/* writes thread into file dbasefp, beginning at subjP->RootArt */
/* returns # of articles written into database                  */
{
  struct art_struct *ActArtP;
  ULONG ArtNr, count_ds= 0UL;
  int len;
  UCHAR subjlen;

  if(!(ActArtP= subjP->RootArt))
  {
    ErrorHandle(STRANGE,"NoRoot");
    return 0;
  }

  /* Write all FollowUps */
  for (; ActArtP!=NULL; ActArtP= ActArtP->NextArt)
  {
    ArtNr= GetRelArtNr(ActArtP)+(ULONG) FirstArtNr;

    mask.ArtNr=(long) ArtNr;    /* artnr */

    mask.ArtStat= (UCHAR) (ActArtP->Status &
       (PARENTART | PA_WITHF | FOLLOWUPART | FUP_WITHF | XREF));

    /* mask.ArtStat&= (UCHAR) ~NONEXISTENT; */
    if (ActArtP->NrofFollowUps) /* Are there followups? Modify Status */
    {
#ifdef DEBUG
            printf("NrofFollowUps: %lu\n",ActArtP->NrofFollowUps);
#endif
      if (mask.ArtStat & (UCHAR) PARENTART)
      {
        mask.ArtStat&= (UCHAR) ~PARENTART;
        mask.ArtStat|= (UCHAR) PA_WITHF;
      }
      else
      if (mask.ArtStat & (UCHAR) FOLLOWUPART)
      {
        mask.ArtStat&= (UCHAR) ~FOLLOWUPART;
        mask.ArtStat|= (UCHAR) FUP_WITHF;
      }
    }
    else /* no followups - reset status if necessary */
    {
#ifdef DEBUG
            printf("NoFollowUps!\n");
#endif
      if (mask.ArtStat & (UCHAR) PA_WITHF)
      {
        mask.ArtStat&= (UCHAR) ~PA_WITHF;
        mask.ArtStat|= (UCHAR) PARENTART;
      }
      else
      if (mask.ArtStat & (UCHAR) FUP_WITHF)
      {
        mask.ArtStat&= (UCHAR) ~FUP_WITHF;
        mask.ArtStat|= (UCHAR) FOLLOWUPART;
      }
    }
#ifdef DEBUG
          printf("#%lu Stat:%x->%x\n",ArtNr, ActArtP->Status, mask.ArtStat);
#endif

    mask.Lines=ActArtP->Lines;      /* lines */
    mask.Date= ActArtP->Date;       /* Date */

    len= strlen(ActArtP->From);      /* Fromstr-length */
    if (len>MAXFROM) len= MAXFROM;
    mask.FromLen= len;

    if (len==0)
    {
      ActArtP->From= NULL;
      sprintf(fnametmp,"%lu",mask.ArtNr);
      ErrorHandle(NOFROM,fnametmp);
      break;
    }

    fwrite((char *) &mask, sizeof(mask), 1, dbasefp);
    fwrite((char *) ActArtP->From, mask.FromLen, 1, dbasefp);

    count_ds++;

    if (subjP->RootArt==ActArtP) /* write subject once if rootart */
    {
      len= strlen(subjP->Subject);
      if (len>255) len= 255;
      subjlen=(UCHAR) len;
      /* write length then subject string */
      fwrite((char *) &subjlen, sizeof(UCHAR), 1, dbasefp);
      fwrite((char *) subjP->Subject, len, 1, dbasefp);

      if (ActArtP->NrofFollowUps) /* if we have followups write # of FUPs */
      {
        fwrite((char *) &ActArtP->NrofFollowUps, sizeof(ULONG), 1, dbasefp);
      }
    }
  } /* end of follow thread */
  return count_ds;
}



int
WriteDataBase(struct GroupInfoList *pGI)
/* Dumps out the database from memory */
{
  ULONG count_ds;
  struct subj_struct *subjP;
  FILE *dbasefp;                  /* file pointer */

  if (!pGI) return 1; /* Error! */

  count_ds= 0UL;

  strcpy(fnametmp,pGI->Name);
  if (!(switches & SW_SINGLEDIR))
  {
    MakeDirName(tmp,dbasedir ? dbasedir : newsdir,fnametmp);
#ifdef DEBUG
    printf("CreateDir(?)\n");
#endif
    CreateDir_IfNotExists(tmp,tmp+strlen(tmp)); /* create dir if necessary */
  }
  else
    strcpy(tmp,fnametmp);

  strcat(tmp,"/.newsdb");
  if (pGI->RootSubjP==NULL) /* no database to write, all articles expired */
  {
    DeleteFile(tmp);
    return 0;
  }

  if (dbasefp=fretryopen(tmp,"w"))
  {
    setvbuf(dbasefp, mBuffer, _IOFBF, M_BUFFERSIZE); /* for faster IO! */

    /* Write all Subject threads */
#ifdef DEBUG
    PrintSubjectTree(RootSubjP);
#endif
    for(subjP= pGI->RootSubjP; subjP!=NULL;)
    {
      if (left(subjP))
      {
        subjP= left(subjP);
      }
      else
      if (right(subjP))
      {
        if (left(subjP)==NULL) /* left subtree is empty, so all following nodes are greater */
           count_ds+= WriteThread(dbasefp, subjP);
        subjP= right(subjP);
      }
      else
      { /* we reached a leaf */
        count_ds+= WriteThread(dbasefp, subjP); /* print leaf */

        while (subjP) /* climbing up or entering right child */
        {
          if (parent(subjP) && subjP==left(parent(subjP))) /* subjP is left child */
          {  /* climb up until we can proceed with a right child */
             subjP= parent(subjP);
             count_ds+= WriteThread(dbasefp, subjP); /* print node */
             if (right(subjP))
             {
                subjP= right(subjP);  /* enter right child */
                break;
             }
          }
          else /* subjP is right child or root */
            subjP= parent(subjP);
        }
      }
    } /* next different subject */
    fclose(dbasefp);
  }
  if (count_ds!=valid_ds)
  {
    ErrorHandle(WRITE_CHECK,pGI->Name);
    printf("%lu %lu\n", count_ds, valid_ds);
  }
}



struct subj_struct *
FindSubject(char *Subject)
/* returns the pointer to the subject structure that matches Subject    */
/* Because this is an iterative in the near balanced binary search tree */
/* it is garantueed to run in O(log n)                                  */
{
  struct subj_struct *subjp;
  int result=-1;

  /* RootSubjP MUST BE SET (or NULL)!!! */

#ifdef DEBUG
  printf("Searching:%s\n",Subject);
#endif
  for(subjp= RootSubjP; subjp!=NULL && (result= stricmp(subjp->Subject,Subject));)
  {
#ifdef DEBUG
  printf("List:%s\n",subjp->Subject);
#endif
    if (result>0)
       subjp= left(subjp);
    else
       subjp= right(subjp);
  }

#ifdef DEBUG
  if (subjp)     printf("subjp:%s\n",subjp->Subject);
  printf("result: %d\n",result);
#endif

  return subjp; /* Only != NULL if exactly found */
}


int
ScanArticle(ULONG ArtNr)
/* opens article ArtNr in group ->pathtogroup<- and sorts it into the */
/* (existing) database.           ~~~~~~~~~~~                         */
/* returns 0 if ok, 1 for non existing article, 2 for invalid ArtNr   */
/* 3 for break signal/abort                                           */
{
  FILE *ArtFP;
  char *c;
  struct subj_struct *subjp= NULL;
  struct art_struct *ActArtP, *TmpArtP, *Tmp2ArtP;
  BOOL Got_Subject= false, Got_From=false, Got_Lines=false;
  ULONG index;
  ULONG ArtDate;

  sprintf(tmp, "%s/%lu", pathtogroup, ArtNr); /* pathtogroup must be set correctly */

  if((index= ArtNr-FirstArtNr)<0) return 2; /* invalid index */

  ActArtP= &RootField[index]; /* FirstArtNr must be set correctly! */
  if (ActArtP->From) return 0; /* We already got this entry */

  if (ArtFP= fretryopen(tmp,"r")) /* tmp is directly overwritten afterwards!!! */
  {
    setvbuf(ArtFP, HeaderBuf, _IOFBF, sizeofHeaderBuf); /* for faster IO! */

    while (fgets(tmp, SUPERLINE, ArtFP))
    {
      if ((c= firststr(tmp,"Subject:")) || (!Got_Subject && *tmp=='\n'))
      { /* process Subject line or: end of header and no subject so far -> use (none) as subject */
        if (c) /* process this body if first condition is true */
        {
          cuteoln(c);
          c= skpwspc(c);
#ifdef DEBUG
          printf("%lu: Subject: %s\n",ArtNr,c);
#endif
          if (Got_Subject) continue; /* we got 2 Subject: lines! */
          else
             Got_Subject= true;

          c= StripSubject(c,&ActArtP->Status); /* Sets article Status accordingly! */
#ifdef DEBUG
          printf("Stripped: %s STAT:%x\n",c, ActArtP->Status);
#endif
          if (*c=='\0') c= none;     /* if empty subject --> use default Subject (none) */
        }
        else
          c= none; /* we had no subject so far at the end of header */

        if (subjp= FindSubject(c)) /* subject already exists */
        {
          ActArtP->Subject= subjp; /* This Subject is ours, too */
          /* the rest of the procedure is below, not in the loop! */
        }
        else
        { /* NewSubject to allocate, ROOTART */
          if (subjp= AllocMem((ULONG) sizeof(struct subj_struct),MEMTYPE))
          {
            subjp->RootArt= ActArtP;
            subjp->LeftSubject=
            subjp->RightSubject=
            subjp->ParentSubject= NULL;
            subjp->Subject= CopyString(c);
            ActArtP->Subject= subjp; /* This is our Subject structure now */
            InsertIntoSubjectTree(&RootSubjP, subjp);
            subjp= NULL; /* Show that we don't need insertion into a thread */
          }
          else
          {
            ErrorHandle(NOMEM,NULL);
            break;
          }
        } /* create new thread */
      } /* Subject: */
      else
      if (c= firststr(tmp,"From:"))
      {
        cuteoln(c);
        GetAuthor(c, FromTmp);    /* is linked into ActArtP below */
        if (*FromTmp!='\0') Got_From=true;
      }
      else
      if (c= firststr(tmp,"Date:"))
      {
        c= skpwspc(c);
        ActArtP->Date= ParseDate(c);
      }
      else
      if (c= firststr(tmp,"Lines:"))
      {
        if (*(c= skpwspc(c))!='\0')
        {
          sscanf(c,"%lu",&ActArtP->Lines);
          Got_Lines= true;
        }
      }
      else
      if (c= firststr(tmp,"Xref:"))
      {
        ActArtP->Status|= XREF;
      }

      if (SetSignal(0L,0L) & (SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D))
      {
        ErrorHandle(ABORT,NULL);
        fclose(ArtFP);
        return 3;
      }

      if (*tmp=='\n') break; /* only scan header */
    } /* end of while */

#ifdef DEBUG
    printf("#%lu From:%16s %lx %s\n",ArtNr,FromTmp,ActArtP->Date,ActArtP->Subject->Subject);
#endif

    if (!Got_Lines)
    {
      if (switches & SW_COUNTLINES)
      {
        ULONG lines= 0UL;

        recounted_lines++;
        while (fgets(tmp, SUPERLINE, ArtFP)) lines++;
        ActArtP->Lines= lines;
      }
      else
       ActArtP->Lines= -1L;  /* no Lines: specified */
    }

    ArtDate= ActArtP->Date;

    /* We got a Subject and there was already a thread. */
    /* To sort it in we need the Date information which */
    /* we should have now */
    /* subjp is NULL if a new thread was created, so subjp should
       still point to the right subject-structure                 */
    if (subjp)
    {
      if (TmpArtP=subjp->RootArt)  /* Get the root-article */
      {
        /* Get the right position in thread (priority by Date) */
        while ((Tmp2ArtP= TmpArtP->NextArt) && (TmpArtP->Date<=ArtDate))
        {
          TmpArtP= Tmp2ArtP;
        }
        if (TmpArtP->Date<=ArtDate) /* link behind last node */
        {
          ActArtP->PrevArt= TmpArtP;
          ActArtP->NextArt= Tmp2ArtP; /* NULL */
          TmpArtP->NextArt= ActArtP;
        }
        else
        {
          /* TmpArtP is now older than this Article */
          /* so insert this article before TmpArtP into list */
          ActArtP->PrevArt= Tmp2ArtP= TmpArtP->PrevArt;
          ActArtP->NextArt= TmpArtP;
          TmpArtP->PrevArt= ActArtP;
          if (Tmp2ArtP==NULL) /* ActArt->PrevArt==NULL => now we're RootArt */
          {
            Tmp2ArtP=subjp->RootArt;
            ActArtP->NrofFollowUps= Tmp2ArtP->NrofFollowUps;
            Tmp2ArtP->NrofFollowUps= 0;
            subjp->RootArt= ActArtP;
          }
          else
            Tmp2ArtP->NextArt= ActArtP;
        }
        subjp->RootArt->NrofFollowUps++;
#ifdef DEBUG
        printf("NofFUps:%lu\n",subjp->RootArt->NrofFollowUps);
        for(TmpArtP=subjp->RootArt; TmpArtP; TmpArtP=TmpArtP->NextArt)
           printf("#%lu From:%-16s %lx\n",FirstArtNr+GetRelArtNr(TmpArtP),TmpArtP->From,TmpArtP->Date);
#endif
      }
      else
        ErrorHandle(STRANGE,"NoTmpArtP");
    }

    if (!Got_From)
      strcpy(FromTmp,nobody);

    ActArtP->From= CopyString(FromTmp);

    valid_ds++;

    fclose(ArtFP);
  }
  else
   return 1; /* article not found! */

  return 0;
}
#undef DEBUG

void PrintThread (struct subj_struct *subjP)
{
  struct art_struct *ActArtP;
  ULONG ArtNr;

  strncpy(tmp,subjP->Subject,35);
  tmp[35]='\0';
  for (ActArtP= subjP->RootArt; ActArtP!=NULL; ActArtP=ActArtP->NextArt)
  {
    ArtNr= GetRelArtNr(ActArtP)+(ULONG) FirstArtNr;
    if (switches & SW_DEBUG)
       printf("%6lu: %-35s %-16s %6lu %08lx\n",ArtNr, tmp, ActArtP->From, ActArtP->Lines, ActArtP->Date);
  }
}


void
PrintGroup(struct GroupInfoList *pGI)
/* Displays a database from a group on screen - for debugging */
{
  ULONG NrOfArticles;

  if (!pGI) return; /* Error! */

  NrOfArticles= pGI->lastArtNr - pGI->firstArtNr + 1L;
  if (switches & SW_DEBUG)
     printf("%lu Articles in %s:\nactive:%lu-%lu active.bak:%lu-%lu\n",
             NrOfArticles, pGI->Name,
                                      pGI->firstArtNr, pGI->lastArtNr,
                                      pGI->OldfirstArtNr, pGI->OldlastArtNr);
  if (NrOfArticles<=0) return;
  TraverseSubjectTree(pGI->RootSubjP, PrintThread);
}


void
ScanGroup (struct GroupInfoList *pGI)
/* scans a single group unconditionally */
/* invoked if there are new articles to scan */
{
  ULONG NrOfArticles, ArtNr, LastArtNr, NewArts;
  struct art_struct *ActArtP;

  if (!pGI) return; /* Error! */
  NrOfArticles= pGI->lastArtNr - pGI->firstArtNr + 1L;
  LastArtNr= pGI->lastArtNr;
  FirstArtNr= pGI->firstArtNr;  /* set for scanarticle!! */

  if (NrOfArticles<=0)
  {
    pGI->RootSubjP= NULL;
    return;
  }
  /* pathtogroup and pathtodbase are set here */
  if (!(switches & SW_SINGLEDIR))
  {
    strcpy(tmp,pGI->Name);
    MakeDirName(fnametmp, newsdir, tmp);
    pathtogroup= CopyString(fnametmp);
    if (dbasedir) /* DBASEDIR specified */
    {
      MakeDirName(fnametmp, dbasedir, tmp);
      pathtodbase= CopyString(fnametmp);
    }
    else
      pathtodbase= CopyString(pathtogroup);
  }
  else
  {
    pathtogroup= CopyString(pGI->Name);
    pathtodbase= CopyString(pGI->Name);
  }

  valid_ds= 0;

  /* allocate dynamic array of article structures */
  RootField= AllocMem((ULONG) sizeof(struct art_struct)*NrOfArticles, MEMTYPE);
  if (!RootField)
  {
    ErrorHandle(NOMEM, NULL);
    return;
  }
  /* Initialize Array */
  for (ArtNr= 0, ActArtP= RootField; ArtNr<NrOfArticles; ArtNr++, ActArtP= &RootField[ArtNr])
  {
    ActArtP->Status= NONEXISTENT;
    ActArtP->NextArt= ActArtP->PrevArt= NULL;
    ActArtP->From= NULL;
    ActArtP->Subject= NULL;
    ActArtP->NrofFollowUps=
    ActArtP->Date= 0UL;
    ActArtP->ExtData=NULL;
  }

  RootSubjP= ReadDataBase(pGI,RootField); /* read in DataBase if exists */
  /* RootSubjP is set to NULL, if no dbase exists or all articles are
     no longer valid (i.e. expired) */

  pGI->RootSubjP= RootSubjP;
  if (switches & SW_DEBUG)
  {
    printf("Read Database in and got:\n");
    PrintGroup(pGI);
  }

  if (pGI->OldlastArtNr==0L)  /* complete new scan, because active.bak is missing */
  {
    pGI->OldlastArtNr= pGI->firstArtNr-1;
  }

  /* scan only new articles */
  for (ArtNr=pGI->OldlastArtNr+1; ArtNr<=LastArtNr; ArtNr++)
  {
    if (ScanArticle(ArtNr)==3) break; /* break signalled */
  }

#ifdef MYFORMAT
  len= strlen(pGI->Name);
  if (len>MAXGROUPLEN_DISPLAY)
  {
    sprintf(groupname_tmp,formatstr,&(pGI->Name)[len-MAXGROUPLEN_DISPLAY]);
  }
  else
    sprintf(groupname_tmp,formatstr,pGI->Name);
#endif

  if (valid_ds!=NrOfArticles && GlobalERROR!=ABORT && (switches & SW_CONSISTENCY_CHECK))
  {
    for (ArtNr= 0, ActArtP= RootField; ArtNr<NrOfArticles; ArtNr++, ActArtP= &RootField[ArtNr])
    {
      if (ActArtP->From==NULL && ((ArtNr+FirstArtNr)!=0))
      {
        if (ScanArticle(ArtNr+FirstArtNr)==0) /* scan missing articles */
        { /* article was successfully scanned */
#ifdef MYFORMAT
          if (logFP)
             fprintf(logFP, "%s%s -(%s)- %s: missing article:%lu\n",
                             prgname, Version, TimeStr(), groupname_tmp, ArtNr+FirstArtNr);
          printf("%s: missing article:%lu\n", groupname_tmp, ArtNr+FirstArtNr);
#else
          if (logFP)
             fprintf(logFP, "%s%s -(%s)- %-25s: article %5lu was missing\n",
                             prgname, Version, TimeStr(), pGI->Name, ArtNr+FirstArtNr);
          printf("%-25s: article %5lu was missing\n", pGI->Name, ArtNr+FirstArtNr);
#endif
        }
      }
    }
  }


  NewArts=LastArtNr-pGI->OldlastArtNr;

#ifdef MYFORMAT
  if (logFP)
     fprintf(logFP, "%s%s -(%s)- %s: %6lu/%lu:%lu\n",
                     prgname, Version, TimeStr(), groupname_tmp, NewArts, valid_ds, NrOfArticles);
#else
  if (logFP)
  {
    if (NewArts>0 || (switches & SW_CONSISTENCY_CHECK))
       fprintf(logFP, "%s%s -(%s)- %-25s: %6lu/%lu:%lu\n",
                      prgname, Version, TimeStr(), pGI->Name, NewArts, valid_ds, NrOfArticles);
  }
#endif
  TotalScanned+= NewArts;
  ScannedGroups++;

  FreeString(pathtogroup);
  FreeString(pathtodbase);
  pathtogroup= pathtodbase= NULL;

  pGI->RootSubjP= RootSubjP;
}


void
ScanAll(void)
/* scans all groups for new articles, updates all dbases */
{
  struct GroupInfoList *pGI;


  for (pGI= RootGI; pGI!=NULL; pGI= pGI->NextGI)
      if (pGI->OldlastArtNr < pGI->lastArtNr ||  /* new articles */
          pGI->OldfirstArtNr < pGI->firstArtNr || /* expired arts */
          switches & SW_CONSISTENCY_CHECK)
      {
        ScanGroup(pGI);
        if (switches & SW_DEBUG)
        {
          PrintGroup(pGI);
          printf("writing database...\n");
        }
        WriteDataBase(pGI);
        if (switches & SW_DEBUG)
        {
          printf("Freeing mem...\n");
        }
        FreeGroup(pGI);
        if (SetSignal(0L,0L) & (SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D))
        {
          ErrorHandle(ABORT,NULL);
        }
      }
}


void CleanUp(void) /* Erase all .newdb files */
{
  struct GroupInfoList *pGI;


  printf("Deleting all databases!...");

  for (pGI= RootGI; pGI!=NULL; pGI= pGI->NextGI)
  {
    strcpy(tmp,pGI->Name);
    MakeDirName(fnametmp, (dbasedir!=NULL ? dbasedir: newsdir), tmp);
    strcat(fnametmp,"/.newsdb");

    DeleteFile(fnametmp);
  }

  sprintf(fnametmp,"%s.bak",activefile);
  DeleteFile(fnametmp);

  printf("ready.\n");

  if (logFP)
    fprintf(logFP, "%s%s -(%s)- DELETED ALL databases!!\n",
                    prgname, Version, TimeStr());
}


int
CloseMaster(void)
{
  if (logFP)
  {
    if (TotalScanned) fprintf(logFP, "%s%s -(%s)- Total articles scanned: %lu in %lu/%lu groups\n",
                      prgname, Version, TimeStr(), TotalScanned, ScannedGroups, TotalGroups);
  }

  if (switches & SW_DEBUG)
     Display("\nArnMaster - Shutdown: ");
  FreeGIL();

  if (switches & SW_DEBUG)
     Display("1..");

  if (tmp)          FreeMem(tmp, (ULONG) SUPERLINE+1UL);
  if (fnametmp)     FreeMem(fnametmp, (ULONG) MAXLINE+1UL);
  if (mBuffer)      FreeMem(mBuffer, M_BUFFERSIZE);
  if (HeaderBuf)    FreeMem(HeaderBuf, sizeofHeaderBuf);
  FreeString(activefile);
  FreeString(newsdir);
  FreeString(logfile);

  if (switches & SW_DEBUG)
     Display("2..");

  if (switches & SW_PRIORITY)
     SetTaskPri(FindTask(NULL),OldPriority);

  if (switches & SW_DEBUG)
     Display("3.\n");

  return (0);
}



void InitMaster(void)
{
  RootGI= NULL;
#ifdef OLD
  RootNI= NULL;
#endif
  TotalScanned= TotalGroups= ScannedGroups= 0L;

  tmp= (char *) AllocMem((ULONG) SUPERLINE+1L, MEMTYPE);
  fnametmp= (char *) AllocMem((ULONG) MAXLINE+1L, MEMTYPE);
  mBuffer= (char *) AllocMem(M_BUFFERSIZE, MEMTYPE);
  HeaderBuf= (char *) AllocMem(sizeofHeaderBuf, MEMTYPE);

  if (!(tmp && fnametmp && mBuffer && HeaderBuf))
  {
    ErrorHandle(NOMEM,NULL);
    CloseMaster();
    exit(RETURN_ERROR);
  }
  sprintf(tmp, InitMessage, Version, Last_UpdateD, Last_UpdateT);

  Display(tmp);
}


int
SetPriority(char *prio)
{
  char buf, *t;
  long priority;

  for (t=prio; *t!='\0' && (isdigit(*t) || *t=='-'); t++);

  buf=*t;
  *t='\0';
  if (*prio=='\0')
  {
    *t=buf;
    return 0;
  }
  else
   priority= atol(prio);

  if (priority>=-127 || priority<=128)
  {
     OldPriority= SetTaskPri(FindTask(NULL),priority);
     switches|=SW_PRIORITY;
  }

  *t=buf;
  return (int) (t-prio);
}


checkswitches(short *sw, char **argvp, int argc)
{
  int i, j;

  onlyonegroup= NULL;

  for (*sw= 0,j=1; j<argc;j++)
  {
    if (*argvp[j]=='-')
    {
      for (i=1; argvp[j][i]!='\0'; i++)
      {
        switch(argvp[j][i])
        {
          case '?':
          case 'i': *sw|= SW_INFO;
                    break;
          case 'd': *sw|= SW_DEBUG;
                    break;
          case 'G':
          case 'g': if (argvp[j][i]=='G') *sw|= SW_SINGLEDIR;
                    *sw|= SW_GOTOGROUP;
                    if (j+1<=argc)
                    {
                      onlyonegroup=argvp[j+1];
                      /* printf("group: %s\n",onlyonegroup); */
                    }
                    else
                       *sw&=~SW_GOTOGROUP;
                    break;
          case 'P': i+=SetPriority(&(argvp[j][i+1]));
                    /* SW_PRIORITY is set if prio is changed */
                    break;
          case 'X': *sw|= SW_CLEANUP;
                    break;
          case 'A': *sw|= SW_UPDATEACTIVE;
                    break;
          case 'L': *sw|= SW_COUNTLINES;
                    break;
          case 'C': *sw|= SW_CONSISTENCY_CHECK;
                    break;
           default: ErrorHandle(USAGE,NULL);
                    return 1;
                    break;
        }

        if (onlyonegroup)
        {
          j++; /* skip groupname */
          break;
        }
      }
    }
    else
    {
      ErrorHandle(USAGE,NULL);
      return 1;
      break;
    }
  }
  return 0;
}


void GetMasterConfig(void)
{
  FILE *conffile;
  char *c, *defnewsdir=DEFAULT_NEWSDIR;


  conffile= fretryopen(MASTERCONFIG_1,"r");
  if (!conffile) conffile= fretryopen(MASTERCONFIG_2,"r");

  if (conffile)
  {
    while (fgets(tmp, SUPERLINE, conffile))
    {
      if (c= firststr(tmp, CONF_ACTIVEFILE))
      {
        Entry (c, &activefile);
      }
      else
      if (c= firststr(tmp, CONF_NEWSDIR))
      {
        Entry (c, &newsdir);
      }
      else
      if (c= firststr(tmp, CONF_DBASEDIR))
      {
        Entry (c, &dbasedir);
      }
      else
      if (c= firststr(tmp, CONF_LOGFILE))
      {
        Entry (c, &logfile);
      }
      if (c= firststr(tmp, CONF_DOTDIRS))
         Dots_to_Slashes= false;
      else
      if (c= firststr(tmp, CONF_COUNTLINES))
         switches|=SW_COUNTLINES;
      else
      if (c= firststr(tmp, CONF_UPDATEACTIVE))
         switches|=SW_UPDATEACTIVE;
    }
    fclose(conffile);
  }
  else
  {
    activefile=NULL;
    ErrorHandle(NOCONFIG,NULL);
  }
  if (newsdir==NULL)
  {
    newsdir=CopyString(defnewsdir);
    strcpy(newsdir,defnewsdir);
  }
}


#ifdef __SASC
int nop_func(void)
{
  return(0);
}
#endif


main(int argc, char **argv)
{
  struct GroupInfoList *pGI;
  ULONG scantime;
  FILE *fp;
  BOOL fixactive_mode;

  prgname= *argv;
  GlobalERROR= OK;

#ifdef __SASC
  onbreak(&nop_func);
#endif

  InitMaster();

  checkswitches(&switches, argv, argc);

  if(fixactive_mode= (strnicmp(prgname,"fixactive",9)==0))
     switches|= SW_UPDATEACTIVE;

  GetMasterConfig();

  if (switches & SW_INFO)
  {
    ErrorHandle(USAGE,NULL);
    CloseMaster();
    exit(RETURN_OK);
  }

  if (logFP= fretryopen(logfile,"a"))
     fprintf(logFP,"%s%s -(%s)- STARTED\n", prgname, Version, TimeStr());

  if (logFP && (switches & SW_CONSISTENCY_CHECK))
     fprintf(logFP,"%s%s -(%s)- CONSISTENCY CHECK\n", prgname, Version, TimeStr());

  sprintf(formatstr,"%%%d.%ds",MAXGROUPLEN_DISPLAY,MAXGROUPLEN_DISPLAY);

  stopwatch= GetTime();

  if (switches & SW_SINGLEDIR)
  {
    if (RootGI= AllocMem((ULONG) sizeof(struct GroupInfoList),MEMTYPE))
    {
      RootGI->NextGI= RootGI->PrevGI= NULL;
      RootGI->Name= CopyString(onlyonegroup);
      RootGI->RootSubjP= NULL;
      RootGI->firstArtNr= RootGI->lastArtNr=
      RootGI->OldlastArtNr= RootGI->OldfirstArtNr= 0L;
      ScanArticles(RootGI,NULL);
      if (fp=fretryopen(MASTERINFO,"w"))
      {
        fprintf(fp,"%lu %lu\n",RootGI->lastArtNr,RootGI->firstArtNr);
        fclose(fp);
      }
    }
    else
     ErrorHandle(NOMEM,"RootGI");
  }
  else
  {
    ReadActiveFile();
    ReadOldActiveFile();
    /* UpdateActive is only called if we want to examine more than one group */
    if ((switches & SW_UPDATEACTIVE) && !(switches & SW_GOTOGROUP))
    {
      scantime= GetTime();
      UpdateActive();
      scantime= GetTime()-scantime;
      if (logFP)
         fprintf(logFP,"%s%s -(%s)- Updated ACTIVEFILE: Elapsed time %02lu:%02lus\n", prgname, Version, TimeStr(), scantime/60, scantime%60);
    }
  }

  if (switches & SW_GOTOGROUP) /* scan single group */
  {
    pGI= FindGroup(onlyonegroup);
    if (pGI)
    {
      if (switches & SW_UPDATEACTIVE) /* scan this specific group only! */
         ScanArticles(pGI,NULL);

      if (pGI->OldlastArtNr<pGI->lastArtNr || (switches & (SW_CONSISTENCY_CHECK|SW_UPDATEACTIVE)))
      {
        ScanGroup(pGI);
        if (switches & SW_DEBUG)
        {
          PrintGroup(pGI);
          printf("writing database...\n");
        }
        WriteDataBase(pGI);
        if (switches & SW_DEBUG)
        {
          printf("Freeing mem...\n");
        }
        FreeGroup(pGI);
      }
    }
  }
  else
  if (switches & SW_CLEANUP) /* remove all databases */
  {
    CleanUp();
  }
  else
  if (!fixactive_mode)
  {
    ScanAll();
    if (GlobalERROR==OK)
    {
      sprintf(fnametmp,"copy %s %s.bak", activefile, activefile);
      Execute(fnametmp,NULL,NULL);
    }
  }

  stopwatch= GetTime()-stopwatch;

  CloseMaster();

  if (!fixactive_mode && logFP)
  {
    if (expired)
       fprintf(logFP,"%s%s -(%s)- Expired : %lu\n", prgname, Version, TimeStr(), expired);
    if (switches & SW_COUNTLINES)
       fprintf(logFP,"%s%s -(%s)- CREATED 'Lines:' : %lu\n", prgname, Version, TimeStr(), recounted_lines);
    if (switches & SW_CONSISTENCY_CHECK)
       fprintf(logFP,"%s%s -(%s)- INFO: Databases use %lu bytes\n", prgname, Version, TimeStr(), TotalDBbytes);

    fprintf(logFP,"%s%s -(%s)- ENDED: Elapsed time %02lu:%02lus (%lu/min)\n", prgname,
                     Version, TimeStr(), stopwatch/60, stopwatch%60, stopwatch?(TotalScanned*60)/stopwatch:0);
  }

  if (logFP) fclose(logFP);
}


