/************************************************************************
 * Choosegroups - Newsgroups-selector                                   *
 ************************************************************************
 *                  Arn - (C)opyright by Roland Bless                   *
 *                All rights reserved, written 1990-1992                *
 * -------- YOU ARE NOT ALLOWED TO DISTRIBUTE A CHANGED SOURCE -------- *
 ************************************************************************/

#ifndef ENUMS_H
 #include "enums.h"
#endif

#ifndef GLOBALVARS_H
#include "globalvars.h"
#endif

#undef DEBUG

/* Prototypes for functions defined in choosegroups.c */
/* proto$start */
void EditGlobalKill(void);
BOOL CatchUpGroup(struct GroupInfoList *pGI);
ULONG longceil(ULONG dividend,
               ULONG divisor);
void SetGroupCursor(void);
void SelectLine(int linenr,
                int *cursory);
void DeSelectLine(int linenr,
                  int *cursory);
int OptionsMenu(int depth);
BOOL GroupSelCommand(char key,
                     int *command);
void UpdateGroupMenu(ULONG firstgroup,
                     int firstline,
                     int lines,
                     struct GroupInfoList *grouparray[],
                     BOOL only_refresh_set);
void RefreshGroupLines(ULONG firstart,
                       int lines,
                       struct GroupInfoList *artarray[],
                       BOOL only_refresh_set);
int  GetLineNrFromKey(char key, int lines);
void GetKeyFromLine(int line);
BOOL SelectGroups(ULONG firstgroup,
                  int lines,
                  struct GroupInfoList **grouparray,
                  int *command);
ULONG FilterGroups(struct GroupInfoList **grouparray,
                  UWORD unsetbits,
                  UWORD setbits);
ULONG FindGroupArrayNum(struct GroupInfoList **grouparray,
                        char *groupname);
int ChooseGroups(int *choosegroup_cmd);
/* proto$end */

#include "proto/arnhelp_proto.h"
#include "proto/ArnIO_proto.h"
#include "proto/Arn_proto.h"
#include "proto/functions_proto.h"
#include "proto/followup_proto.h"
#include "proto/grouplist_proto.h"
#include "proto/kill_proto.h"
#include "proto/pager_proto.h"
#include "proto/strprocs_proto.h"
#include "proto/cmdline_proto.h"
#include "proto/readgroup_proto.h"


/* from version.c */
extern char *Version;

extern ULONG NrOfArticles, FirstArtNr, LastArtNr;

extern char *noaction;
extern int  art_lines;

                   /*************  global for this file *************/
char *Menulines[MAX_SELECT_LINES], /* array of pointers to MenuLines */
     *GroupMenuTemplate;           /* Template for printing the overview */
int group_cursory,
    max_grouplines,  /* maximum of lines that can be used by groups */
    grouplines,      /* actual maximum of lines available for groups (title toggle) */
    max_groupnamelen;
ULONG NrOfActGroups,     /* Nr of actual Groups that are displayed */
     UserlastArtNr, NrOfNewArts;
                                   /*******************/


void
EditGlobalKill(void)
{
  BOOL titleset= false;

  if (!(GeneralBool & SCREENTITLE))
  {
    ShowTitle(ArnScreen, TRUE);
    titleset= true;
  }

  StatusLine();
  DispMsg("Invoking editor...");
  if (GeneralBool & EDSCRSWITCH)
     WBenchToFront();
  sprintf(tmp,"%s %s", NewsEditor, globalkillfile);
  Execute(tmp,(BPTR) NullFH,(BPTR) NullFH);
  ScreenToFront(ArnScreen);

  FreeKill_List(GlobalKill_List);
  DispMsg("rereading global X-file...");
  ReadKillFile(&GlobalKill_List,globalkillfile);
  DispMsg("ready.\n");

  if (titleset) ShowTitle(ArnScreen, FALSE);
}


BOOL
CatchUpGroup(struct GroupInfoList *pGI)
/* CatchUps and marks the group */
{
  struct Markers *pmark;
  int no= 1;
  ULONG ArtNr;

  if (GeneralBool & CATCHUP_XREFS)
  {
    LastArtNr= pGI->lastArtNr;
    FirstArtNr= pGI->firstArtNr;
    NrOfArticles= LastArtNr - FirstArtNr + 1L;

    if (NrOfArticles<=0L)
    {
      StatusLine(); DispMsg("Empty group!\n");
      Delay(10L);
      return false;
    }

    /* Reserve array of article-structures */
    RootArt= AllocMem((ULONG) sizeof(struct art_struct)*NrOfArticles, MEMTYPE);
    if (!RootArt)
    {
      ErrorHandle(NOMEM, NULL);
      return false;
    }
    /* Initialize Array */
    InitArtArray(RootArt, &LastArt);


    strcpy(newsgroup,pGI->Name);

    MakeDirName(tmp, newsdir, newsgroup);

    strncpy(pathtogroup,tmp,MAXLINE); /* In pathtogroup is now the path
                                         to the group! */
    pathtogroup[MAXLINE]= '\0';

    /* get XREF-flags from Database */
    pGI->RootSubjP= ReadDataBase(pGI,RootArt,&ArtNr);
    for (pmark= pGI->Marks; pmark!=NULL; pmark= pmark->NextMark)
        if (pmark->Mark)
          CheckLibMarks(pmark->Mark,pGI);        /* Get the Marks */
    no= CatchUp(true); /* no==1 if not catchuped */

    FreeSubjandFrom(pGI);
    FreeMem(RootArt, (ULONG) sizeof(struct art_struct)*(ULONG) NrOfArticles);
    RootArt= NULL;
  }
  else
  {
    sprintf(tmp,"Really mark all articles as read? [y,N]:");
    StatusLine();
    DispMsg(tmp);
    *tmp=GetUpKey();
    ClrLine();
    if (*tmp=='Y')
      no= 0;
    else
      no= 1;
  }

  if (no==0) /* Catchup */
  {
    FreeMarks(pGI->Marks);
    pGI->Marks= NewMark(pGI);  /* Mark the whole group */
  }

  return (BOOL) ((no==0) ? true : false);
}


ULONG
longceil(ULONG dividend, ULONG divisor)
{
  ULONG result= dividend/divisor;

  return (dividend%divisor == 0) ? result : result+1;
}


#ifdef DOUBLECURSOR
void UnsetGroupCursor(void)
{
  (*ConWriteChar)(group_cursory<MENU_OFFSET_1 ? ' ' : OFFSET_1_CHAR);
  GotoXY(MaxChars-1,group_cursory+2);
  (*ConWriteChar)(' ');
}
#endif


void SetGroupCursor(void)
{
#ifdef DOUBLECURSOR
  GotoXY(MaxChars-1,group_cursory+2);
  (*ConWriteChar)('<');
#endif
  GotoXY(1,group_cursory+2);
  (*ConWrite)(">\r");
  /* (*ConWriteChar)('>');
    (*ConWrite)(tc_CURSOR_LEFT); */
}


void
SelectLine(int linenr, int *cursory) /* Select linenr */
{
  (*ConWriteChar)(*cursory+2<=MENU_OFFSET_1 ? ' ' : OFFSET_1_CHAR);
  *cursory= linenr;
  GotoXY(2,*cursory+2);
  InvON();
  (*ConWrite)(Menulines[linenr]);
  InvOFF();
}


void
DeSelectLine(int linenr, int *cursory) /* Select linenr */
{
  (*ConWriteChar)(*cursory+2<=MENU_OFFSET_1 ? ' ' : OFFSET_1_CHAR);
  *cursory= linenr;
  GotoXY(2,*cursory+2);
  (*ConWrite)(Menulines[linenr]);
}


int
OptionsMenu(int depth)
{
  char key;

  StatusLine();
  DispMsg("H)eaderstop");
  if (GeneralBool & HEADERSTOP)
    DispMsg(" off, ");
  else
    DispMsg(" on, ");
  DispMsg("L)ocal Kill");
  if (GeneralBool & LOCALKILLON)
    DispMsg(" off, ");
  else
    DispMsg(" on, ");
  DispMsg("G)lobal Kill");
  if (GeneralBool & GLOBALKILLON)
    DispMsg(" off, ");
  else
    DispMsg(" on, ");
  DispMsg("C)learScreen");
  if (GeneralBool & CLEARSCREEN)
    DispMsg(" off, ");
  else
    DispMsg(" on, ");

  DispMsg("Q)uit:");

  key= GetUpKey();
  ClrLine();
  switch (key)
  {
    case 'H':headerstop();
             break;
    case 'L':KillOnOff(false);
             break;
    case 'G':KillOnOff(true);
             break;
    case 'C':clrscreenmode();
             break;
    default: DispMsg(noaction);
             break;
  }
  if (depth==PAGER) ConCR();

  return NOP;
}


BOOL GroupSelCommand(char key, int *command)
{
  BOOL leave= false;
  short int cu_num;

  if (cu_num= CursorKeys(&key))
  {
    switch (cu_num)
    {
      case CUNUM_UP: *command= LINEUP;
                return false;
                break;
      case CUNUM_DOWN: *command= LINEDOWN;
                return false;
                break;
      case CUNUM_RIGHT: *command= PAGEFORWARD;
                return true;
                break;
      case CUNUM_LEFT: *command= PAGEBACK;
                return true;
                break;
    }
  }

  switch(key)
  {
    case '-': *command= SELECTRANGE;
              break;
    case '<': *command= PAGEBACK;
              break;
    case '>': *command= PAGEFORWARD;
              break;
    case ' ': *command= SPACE;
              break;
    case '^': *command= FIRSTPAGE;
              break;
    case '$': *command= LASTPAGE;
              break;
    case '+': *command= SELECT_NEW;
              break;
    case '@': *command= SELECT_INVERSE;
               break;
    case '#': *command= SELECT_ALL;
               break;
    case '%': *command= UNSELECT_ALL;
               break;
    case 'A': *command= SHOWALL;
              break;
    case '\x0d': *command= ENTERGROUP;  /* ENTER */
                 break;
    case 'R':*command= REREAD_ACTIVE;
             break;
    case ctrl('T'):*command= TOGGLE_SCREENTITLE;
                   break;
    case ctrl('L'): *command= REFRESHSCREEN;
                  break;
    case 'O': *command= OptionsMenu(GROUPSEL);
              break;
    case 'K': *command= EDIT_KILL;
              break;
    case 'E': *command= POSTARTICLE;
                 break;
    case 'U': *command= UNSUB;
              break;
    case 'B': versioninfo();
              break;
    case 'H': dohelp(H_GROUPSEL);
              *command= REFRESHSCREEN;
              break;
    case 'L': *command= LISTGROUPS;
              break;
    case 'C': *command= CATCHUP;
              break;
    case 'V': *command= GROUPMENU;
              break;
    case 'X': *command= QUIT;                /* Don't update newmarks! */
              break;
    case 'Q': *command= QUIT;
              GeneralBool|= UPDATE_NEWMARKS; /* Update Newmarks! */
              break;
    case 'Z': *command= READNOW;
              break;
    case ctrl('C'): *command= BREAK;  /* CTRL-C Quick exit */
               GlobalERROR= QUITARN;
               break;
     default: *command= NOP;
              break;
  }

  switch (*command) /* which command aborts THIS newsgroup selection-page? */
  {
    case READNOW:
    case ENTERGROUP:
    case PAGEFORWARD:
    case SPACE:
    case PAGEBACK:
    case FIRSTPAGE:
    case LASTPAGE:
    case SHOWALL:
    case LISTGROUPS:
    case REFRESHSCREEN:
    case REFRESHLINE:
    case REREAD_ACTIVE:
    case CATCHUP:
    case POSTARTICLE:
    case EDIT_KILL:
    case TOGGLE_SCREENTITLE:
    case QUIT:
    case BREAK:
         leave= true;
         break;
    default: leave= false; /* These commands MUST BE handled by SelectGroups */
  }

  return leave;
}


void
UpdateGroupMenu(ULONG firstgroup, int firstline, int lines,
                struct GroupInfoList *grouparray[], BOOL only_refresh_set)
/* Update GroupArray: firstgroup is the index of the first group to update
   (this is viewed as the first group on page {Menulines[0]})
   firstline is the "offset" to firstgroup (page), lines: nr of groups
   to update from offset on. grouparray: array containing all pointers to
   filtered groups. The RESULT is found in MENULINES[].
   only_refresh_set is true if only those lines should be "refreshed"
   which need refresh (indicated by G_NEEDS_REFRESH-flag)                  */

{
  char key, GroupFlags[5];
  ULONG NrOfUnreadArts, TotalArts, NrOfReadArts,
       i, lastline= firstline+lines;
  struct Markers *Marks;
  struct GroupInfoList *pGI;

  for (i= firstline; i<lastline; i++) /* begin with line firstline, then refresh */
                                  /* number of lines entries */
  {
    if (firstgroup+i<NrOfActGroups)
       pGI=grouparray[firstgroup+i];
    else
       Menulines[i][0]= '\0';

    if (!only_refresh_set || (pGI->Status & G_NEEDS_REFRESH))
    {
      FirstArtNr= pGI->firstArtNr; /* set for CountMarks! */
      LastArtNr = pGI->lastArtNr;  /* set for CountMarks! */
      UserlastArtNr= pGI->UserlastArtNr;

      NrOfNewArts= LastArtNr-UserlastArtNr;

      if (NrOfNewArts==0UL)
          pGI->Status&= ~NEWARTS_GROUP;

      if (LastArtNr==0UL && FirstArtNr==0UL) /* empty group */
         TotalArts= 0UL;
      else
      {
        TotalArts= LastArtNr-FirstArtNr+1UL;
      }

      NrOfReadArts= 0L;
      for (Marks= pGI->Marks; Marks!=NULL && Marks->Mark; Marks= Marks->NextMark)
          NrOfReadArts+= CountMarks(Marks->Mark);

      NrOfUnreadArts= TotalArts-NrOfReadArts;

      strcpy(GroupFlags,"    "); /* reset groupflags */
      if (pGI->Status & NEWARTS_GROUP)
         GroupFlags[0]= 'N';
      if (NrOfUnreadArts==0)
      {
        GroupFlags[1]= 'R';
        pGI->Status&= ~UNREAD_GROUP;
        NrOfNewArts= 0L;
      }

      if (pGI->Status & NEW_GROUP)    GroupFlags[1]= '+'; /* New group */
      if (pGI->Status & UNSUBSCRIBED) GroupFlags[1]= 'U'; /* Unsubscribed Group */
      if (pGI->Status & MODERATED)    GroupFlags[2]= 'M'; /* Moderated Group */

      if (i<MENU_OFFSET_1)
         key= i+'a';
      else
      if (i<=MENU_OFFSET_1+'z'-'a')
         key= i+'a'-MENU_OFFSET_1;
      else
         key= i+'A'-(MENU_OFFSET_1+'z'-'a'+1);

      sprintf(Menulines[i], GroupMenuTemplate, key,
              GroupFlags[0], pGI->Name, GroupFlags[1], GroupFlags[2],
              NrOfNewArts, NrOfUnreadArts, TotalArts);
      Menulines[i][MaxChars]= '\0';

    }
  } /* for{i} */
}


void
RefreshGroupLines(ULONG firstgroup, int lines,
                  struct GroupInfoList *grouparray[], BOOL only_refresh_set)
/* only printfs those lines from Menulines with G_NEEDS_REFRESH set */
{
  int i;
  BOOL tmpbool;
  UWORD stat;
  ULONG actgroup;

  if (!only_refresh_set) /* start with headerline */
  {
    /* headerline */
    sprintf(tmp," Arn by R.Bless - %4ld/%-4ld Newsgroups - page %3ld of %-3ld New/Unread/Total",
                NrOfActGroups, NrOfGroups, longceil(firstgroup+1,grouplines),
                longceil(NrOfActGroups,grouplines));

    /* fill to end of line with spaces, so we get a black bar */
    for(i= strlen(tmp); i<MaxChars-1; i++)
       tmp[i]= ' ';
    tmp[i]='\0';

    /* print headerline */
    ClrScr();
    InvON();
    (*ConWriteLn)(tmp);
    InvOFF();
  }

  /* display page of groups */
  for(i= 0,actgroup= firstgroup; i<lines; i++,actgroup++)
  {
    stat= grouparray[actgroup]->Status;

    if (!only_refresh_set)
    {
      (*ConWriteChar)(i<MENU_OFFSET_1 ? ' ' : OFFSET_1_CHAR);
    }
    else
    {
      if (stat & G_NEEDS_REFRESH)
      {
        grouparray[actgroup]->Status&= ~G_NEEDS_REFRESH; /* delete flag */
        GotoXY(2,i+2); /* position cursor */
      }
      else
        continue; /* don't print the line */
    }

    if (tmpbool= (stat & SELECTED_GROUP))
       InvON();

    (*ConWrite)(Menulines[i]); /* Display all lines */

    if (tmpbool) InvOFF();
  }
}


int
GetLineNrFromKey(char key, int lines)
/* returns number from 0..lines-1 or -1 on error
   valid inputs: 'a'-'z','.a'-'.z','.A'-'.Z'
   if c is OFFSET_1_CHAR the second char is read in
*/
{
  int return_num;

  return_num= 0;

  if (key==OFFSET_1_CHAR) /* read in next character */
  {
    return_num+= MENU_OFFSET_1; /* the OFFSET_1_CHAR steps to the next range */
    key= GetKey();
    if (key>='a' && key<='z') /* 2nd range */
    {
      return_num+= key-'a';
    }
    else
    if (key>='A' && key<='Z') /* 3rd range */
    {
      return_num+= key-'A'+'z'-'a'+1;
    }
  }
  else
  {
    if (key>='a' && key<='z')
       return_num+= key-'a';
    else return -1;
  }

  if (return_num>=0 && return_num<lines)
     return return_num;
  else
     return -1; /* error */
}


void
GetKeyFromLine(int line)
/* writes the corresponding char(s) that belong(s) to line */
{
  if (line>=MENU_OFFSET_1)
  {
    (*ConWriteChar)(OFFSET_1_CHAR); /* offset char */
    line-= MENU_OFFSET_1;
  }
  if (line>'z'-'a')
  {
    line-= 'z';
    (*ConWriteChar)('A'+(char) line);
  }
  else
    (*ConWriteChar)('a'+(char) line);
}


BOOL
SelectGroups(ULONG firstgroup, int lines, struct GroupInfoList *grouparray[],
             int *command)
/* NOTE: uses tmp
   This procedure returns the "leave" condition,
   <firstgroup> is the index number from the first group shown on the page (MenuLines[0])
   <lines>      is the number of lines that are available for selection on this page
   <grouparray> is the array of pointers to all actually listed groups
   <command>    is the command that is returned
*/
{
  int i, begin, end, selected, seltmp;
  char key;
  BOOL leave= false;
  ULONG actgroup, index;

  *command= NOP;

  /* Built Screen */
  UpdateGroupMenu(firstgroup, 0, lines, grouparray, false);
  RefreshGroupLines(firstgroup, lines, grouparray, false);

  leave= false;
  begin= selected= end= -1; /* begin & selected:0..25 ! */
  SetGroupCursor();
  do
  {
    *command= NOP;
    key= (*GetKey)();

    /* for a range last <selected> must be retained, so don't call GetLineNrFromKey() */
    if ((seltmp= GetLineNrFromKey(key, lines))>=0)
    {
      selected= seltmp;
      if (begin>=0)  /* range - got end value */
      {
        end= selected;
        GetKeyFromLine(end);
        ClrLine();
        SetGroupCursor();
        if (begin>end) /* swap range */
        {
          int help= begin;

          begin= end-1;
          end= help-1;
        }
        for (i= begin+1,actgroup= firstgroup+i; i<=end; i++,actgroup++)
        {
          if (grouparray[actgroup]->Status & SELECTED_GROUP)
          {
            grouparray[actgroup]->Status&= ~SELECTED_GROUP;
            DeSelectLine(i, &group_cursory);
          }
          else
          {
            grouparray[actgroup]->Status|= SELECTED_GROUP;
            SelectLine(i, &group_cursory);
          }
        }
        end= begin= -1;
      }
      else /* select single line */
      {
        actgroup= firstgroup+selected;
        if (grouparray[actgroup]->Status & SELECTED_GROUP)
        {
          grouparray[actgroup]->Status&= ~SELECTED_GROUP;
          DeSelectLine(selected, &group_cursory);
        }
        else
        {
          grouparray[actgroup]->Status|= SELECTED_GROUP;
          SelectLine(selected, &group_cursory);
        }
      }
      SetGroupCursor();
    } /* if selected>=0 */
    else /* no direct selection -> interpret other command */
    {
      if (begin>=0 && *command==NOP)/* no real range */
      {
        begin= selected= -1;
        ClrLine();
        *command= NOP;
      }
      else
      {
        leave= GroupSelCommand(key,command);
        if (key!=ctrl('Z')) cmdline[0]='\0';       /* check for command-line */
        EnterCmdLine(key, GROUPSEL);            /* This MUST hold a disjoint set
                                                   of keys with GroupSelComm */
        if (*cmdline!='\0')
           leave= InterpretCmdLine(cmdline, command, GROUPSEL);
      }
    }

    switch(*command) /* Which action can we examine now? */
    {
      case SELECTRANGE: if (selected>=0)
                        {
                          begin= selected;
                          (*ConWriteChar)(group_cursory<MENU_OFFSET_1 ? ' '
                                                                      : OFFSET_1_CHAR);
                          StatusLine();
                          (*ConWrite)(">Select range: ");
                          GetKeyFromLine(selected);
                          (*ConWriteChar)('-');
                        }
                        break;
      case LINEUP: if (group_cursory>0)
                   {
                     (*ConWriteChar)(group_cursory<MENU_OFFSET_1 ? ' ' : OFFSET_1_CHAR);
                     group_cursory--;
                     SetGroupCursor();
                   }
                   break;
      case LINEDOWN: if (group_cursory<lines-1)
                     {
                       (*ConWriteChar)(group_cursory<MENU_OFFSET_1 ? ' ' : OFFSET_1_CHAR);
                       group_cursory++;
                       SetGroupCursor();
                     }
                     break;
      case POSTARTICLE: ClrScr();
                        PostArticle(NULL);
                        *command= NOP;
                        break;
      case EDIT_KILL: EditGlobalKill();
                      *command= NOP;
                      break;
      case GROUPMENU: leave= ViewGroupMenu(grouparray[firstgroup+group_cursory], command);
                      break;
      case UNSUB:     leave= UnSubscribeMenu(grouparray[firstgroup+group_cursory], command);
                      break;
      case SELECT_NEW: for(index= 0; index<NrOfActGroups; index++)
                         if (grouparray[index]->Status & NEWARTS_GROUP)
                             grouparray[index]->Status|= (SELECTED_GROUP | G_NEEDS_REFRESH);
                       *command= REFRESHLINES;
                       break;
      case SELECT_ALL: for (index= 0L; index<NrOfActGroups; index++)
                         if (!(grouparray[index]->Status & SELECTED_GROUP))
                         {
                            grouparray[index]->Status|= (G_NEEDS_REFRESH|SELECTED_GROUP);
                         }
                       *command= REFRESHLINES;
                       break;
      case UNSELECT_ALL: for (index= 0L; index<NrOfActGroups; index++)
                             if (grouparray[index]->Status & SELECTED_GROUP)
                             {
                               grouparray[index]->Status&= ~SELECTED_GROUP;
                               grouparray[index]->Status|= G_NEEDS_REFRESH;
                             }
                       *command=REFRESHLINES;
                       break;

      case SELECT_INVERSE: for (index= firstgroup; index<firstgroup+lines; index++)
                             if (grouparray[index]->Status & SELECTED_GROUP)
                                 grouparray[index]->Status&= ~SELECTED_GROUP;
                              else
                                 grouparray[index]->Status|= SELECTED_GROUP;
                           *command= REFRESHSCREEN;
                           leave= true;
                           break;
      case CATCHUP: if (CatchUpGroup(grouparray[firstgroup+group_cursory]))
                    {
                      *command= REFRESHLINE;
                    }
                    else
                      *command= NOP;
                    break;
      case PAGEFORWARD: if (firstgroup+lines>=NrOfActGroups)
                        {
                          *command= NOP;
                          leave= false;
                        }
                        break;
      case PAGEBACK:    if (firstgroup<grouplines)
                        {
                          *command= NOP;
                          leave= false;
                        }
                        break;
    } /* end of switch */

    if (*command==REFRESHLINE) /* updates menu-line at cursor-pos */
    {
      UpdateGroupMenu(firstgroup, group_cursory, 1, grouparray, true);
      GotoXY(2,group_cursory+2);
      if (grouparray[firstgroup+group_cursory]->Status & SELECTED_GROUP)
         InvON();
      (*ConWrite)(Menulines[group_cursory]);
      InvOFF();
      *command= NOP; /* SetGroupCursor */
    }
    else
    if (*command==REFRESHLINES)
    {
      UpdateGroupMenu(firstgroup, 0, lines, grouparray, true);
      RefreshGroupLines(firstgroup,lines,grouparray,true);
      *command= NOP;
    }

    if (*command==NOP)
    {
      SetGroupCursor();
    }
  }
  while (!leave);

  switch (*command) /* which commands terminate GROUPSELECTION? */
  {
    case QUIT:
    case BREAK:
    case REREAD_ACTIVE:
    case GOTO_GROUP:
    case SEARCH_GROUP_F:
    case SEARCH_GROUP_B:
         leave= true;
         break;
    default: leave= false;
             break;
  }

  return leave;
}


ULONG
FilterGroups(struct GroupInfoList **grouparray, UWORD unsetbits, UWORD setbits)
/* fills grouparray with groups that have <unsetbits> not set and <setbits>
   set in their Status byte. Read marks are checked to set UNREAD_GROUP
   if needed. The NEWG_GROUP flag is set accordingly, too. */
{
  ULONG i, NrOfReadArts, TotalArts;
  struct GroupInfoList *pGI;
  struct Markers *Marks;

  for (i= 0L,pGI= RootGI; pGI; pGI= pGI->NextGI)
  {
    FirstArtNr= pGI->firstArtNr; /* set for CountMarks! */
    LastArtNr = pGI->lastArtNr;  /* set for CountMarks! */

    NrOfReadArts= 0UL;

    if (LastArtNr!=0 || FirstArtNr!=0)      /* non empty group */
       TotalArts= LastArtNr - FirstArtNr+1;
    else
       TotalArts= 0UL;

    for (Marks= pGI->Marks; Marks!=NULL; Marks=Marks->NextMark)
    {
      NrOfReadArts+= CountMarks(Marks->Mark);
    }

    if (NrOfReadArts<TotalArts)
    {
       pGI->Status|= UNREAD_GROUP;
    }

    if ((pGI->Status & setbits) && !(pGI->Status & unsetbits))
    {
      grouparray[i++]= pGI;
    }
  }

  return i; /* return NrOfActGroups */
}


ULONG
FindGroupArrayNum(struct GroupInfoList **grouparray, char *groupname)
/* returns index of group that matches groupname, returns NrOfActGroups if
   no group matched */
{
  ULONG i;
  int result;

  for (i= 0L; i<NrOfActGroups && (result= stricmp(grouparray[i]->Name,groupname)); i++);

  return i;
}


int
ChooseGroups(int *choosegroups_cmd)
/* choosegroups_cmd is a command that is returned by choosegroups.
   REREAD_ACTIVE e.g. calls RereadActive and then enters ChooseGroups()
   once again */
{
  int lines, command,
      oldgrouplines;  /* to recognize changes by ReadGroup() caused by ToggleScreenTitle */

  UWORD setbits= (UNREAD_GROUP | NEW_GROUP), unsetbits= UNSUBSCRIBED;
  ULONG i, firstgroup;
  struct GroupInfoList **grouparray, *pGI;
  char *c;
  BOOL leave= false;
  ULONG index;

  command= NOP;
  group_cursory= 0;
  max_grouplines= MaxScrLines>=(MAX_SELECT_LINES+2) ? MAX_SELECT_LINES : MaxScrLines-2;
  grouplines= oldgrouplines= max_grouplines;
  max_groupnamelen= 45;
  /* max_groupnamelen= MaxChars-30-1; */
  /* template: "%c %c %-45s %c%c %6lu/%6lu/%6lu\n" */
  sprintf(tmp,"%%c %%c %%-%ds %%c%%c %%6lu/%%6lu/%%6lu\n", max_groupnamelen);
  GroupMenuTemplate= CopyString(tmp);
  grouparray= AllocMem((ULONG) NrOfGroups*sizeof(struct GroupInfoList *), MEMTYPE);
  /* allocate MenuLines en bloc */
  c= AllocMem((ULONG) (max_grouplines * ((ULONG)MaxChars+1UL)), MEMTYPE);

  if (c && grouparray && GroupMenuTemplate)
  {
    for (i= 0; i<grouplines; i++) /* set pointers to Menulines */
       Menulines[i]= &c[i*(MaxChars+1)];

    /* Fill GroupArray with NEW or UNREAD groups */
    NrOfActGroups= FilterGroups(grouparray, unsetbits, setbits); /* Nr of articles that we should display */
    if (NrOfActGroups==0L)
    {
      setbits= LASTBIT;
      /* Show all subscribed groups, because LASTBIT is always set! */
      NrOfActGroups= FilterGroups(grouparray, unsetbits, setbits);
      if (NrOfActGroups==0L)
      {
        unsetbits= SUBSCRIBED;
        /* Show all unsubscribed groups too, because LASTBIT is always set and
            SUBSCRIBED then 0! */
        NrOfActGroups= FilterGroups(grouparray, unsetbits, setbits);
      }
    }

    lines= grouplines;
    firstgroup= 0L;

    if (switches & SW_GOTOGROUP)  /* -g flag GOTO GROUP */
    {
      *tmp= '\0';
      StatusLine();
      DispMsg("Newsgroup:"); Getline(tmp,256);
      ClrLine();
      strcpy(cmdline,"g ");
      strcat(cmdline,tmp);
      command= GOTO_GROUP;
    }

    while (command!=QUIT && command!=BREAK)
    {
      if (command==GOTO_GROUP)  /* Goto Newsgroup */
      {
        struct GroupInfoList *tmppGI;
        char *cmdlp;
        BOOL refresh_array= false;

        cmdlp= skpwspc(cmdline);
        cmdlp= NextSpc(cmdlp);
        cmdlp= skpwspc(cmdlp);

        if (*cmdlp!='\0')
        {
          tmppGI= FindGroup(cmdlp);
          if (tmppGI)
          {
            strcpy(newsgroup, cmdlp);
            pGI= tmppGI;
            if (pGI->Status & UNSUBSCRIBED)
            {
              sprintf(tmp, "%s currently unsubscribed. Do you want to resubscribe? (y/n):", newsgroup);
              StatusLine();
              DispMsg(tmp);
              tmp[0]= GetUpKey();
              ClrLine();
              if (tmp[0]=='Y')
              {
                pGI->Status&= ~UNSUBSCRIBED;
                DispMsg("Resubscribed.");
                Delay(5);
                refresh_array= true;
              }
              else
              {
                DispMsg("No change. Still unsubscribed.");
                Delay(5);
                command= NOP; /* nevertheless scannewsgroup */
              }
            }

            if (refresh_array)
            {
              NrOfActGroups= FilterGroups(grouparray, unsetbits, setbits);
            }
          } /* if group was found command is still GOTO_GROUP */
          else
          {
            DispMsg("No such newsgroup!\n");
            command= NOP;    /* not found? stay! */
          }
        }
        else
          command= NOP;

        leave= false;
      }  /* goto newsgroup */

      while (!leave) /********* WHILE *********/
      {
        if (NrOfActGroups==0L)
        {
          DispMsg("There is currently no group subscribed!");
        }
        lines= NrOfActGroups-firstgroup;
        if (lines>grouplines) lines= grouplines;
        else
        if (group_cursory>=lines) group_cursory= lines/2;

        switch(command)
        {
          case PAGEFORWARD:
          case PAGEBACK:
          case SPACE:
          case LASTPAGE:
          case FIRSTPAGE: group_cursory= lines/2;
                          break;
        }

        if (command==GOTO_GROUP || command==CURSOR_TO_GROUP)
        {
          index= FindGroupArrayNum(grouparray, newsgroup);
          if (index==NrOfActGroups) /* not found! */
          {
            setbits= LASTBIT;
            NrOfActGroups= FilterGroups(grouparray, unsetbits, setbits);
            index= FindGroupArrayNum(grouparray, newsgroup);
          }
          if (index==NrOfActGroups) /* group not found */
          {
            StatusLine();
            DispMsg("Group not found!");
            Delay(50L);
            command= NOP;
          }
          else
          {
            /* set page offset and cursor position */
            firstgroup= (index/grouplines)*grouplines;
            group_cursory= index-firstgroup;
            if (command!=CURSOR_TO_GROUP)
               command= ENTERGROUP;  /* directly enter the group */
            else
               command= NOP;
            leave= false;
          }
        }
        else /* Normally Select Groups */
          leave= SelectGroups(firstgroup, lines, grouparray, &command); /*** ### ***/
          /********************* main call *********************/

        if (!leave) /********** If !LEAVE **************/
        {
          if (command==SHOWALL || command==SHOW_UNSUBS || command==REBUILD)
          {
            if (command==SHOWALL)
            {
               unsetbits= UNSUBSCRIBED;

               if (setbits==LASTBIT)
                  setbits= (UNREAD_GROUP);
               else
                  setbits= LASTBIT; /* This bit should be always set, so we get all groups */
            }
            else
            if (command==SHOW_UNSUBS)
            {
              setbits= LASTBIT;
              unsetbits= SUBSCRIBED; /* show all unsubscribed groups */
            }

            NrOfActGroups=FilterGroups(grouparray, unsetbits, setbits);

            if (NrOfActGroups==0L) /* No new or unread groups */
            {
              if (setbits==LASTBIT)     /* No subscribed groups */
                 unsetbits= SUBSCRIBED; /* show all unsubscribed groups */
              else
                 setbits= LASTBIT;
              /* Show all subscribed groups, because LASTBIT is always set! */
              NrOfActGroups= FilterGroups(grouparray, unsetbits, setbits);
            }

            if (firstgroup>NrOfActGroups)
               command= FIRSTPAGE;
            else
               command= REFRESHSCREEN;
          }

          if (command==PAGEBACK)
          {
            if (firstgroup>=grouplines) firstgroup-= grouplines;
            else
               firstgroup= 0UL;
          }
          else
          if (command==PAGEFORWARD || command==SPACE)
          {
            if (firstgroup+lines<NrOfActGroups) firstgroup+= lines;
            else
            {
              if (command==SPACE)
                 command= READNOW;
            }
          }
          else
          if (command==LASTPAGE)
          {
            firstgroup= ((NrOfActGroups-1)/grouplines)*grouplines;
          }
          else
          if (command==FIRSTPAGE)
          {
            firstgroup= 0UL;
          }
          else
          if (command==ENTERGROUP)
          { /**** <== ****/
            oldgrouplines= grouplines;
            leave= ReadGroup(grouparray[firstgroup+group_cursory], &command);
          }
          else
          if (command==LISTGROUPS)
          {
            ListGroups();
            command=REFRESHSCREEN;
          }

          if (command==READNOW)
          {
            i= 0UL;
            while (!leave && i<NrOfActGroups)
            {
              if (grouparray[i]->Status & SELECTED_GROUP)
              {
#ifdef DEBUG
           printf("%s\n",grouparray[i]->Name);
#endif
                oldgrouplines= grouplines;
                leave= ReadGroup(grouparray[i], &command); /**** <== ****/

                if (command==QUIT)
                {
                  leave= false;
                  command= NEXTGROUP;
                } /* QUIT returned by READGROUP doesn't mean to quit Arn!*/

                if (command==TOGROUPDIR) /* exit while */
                   i= NrOfActGroups;
              }

              if (command==PREVGROUP)
              {
                 if (i>0L) i--;
                 else
                   command= NOP;
              }
              else
              if (command==NEXTGROUP || !leave) i++;
            }

          /* for (i=0; i<NrOfActGroups; i++)
                grouparray[i]->Status&= ~SELECTED_ART; */
          }
          else
          if (command==REFRESHSCREEN)
          {
            command= NOP;
          }
        } /********** If !LEAVE **************/

        if (command==SEARCH_GROUP_F || command==SEARCH_GROUP_B)
        {
          struct GroupInfoList *tmppGI;
          int patlength= strlen(cmdline+1);

          casedep= 0x5F;

          pGI= grouparray[firstgroup+group_cursory];

          if (searchpat.patlen && patlength) /* new pattern given -> remove old ones */
          {
            if (searchpat.kmp_table)
            {
              free(searchpat.pattern.kmppat);
              free(searchpat.kmp_table);
              searchpat.kmp_table= NULL;
              searchpat.pattern.kmppat= NULL;
            }
            else
            {
              if (searchpat.pattern.regpat) free(searchpat.pattern.regpat);
              searchpat.pattern.regpat= NULL;
            }
            searchpat.patlen= 0;
          }

          if (patlength) /* new pattern given */
          {
            searchpat.kmp_table= malloc(patlength*sizeof(short int));
            KMP_ComputeNextTable(cmdline+1,patlength,searchpat.kmp_table);
            searchpat.pattern.kmppat= malloc(patlength+1);
            strcpy(searchpat.pattern.kmppat,cmdline+1);
            searchpat.patlen= patlength;
            pGI=RootGI;
          }
          else /* simply repeat the search */
            patlength= searchpat.patlen;


          if (command==SEARCH_GROUP_F)        /* search forward */
            for (tmppGI= pGI->NextGI; tmppGI!=NULL &&
                 KMP_Matcher(tmppGI->Name,strlen(tmppGI->Name),
                             searchpat.pattern.kmppat,patlength,
                             searchpat.kmp_table)==-1; tmppGI= tmppGI->NextGI);
          else                                /* search backwards */
            for (tmppGI= pGI->PrevGI; tmppGI!=NULL &&
                 KMP_Matcher(tmppGI->Name,strlen(tmppGI->Name),
                             searchpat.pattern.kmppat,patlength,
                             searchpat.kmp_table)==-1; tmppGI= tmppGI->PrevGI);

          if (tmppGI) /* found! */
          {
            pGI= tmppGI;
            strcpy(newsgroup,pGI->Name);
            command= CURSOR_TO_GROUP;
          }
          else /* not found */
          {
            DispMsg("No such newsgroup!");
            Delay(50L);
            command= NOP;   /* not found? stay! */
          }

          leave= false;
        } /* SEARCH */

        if (command==TOGGLE_SCREENTITLE || grouplines!=oldgrouplines)
        { /* we must be careful, because if at the last group we must
             go to the next page and reset the cursor */
          index= group_cursory+firstgroup; /* store old cursor position */
          if (command==TOGGLE_SCREENTITLE) ToggleScreenTitle();
          grouplines= oldgrouplines= MaxScrLines-2;
          firstgroup= (index/grouplines)*grouplines; /* calculate new page */
          group_cursory= index-firstgroup;           /* set new cursor position */
          command= REFRESHSCREEN;
        }
      } /* while !leave */

      switch (command)
      {
        case REREAD_ACTIVE:
                                 *choosegroups_cmd= command;
                                 command= QUIT; /* leave choosegroups and enter once again */
                                 break;
      }
    } /* while !QUIT */
  }
  else
    ErrorHandle(NOMEM,NULL);

  if (grouparray) FreeMem (grouparray, (ULONG) NrOfGroups*sizeof(struct GroupInfoList *));
  if (c) FreeMem(c, (ULONG) (max_grouplines * ((ULONG)MaxChars+1UL)));
  FreeString(GroupMenuTemplate);

  return GlobalERROR;
}
