/***********************************************************************
 * functions for command-line handling/parsing                         *
 ***********************************************************************
 * Arn - (C)opyright by Roland Bless, All rights reserved,             *
 * -------- YOU ARE NOT ALLOWED TO DISTRIBUTE A CHANGED SOURCE ------- *
 ***********************************************************************/

#ifndef GLOBALVARS_H
#include "globalvars.h"
#endif

#ifndef ENUMS_H
 #include "enums.h"
#endif

/* Prototypes for functions defined in cmdline.c */
/* proto$start */
BOOL InterpretCmdLine(char *cmdline,
                      int *command,
                      int depth);
int EnterCmdLine(char key,
                 int depth);
/* proto$end */
#include "proto/ArnIO_proto.h"
#include "proto/functions_proto.h"
#include "proto/strprocs_proto.h"

/*************************/


BOOL InterpretCmdLine(char *cmdline, int *command, int depth)
{
  char *c;
  BOOL leave= false;

  SubjectOrder=false;

  if (cmdline)
     cmdline= skpwspc(cmdline);  /* SkipWhiteSpaces */

  if (cmdline==NULL || cmdline[0]=='\0')   /* Nothing? ->NOP */
  {
    *command= NOP;
    return false;
  }
  c= cmdline;
  *command=NOP;

  if (isdigit(*c) || *c=='^' || *c=='.' || *c=='$')   /* Ranges are first */
  {
    *command= GOTO;
    c= NextSpc(c);
    c= skpwspc(c);
  }

  if (*c=='o')
  {
    c++;
    SubjectOrder=true;
  }

  switch (*c)
  {
    case 's':
    case 'S':
    case '|': *command= SAVEART_WH;
              break;
    case 'w':
    case 'W': *command= SAVEART_NH;
              break;
    case '?': if (depth==GROUPSEL)
                 *command= SEARCH_GROUP_B;
              else
                 *command= SEARCH_INART_B;
              break;
    case '/': if (depth==GROUPSEL)
                 *command= SEARCH_GROUP_F;
              else
                 *command= SEARCH_INART_F;
              break;
    case 'G': if (depth==GROUPSEL)
               *command= GOTO_GROUP;
    case 'g': if (depth==PAGER)
                *command= SEARCH_PAT;
              break;
    case 'j':
    case 'r': *command= JUNK;
              break;
    case '=': *command= LIST_SUBJ;
              break;
    case 'q': *command= QUIT;
              break;
    case 'U':
    case 'u': *command= UNREADART;
              break;
    case 'm': *command= SELECT;
              break;
    case 'M': *command= UNSELECT;
    case 'c': *command= CATCHUP;
              break;
    case 'D': ChangeRotNumber(c);
              *command= NOP;
              break;
     default: if (*command!=GOTO)
                 *command= NOP;
              break;
  }

  if (*++c!='\0' && *c==';' && *++c!='\0')
     nextcmd= c;
  else
     nextcmd= NULL;

  if (depth==PAGER)
  {
    switch (*command)
    {
      case           QUIT:
      case     SAVEART_WH:
      case     SAVEART_NH:
      case      LIST_SUBJ:
      case         SELECT:
      case       UNSELECT:
      case           JUNK:
      case      UNREADART:
      case           GOTO:
      case SEARCH_INART_F:
      case SEARCH_INART_B:
                          leave= true;   /* leave pager */
                          break;
       default:
                leave= false;
                break;
    }
  }
  else
  if (depth==ARTSEL)
  {
    switch (*command)
    {
      case           QUIT:
      case     SAVEART_WH:
      case     SAVEART_NH:
      case      LIST_SUBJ:
      case SEARCH_INART_F:
      case SEARCH_INART_B:
      case         SELECT:
      case       UNSELECT:
      case           JUNK:
      case      UNREADART:
      case           GOTO:
                          leave= true;   /* leave pager */
                          break;
       default:
                leave= false;
                break;
    }
  }
  else

  if (depth==GROUPSEL)
  {
    switch (*command)
    {
      case       QUIT:
      case GOTO_GROUP:
      case SEARCH_GROUP_F:
      case SEARCH_GROUP_B:
                      leave= true;   /* leave this group-selection page */
                      break;
       default:
                leave= false;
                break;
    }
  }

  return leave;
}


int EnterCmdLine(char key, int depth)
{
  if (depth==GROUPSEL) /* These keys cause entering of the command-line-modus */
  switch (key)         /* If you're in the groupselector! */
  {
    case 'G':
    case '/':
    case '?':
    case ctrl('Z'):
              break;
    default: return 0;
             break;
  }
  else
  if (depth==ARTSEL)
  switch (key)
  {
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
    case '0':
    case 's':
    case 'w':
    case 'S':
    case 'W':
    case 'o':
    case '|':
    case '/':
    case '?':
    case ctrl('Z'):
              break;
    default: return 0;
             break;
  }
  else
  if (depth==PAGER)
  switch (key)
  {
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
    case '0':
    case 'g':
    case 's':
    case 'w':
    case 'S':
    case 'W':
    case '|':
    case '/':
    case '?':
    case 'D':
    case ctrl('Z'):
              break;
    default: return 0;
             break;
  }


  if (key!=ctrl('Z'))
  {
    cmdline[0]=key;
    if (!isdigit(key) && key!='/' && key!='?' && key!='.' && key!='o')
    {
      cmdline[1]=' ';
      cmdline[2]='\0';
    }
    else
      cmdline[1]='\0';
  } /* ^Z prints last cmdline */
  if (depth!=PAGER) StatusLine();
  ClrLine();
  DispMsg("Command: ");
  Getline(cmdline,MAXLINE);
  ClrLine();

  return 1;
}
