/************************************************************************
 * followup.c -- followup functions and other article goodies for arn   *
 *                                                                      *
 * Arn - (C)opyright by Roland Bless, All rights reserved,              *
 * -------- YOU ARE NOT ALLOWED TO DISTRIBUTE A CHANGED SOURCE -------- *
 ************************************************************************/

#ifndef GLOBALVARS_H
#include "globalvars.h"
#endif

#include <time.h>
#include <limits.h>
#include <dos.h>
#ifndef ENUMS_H
 #include "enums.h"
#endif

#define MAXUNIQUENUM 14
#define CANCELLED_ARTNAME "T:CancelledArt"
#define POSTING  1
#define FOLLOWUP 2
#define CANCEL   3

/* Prototypes for functions defined in followup.c */
/* proto$start */
char *mytmp(char *mask);
char *uniquenum(char *s,
                long time,
                int seq);
void RenameTmpFile(char *tmpname);
char *GetBufField(char *s);
char *GetBufField_nwspc(char *s);
void splitline(FILE *fp,
               char *line);
void AppendHeader(FILE *fp);
BOOL AppendSignature(char *fname);
void CheckHeader(char *oldartname,
                 char *newartname);
int EditNewsArticle(int seq,
                    char *ArtName,
                    int type);
void StandardHeader(FILE *Article);
char *Sender(char *s);
char *GetInitials(void);
BOOL MailToModerator(ULONG ArticleNr);
char *SearchModerator(char *newsgroup);
BOOL OneGroupKnown(char *newsgroups);
int FollowUp(struct art_struct *pArt,
             char *quotestring,
             BOOL quote_header);
int PostArticle(char *newsgroup);
int Cancel(char *newsgroup,
           FILE *fp);
void DoFollowUp(struct art_struct *pArt,
                int rcommand);
/* proto$end */

                /*** external functions used ***/
#include "proto/ArnIO_proto.h"
#include "proto/functions_proto.h"
#include "proto/kill_proto.h"
#include "proto/pager_proto.h"
#include "proto/time_proto.h"
#include "proto/reply_proto.h"
#include "proto/strprocs_proto.h"
#include "proto/grouplist_proto.h"
#ifdef LATTICE_50
#include "proto/raw_proto.h"
#endif

                              /****/

extern char *pend, *actual_pos;
extern char *Version;
extern struct kill_field *LocalKill_List; /* readgroup.c */
extern char *pathtogroup;
extern char *waitmsg;
extern char *mimefile; /* pager.c */

                  /********************************/

char uniquenumber[MAXUNIQUENUM];
int pathtype=2;
char *warn_moderated_group= "\n** WARNING: This group is moderated! **\n\
Do not post to it, but send a mail to its moderator (try 'r')\n\
or change the Newsgroups:-line!\n";


                  /********************************/

char *
mytmp(char *mask)
{
  char tmpname[9], *t, *orgmask= mask;

  sprintf(tmpname,"%08lx",(long) FindTask(NULL));
  for (t=tmpname; *mask!='\0'; mask++)
  {
    if (*mask=='X')
    {
      *mask=toupper(*t);
      t++;
    }
    else
    if (*mask=='x') *mask=*t++;

    if (t>=&tmpname[9]) t=tmpname;
  }
  return orgmask;
}


char *
uniquenum(char *s, long time, int seq)
{
  int len;

  ToBase(time,s,64);
  len=strlen(s);
  s[len]='*';len++;
  ToBase((unsigned long) seq,s+len,64);
  seq=strlen(s);
  if (seq-len <= 2) /* pad with 0 */
  {
    while (seq<len+3) s[seq++]='0';
    s[seq]='\0';
  }
  s[MAXUNIQUENUM]='\0'; /* Just to be sure */
  return s;
}

void
RenameTmpFile(char *tmpname)
{
  char newtmp[40];

  sprintf(newtmp,"%s.%s",CANCELLED_ARTNAME,UserName);
  DeleteFile(newtmp);
  Rename(tmpname,newtmp);
}


/* returns pointer to string in HeaderBuf, which begins with s.
   The returned string is null terminated at the first "\n" which
   is not followed by a whitespace (line continuation).
   Because one character is always overwritten by the terminating \0 byte,
   it has to be restored. This is done automatically at each call, but
   the last time _you have to call_ GetBufField(NULL) which restores the last
   overwritten byte.
*/
char *
GetBufField(char *s)
{
  char *c,*r= NULL;
  static char mem='\0';  /* store for overwritten byte */

  if (pend)      /* pointer to end */
     *pend=mem;  /* restore overwritten byte */

  if (s==NULL)   /* "last" call restores overwritten byte */
  {
    pend=NULL;
    return NULL; /* restore Buffer */
  }

  c= HeaderBuf; /* search from start of HeaderBuf */
  pend=NULL;
  while (*c!='\0')
  {
    if (r= firststr(c,s))
    {
      c= hlineend(c);
      if (*c=='\n' || *c=='\0')
      {
        if (*c=='\0' && c>r) c--;
        pend= c;
        mem=*c;
        *c='\0';
        break;
      }
      else
      {
        ErrorHandle(STRANGE,"getbuffield");
        break;
      }
    }
    else
    {
      c= hlineend(c);
      if (*c!='\0')
         c++;
    }
  }
  return r;
}


char *
GetBufField_nwspc(char *s)
/* Gets next non white space string after field s from HeaderBuf */
{
  char *c;

  c= GetBufField(s);
  return skpwspc(c);
}


void
splitline(FILE *fp, char *line)
{
  long count= 12L, linebreak= 75L;               /* References: */
  char *lastmsgid= line, ctmp;

  lastmsgid= line= skpwspc(line);

  while (*line!='\0')
  {
    count+= ((line= NextWhiteSpc(line))-lastmsgid);

    ctmp= *line; *line='\0';
    if (count>=linebreak)
    {
      fprintf(fp, " %s\n", lastmsgid);
      count=0L;
    }
    else
    {
      fprintf(fp, " %s", lastmsgid);
     count++;
    }
    *line=ctmp;
    lastmsgid= line=skpwspc(line);
  }
}


void
AppendHeader(FILE *fp)
{
  FILE *headerfp;

  if (headerfp=fopen(headerfile,"r"))
  {
    while (fgets(tmp, SUPERLINE, headerfp)) fprintf(fp, "%s", tmp);
    fclose(headerfp);
  }
}


BOOL
AppendSignature(char *fname)
{
  FILE *fpart, *fpsign;

  if ((fpart=fopen(fname,"a"))==NULL)
  {
    ErrorHandle(STRANGE,"sig");
    return false;
  }
  else
  if (*signfile!='\0') /* No signature specified */
  {
    if ((fpsign= fopen(signfile,"r"))==NULL)
    {
      DispMsg("\n Signature not appended! (not found?)\n");
      fclose(fpart);
      return false;
    }
    else
    {
      while (fgets(tmp, SUPERLINE, fpsign))
            fprintf(fpart, "%s", tmp);
    }
  }
  else
   return false;

  if (fpart) fclose (fpart);
  if (fpsign) fclose(fpsign);

  return true;
}


/* Searches for blank lines and followup-field, inserts Lines: */
void
CheckHeader(char *oldartname, char *newartname)
{
  FILE *oldart,*newart;
  long fposi;
  ULONG number= 0UL;
  char *c=NULL, *newsgroup=NULL, *stmp;
  char *X_Newsreader="E-Uldzylhkly: Hyu";

  stmp= (char *) AllocMem((ULONG) SUPERLINE+1L, MEMTYPE);

  strcpy(newartname,"T:ARN.xxxx.XXXX.2");
  mytmp(newartname);

  if (oldart= fopen(oldartname,"r"))
  {
    if (newart= fopen(newartname,"w"))
    {
      while (fgets(tmp, SUPERLINE, oldart) && *tmp!='\n') /* while in header */
      {
        if (!newsgroup) /* If we didn't get the Newsgroups-line yet */
        {
          if (c= firststr(tmp,HeaderLines[HL_NEWSGROUPS]))
          {
            c=skpwspc(c);
            newsgroup=CopyString(c);
          }
        }

        if (c= firststr(tmp,HeaderLines[HL_FOLLOWUP_TO]))
        {
          if (!newsgroup && stmp) /* Didn't get the newsgroup-field yet */
          {              /* so look for it first */
            char *d;

            fposi=ftell(oldart);
            while (fgets(stmp, SUPERLINE, oldart) && *stmp!='\n')
            {
              if (d=firststr(stmp,HeaderLines[HL_NEWSGROUPS]))
              {
                d=skpwspc(d);
                newsgroup=CopyString(d);
              }
            }
            fseek(oldart,fposi,0);
          }
          c=skpwspc(c);
          if (*c) /* This line is not empty */
             if (strcmp(c,newsgroup)) /* If we have a different followup-to line */
               fprintf(newart,"%s",tmp);
        }
        else
        if (c= strchr(tmp,':'))
        {
#ifdef DEBUG
          printf("%s|",c);
#endif
          c++;
          c=skpwspc(c);
#ifdef DEBUG
          printf("->%s<",c);
#endif
          if (*c!='\0') /* If there are not only blanks after the colon */
            fprintf(newart,"%s",tmp);
        }
        else /* line continuation */
         fprintf(newart,"%s",tmp);
      }
      /* We're now at the beginning of the message-body */
      fposi=ftell(oldart);
      while (fgets(tmp, SUPERLINE, oldart)) number++;  /* count body lines */
      strcpy(tmp,X_Newsreader);
      fprintf(newart,"%s %s\n",rot(tmp,19),Version);
      fprintf(newart,"%s %lu\n\n",HeaderLines[HL_LINES],number);
      fseek(oldart,fposi,0);
      while (fgets(tmp, SUPERLINE, oldart)) fprintf(newart,"%s",tmp);
      fclose(newart);
    }
    else
     ErrorHandle(STRANGE,"notmp_chW");

    fclose(oldart);
  }
  else
   ErrorHandle(STRANGE,"notmp_chR");

  FreeString(newsgroup);
  if (stmp) FreeMem(stmp, (ULONG) SUPERLINE+1L);
}


/* Invokes Editor and asks for post,abort,edit,etc */
int
EditNewsArticle(int seq, char *ArtName, int type)
/* type is POSTING,FOLLOWUP or CANCEL */
{
  FILE *fp= NULL;
  long result= 0L, filedate=0L;
  BOOL titleset=false, wrongchar=false;
  char *SubjectTmp= NULL, *GroupTmp= NULL;

  AppendSignature(ArtName); /* preinclude the signature */

  filedate= getft(ArtName);

  if (!(GeneralBool & SCREENTITLE)) /* if title not already set */
  {
    ShowTitle(ArnScreen, TRUE);
    titleset=true;
  }

  do
  {
    DispMsg("invoking editor...");
    if (GeneralBool & EDSCRSWITCH)
       WBenchToFront();
    if (auxmode && AuxEditor)
      sprintf(tmp,"%s %s", AuxEditor, ArtName);
    else
      sprintf(tmp,"%s %s", NewsEditor, ArtName);

    /* Execute(tmp,(BPTR) InFH,(BPTR) OutFH); */
    system(tmp);

    if (auxmode) rawcon(1); /* get sure we're in raw mode */

    ScreenToFront(ArnScreen);
    ActivateWindow(ArnWin);
    DispMsg("ready.\n");

    if (getft(ArtName)==filedate)
    {
      *tmp='a'; /* abort */
      DispMsg("article not modified, aborted. ");
      break;
    }

    /* final header check */
    if (fp= fopen(ArtName,"r"))
    {
      char *tmp2;

      while (fgets(tmp,SUPERLINE,fp) && *tmp!='\n')
      {
        if (firststr(tmp,HeaderLines[HL_FOLLOWUP_TO]))
        {
          if (strchr(tmp,',')!=NULL)
             DispMsg("\n\x7** WARNING: Ambiguous FollowUp-To:-Line! Please (e)dit! **\n");
        }
        else
        if (tmp2= firststr(tmp,HeaderLines[HL_SUBJECT]))
        { /* get subject for autoselect X-files */
          /* Don't deal with such long subjects in X-files */
          cuteoln(tmp2);
          tmp2= StripSubject(skpwspc(tmp2));
          tmp2[DEFAULT_SUBJECTLEN_FOR_XFILES]= '\0';
          SubjectTmp= CopyString(tmp2); /* store subject, we might need it */
        }
        else
        if (tmp2= firststr(tmp,HeaderLines[HL_NEWSGROUPS]))
        { /* Get final newsgroup-name for AUTOSELECT */
          /* newsgroup may not be set if posting from group-level */
          if (OneGroupKnown(tmp2)==false)
          {
            DispMsg("\nWARNING: ");
            DispMsg(HeaderLines[HL_NEWSGROUPS]);
            DispMsg(" contained no known newsgroups, please check!\n");
          }
          GroupTmp= tmp2= skpwspc(tmp2);
          while (*GroupTmp && *GroupTmp!=',' && *GroupTmp!='\n') GroupTmp++;
          *GroupTmp='\0';
          GroupTmp= CopyString(tmp2); /* store subject, we might need it */
        }
      }
      fclose(fp);
    }
    else
      ErrorHandle(STRANGE,"notmp_ed");

    do
    {
      do
      {
        DispMsg("\np)ost, e)dit, f)ilter, a)bort:");
        ConClear();
        wrongchar= false;
        tmp[0]= (*GetKey)();
        ConCR();
        tmp[1]= '\0';
        if(strcspn(tmp,"pefa"))
        {
          ClrLine();
          DispMsg("Wrong input. Try again.");
          wrongchar=true;
        }
      }
      while (wrongchar);

      if (*tmp=='f')
      {
        sprintf(tmp, filter, ArtName);
        result= system(tmp);
        if (auxmode) rawcon(1); /* get sure we're in raw mode */
        *tmp='f';
      }
    }
    while (tmp[0]=='f');
  }
  while (tmp[0]=='e');

  if (tmp[0]=='p')
  {
    CheckHeader(ArtName,fnametmp);
    DeleteFile(ArtName);

    DispMsg("sending article...");
    if ((fp= fopen(seqf,"w")) != NULL)  /* increase the seq number */
    {
      if (seq<0 || seq==SHRT_MAX) seq= 0;
      else
         seq++;
#ifdef DEBUG
      printf("seq: %d %u %x\n", seq, seq, seq);
      if (seq<0) seq= 0;
#endif
      fprintf(fp,"%d",seq);
      fclose(fp);
    }
    if (NewsFeed)
       sprintf(tmp,uuxname, fnametmp, NewsFeed);
    else
       sprintf(tmp,uuxname, fnametmp);

    result= system(tmp);
    if (auxmode) rawcon(1); /* get sure we're in raw mode */

    if (rnews)
    {
      DispMsg("to ");
      DispMsg(NodeName);
      sprintf(tmp,rnews, fnametmp);
      result= system(tmp);
      if (auxmode) rawcon(1); /* get sure we're in raw mode */
    }
    DispMsg("...ready.\n");
    if (result)
    {
      DispMsg("Some trouble! Tmp-File is not deleted!\n");
      RenameTmpFile(fnametmp);
      result=1L;
    }
    else
    {
      BOOL readkill= false;

      DeleteFile(fnametmp);
      DispMsg("Your article will not be displayed immediately!\n");
      /* look if we already have this subject in the local X-file */
      if (LocalKill_List==NULL)
      {
        MakeDirName(tmp,newsdir,GroupTmp);
        sprintf(fnametmp,"%s/.X%s",tmp,UserName);
        ReadKillFile(&LocalKill_List, fnametmp);
        readkill=true;
      }
      else /* get localXfilename into fnametmp */
      {
        sprintf(fnametmp,"%s/.X%s",pathtogroup,UserName);
      }
      if ( ((type==FOLLOWUP && (GeneralBool & ASK_FOR_AUTOSELECT)) || type==POSTING)
           && SubjectTmp && !FoundSubjectInKillList(LocalKill_List, SubjectTmp))
      {
        DispMsg("Do you want to autoselect this subject? (y/N):");
        ConClear();
        tmp[0]= GetUpKey();
        ClrLine();
        if (tmp[0]=='Y')
          AppendKillFile(&LocalKill_List, fnametmp, NULL, SubjectTmp, "s");
      }
      else
        Delay(50L);

      if (readkill)
      {
        FreeKill_List(LocalKill_List);
        LocalKill_List= NULL;
      }

      result=0L;
    }
  }
  else
  {
    DispMsg("article not sent!\n");
    RenameTmpFile(ArtName);
    result=2L;
  }

  if (titleset) ShowTitle(ArnScreen, FALSE);

  FreeString(SubjectTmp);
  FreeString(GroupTmp);

  return (int) result;
}


/* Writes the constant header-part */
void
StandardHeader(FILE *Article)
{
  if (pathtype!=PATH_NONE)
  switch (pathtype)
  {
    case PATH_USER: fprintf(Article,"%s %s\n", HeaderLines[HL_PATH], UserName);
                    break;
    case PATH_DOMAIN: fprintf(Article,"%s %s%s!%s\n", HeaderLines[HL_PATH], NodeName, DomainName, UserName);
                    break;
    case PATH_FULL: fprintf(Article,"%s %s!%s\n", HeaderLines[HL_PATH], NodeName, UserName);
                    break;
  }
  fprintf(Article,"%s %s@%s%s (%s)\n", HeaderLines[HL_FROM], UserName, NodeName,
                                       DomainName, RealName);
}


/* WARNING: Sender destroys s with \0!!! (TODO) */
char *
Sender(char *s)
{
  int i=0;
  char *c;

  s= skpwspc(s);
  if ((c= NextSpc(s))=='\0')    /* No further space: foo@bar */
     return s;                  /* return foo@bar */
  else
  {
    if (*(++c)=='(')            /* That's: foo@bar (fred foobar) */
    {
      /* where's the closing brace? */
      while (c[i]!='\0' && c[i]!=')') i++;

      if (c[i]==')')
      {
        c[i]= '\0';
      }
      else
      {     /* No closing brace found! */
        cuteoln(c);
      }
      c++;
      return c;                          /* return "fred foobar" */
    }
    else                       /* Should be: Fred Foobar <foo@bar> */
    {
      while (c[i]!='\0' && c[i]!='<') i++;
      if (c[i--]=='<');
         c[i]= '\0';

      return s;                /* return: Fred Foobar */
    }
  }
}


char *
GetInitials(void)
/* returns the first two initials from a Senders name, warning: uses tmp! */
/* HeaderBuf must contain something */
{
  char *from, *c;
  static char initials[3];
  int i;

  initials[2]='\0';

  from= GetBufField_nwspc(HeaderLines[HL_FROM]);
  strcpy(tmp,from);
  GetBufField(NULL); /* restore HeaderBuf */

  for(i=0, c= Sender(tmp); c!=NULL && *c!='\0' && i<2; c= skpwspc(NextWhiteSpc(c)))
     initials[i++]=*c;

  initials[i]='\0';

  return initials;
}


/* get string with newsgroups and returns
   true - if one group is found in sortedgrouplist
   false - of none of the groups is found in sortedgrouplist */
BOOL OneGroupKnown(char *newsgroups)
{
  char *groupstart, *groupend, ctmp;
  struct GroupInfoList *pGI;

  groupend= NULL;
  for (pGI= NULL, groupstart= newsgroups; *groupstart && *groupstart!='\n';)
  {
    groupstart= skpwspc(groupstart);
    for (groupend= groupstart; *groupend && *groupend!=' '
                            && *groupend!=',' && *groupend!='\n'; groupend++);

    /* groupend points to \0 or SPACE or ','/'\n' */
    ctmp= *groupend;
    *groupend= '\0';

    if (FindGroup(groupstart)) /* search group */
    {
      /* found! */
      *groupend= ctmp;
      return true;
    }

    if (ctmp==' ') /* space before comma -> search comma */
    {
      groupstart= skpwspc(groupend);
      if (*groupstart==',') groupstart++;
    }
    else
    if (ctmp==',' || ctmp=='\n')
       groupstart= groupend+1;
    else /* end of string */
       groupstart= groupend;

    /* restore old terminating byte */
    if (groupend)
       *groupend= ctmp;
  }

  return false;
}



/* Generates FollowUps, set quotestring=NULL if short followup (no quotations) */
int
FollowUp(struct art_struct *pArt, char *quotestring, BOOL quote_header)
{
  ULONG ArticleNr= pArt->ArticleNr;
  FILE *OrigArt= NULL, *QArt=NULL;
  BOOL Error= false, OneKnownGroup;
  char *newsgrp, *subject, *refs, *distrib, *MyTmp;
  int seq;
  unsigned long time;

  pend= NULL;

  strcpy(tmp,"T:ARN1.XXXXXXXX");
  MyTmp= CopyString(mytmp(tmp));


  if (quotestring) DispMsg("Quoting article...");
  else
   DispMsg("Generating header...");

  if ((OrigArt= (FILE *)fopen(seqf,"r")) != NULL)     /* get the seq number */
  {
    fscanf(OrigArt,"%d",&seq);
    if (seq<0) seq=0;
    fclose(OrigArt);
  }
  else
  {
    ErrorHandle(SEQFNOTFOUND,NULL);
    return 1;
  }
  OrigArt=NULL;

  if (mimefile) /* we have a decoded article */
  {
    strcpy(fnametmp, mimefile);
  }
  else
  {
    MakeDirName(tmp,newsdir,newsgroup);
    sprintf(fnametmp,"%s/%lu",tmp,ArticleNr);
  }

  OrigArt= fopen(fnametmp, "r");
  if (OrigArt)
  {
#ifdef __SASC
  /* fmode(OrigArt,0);*/ /* SAS-C Mode A specifies to delete CRs at EOLs */
#endif
    while (fgets(tmp,SUPERLINE,OrigArt) && *tmp!='\n' );
    /* skip header */
  }
  else
    Error=true;

  if (!Error)
  {
    QArt= fopen(MyTmp,"w");
    if (QArt)
    {
      StandardHeader(QArt);
      newsgrp= GetBufField_nwspc(HeaderLines[HL_FOLLOWUP_TO]);
      if (newsgrp)
      {
        if (!stricmp(newsgrp,"poster"))
        {
          sprintf(tmp,"WARNING: %s was set to 'poster', maybe a reply was desired\n",
                  HeaderLines[HL_FOLLOWUP_TO]);
          DispMsg(tmp);
          newsgrp= NULL;
        }
        else /* check for existing newsgroups */
        if ((OneKnownGroup= OneGroupKnown(newsgrp))==false)
        {
          sprintf(tmp,"\nWARNING: %s contained no known newsgroups, please check!\n %s was %s\n",
                  HeaderLines[HL_FOLLOWUP_TO], HeaderLines[HL_FOLLOWUP_TO], newsgroup);
          DispMsg(tmp);
          newsgrp= NULL;
        }

        if (!newsgrp)
        {
          sprintf(tmp,"Using %s instead! Noticed?\n",HeaderLines[HL_NEWSGROUPS]);
          DispMsg(tmp);
          WaitForSpace(NULL);
        }
      }

      if (!newsgrp)
         newsgrp= GetBufField_nwspc(HeaderLines[HL_NEWSGROUPS]);
      if (newsgrp)
         fprintf(QArt,"%s %s\n", HeaderLines[HL_NEWSGROUPS], newsgrp);

      fprintf(QArt,"%s ",HeaderLines[HL_SUBJECT]);
      subject= GetBufField_nwspc(HeaderLines[HL_SUBJECT]);
      if (!subject)
        subject= GetBufField_nwspc(HeaderLines[HL_TITLE]);   /* old format */

      if (!firststr(subject, "Re:") && !firststr(subject, "Re^"))
      {
        fprintf(QArt,"Re: ");
      }
      fprintf(QArt,"%s\n", subject);
      fprintf(QArt,"%s ", HeaderLines[HL_MESSAGE_ID]);
      if (time= GetTime())
         fprintf(QArt,"<%s@%s%s>\n", uniquenum(uniquenumber, time,seq),
                           NodeName, DomainName);
      else
         fprintf(QArt,"<%04x.ARN@%s%s>\n", seq, NodeName, DomainName);

      fprintf(QArt,"%s %s\n", HeaderLines[HL_DATE], TimeStr(0));
      if (replyto)
         fprintf(QArt,"%s %s\n", HeaderLines[HL_REPLY_TO], replyto);

      newsgrp= GetBufField_nwspc(HeaderLines[HL_FOLLOWUP_TO]); /* sorry, but once again! */
      if (!newsgrp || !OneKnownGroup)
         newsgrp= GetBufField_nwspc(HeaderLines[HL_NEWSGROUPS]);
      fprintf(QArt,"%s %s\n", HeaderLines[HL_FOLLOWUP_TO], newsgrp);
      fprintf(QArt,HeaderLines[HL_REFERENCES]);
      refs= GetBufField_nwspc(HeaderLines[HL_REFERENCES]); /* If refs-line exists get it ...*/
      if (refs)
      {
        splitline(QArt, refs); /* splitline outputs it */
      }

      refs= GetBufField_nwspc(HeaderLines[HL_MESSAGE_ID]);  /* put msg-id on refs-line */
      if (!refs)
        refs= GetBufField_nwspc(HeaderLines[HL_ARTICLE_ID]);  /* old format */
      if (refs)
         fprintf(QArt, " %s", refs);

      fprintf(QArt,"\n%s \n%s \n%s \n", HeaderLines[HL_KEYWORDS],
                                        HeaderLines[HL_SUMMARY],
                                        HeaderLines[HL_EXPIRES]);
      fprintf(QArt,"%s ", HeaderLines[HL_DISTRIBUTION]);
      distrib= GetBufField_nwspc(HeaderLines[HL_DISTRIBUTION]);
      if (distrib)
         fprintf(QArt,"%s",distrib);
      fprintf(QArt,"\n");
      if (OrgaName)
         fprintf(QArt,"%s %s\n", HeaderLines[HL_ORGANIZATION], OrgaName);

      if (headerfile)
         AppendHeader(QArt);    /* include the headerfile */

      fprintf(QArt,"\n"); /* separate header from the body */


      if (quotestring)
      {
        refs= GetBufField_nwspc(HeaderLines[HL_MESSAGE_ID]);
        if (!refs)
           refs= GetBufField_nwspc(HeaderLines[HL_ARTICLE_ID]);

        fprintf(QArt,"In article %s, ", refs);

        refs= GetBufField_nwspc(HeaderLines[HL_FROM]);

        if (refs)
        {
          strcpy(tmp, refs);
          fprintf(QArt,"%s writes:\n\n",Sender(tmp));
        }

        if (quote_header)
        {
          char *headerline;

          GetBufField(NULL); /* restore Headerbuffer */
          actual_pos=HeaderBuf;
          while (GetNextHeaderLine(&headerline))
            fprintf(QArt,"%s%s\n", quotechar, headerline);
          fprintf(QArt,"\n");
        }
        while (fgets(tmp,SUPERLINE,OrigArt))
           fprintf(QArt,"%s%s",quotestring,tmp);
      }
      fclose (QArt);
    }
  }
  if (OrigArt) fclose(OrigArt);

  GetBufField_nwspc(NULL); /* restore Buffer */
  pend=NULL;

  EditNewsArticle(seq, MyTmp, FOLLOWUP);

  FreeString(MyTmp);
}


int
PostArticle(char *newsgroup)
{
  FILE *PArt=NULL;
  char *newsgrp= NULL, *subject= NULL, *distrib= NULL, *MyTmp;
  int seq;
  long time;

  if (GroupModerated(newsgroup))
     if (!MailToModerator(0))
     {
       DispMsg("Nevertheless post? (y/N):");
       if (GetUpKey()!='Y')
          return 1;
     }
     else
      return 0;

  strcpy(tmp,"T:ARN1.XXXXXXXX");

  MyTmp= CopyString(mytmp(tmp));

  DispMsg("Post a new article to the net:\n");

  if (!newsgroup)
  {
    *tmp='\0';
    do
    {
      ClrLine();
      DispMsg("Newsgroup(s):");
      Getline(tmp, MAXLINE);
    }
    while(*tmp=='\0');
    ConCR();
    newsgrp=CopyString(tmp);
    newsgroup=newsgrp;
    if (GroupModerated(newsgroup))
    {
      DispMsg(warn_moderated_group);
      WaitForSpace(waitmsg);
      return 1;
    }
  }

  *tmp='\0';
  DispMsg(HeaderLines[HL_SUBJECT]);
  Getline(tmp,MAXLINE);
  ConCR();
  subject= CopyString(tmp);

  *tmp='\0';
  DispMsg(HeaderLines[HL_DISTRIBUTION]);
  Getline(tmp,MAXLINE);
  ConCR();
  distrib= CopyString(tmp);

  if ((PArt= (FILE *)fopen(seqf,"r")) != NULL)     /* get the seq number */
  {
    fscanf(PArt,"%d",&seq);
    if (seq<0) seq=0;
    fclose(PArt);
  }
  else
  {
    ErrorHandle(SEQFNOTFOUND,NULL);
    return 1;
  }

  PArt= fopen(MyTmp, "w");
  if (PArt)
  {
    StandardHeader(PArt);
    fprintf(PArt,"%s %s\n", HeaderLines[HL_NEWSGROUPS], newsgroup);
    fprintf(PArt,"%s %s\n", HeaderLines[HL_SUBJECT], subject);
    fprintf(PArt,"%s ",HeaderLines[HL_MESSAGE_ID]);
    if (time= GetTime())
       fprintf(PArt,"<%s@%s%s>\n", uniquenum(uniquenumber,time,seq),
                         NodeName, DomainName);
    else
       fprintf(PArt,"<%04x.ARN@%s%s>\n", seq, NodeName, DomainName);

    fprintf(PArt,"%s %s\n", HeaderLines[HL_DATE], TimeStr(0));
    if (replyto)
         fprintf(PArt,"%s %s\n", HeaderLines[HL_REPLY_TO], replyto);

    fprintf(PArt,"%s %s\n", HeaderLines[HL_FOLLOWUP_TO], newsgroup);
    fprintf(PArt,"%s \n%s \n%s \n", HeaderLines[HL_KEYWORDS],
                                    HeaderLines[HL_SUMMARY],
                                    HeaderLines[HL_EXPIRES]);

    fprintf(PArt,"%s %s\n", HeaderLines[HL_DISTRIBUTION], distrib);
    if (OrgaName)
       fprintf(PArt,"%s %s\n", HeaderLines[HL_ORGANIZATION], OrgaName);

    if (headerfile)
       AppendHeader(PArt);    /* include the headerfile */

    fprintf(PArt,"\n");

    fclose (PArt);
  }

  FreeString(newsgrp);
  FreeString(subject);
  FreeString(distrib);

  EditNewsArticle(seq, MyTmp, POSTING);

  FreeString(MyTmp);

  return 0;
}




int
Cancel(char *newsgroup, FILE *fp)
{
  char *MsgID= NULL,*fromline= NULL, *distrib= NULL, *MyTmp= NULL;
  FILE *PArt=NULL;
  int seq, rc=0;
  long time;

  if ((PArt= (FILE *)fopen(seqf,"r")) != NULL)     /* get the seq number */
  {
    fscanf(PArt,"%d",&seq);
    if (seq<0) seq=0;
    fclose(PArt);
  }
  else
  {
    ErrorHandle(SEQFNOTFOUND,NULL);
    return 2;
  }

  pend=NULL;
  fromline= GetBufField_nwspc(HeaderLines[HL_FROM]);

  sprintf(tmp,"%s@%s%s",UserName,NodeName,DomainName);

  if (strnicmp(fromline,tmp, strlen(tmp))) /* Check for author */
  {
    DispMsg("SORRY, YOU ARE NOT THE AUTHOR of this article! Won't cancel this!\n");
    rc=1;
  }
  else
  {
    strcpy(tmp,"T:ARN.XXXXXXX.C");
    MyTmp= CopyString(mytmp(tmp));

    PArt= fopen(MyTmp, "w");
    if (PArt)
    {
      StandardHeader(PArt);
      fprintf(PArt,"%s %s\n", HeaderLines[HL_NEWSGROUPS], newsgroup);

      MsgID=GetBufField_nwspc(HeaderLines[HL_MESSAGE_ID]);
      if (!MsgID)
         MsgID=GetBufField_nwspc(HeaderLines[HL_ARTICLE_ID]);

      fprintf(PArt,"%s cancel %s\n", HeaderLines[HL_SUBJECT], MsgID);

      fprintf(PArt,"%s ",HeaderLines[HL_MESSAGE_ID]);
      if (time= GetTime())
         fprintf(PArt,"<%s@%s%s>\n", uniquenum(uniquenumber,time,seq),
                           NodeName, DomainName);
      else
         fprintf(PArt,"<%04x.ARN@%s%s>\n", seq, NodeName, DomainName);

      fprintf(PArt,"%s %s\n", HeaderLines[HL_DATE], TimeStr(0));

      fprintf(PArt,"%s cancel %s\n", HeaderLines[HL_CONTROL], MsgID);

      if (replyto)
         fprintf(PArt,"%s %s\n", HeaderLines[HL_REPLY_TO], replyto);

      fprintf(PArt,"%s %s\n", HeaderLines[HL_FOLLOWUP_TO], newsgroup);
      fprintf(PArt,"%s \n", HeaderLines[HL_EXPIRES]);
      fprintf(PArt,"%s \n", HeaderLines[HL_KEYWORDS]);

      distrib=GetBufField_nwspc(HeaderLines[HL_DISTRIBUTION]);

      fprintf(PArt,"%s %s\n",HeaderLines[HL_DISTRIBUTION], distrib);
      if (OrgaName)
         fprintf(PArt,"%s %s\n", HeaderLines[HL_ORGANIZATION], OrgaName);

      if (headerfile)
         AppendHeader(PArt);    /* include the headerfile */

      fprintf(PArt,"\n");

      MsgID=GetBufField_nwspc(HeaderLines[HL_MESSAGE_ID]);
      if (!MsgID)
         MsgID=GetBufField_nwspc(HeaderLines[HL_ARTICLE_ID]);

      if (MsgID)
         fprintf(PArt,"cancel %s\n\n", MsgID);
      fprintf(PArt,"This message was cancelled from within Arn.\n");

      fclose (PArt);
      fclose (fp);   /* Close Article to allow deletion! */

      if (EditNewsArticle(seq, MyTmp, CANCEL)) rc=1; /* Abort or error */
    }
  }
  FreeString(MyTmp);
  GetBufField_nwspc(NULL);

  return rc;
}


char *
SearchModerator(char *newsgroup)
/* searches in moderators-file for matching moderator e-mail address.
   "all" is a default entry which is taken if no matching entry was found.
   all should contain a %s for the newsgroup name (which has dashes instead of dots).
   example:
   all:       %s@uunet.uu.net
   the moderators-file should have the format:
   groupname  e-mail-address
   groupname is either the groupname or the groupname with dashes instead of dots.
   any trailing stuff after newsgroup is ignored, so the moderators file
   should be sorted, because of partial matches.
*/

{
  FILE *mod_fp;
  char *tmp_group, *c, *default_moderator= NULL;

  if (Moderators==NULL) return NULL;
  if (tmp_group= CopyString(newsgroup)) /* convert dots to dashes */
  {
     for(c= tmp_group; *c!='\0'; c++)
        if(*c=='.') *c= '-';
  }
  else
     return NULL;

  c= NULL;
  if (mod_fp= fopen(Moderators, "r"))
  {
    while(fgets(tmp, SUPERLINE, mod_fp))
    {
      if (c= firststr(tmp, tmp_group)) /* look for "news-group-with-dashes" */
      {
        c= skpwspc(NextWhiteSpc(c)); /* get second word */
        break;
      }
      else
      if (c= firststr(tmp, newsgroup)) /* look for "news.group.with.dots" */
      {
        c= skpwspc(NextWhiteSpc(c));
        break;
      }
      else
      if (c= firststr(tmp, "all")) /* get default entry */
      {
        c= skpwspc(NextWhiteSpc(c));
        if (default_moderator==NULL)
           default_moderator= CopyString(c);
        c= NULL;
      }
    }
    fclose(mod_fp);
  }

  if (c==NULL && default_moderator) /* only default_mod found */
  {
    sprintf(tmp, default_moderator, tmp_group);
    c= tmp;
  }
  FreeString(default_moderator);
  FreeString(tmp_group);

  cuteoln(c);

  return c; /* note that a pointer to tmp is returned! */
}


BOOL
MailToModerator(ULONG ArticleNr)
/* returns false if not mailed or no moderator exists */
{
  char key, *Moderator;

  DispMsg(warn_moderated_group);
  Moderator= CopyString(SearchModerator(newsgroup));
  if (Moderator)
  {
    DispMsg("Moderators e-mail address is: ");
    DispMsg(Moderator);
    DispMsg("\nSend mail to this address? (Y/n):");
    key= GetUpKey();
    ClrLine();
    if (key!='N')
    {
      if (ArticleNr) /* followup */
         Reply(ArticleNr, Moderator, QUOTE);
      else           /* posting */
         MailTo(Moderator);
    }
    else
       return false;

    FreeString(Moderator);
  }
  else
    return false;

  return true;
}


void
DoFollowUp(struct art_struct *pArt, int rcommand)
{
  BOOL quote_header= false, fup_ok= false;
  char key, quotechars[20];
  FILE *tmp_artfp;
  char *c;

  strcpy(quotechars,quotechar);

  if (GroupModerated(NULL))
  {
    /* check for a correct followupto group */
    sprintf(fnametmp,"%s/%lu",pathtogroup,pArt->ArticleNr);
    if (tmp_artfp= fopen(fnametmp,"r"))
    {
      while (fgets(tmp, SUPERLINE, tmp_artfp) && *tmp!='\n') /* while in header */
      {
        if (c= firststr(tmp,HeaderLines[HL_FOLLOWUP_TO]))
        {
          cuteoln(c);
          if (stricmp(c,newsgroup)==0) /* no difference */
             fup_ok= false;
          else
          {
            if (GroupModerated(c)) /* FUpGroup is moderated, too? */
               fup_ok= false;
            else
               fup_ok= true;
          }
        }
      }
      fclose(tmp_artfp);
    }

    if (!fup_ok)
    {
      if (!MailToModerator(pArt->ArticleNr))
      {
        DispMsg("Nevertheless want to followup? y/N:");
        key= GetUpKey();
        ClrLine();
        if (key!='Y')
           return;
      }
      else
        return;
    }
  }
  if (rcommand==QSP_FOLLOWUP)
  {
    DispMsg("Quote header? y)es or N)o:");
    if (GetUpKey()=='Y') quote_header= true;
    ClrLine();
    DispMsg("Quotechars: S)pecify, I)nitials, else normal:");

    if ((key=GetUpKey())=='I')
       sprintf(quotechars,"%s%s",GetInitials(),quotechar);

    ClrLine();

    if (key=='S' || key=='I')
    {
      DispMsg("Quotechars:");
      Getline(quotechars,20);
      ClrLine();
    }
    rcommand= Q_FOLLOWUP;
  }
  FollowUp(pArt, rcommand==Q_FOLLOWUP ? quotechars : NULL, quote_header);
}
