/*************************************************************************
 * functions.c -- contains important functions for ARN                   *
 *  -------------------------------------------------------------------  *
 *         Arn - (C)opyright by Roland Bless, All rights reserved,       *
 *  -------- YOU ARE NOT ALLOWED TO DISTRIBUTE A CHANGED SOURCE -------  *
 *   (C)opyright 1990-93 by Roland Bless of Blessed Software Products    *
 *************************************************************************/

#ifndef ENUMS_H
 #include "enums.h"
#endif

#ifndef GLOBALVARS_H
#include "globalvars.h"
#endif

#define CP_BUFSIZE 0x4000L                 /* 16kB Copy-Buffer */

#define SE_UNREADONLY 0x01
#define SE_FORWARD    0x02
#define SE_HEADERONLY 0x04
#define SE_SELECTED   0x08

#define FOPEN_RETRY_MAX 5

/* Prototypes for functions defined in functions.c */
/* proto$start */
int
cdtogroup(void);
void cdback(void);
ULONG GetRelArtNr(struct art_struct *ArtP);
int NextRange(char *cmdline,
              ULONG *brange,
              ULONG *erange,
              short *checkbits,
              short *resultbits);
int ProcessRange(struct art_struct *pArtcurr,
                 int command,
                 int (*function)());
int listartsubj(struct art_struct *pArt,
                int command);
int ListSubjects(struct art_struct *pArtcurr);
int CatchUp(BOOL ask);
int appendarticle(struct art_struct *pArt,
                  int command);
int SaveArtToPipe(struct art_struct *pArt,
                  char *cmdline,
                  int command);
int SaveArticle(struct art_struct *pArt,
                char *cmdline,
                int command);
int markrangeas(struct art_struct *pArt,
                int command);
void MarkRange(struct art_struct *pArt,
               int command);
ULONG CountArticles(UWORD testbit,
                   UWORD result);
void StatusReport(struct art_struct *pArt);
char *Matchregexp(char *s, struct search_struct *spattern);
char *MatchKMP(char *s, struct search_struct *spattern);
ULONG
ScanArticle(struct art_struct *artarray[],
            ULONG NrofArticles,
            ULONG actartindex,
            struct search_struct *searchfor,
            char *headerfield,
            char * (*matchroutine)(char *string, struct search_struct *pattern),
            short int options,
            int command);
ULONG
SearchInArt(struct art_struct *artarray[],
            ULONG NrofArticles,
            ULONG actartindex,
            int command);

char *GetMsgID(void);
char *GetParentMsgID(void);
char *GetCurrentSubject(void);
char *GetAuthor(void);
BOOL GroupModerated(char *newsgroup);
int ExternPrg(struct art_struct *pArt);
void ChangeRotNumber(char *s);
struct art_struct *FirstUnread(void);
BOOL IsInMark(struct GroupInfoList *pGI,
              char *number,
              char *markstr);
void AppendMark(struct GroupInfoList *pGI,
                char *s);
char *GetXRef(FILE *artfp);
void MarkXRefs(FILE *fpart,
               BOOL inside_article);
void MakeDirName(char *dest,
                 char *newsdir,
                 char *newsgroup);
FILE *
fretryopen(char *Name, char *mode);
/* proto$end */

#include "proto/followup_proto.h"
#include "proto/strprocs_proto.h"
#include "proto/ArnIO_proto.h"
#include "proto/pager_proto.h"
#include "proto/cmdline_proto.h"
#include "proto/grouplist_proto.h"

/****************************************************/


                   /******* extern references *********/
extern char *p_fbuf, *p_fbufEOF, *p_fbufEND, *fbuffer,
            *shellwin, *externprg;
extern char *pend;
extern ULONG FirstArtNr, LastArtNr;
extern struct subj_struct *RootSubjP;
extern int _OSERR;
             /****************************************************/


                     /* global for this file */

          /* remove static keyword/vars for re-entrant code */
static struct FileHandle *savefp= NULL;
static char *cp_Buffer;
static int ls_lcount= 0;
static struct FileLock *FL;
static short int search_options= 0;
static BOOL regularexp= false;

char *stmp=NULL, *subj=NULL, *from=NULL, *c, *ttmp=NULL;
char *waitmsg="[Type any key to continue]";

BOOL headeronly, unreadonly;


             /****************************************************/

FILE *
fretryopen(char *Name, char *mode)
/* This function tries an fopen(Name,mode) and if it fails
   because the file is in use, it tries again after waiting a
   second. After FOPEN_RETRY_MAX it gives up and returns NULL */
{
  int count= 0;
  FILE *fp;

  while ((fp= fopen(Name, mode))==NULL &&
         _OSERR==ERROR_OBJECT_IN_USE &&
         count<FOPEN_RETRY_MAX)
  {
    Delay(50L);
    count++;
  }

  return fp;
}



int
cdtogroup(void)
/* make the directory to "pathtogroup" (global variable) to current directory */
{
  if (!(FL= (struct FileLock *) Lock(pathtogroup, (ULONG) ACCESS_READ)))
  {
    ErrorHandle(NONEWSGROUP,pathtogroup);
    return NONEWSGROUP;  /* NONEWSGROUP must be >1 !*/
  }

  FL= (struct FileLock *) CurrentDir((BPTR) FL);
}


void cdback(void)
/* restores the original dir before cdtogroup() */
{
  FL= (struct FileLock *) CurrentDir((BPTR) FL);

  if (FL) UnLock((BPTR) FL);         /* Unlock my Lock */
}


ULONG
GetRelArtNr(struct art_struct *ArtP)
{
  if (ArtP)
     return ((ULONG)ArtP-(ULONG)RootArt)/sizeof(struct art_struct);
  else
     return 0;
}


int
NextRange(char *cmdline, ULONG *brange, ULONG *erange, short *checkbits, short *resultbits)
{
  static char *cmdlp;

  ULONG x, current_artnr;

  x= 0L; current_artnr=*brange;  /* current-articleNR for . in brange */
  *checkbits= *resultbits= 0;

  if (cmdline)
     cmdlp= cmdline;

  if (*cmdlp==',') cmdlp++;

  if (*cmdlp=='^')
  {
     x=FirstArtNr;
     cmdlp++;
  }
  else
  if (*cmdlp=='.')
  {
    x=current_artnr;
    cmdlp++;
  }
  while (isdigit(*cmdlp))
  {
    x= x*10L + (ULONG) (*cmdlp-'0');
    cmdlp++;
  }
  *brange= *erange= x;

  if (*cmdlp=='-')
  {
    cmdlp++;
    x= 0L;
    if (*cmdlp=='$')
    {
      x= LastArtNr;
      cmdlp++;
    }
    else
    if (*cmdlp=='.')
    {
      x=current_artnr;
      cmdlp++;
    }

    while (isdigit(*cmdlp))
    {
      x= x*10L + (ULONG) (*cmdlp-'0');
      cmdlp++;
    }
    *erange= x;
  }

  if (*cmdlp==':')
  {
    cmdlp++;
    while(isalpha(*cmdlp))
    {
      switch (*cmdlp)
      {
        case 'n': *checkbits|= NEW;    /* new */
                  *resultbits|= NEW;
                  break;
        case 'N': *checkbits|= NEW;    /* not new */
                  break;
        case 's': *checkbits|= SELECTED_ART;  /* selected */
                  *resultbits|= SELECTED_ART;
                  break;
        case 'S': *checkbits|= SELECTED_ART;  /* not selected */
                  break;
        case 'U':
        case 'u':
        case 'R':
                  *checkbits|= READ;          /* unread */
                  break;
        case 'r': *checkbits|= READ;          /* read */
                  *resultbits|= READ;
                  break;
      }
      cmdlp++;
    }
  }
  if (*erange<*brange)    /* Not a range -> error */
  {
    *brange= *erange= 0L;
  }

  if (!brange)
     return 0;
  else
     return 1;

  /* if erange==brange then single nr,
     or brange==0L -> error or no range found */
}


int
ProcessRange(struct art_struct *pArtcurr, int command, int (*function)(struct art_struct *, int))
{
  ULONG brange=1L, erange=1L;
  UWORD checkb, resultb;
  struct art_struct *pArt;

  if (pArtcurr) brange=pArtcurr->ArticleNr;  /* x-mit actual artikel-nr */

  NextRange(cmdline, &brange, &erange, &checkb, &resultb);

  if (brange)
     while (brange)
     { /* TODO: optimize! */
       if (brange<FirstArtNr) brange=FirstArtNr;
       if (erange>LastArtNr) erange=LastArtNr;
       for(pArt= &RootArt[brange-FirstArtNr]; pArt<=LastArt && brange<=erange; pArt++,brange++)
       {
         if ((pArt->Status & checkb)== resultb)
         {
           if ((*function)(pArt, command)) return 2; /* Command breaks */
         }
       }
       if (pArtcurr) brange=pArtcurr->ArticleNr;  /* x-mit actual artikel-nr */
       NextRange(NULL, &brange, &erange, &checkb, &resultb);
     }
  else
     return 0; /* No range */

  return 1; /* range okay */
}


int
listartsubj(struct art_struct *pArt,int command)
{

#ifdef MYFILE
  struct FileHandle *Art= NULL;
#else
  FILE *Art=NULL;
#endif
  int stop= false;

  ltostr(pArt->ArticleNr, stmp, MAXLINE);
  subj[0]='\0';
  from[0]='\0';

#ifdef MYFILE
  initmyf();
  Art= (struct FileHandle *) Open(stmp, MODE_OLDFILE);
  if (Art)
  {
    while (myfgets(ttmp, SUPERLINE, Art) && *ttmp!='\n')
    {
#else
  Art= fopen(stmp, "r");
#endif
#ifndef MYFILE
  if (Art)
  {
#ifdef __SASC
    /* fmode(Art,0); */
#endif
    while (fgets(ttmp, SUPERLINE, Art) && *ttmp!='\n')
    {
#endif
      if (c= firststr(ttmp, "From: "))
      {
        strncpy(from, c, MAXLINE);
        from[MAXLINE]='\0';
      }
      if (c= firststr(ttmp, "Subject: "))
      {
        strncpy(subj, c, MAXLINE);
        subj[MAXLINE]='\0';
      }
    }

    if (subj[0]!='\0')
    {
      cuteoln(subj);
      subj[49]= '\0';
    }
    else
      strcpy(subj,"(none)");

    if (from[0]!='\0')
    {
      c= Sender(from);
      cuteoln(c);
      c[18]='\0';
    }
    else
    {
      strcpy(from,"(nobody)");
      c= from;
    }

    sprintf(stmp, "%6ld: %-50s %s", pArt->ArticleNr, subj, c);
    (*ConWriteLn)(stmp);
    ls_lcount++;
    if (ls_lcount==MaxScrLines-1)
    {
      char c;

      c=WaitForSpace(NULL);
      if (toupper(c)=='Q') stop=true;
      else
      if (isdigit(c))
      {
         EnterCmdLine(c,ARTSEL);
         stop=true;  /* abort under all circumstances, because cmdline */
                     /* is now destroyed!!! */
      }
      ls_lcount=0;
    }
#ifdef MYFILE
    Close((BPTR) Art);
#else
    fclose(Art);
#endif
  }

  return stop;
}



int   /* rc: 0=Okay, 1=quit  */
ListSubjects(struct art_struct *pArtcurr)
{
  struct art_struct *pArt;
  struct FileLock *tmplock= NULL;
  int rc=0; /* return-code */
  BOOL endprompt=true;

  ls_lcount=0;
  pArt=RootArt;

  ttmp= (char *) AllocMem((ULONG) 4*(MAXLINE+1L), MEMTYPE);
  if (!ttmp)
  {
    ErrorHandle(NOMEM,NULL);
    return 1;
  }
  stmp= ttmp+MAXLINE+1;
  subj= stmp+MAXLINE+1;
  from= subj+MAXLINE+1;

  (*ConWriteLn)(newsgroup);

  strcpy(fnametmp,newsgroup);

  if (GeneralBool & MOREDIRS)
     makeslashes(fnametmp);

  if (newsdir[strlen(newsdir)-1]!=':')
     sprintf(ttmp, "%s/%s",newsdir, fnametmp);
  else
     sprintf(ttmp, "%s%s",newsdir, fnametmp);

  if (!(MyDir= (struct FileLock *) Lock(ttmp, (ULONG) ACCESS_READ)))
  {
    if (ttmp) FreeMem (ttmp, (ULONG) 4*(MAXLINE+1L));
    return 1;
  }
  else
    tmplock= (struct FileLock *) CurrentDir((BPTR) MyDir);

  endprompt=true;

  if (*cmdline=='\0')
  {
    for (pArt= RootArt; pArt<=LastArt; pArt++)
        if ((pArt->Status & READ) == UNREAD)
           if (listartsubj(pArt, LIST_SUBJ)) break;
    if (pArt) endprompt=false; /* we aborted earlier */
  }
  else
    endprompt= (ProcessRange(pArtcurr, LIST_SUBJ, listartsubj)!=2);

  if (endprompt)
  {
    char c;

    c=WaitForSpace("[Type number, q quit or any key to continue]:");
    if (isdigit(c))
    {
       EnterCmdLine(c,ARTSEL);
    }
    else
    if (toupper(c)=='Q')
      rc=1; /* leave Group */
  }
  ls_lcount=0;

  if (ttmp) FreeMem (ttmp, (ULONG) 4*(MAXLINE+1L));
  if (tmplock) MyDir= (struct FileLock *) CurrentDir((BPTR) tmplock);
  if (MyDir) UnLock ((BPTR) MyDir);

  return rc;
}


int CatchUp(BOOL ask)
  /* marks all articles as read, assumes that RootArt and LastArt are valid */
{
  struct art_struct *pArt;
  FILE *article;

  if (ask)
  {
    sprintf(tmp,"Really mark all articles as read? [y,N]:");
    StatusLine();
    DispMsg(tmp);
    *tmp=(*GetKey)();
    ClrLine();
  }
  else
    *tmp='Y';

  if (toupper(*tmp)=='Y')
  { /* cd into group */
    cdtogroup();

    strcpy(tmp,"Marking");
    if (GeneralBool & CATCHUP_XREFS)
        strcat(tmp," and checking for crosspostings");
    strcat(tmp,"...");
    StatusLine();
    DispMsg(tmp);

    for (pArt=RootArt; pArt<=LastArt; pArt++)
    {
      if (!(pArt->Status & READ))
      {
        pArt->Status|= READ;
        if (GeneralBool & CATCHUP_XREFS)
        {
          if (pArt->Status & XREF)
          {
            ltostr(pArt->ArticleNr, fnametmp, 32);
            if (article=fopen(fnametmp,"r"))
            {
              MarkXRefs(article, false);
              fclose(article);
            }
          }
        }
      }
    }
    cdback();
  }
  else
   return 1; /* Not marked */

  return 0;  /* Marked */
}


int
appendarticle(struct art_struct *pArt, int command)
{
  struct FileHandle *artfp;
  long readbytes;

  sprintf(tmp,"%s/%lu",pathtogroup,pArt->ArticleNr);

  stmp= (char *) AllocMem ((ULONG) MAXLINE+1L, MEMTYPE);
  if (!stmp)
  {
    ErrorHandle(NOMEM,NULL);
    return GlobalERROR;
  }

  artfp= (struct FileHandle *) Open(tmp, MODE_OLDFILE);

  if (artfp)
  {
    initmyf();

#ifdef OLD /* no longer dividing spaces */
    Write((BPTR) savefp,"\n",1L);
#endif
    if (command==SAVEART_NH)  /* skip header! */
    {
      while (myfgets(stmp, MAXLINE, artfp) && *stmp!='\n');
      if (p_fbufEND<p_fbufEOF) /* flush whole buffer */
      {
        Write((BPTR) savefp, p_fbuf, (long) (p_fbufEND-p_fbuf));
      }
      else /* flush until EOF, p_fbufEOF is one BEHIND last (valid) byte */
        Write((BPTR) savefp, p_fbuf, (long) (p_fbufEOF-p_fbuf));
    }

    while (readbytes= Read((BPTR) artfp, cp_Buffer, CP_BUFSIZE))
    {
      Write((BPTR) savefp, cp_Buffer, readbytes);
    }

    if (!readbytes)
    {
      sprintf(tmp,"appended article %lu to %s.\n",
                   pArt->ArticleNr, fnametmp);
      DispMsg(tmp);
    }
    else
    {
      sprintf(tmp,"cannot save article %lu!\n", pArt->ArticleNr);
      DispMsg(tmp);
    }
    Close((BPTR) artfp);
  }
  else
  {
    sprintf(tmp,"cannot open article %lu!\n", pArt->ArticleNr);
    DispMsg(tmp);
  }
  if (stmp) FreeMem(stmp, (ULONG) MAXLINE+1L);

  return 0;
}


SaveArtToPipe (struct art_struct *pArt, char *cmdline, int command)
{
  struct FileHandle *ConFH;
  char *rest, *cmdlp;
  struct subj_struct *subjp;
  struct art_struct  *tmp_Art;

  cp_Buffer= (char *) AllocMem(CP_BUFSIZE, MEMTYPE);

  if (!cp_Buffer)
  {
    ErrorHandle(NOMEM,NULL);
    return 1;
  }

  sprintf(fnametmp,"T:Arn_PIPE%lx",FindTask(NULL));

  savefp= (struct FileHandle *) Open(fnametmp,MODE_NEWFILE);
  if (!savefp)
  {
    sprintf(tmp,"cannot save article %lu to %s!\n", pArt->ArticleNr, fnametmp);
    DispMsg(tmp);
    return 0;
  }

  if (SubjectOrder)
  {
    for (subjp= RootSubjP; subjp!=NULL; subjp=subjp->NextSubject)
    {
      for (tmp_Art= subjp->RootArt; tmp_Art!=NULL; tmp_Art=tmp_Art->NextArt)
      if (tmp_Art->Status & SELECTED_ART)
          appendarticle(tmp_Art,command);
    }
  }
  else
  if (!ProcessRange(pArt, command, appendarticle))
      appendarticle(pArt,command);

  Close((BPTR) savefp);

  if (cp_Buffer) FreeMem(cp_Buffer, (ULONG) CP_BUFSIZE);

  cmdlp=NextSpc(cmdline);    /* skip command */
  cmdlp=skpwspc(cmdlp);      /* until fname  */

  if (*cmdlp=='|')
    cmdlp=skpwspc(cmdlp+1); /* cmdlp=comm from (|comm) or (| comm) */

  rest=NextSpc(cmdlp);      /* skip first command */
  rest=skpwspc(rest);
  *(NextSpc(cmdlp))='\0';

  sprintf(tmp,"%s <%s %s",cmdlp,fnametmp,rest);

  ConFH=NULL;

  if (*cmdline=='S' || *cmdline=='W' || *cmdline=='|')
     if (shellwin)
        ConFH= (struct FileHandle *) Open(shellwin, MODE_NEWFILE);

  if (ConFH)
  {
    if (GeneralBool & SHELLWINSW)
       WBenchToFront();
    Execute(tmp,(BPTR) ConFH,(BPTR) ConFH);
    Close((BPTR) ConFH);
    ScreenToFront(ArnScreen);
    ActivateWindow(ArnWin);
  }
  else
    Execute(tmp,(BPTR) NullFH,(BPTR) NullFH);

  DeleteFile(fnametmp);
  return 0;
}


SaveArticle (struct art_struct *pArt, char *cmdline, int command)
{
  char c, *cmdlp, *hcmdlp;
  int i;
  struct subj_struct *subjp;
  struct art_struct  *tmp_Art;

  cmdlp= skpwspc(cmdline);

  if (toupper(*cmdlp)!='S' && toupper(*cmdlp)!='W' && *cmdlp!='|' && *cmdlp!='o')
                                  /* Range or search pattern given! */
  {
    cmdlp= NextSpc(cmdlp);            /* skip range or search pattern */
    cmdlp= skpwspc(cmdlp);            /* skip till 's' or 'w' */
  }

  hcmdlp=cmdlp;

  if (*cmdlp=='S' || *cmdlp=='W' || *cmdlp=='|')
  {
    SaveArtToPipe(pArt,hcmdlp,command);
    return 0;
  }

  cmdlp= NextSpc(cmdlp);            /* skip command */
  cmdlp= skpwspc(cmdlp);            /* skip till '|cmd' or filename */

  if (*cmdlp=='|')
  {
    SaveArtToPipe(pArt,hcmdlp,command);
    return 0;
  }

  if (*cmdlp=='\0')
  {
    strncpy (cmdlp,newsgroup, 32);  /* no filename given, use newsgroupname */
    cmdlp[32]= '\0';
  }

  for (i= 0; cmdlp[i]!='\0' && cmdlp[i]!=' ' && cmdlp[i]!=':' && cmdlp[i]!=';'; i++);

  if (cmdlp[i]==':')
     strcpy(fnametmp, cmdlp);
  else
  {
    strcpy(fnametmp,savedir);
    AddPart(fnametmp,cmdlp,MAXLINE);
  }

  savefp= (struct FileHandle *) Open(fnametmp, MODE_OLDFILE);

  c='y';
  if (savefp==DOSFALSE)
  {
    sprintf(tmp, "File %s doesn't exist! Create? [Y,n]:", fnametmp);
    DispMsg(tmp);
    c= (*GetKey)();
    ClrLine();
    ConCR();
    if (toupper(c)!='N')
       savefp= (struct FileHandle *) Open(fnametmp, MODE_NEWFILE);
  }

  if (toupper(c)=='N')
  {
    strcpy(tmp, "article(s) is/are not saved!\n");
    DispMsg(tmp);
    WaitForSpace(waitmsg);
    return 0;
  }

  if (!savefp)
  {
    strcpy(tmp, "cannot save article(s)!\n");
    DispMsg(tmp);
    WaitForSpace(waitmsg);
    return 0;
  }
  else
  {
    Seek((BPTR) savefp, 0L, OFFSET_END); /* append it */

    cp_Buffer= (char *) AllocMem(CP_BUFSIZE, MEMTYPE);

    if (!cp_Buffer)
    {
      ErrorHandle(NOMEM,NULL);
      Close((BPTR) savefp);
      return 1;
    }

    if (SubjectOrder)
    {
      for (subjp= RootSubjP; subjp!=NULL; subjp=subjp->NextSubject)
      {
        for (tmp_Art= subjp->RootArt; tmp_Art!=NULL; tmp_Art=tmp_Art->NextArt)
        if (tmp_Art->Status & SELECTED_ART)
            appendarticle(tmp_Art,command);
      }
    }
    else
    if (!ProcessRange(pArt, command, appendarticle))
       appendarticle(pArt,command);

    Close((BPTR) savefp);

    if (cp_Buffer) FreeMem(cp_Buffer, (ULONG) CP_BUFSIZE);
  }
}


int
markrangeas(struct art_struct *pArt,int command)
{
  switch (command)
  {
    case   UNREADART: pArt->Status&= ~READ;
                      break;
    case        JUNK: pArt->Status|= READ;
                      break;
    case SELECT     : pArt->Status|= SELECTED_ART;
                      break;
    case UNSELECT     : pArt->Status&= ~SELECTED_ART;
                        break;

  }
  return 0;
}


void
MarkRange(struct art_struct *pArt, int command)
       /* current articlenr for . */
{
  ProcessRange(pArt, command, markrangeas);
}


/* Counts articles in the art_struct (NewsGroup), returns result */
ULONG
CountArticles(UWORD testbit,UWORD result)
{
  struct art_struct *pArt;
  register ULONG count= 0L, i=0L;

  if (testbit==ALL && result==ALL)
        count=NrOfArticles;
  else
     for(pArt= RootArt; i<NrOfArticles; pArt=&RootArt[++i])
        if ((pArt->Status & testbit) == result)
        {
           count++;
        }
  return count;
}


void
StatusReport(struct art_struct *pArt)
{
  InvON();
  strcpy(tmp, "Newsgroup STATUS:\n");
  (*ConWrite)(tmp);
  sprintf(tmp, "You're reading %s", newsgroup);
  (*ConWrite)(tmp);
  if (pArt && LastArt)
  {
    sprintf (tmp,", article %lu %s (of %lu).", pArt->ArticleNr,
                 ArtStat(pArt), LastArt->ArticleNr);
    (*ConWriteLn)(tmp);
  }
  else
    ConCR();
  sprintf(tmp, "Unread: %lu Read: %lu TOTAL: %lu\n",
          CountArticles(READ,UNREAD),
          CountArticles(READ,READ),
          CountArticles(ALL,ALL));
  (*ConWrite)(tmp);
  InvOFF();
  WaitForSpace(waitmsg);
}


char *Matchregexp(char *s, struct search_struct *spattern)
{
  if (regexec(spattern->pattern.regpat, s))
     return (spattern->pattern.regpat)->startp[0];
  else
     return NULL;
}


char *MatchKMP(char *s, struct search_struct *spattern)
{
  int index;

  if ((index= KMP_Matcher(s,strlen(s),
                         spattern->pattern.kmppat, spattern->patlen,
                         spattern->kmp_table))!=-1)
     return s+index;
  else
     return NULL;
}


/*** ScanArticle(artarray, nrofarticles, actartindex, searchfor, headerfield, *
 *               matchroutine, options)                                       *
 * pArt - array of article pointers, searchfor - pattern to search for,       *
 * headerfield - not NULL if only header scan,                                *
 * matchroutine - regexp() or KMP_Matcher()                                   *
 * options - bitflags (forward/backward,unread/read)                          *
 **  returns index-number (according to artarray) of found article           **
 ******************************************************************************/

ULONG
ScanArticle(struct art_struct *artarray[],
            ULONG NrofArticles,
            ULONG actartindex,
            struct search_struct *searchfor,
            char *headerfield,
            char * (*matchroutine)(char *string, struct search_struct *pattern),
            short int options,
            int command)
{
  /* WARNING: NrOfArticles must be <= NrOfActSubj */
  /* * * * * * * * * * * * * * * * * * * * * * *  */
  struct art_struct *pArt;
  ULONG index;
#ifdef MYFILE
  struct FileHandle *Art= NULL;
#else
  FILE *Art=NULL;
#endif
  char *s, c;

  index= actartindex;

  if (!searchfor)
  {
    return index;
  }
  else
  {
    DispMsg("searching..");
    if (!auxmode)
       ConDevRead(conReadReq, &c, 1L);    /* Check for Key-pressed */

    if (options & SE_FORWARD)          /* Direction of scan */
    {
      index++;

      if (index>=NrofArticles)
         index= 0UL;

      if (index==actartindex)
      {
        return index;
      }
    }
    else
    {
      if (index==0)
         index= NrOfArticles-1;
      else
         index--;

      if (index==actartindex)
      {
        return index;
      }
    }

    cdtogroup();

    while (index!=actartindex)
    {
      pArt= artarray[index];

      if (!auxmode && GetMsg(conReadPort))
      {
        ClrLine();
        cdback();
        DispMsg("Search aborted!\n");
        SetSignal(0L,1L<<conReadPort->mp_SigBit);
        return actartindex;
      }
      if (options & SE_UNREADONLY)
         if (pArt->Status & READ) goto nextloop;

      ltostr(pArt->ArticleNr, tmp, 32);

      if (Art= fopen(tmp, "ra"))
      {
        while (fgets(tmp,SUPERLINE, Art))
        {
          if (options & SE_HEADERONLY) /* don't scan the text */
          {
            if (*tmp=='\n') break;     /* skip body */

            if (headerfield==NULL) /* no special headerfield given */
            {
              if ((*matchroutine)(tmp, searchfor)) /* found! */
              {
                fclose(Art);
                cdback();
                ClrLine();
                if (!auxmode)
                {
                  AbortIO((struct IORequest *) conReadReq);
                  WaitIO((struct IORequest *) conReadReq);
                  SetSignal(0L,1L<<conReadPort->mp_SigBit);
                }
                return index;
              }
            }
            else
            if (s=firststr(tmp,headerfield))
            {
              do /* match line-continuation, too */
              {
                if ((*matchroutine)(s, searchfor))
                {
                  fclose(Art);
                  cdback();
                  ClrLine();
                  if (!auxmode)
                  {
                    AbortIO((struct IORequest *) conReadReq);
                    WaitIO((struct IORequest *) conReadReq);
                    SetSignal(0L,1L<<conReadPort->mp_SigBit);
                  }
                  return index;
                }
              }
              while ((s=fgets(tmp,SUPERLINE, Art)) &&
                     (*s==' ' || *s=='\t'));
              break;
            }
          }
          else
          if ((*matchroutine)(tmp, searchfor))
          {
            fclose(Art);
            cdback();
            ClrLine();
            if (!auxmode)
            {
              AbortIO((struct IORequest *) conReadReq);
              WaitIO((struct IORequest *) conReadReq);
              SetSignal(0L,1L<<conReadPort->mp_SigBit);
            }
            return index;
          }
        } /* while fgets */
        fclose(Art);
      } /* endif opened article */
nextloop:
      if (index % 100 == 0)
      {
        sprintf(tmp, "%lu..", index);
        DispMsg(tmp);
      }

      if (options & SE_FORWARD)          /* Direction of scan */
      {
        index++;

        if (index>=NrofArticles)
           index= 0UL;
      }
      else
      {
        if (index==0)
           index= NrOfArticles-1;
        else
           index--;
      }
    } /* while index!= actartindex */

    if (!auxmode)
    {
      AbortIO((struct IORequest *) conReadReq);
      WaitIO((struct IORequest *) conReadReq);
      SetSignal(0L,1L<<conReadPort->mp_SigBit);
    }
    ClrLine();
    DispMsg("NOT found!");
    Delay(50L);
    cdback();
    return actartindex;
  }
}




ULONG
SearchInArt(struct art_struct *artarray[],
            ULONG NrofArticles,
            ULONG actartindex,
            int command)
/* Called from pager or article-selection */
/* NrOfArticles must contain the number of articles that are stored in artarray[],
   so it must be<=NrOfActSubj, actartindex is the current article                 */
/* parses cmdline and invokes ScanArticle */
{
  /* <search_options> and <regularexp> are static global, because they
     must be retained after one call, because of possible contituation;
     <NrOfArticles> must be the least upper bound of elements in artarray */

  if (command==SEARCH_INART_F)
    search_options|= SE_FORWARD;
  else
    search_options&= ~SE_FORWARD;


  if (*(cmdline+1))   /* New search-pattern */
  {
    casedep= 0x5F;     /* Case-independent as default */
    search_options|= SE_HEADERONLY;     /* Subject:-line as default */
    search_options|= SE_UNREADONLY;  /* default */
    regularexp= false;

    c=strmetachr(cmdline+1,'~');

    if (c && *c=='~')       /* Any modifiers? */
    {
      *c++='\0';             /* Cut off the search-pattern at the delimiter */
      while (*c!='\0')
      {
        switch (*c)
        {
          case 'a': search_options&= ~SE_HEADERONLY;
                    break;
          case 'h': search_options|= SE_HEADERONLY;
                    break;
          case 'c': casedep=0xff;      /* case-dependent search */
                    break;
          case 'r': search_options&= ~SE_UNREADONLY;
                    break;
          case 'R': regularexp= true;
                    break;
        }
        c++;
      }
    }

    /* free previous search-pattern before allocating a new one */
    if (searchpat.patlen)
    {
      if (searchpat.kmp_table)
      {
        free(searchpat.pattern.kmppat);
        free(searchpat.kmp_table);
        searchpat.kmp_table= NULL;
        searchpat.pattern.kmppat= NULL;
      }
      else
      {
        if (searchpat.pattern.regpat) free(searchpat.pattern.regpat);
        searchpat.pattern.regpat=NULL;
      }
      searchpat.patlen= 0;
    }

    if (regularexp)
    { /* compile regular expression */
      if((searchpat.pattern.regpat=regcomp(cmdline+1))==NULL) /* Regular Expression */
      {
        ErrorHandle(REGEXP,NULL);
        return actartindex;
      }
      searchpat.patlen= 1; /* not the real patlen, only to signal that
                             there is a searchpattern */
    }
    else
    { /* compile KMP-table */
      int patlen= strlen(cmdline+1);

      searchpat.kmp_table= malloc(patlen*sizeof(short int));
      KMP_ComputeNextTable(cmdline+1,patlen,searchpat.kmp_table);
      searchpat.pattern.kmppat= malloc(patlen+1);
      strcpy(searchpat.pattern.kmppat,cmdline+1);
      searchpat.patlen= patlen;
    }
  }
  else /* no new search pattern */
  if (searchpat.patlen==0)
  {
    DispMsg("No pattern to search for...!\n");
    return actartindex;
  }

  return ScanArticle(artarray, NrofArticles, actartindex, &searchpat,
                     NULL, regularexp ? Matchregexp : MatchKMP ,
                     search_options, NOP);
}


char *
GetMsgID(void)
{
  char *s= NULL,*t= NULL;

  pend= NULL;

  if ((s= GetBufField(HeaderLines[HL_MESSAGE_ID]))==NULL)
    s=GetBufField(HeaderLines[HL_ARTICLE_ID]);

  if (s)
  {
    s= skpwspc(s);
    t= CopyString(s);
  }

  GetBufField(NULL);

  return t;
}


char *
GetParentMsgID(void)
{
  char *s= NULL,*t= NULL, mem, *memp;

  pend= NULL;

  if (s= GetBufField(HeaderLines[HL_REFERENCES]))
  {
    s=skpwspc(s);
    for (t=s; *t!='\0' && *t!=' ' && *t!='\t' && *t!='\n'; t++);

    memp= t;mem= *memp;*t= '\0'; /* Overwrite with \0 */
    t= CopyString(s);
    *memp= mem;                /* and restore */
  }

  GetBufField(NULL);

  return t;
}



char *
GetCurrentSubject(void)
{
  char *s= NULL,*t= NULL;

  pend=NULL;

  if (s= GetBufField(HeaderLines[HL_SUBJECT]))
  {
    s=skpwspc(s);
    t= CopyString(s);
  }

  GetBufField(NULL);

  return t;
}



char *
GetAuthor(void)
{
  char *s= NULL,*t= NULL;

  pend= NULL;

  if (s= GetBufField(HeaderLines[HL_FROM]))
  {
    s= skpwspc(s);
    t= CopyString(s);
  }

  GetBufField(NULL);

  return t;
}


BOOL
GroupModerated(char *newsgroup)
/* Checks if newsgroup is moderated. If newsgroup==NULL only GeneralBool is checked
   for the G_MODERATED flag */
{
  struct GroupInfoList *pGI;

  if (newsgroup)
  {
    if (pGI= FindGroup(newsgroup))
    {
      if (pGI->Status & MODERATED)
         return true;
    }
  }
  else
     if (GeneralBool & G_MODERATED) return true;

  return (BOOL) false;
}


int
ExternPrg (struct art_struct *pArt)
{
  struct FileHandle *ConFH=NULL;

  if (externprg==NULL || *externprg=='\0') return 0;

  stmp= (char *) AllocMem ((ULONG) MAXLINE+1L, MEMTYPE);
  if (!stmp)
  {
    ErrorHandle(NOMEM,NULL);
    return GlobalERROR;
  }
  /* create filename of current article */
  strcpy(stmp,newsgroup);
  if (GeneralBool & MOREDIRS)
     makeslashes(stmp);
  if (newsdir[strlen(newsdir)-1]!=':')
     sprintf(fnametmp, "%s/%s/%lu",newsdir, stmp, pArt->ArticleNr);
  else
     sprintf(fnametmp, "%s%s/%lu",newsdir, stmp, pArt->ArticleNr);

  sprintf(tmp,externprg,fnametmp);

  if (GeneralBool & EXTERNWIN)
     if (shellwin)
        ConFH= (struct FileHandle *) Open(shellwin, MODE_NEWFILE);

  DispMsg("executing external program...");

  if (ConFH)
  {
    if (GeneralBool & SHELLWINSW) WBenchToFront();
    Execute(tmp,(BPTR) ConFH,(BPTR) ConFH);
    Close((BPTR) ConFH);
  }
  else
  {
    if (GeneralBool & EXTERNSCRSW) WBenchToFront();
    Execute(tmp,(BPTR) NullFH,(BPTR) NullFH);
  }

  ScreenToFront(ArnScreen);
  ActivateWindow(ArnWin);
  DispMsg("ready.\n");

  if (stmp) FreeMem(stmp, (ULONG) MAXLINE+1L);
}


void
ChangeRotNumber(char *s)
{
   s= skpwspc(NextSpc(s));
   rotnumber= (int) atol(s);
   if (rotnumber<=0) rotnumber=13;
   sprintf(tmp,"Changed ROT-Number to %d\n",rotnumber);
   DispMsg(tmp);
}


struct art_struct *
FirstUnread(void)
{
  struct art_struct *pArt;

  for (pArt= RootArt; pArt<=LastArt && (pArt->Status & READ); pArt++);

  return pArt; /* Returns NULL or first occurence of UNREAD */
}


BOOL
IsInMark(struct GroupInfoList *pGI, char *number, char *markstr)
 /* Check whether number s is already in the markstr */
{
  int i;
  ULONG brange= 0UL, erange= 0UL;
  register ULONG x=0L, nrtocheck=0L;
  ULONG FirstArtNr, LastArtNr;


  FirstArtNr=pGI->firstArtNr;
  LastArtNr=pGI->lastArtNr;

  nrtocheck=(ULONG) atol(number);

  /* Parse Markstr */

  if (nrtocheck<FirstArtNr) return true; /* expired article */

  if (markstr!= NULL)
     for (i= 0; markstr[i]!='\0'; i++)
     {
       x= 0UL;

       while (isdigit(markstr[i]))
       {
         x= x*10UL + (ULONG) (markstr[i]-'0');
         i++;
       }
       if (markstr[i]=='-')  /* real range follows */
       {
         brange= x;
         erange= x;
       }
       else                /* end of range or single number */
       {
         if (erange!=0L)   /* we have brange, now we read erange */
         {
           if (x>=brange)
           {
             if (brange<FirstArtNr)
                brange= FirstArtNr;

             if (brange>LastArtNr) brange= erange= x= 0UL;

             if (x<FirstArtNr)    /* erange is not in valid range */
             {
               brange= erange= 0UL; /* do not count it */
             }
             else
             {
               if (x>LastArtNr) x=LastArtNr;
               erange= x;        /* real range */
             }
           }
           else
              erange=brange;    /* nonsens range */
         }
         else                /* single number */
         {
           if (x<FirstArtNr || x>LastArtNr) x=0L;

           erange= brange= x;
         }

         if (erange && brange) /* valid range or number */
         {
            if (nrtocheck>=brange && nrtocheck<=erange)
               return true;
         }
         brange= erange= 0UL;  /* reset range that is to mark */
         if (markstr[i]=='\0')  /* stop */
           i--; /* The next test will end the loop */
       }
     }

  return false;
}


/* #define DEBUG */

/* Appends a mark to Markers-Structure from a group */
void
AppendMark(struct GroupInfoList *pGI,char *s)
{
  struct Markers *pmark;
  int slen, lastpos= 0;

#ifdef DEBUG
  printf("Appendmark %s:\n", pGI->Name);
#endif

  pmark= pGI->Marks;
  if (pmark)
  {
    while (pmark->NextMark)
    {
#ifdef DEBUG
      printf ("Testing: %s\n",pmark->Mark);
#endif
      if (IsInMark(pGI, s, pmark->Mark)) return; /* This mark exists already */
      pmark= pmark->NextMark;  /* get last Mark */
    }
#ifdef DEBUG
      printf ("Testing: %s\n",pmark->Mark);
#endif
    if (IsInMark(pGI, s, pmark->Mark)) return; /* This mark exists already */
#ifdef DEBUG
    printf("Gotit!\n");
#endif
    lastpos= pmark->lastpos;
  }

  slen= strlen(s);

  if (!pmark || lastpos+slen+1>=MAXLINE)
  {              /* it doesn't fit (don't forget the comma!) */
    if (pmark)
    {
      pmark->NextMark= AllocMem((ULONG) sizeof(struct Markers),MEMTYPE);
      pmark= pmark->NextMark; /* actual markstruct */
    }
    else
     pmark= pGI->Marks= AllocMem((ULONG) sizeof(struct Markers),MEMTYPE);

    if (!pmark)
    {
      ErrorHandle(NOMEM,NULL);
    }
    else
    {
      pmark->NextMark= NULL;  /* No further MarkString, yet */
      lastpos= pmark->lastpos= 0;      /* no actual Markstring */
      pmark->Mark= AllocMem((ULONG)MAXLINE+1L,MEMTYPE);
         /* alloc Markstring */
      if (!pmark->Mark)
        ErrorHandle(NOMEM,NULL);
    }
  }

  if (lastpos) /* insert leading comma */
  {
#ifdef DEBUG
    printf("lastpos: %d %c=%x\n",lastpos,*(pmark->Mark+lastpos),*(pmark->Mark+lastpos));
#endif
    *(pmark->Mark+lastpos)=',';
#ifdef DEBUG
    printf("now: %c pad:%s| ",*(pmark->Mark+lastpos),s);
#endif
    strcpy(pmark->Mark+lastpos+1, s);
#ifdef DEBUG
    printf("result:%s\n",pmark->Mark+lastpos+1);
#endif
    pmark->lastpos= lastpos+slen+1;
  }
  else
  {
    if (pmark->Mark)
    {
      strcpy(pmark->Mark,s);
      pmark->lastpos= slen;
    }
  }
#ifdef DEBUG
    printf("%s-m:%s\n",pGI->Name,pmark->Mark);
#endif
}


char *
GetXRef(FILE *artfp)
{
  char *c;

  while (fgets(tmp, SUPERLINE, artfp) && *tmp!='\n')
  {
    if (c= firststr(tmp,"XRef:")) return c;
  }
  return NULL;
}


/* #define DEBUG */
/* Mark all articles in XRef line as read */
/* Format: "XRef: nodename newsgroup:xxxx newsgroup2:yyyy" */
void MarkXRefs(FILE *fpart, BOOL inside_article)
{
  char *s= NULL,*articlenr, *c, ctmp;
  struct GroupInfoList *pGI;

  stmp= (char *) AllocMem ((ULONG) (MAXLINE+1L)+32UL, MEMTYPE);
  if (!stmp)
  {
    ErrorHandle(NOMEM,NULL);
    return;
  }

  articlenr= stmp+MAXLINE+1; /* Use mem of stmp as well */

  pend= NULL;

  if (s= inside_article ? GetBufField("XRef:") : GetXRef(fpart))
  {
#ifdef DEBUG
    printf("Got XREF: %s\n",s);
#endif
    s= skpwspc(s);
#ifdef HARD
    if (!strncmp(s,NodeName,strlen(NodeName))) /* it's really our XRef */
    {
#endif
      s= skpwspc(NextWhiteSpc(s));
      while (*s)
      {
        c= strchr(s,':');
        if (c!=NULL)
        {
          *c= '\0';
          strcpy(stmp,s);
          *c= ':';
          s= c+1;
          c= NextWhiteSpc(s);
          ctmp= *c; *c= '\0';
          strncpy(articlenr,s,31);
          articlenr[31]= '\0';
          *c= ctmp;
#ifdef DEBUG
          printf("Got:%s:%s\n",stmp,articlenr);
#endif
          for (pGI= RootGI; pGI!=NULL && strcmp(pGI->Name,stmp); pGI= pGI->NextGI);
          if (pGI!=NULL)
          {
            /* don't mark current or unsubscribed newsgroups */
            if (strcmp(pGI->Name,newsgroup) && !(pGI->Status & UNSUBSCRIBED))
              AppendMark(pGI,articlenr); /* Insert the mark in the group */
          }
        }
        s= skpwspc(NextWhiteSpc(s)); /* Next Group */
      }
#ifdef HARD
    }
#endif
  }

  GetBufField(NULL); /* restore Buffer */

  if (stmp) FreeMem(stmp, (ULONG) (MAXLINE+1L)+32UL);
}


void
MakeDirName(char *dest, char *newsdir, char* newsgroup)
/* Produces the dirname from newsgroup, result goes to dest */
/*  which must be big enough to hold the result */
/* !!WARNING: fnametmp is used!! */
{
  char *ftmp;

  ftmp= CopyString(newsgroup);
  if (GeneralBool & MOREDIRS) makeslashes(ftmp);
  strcpy(fnametmp, newsdir);
  if (AddPart(fnametmp, ftmp, MAXLINE)) strcpy(dest,fnametmp);
  FreeString(ftmp);
}
