/***********************************************************************
 * getconfig.c -- CONFIG functions for ARN                             *
 *                                                                     *
 * Arn - (C)opyright by Roland Bless, All rights reserved,             *
 * -------- YOU ARE NOT ALLOWED TO DISTRIBUTE A CHANGED SOURCE ------- *
 ***********************************************************************/

#ifndef GLOBALVARS_H
#include "globalvars.h"
#endif

#ifndef ENUMS_H
 #include "enums.h"
#endif

#ifndef GRAPHICS_DISPLAYINFO_H
#include <graphics/displayinfo.h>
#endif

#ifndef GRAPHICS_TEXT_H
#include <graphics/text.h>
#endif

#ifndef LIBRARIES_DISKFONT_H
#include <libraries/diskfont.h>
#endif

#include <proto/graphics.h>
#include <proto/diskfont.h>

 /***************/
#define EDSCREEN     "SCREENSW"    /* Switch back for editor? */
#define NEWSDIRF     "NEWSDIR"     /* Where I can find the newsgroups */
#define DBASEDIR     "DBASEDIR"    /* Where should I store .newsdb and .X? */
#define NEWSDIRT     "DOTDIRS"     /* Where I can find the newsgroups */
#define SAVEDIR      "SAVENEWS"    /* Where should I save the news? */
#define GROUPLISTF   "GROUPLIST"   /* Where I can find the .newsrc */
#define ACTIVEFILE   "ACTIVEFILE"  /* Where's the active-file? */
#define UUCONFIGF    "UUCONFIG"    /* What's our uucp config-file ? */
#define SEQFILE      "SEQFILE"     /* What's the name of our SeqF? */
#define SIGNFILE     "SIGNATURE"   /* Whereis the signfile? */
#define HEADERFILE   "HEADERFILE"  /* Stuff to be appended to the header */
#define QUOTEC       "QUOTECHAR"   /* The quotechar for replies,followups */
#define REPLYTO      "REPLY-TO"
#define TIMEOFFSET   "TIMEOFFSET"    /* TZ-Offset */
#define UUXNAME      "SENDNEWS"      /* uuxname? */
#define RNEWS        "SELFSEND"      /* rnews */
#define SENDMAIL     "SENDMAIL"      /* sendmail-name? */
#define SHELLWINDOW  "SHELLWINDOW"   /* Shell-Windowtype */
#define EXTERNPRG    "EXTERNALPRG"   /* Name of external program */
#define MAILOPTS     "MAILOPTS"      /* Generate From:,Date-line for sendmail? */
#define SCREENDATA   "SCREENDATA"    /* User-defined Screen sizes... */
#define PATHTYPE     "PATHTYPE"      /* Full Path, No Path, Short Path */
#define FILTER       "FILTER"        /* Filter for preparing article to post */
#define XLIFETIME    "XLIFETIME"     /* expiration period for X-file entries */
#define CATCHUPXREFS "CATCHUPXREFS"  /* GlobalBool: Catchup X-Refs, too? */
#define COLORS       "COLORS"        /* Colors: Specify custom colors */
#define HEADERDISPLAY "HEADERDISPLAY" /* String which header-lines should be displayed */
#define PAGEROPTS     "PAGEROPTS"     /* ClearScreen, headerstop */
#define ASKFORAUTOSEL "ASK-FOR-AUTOSELECT-FUPS" /* Ask for autoselect follow ups */
#define CUSTOMFONT    "CUSTOMFONT"      /* Use Diskfont CUSTOMFONT */
#define AUXEDITOR     "AUXEDITOR"      /* Use this editor on Terminal */
#define METAMAIL      "METAMAIL"       /* Specify Metamail-call */
#define ARTDISPLAYORDER  "ARTDISPLAYORDER" /* Article display order */
#define PUBSCREENNAME "PUBSCREENNAME"  /* Public Screen Name */

 /* UUCONFIG-entries */
 /* as variables, because used twice */
char *USERNAME=    "UserName";
char *REALNAME=    "RealName";
char *NODENAME=    "NodeName";
char *DOMAINNAME=  "DomainName";
char *NEWSFEED=    "NewsFeed";
char *NEWSEDITOR=  "NewsEditor";
char *MAILEDITOR=  "MailEditor";
char *ORGANIZATION="Organization";
char *MODERATORS=  "Moderators";
/*************************************/
char *UULIB= "UULIB:";
char *CONFIGFILE= ".arnrc";

/* Prototypes for functions defined in getconfig.c */
/* proto$start */
void Entry(char *c,
           char **var);
void
GetConfFromVar(char *VarName, char **ConfEntry);
int getconfig(void);
BOOL getfield(char *line,
              char *field,
              char *info);
void GetandSetViewMode(char *ModeName);
struct TextFont *GetFont(char *fontname);
int GetArnConfig(BOOL scrdat);
/* proto$end */

#include "proto/ArnIO_proto.h"
#include "proto/strprocs_proto.h"

extern int pathtype;
extern int xlifetime;
extern UWORD displaymode; /* from readgroup.c */

extern struct TagItem Tag1[]; /* from Arn.c */
extern struct TextFont *ArnTxtFont;

extern char *nullstr;
                     /***********************/
void
Entry(char *c,char **var)   /* Allocates Mem for variable (var) and copies c to var */
{
  int len;

  if (*var) return; /* if variable already exists */

  cuteoln(c);
  c= skpwspc(c);
  if (len= strlen(c))
  {
    *var= AllocMem((ULONG) len+1L, MEMTYPE);
    if (*var)
      strcpy(*var,c);
    else
      ErrorHandle(NOMEM,NULL);
  }
}


void
GetConfFromVar(char *VarName, char **ConfEntry)
/* reads local or environment variable if exists (2.0 only) */
{
  char vartmp[MAXLINE];

  if (GetVar(VarName, vartmp, MAXLINE, NULL)>0)
  {
    *ConfEntry= CopyString(vartmp);
  }
}


int
getconfig()
{
  FILE *fp= NULL;
  char *c;

  GetConfFromVar(REALNAME, &RealName);
  GetConfFromVar(NODENAME, &NodeName);
  GetConfFromVar(DOMAINNAME, &DomainName);
  GetConfFromVar(NEWSEDITOR, &NewsEditor);
  GetConfFromVar(MAILEDITOR, &MailEditor);
  GetConfFromVar(ORGANIZATION, &OrgaName);
  GetConfFromVar(MODERATORS, &Moderators);

  if (uuconfigfile)
     fp= fopen(uuconfigfile, "r");
  else
     fp= NULL;

  if (fp)
  {
    while (fgets(tmp,SUPERLINE,fp))
    {
      if (c= firststr(tmp,USERNAME))
        Entry(c,&UserName);
      else
      if (c= firststr(tmp,REALNAME))
        Entry(c,&RealName);
      else
      if (c= firststr(tmp,NODENAME))
        Entry(c,&NodeName);
      else
      if (c= firststr(tmp,DOMAINNAME))
        Entry(c,&DomainName);
      else
      if (c= firststr(tmp,NEWSFEED))
        Entry(c,&NewsFeed);
      else
      if (c= firststr(tmp,NEWSEDITOR))
        Entry(c,&NewsEditor);
      else
      if (c= firststr(tmp,MAILEDITOR))
        Entry(c,&MailEditor);
      else
      if (c= firststr(tmp,ORGANIZATION))
        Entry(c,&OrgaName);
      else
      if (c= firststr(tmp,MODERATORS))
        Entry(c,&Moderators);
    }
  }
  else
   return 1;

  if (fp) fclose(fp);
  if (UserName && RealName && NodeName && NewsEditor && MailEditor)
     return 0;
  else
    return 2;
}


BOOL
getfield (line, field, info)
char *line, *field, *info;
{
  char *c;

  if (c= firststr(line, field))
  {
    if (*c!=' ' && *c!='\t')
    {
      *info= '\0';
      return false;
    }
    c= skpwspc(c);
    strcpy(info, c);
    return true;
  }
  else
  {
    *info= '\0';
    return false;
  }
}


void
GetandSetViewMode(char *ModeName)
/* Set ViewModes if specified, respect OS2.0 Modes */
{
  ULONG ViewModes= DEFAULT_MONITOR_ID, id;
  struct NameInfo namequery;

  cuteoln(ModeName);

  if (!strncmp(ModeName,"IDNUM:",6)) /* read number of viewmode */
  {
    ModeName+= 6;
    ViewModes= strtoul(ModeName,&ModeName,NULL);
    Tag1[0].ti_Data= ViewModes;

    return;
  }

  for (id= NextDisplayInfo(INVALID_ID); id!=INVALID_ID; id= NextDisplayInfo(id))
  {
    GetDisplayInfoData(FindDisplayInfo(id), (UBYTE *) &namequery,
                       sizeof(namequery), DTAG_NAME, NULL);
    if (!stricmp(namequery.Name, ModeName))
    {
      ViewModes= id;
      break;
    }
  }

  if (ModeNotAvailable(ViewModes))
     ErrorHandle(INVALIDMODE,NULL);
  else
     Tag1[0].ti_Data= ViewModes; /* Set SA_Display_ID Tag */
}


struct TextFont *
GetFont(char *fontname)
{
  struct TextAttr fontattr= { NULL, 0, FS_NORMAL, FPF_DISKFONT };
  struct Library *DiskfontBase;
  struct TextFont *TxtFont= NULL;
  char *c;

  if (*fontname)
  {
    c= fontattr.ta_Name= skpwspc(fontname);
    c= NextWhiteSpc(c);
    fontattr.ta_YSize= (UWORD) atol(skpwspc(c));
    *c='\0';
    if (!(DiskfontBase= OpenLibrary("diskfont.library", 1L)))
    {
      ErrorHandle(FATALERR,"Couldn't open diskfont.library");
    }
    else
    {
      TxtFont= OpenDiskFont(&fontattr);
      if (!TxtFont) ErrorHandle(WARNNOFONT,fontattr.ta_Name);
      GlobalERROR= OK;
      CloseLibrary(DiskfontBase);
    }
  }

  return TxtFont;
}


int GetArnConfig(BOOL scrdat)   /* scrdat ? get only screen-info : get full info */
{
  char *tzstr;
  extern FILE *fp;
  extern char *prgname, *c, *NextSpc();
  extern long TZ_Offset;

  pathtype= PATH_FULL; /* default is full Path: line */
  GeneralBool|= MOREDIRS;   /* DOTDIRS No  == UUNEWS:group/name */

  strcpy(fnametmp,UULIB);

  if (UserName==NULL)  /* try env-variable USER first */
  {
    if (GetVar("USER", tmp, MAXLINE, NULL)>0 ||
        GetVar(USERNAME, tmp, MAXLINE, NULL)>0)
    {
      UserName= CopyString(tmp);
    }
  }

  if (UserName)
  {
    strcat(fnametmp,UserName);
  }
  strcat(fnametmp,CONFIGFILE);
  if (!(fp= fopen(CONFIGFILE,"r")))   /* try local .arnrc */
     if (!(fp= fopen(fnametmp, "r")))  /* try UULIB:$USERNAME.arnrc */
     {
       strcpy(fnametmp, UULIB);
       strcat(fnametmp, CONFIGFILE);
       fp= fopen(fnametmp, "r");      /* try UULIB:.arnrc */
     }

  if (fp)
  {
    while (fgets(cmdline, MAXLINE,fp))
    { /* get CUSTOMFONT if ANY */
      if (scrdat)
      {
        if (getfield(cmdline,CUSTOMFONT, tmp)!=NULL)
        {
          ArnTxtFont= GetFont(tmp);
        }
        else
        if (getfield(cmdline,PUBSCREENNAME, tmp)!=NULL)
          Entry(tmp, &PubScreenName);
        /* get special Screen-Modes */
        if (getfield(cmdline,SCREENDATA, tmp))
        {
          if (*tmp!='\0')
          sscanf(tmp,"%hd %hd %hd", &ArnNewScreen.Width,
                                    &ArnNewScreen.Height,
                                    &ArnNewScreen.Depth);
          for(c= tmp; *c!='\0' && (isspace(*c) || isdigit(*c)); c++);
          if (*c)
          {
            GetandSetViewMode(c);
          }
        }
        else  /* check for custom colors */
        if (getfield(cmdline, COLORS, tmp))
           Entry(tmp,&colordata);
      }
      else /* no scrdat, 2nd phase */
      {
       if (getfield(cmdline, EDSCREEN, tmp))
       {
        if (toupper(*tmp)=='N')
           GeneralBool&= ~EDSCRSWITCH;
        else
           GeneralBool|= EDSCRSWITCH;
       }
       else
       if (getfield(cmdline, NEWSDIRF, tmp))
         Entry(tmp,&newsdir);
       else
       if (getfield(cmdline, DBASEDIR, tmp))
         Entry(tmp,&dbasedir);
       else
       if (getfield(cmdline, NEWSDIRT, tmp))             /* NewsDirType */
       {
         GeneralBool&= ~MOREDIRS;  /* DOTDIRS ==> UUNEWS:group.name */
       }
       else
       if (getfield(cmdline, PATHTYPE, tmp))             /* PathType */
       {
         if (toupper(*tmp)=='U')  /* Only Path: user */
            pathtype= PATH_USER;
         else
         if (toupper(*tmp)=='N')  /* No Path: */
            pathtype= PATH_NONE;
         else
         if (toupper(*tmp)=='D')  /* Path: nodename with domain */
            pathtype= PATH_DOMAIN;
         else
            pathtype= PATH_FULL;           /* Full Path: */
       }
       else
       if (getfield(cmdline, SAVEDIR, tmp))
         Entry(tmp,&savedir);
       else
       if (getfield(cmdline, GROUPLISTF, tmp))
         Entry(tmp,&grouplistfile);
       else
       if (getfield(cmdline, UUCONFIGF, tmp))
         Entry(tmp,&uuconfigfile);
       else
       if (getfield(cmdline, UUXNAME, tmp))
       {
         cuteoln(tmp);
         if (*tmp=='\0')
         {
           ErrorHandle(WARNUUX,NULL);
           strcpy(tmp, "uux %s %s!rnews");
         }
         uuxname= CopyString(tmp);
       }
       else
       if(getfield(cmdline, SENDMAIL, tmp))
       {
         cuteoln(tmp);
         if (*tmp=='\0')
         {
           ErrorHandle(WARNSMAIL,NULL);
           strcpy(tmp, "sendmail <%s");
         }
         sendmail= CopyString(tmp);
       }
       else
       if (getfield(cmdline, MAILOPTS, tmp))
       {
         c= tmp;

         GeneralBool&= ~(GENFROMLINE | GENDATELINE);

         do
         {
           if (*c=='F')
             GeneralBool|= GENFROMLINE;
           else
           if (*c=='D')
             GeneralBool|= GENDATELINE;

           c= skpwspc(NextWhiteSpc(c));
         }
         while (*c!='\0');
       }
       else
       if (getfield(cmdline, PAGEROPTS, tmp))
       {
         c= tmp;

         GeneralBool&= ~(HEADERSTOP | CLEARSCREEN);

         do
         {
           if (*c=='H')
             GeneralBool|= HEADERSTOP;
           else
           if (*c=='C')
             GeneralBool|= CLEARSCREEN;

           c=skpwspc(NextWhiteSpc(c));
         }
         while (*c!='\0');
       }
       else
       if (getfield(cmdline, RNEWS, tmp))
         Entry(tmp,&rnews);
       else
       if (getfield(cmdline, FILTER, tmp))
         Entry(tmp,&filter);
       else
       if (getfield(cmdline, EXTERNPRG, tmp))
       {
         if (toupper(*tmp)=='N')
         {
           GeneralBool&= ~EXTERNWIN;
           if (*tmp=='N')
              GeneralBool|= EXTERNSCRSW;
           else
              GeneralBool&= ~EXTERNSCRSW;
         }
         else
           GeneralBool|= EXTERNWIN;
         c= skpwspc(NextSpc(tmp));
         externprg= CopyString(c);
       }
       else
       if (getfield(cmdline, SEQFILE, tmp))
         Entry(tmp,&seqf);
       else
       if (getfield(cmdline, SHELLWINDOW, tmp))
       {
         cuteoln(tmp);
         if (toupper(*tmp)=='N')
           GeneralBool&= ~SHELLWINSW;
         else
           GeneralBool|= SHELLWINSW;

         c= skpwspc(NextSpc(tmp));
         if (*c=='\0')
            strcpy(c, "CON:0/0/640/200/Arn-CLI");  /* default entry */
         shellwin= CopyString(c);
       }
       else
       if (getfield(cmdline, METAMAIL, tmp))
         Entry(tmp,&MetaMail);
       else
       if (getfield(cmdline, CATCHUPXREFS, tmp))
       {
         GeneralBool|= CATCHUP_XREFS;
       }
       else
       if (getfield(cmdline, ASKFORAUTOSEL, tmp))
       {
         GeneralBool|= ASK_FOR_AUTOSELECT;
       }
       else
       if (getfield(cmdline, ARTDISPLAYORDER, tmp))
       {
         cuteoln(tmp);
         if (stricmp(tmp,"ARTICLE-DATE")==0)
            displaymode= ART_DATE;
         else
         if (stricmp(tmp,"SUBJECT-DATE")==0)
            displaymode= SUBJ_DATE;
         else
         if (stricmp(tmp,"COMPACT-ALPHABETICAL")==0)
            displaymode= COMPACT_ALPHABET;
         else
         if (stricmp(tmp,"COMPACT-DATE")==0)
            displaymode= COMPACT_DATE;
         else
            displaymode= SUBJ_ALPHABET;
       }
       else
       if (getfield(cmdline, SIGNFILE, tmp))
         Entry(tmp,&signfile);
       else
       if (getfield(cmdline, AUXEDITOR, tmp))
         Entry(tmp,&AuxEditor);
       else
       if (getfield(cmdline, HEADERFILE, tmp))
         Entry(tmp,&headerfile);
       else
       if (getfield(cmdline, ACTIVEFILE, tmp))
         Entry(tmp,&activefile);
       else
       if (getfield(cmdline, REPLYTO, tmp))
         Entry(tmp,&replyto);
       else
       if (getfield(cmdline, HEADERDISPLAY, tmp))
         Entry(tmp,&headerline);
       else
       if (getfield(cmdline, QUOTEC, tmp))
       {
         cuteoln(tmp);
         if (strlen(tmp)>6)
            ErrorHandle(WARNQUOTE,NULL);
         else
         {
           int i= 0;

           quotechar[i]= '\0';
           if (tmp[i]!='\0')
           {
             for (i=0; tmp[i+1]!='\0' && i<5; i++)
                 quotechar[i]= tmp[i+1];
           }
           if (i>0)
              quotechar[--i]= '\0';
         }
       }
       else
       if(getfield(cmdline, TIMEOFFSET, tmp))
       {
         cuteoln(tmp);
         tzstr= NextSpc(tmp);
         if (*tzstr=='\0' || strlen(tzstr)>6)
         {
           ErrorHandle(WARNTZNAME,NULL);
           strcpy(timez,"GMT");
         }
         else
           strcpy(timez, tzstr+1);

         *tzstr= '\0';
         if (!islong(tmp) && !((tmp[0]=='-' || tmp[0]=='+') && islong(&tmp[1])))
         {
           ErrorHandle(WARNTZOFFSET,NULL);
           TZ_Offset= 0L;
         }
         else
           TZ_Offset= atol(tmp);
       }
       else
       if (getfield(cmdline, XLIFETIME, tmp))
       {
         cuteoln(tmp);
         xlifetime=(int) atol(tmp);
         if (xlifetime <0)
          xlifetime=30;
       }
      }
    } /* while fgets */
  } /* if fp */
  else
    return 1;     /* Error occured */

  fclose(fp);

#ifdef DEBUG
  printf ("NEWSDIR: %s\n",newsdir);
  if (GeneralBool & MOREDIRS)
     printf ("NEWSTYPE: Y\n");
  printf ("GROUPLIST: %s\n",grouplistfile);
  printf ("UUCONFIGF: %s\n",uuconfigfile);
  printf ("SEQF: %s\n",seqf);
  printf ("SIGNFILE: %s %\n",signfile);
  printf ("QUOTECHAR: %s\n",quotechar);
  printf ("TIMEZONEOFFSET: %ld (%s)\n",TZ_Offset, timez);
  printf ("PATHTYPE: %d\n",pathtype);
#endif
  *cmdline= '\0';
  if (!scrdat)
  {
    if (headerline==NULL)
    {
      char *default_headerline= "n#F#A_S/s/K#DIEOL";

      headerline= CopyString(default_headerline);
    }
  }

  return 0;
}
