/*************************************************************************
 * grouplist.c -- contains functions for the "grouplistfile"             *
 * (C)opyright 1990-92 by Roland Bless of Blessed Software Products      *
 *************************************************************************/
#ifndef ENUMS_H
 #include "enums.h"
#endif

#ifndef GLOBALVARS_H
#include "globalvars.h"
#endif

/* Prototypes for functions defined in grouplist.c */
/* proto$start */
void FreeMarks(struct Markers *rootmark);
struct Markers *NewMark(struct GroupInfoList *pGI);
int UpdateGroupList(void);
void CheckLibMarks(char *Markstr, struct GroupInfoList *pGI);
ULONG CountMarks(char *Markstr);
struct GroupInfoList *FindGroup(char *s);
BOOL Unsubscribe(struct GroupInfoList *pGI);
void Resubscribe(struct GroupInfoList *pGI);
void MoveGroups(struct GroupInfoList *dest);
void SortGroups(void);
BOOL ViewGroupMenu(struct GroupInfoList *act_pGI, int *command);
BOOL UnSubscribeMenu(struct GroupInfoList *act_pGI, int *command);
void ListGroups(void);
void FreeNIL(void);
void FreeGIL(void);
int GetGroups(void);
void InitLowestHighest(struct GroupInfoList *pGI, ULONG lowest, ULONG highest);
struct GroupInfoList *
AllocNewGroup(char *newsgroup, ULONG lowest, ULONG highest, char moderated,
              struct GroupInfoList *prevGI, struct GroupInfoList *nextGI);
int ReReadActiveFile(void);
int ReadActiveFile(void);
void ResetGroupList(void);
/* proto$end */

#include "proto/strprocs_proto.h"
#include "proto/ArnIO_proto.h"
#include "proto/functions_proto.h"

extern char *noaction;
extern ULONG FirstArtNr, LastArtNr; /* used in readgroup */
extern ULONG UserlastArtNr, NrOfNewArts;

  /*********************************/

struct GroupInfoList **sortedgrouplist;

  /*********************************/


/* used in ReReadActiveFile */
struct simple_list
{
  struct GroupInfoList *pGI;
  struct simple_list *nextP;
};

                         /*****************************************/
                         /*****************************************/

/** recursive procedure, so watch out for stack **/
void FreeMarks(struct Markers *rootmark)
{
  if (rootmark)
  {
    if (rootmark->NextMark)
       FreeMarks(rootmark->NextMark);

    FreeMem(rootmark->Mark, (ULONG) MAXLINE+1L);
    FreeMem(rootmark, (ULONG) sizeof(struct Markers));
  }
}


struct Markers *NewMark(struct GroupInfoList *pGI)
/* produces new mark-string from art_structs and returns it */
/* if pGI is not NULL, the whole group is marked at once (catchup) */
{
  struct art_struct *pArt;
  int i, stmplen= 0;
  ULONG brange, erange, count;
  char stmp[30], *mytmp;
  struct Markers *pmark, *rootmark;

  rootmark= AllocMem((ULONG) sizeof(struct Markers),MEMTYPE);
  if (!rootmark)
  {
    ErrorHandle(NOMEM,NULL);
    return NULL;
  }

  pmark= rootmark;
  pmark->NextMark= NULL; /* No further MarkString, yet */
  pmark->lastpos= 0;     /* no actual Markstring */

  pmark->Mark= AllocMem((ULONG)MAXLINE+1UL,MEMTYPE); /* alloc Markstring */
  if (pmark->Mark==NULL)
  {
    ErrorHandle(NOMEM,NULL);
    FreeMem(rootmark,(ULONG) sizeof(struct Markers));
  }
  else
  {
    mytmp=pmark->Mark;
    mytmp[0]='\0';

    if (pGI!=NULL) /* catch up */
    {
      sprintf(mytmp,"%lu-%lu", pGI->firstArtNr, pGI->lastArtNr);
      pmark->lastpos= strlen(mytmp);
      pmark->NextMark= NULL;
      return rootmark;
    }

    brange= erange= 0UL;
    for (pArt= RootArt, i= 0, count= 0L; count<NrOfArticles;
                                         count++, pArt= &RootArt[count])
    {
      if (brange==0UL) /* no range so far */
      {
        if (pArt->Status & READ)
           brange= erange= count+1; /* brange must be greater than 1,
                                       otherwise first article is ignored */
      }
      else /* brange was already set */
      if (pArt->Status & READ)
      {
        if (erange==count) /* That's a range */
           erange++;
      }

      if (erange!=0L && (count==NrOfArticles-1 || !(pArt->Status & READ)))
         /* We got an unread article or we are at the end */
      {
        brange+= FirstArtNr-1; /* brange,erange were relative before */
        erange+= FirstArtNr-1; /* so make them absolute */

        if (i!=0)        /* No Commas at the beginning */
           mytmp[i++]= ',';

        if (erange>brange) /* real range */
           sprintf(stmp,"%lu-%lu",brange,erange); /* write it into */
        else                                      /* buffer first */
           ltostr(brange,stmp,30);

        stmplen= strlen(stmp); /* How long is our buffer? */

        if (i+stmplen>MAXLINE) /* new mark doesn't fit, allocate new structure */
        {
          mytmp[i]= '\0'; /* if last char was a comma */
          pmark->lastpos= i; /* lastpos is the pos of the terminating \0 */

          pmark->NextMark= AllocMem((ULONG) sizeof(struct Markers),MEMTYPE);

          if (!pmark->NextMark)
          {
            ErrorHandle(NOMEM,NULL);
            break; /* exit for */
          }
          else
          {
            pmark= pmark->NextMark; /* actual markstruct */
            pmark->NextMark= NULL;  /* No further MarkString, yet */
            pmark->lastpos= 0;      /* no actual Markstring */
            mytmp=pmark->Mark= AllocMem((ULONG)MAXLINE+1L,MEMTYPE);
               /* alloc Markstring */
            if (!mytmp)
            {
              ErrorHandle(NOMEM,NULL);
              break; /* exit for */
            }
          }
          i= 0; /* reset i */
        }

        strcpy(mytmp+i,stmp);  /* yes, copy it into the MarkString */
        i+= stmplen;            /* set i to pos of \0 */
        pmark->lastpos= i; /* lastpos is the pos of the terminating \0 */
        brange= erange= 0UL; /* reset range */
      } /* end unread article or end */
    } /* end of for-loop */

    if (rootmark->lastpos==0) /* nothing to mark */
    {
      if (rootmark->Mark)
          FreeMem(rootmark->Mark,(ULONG)MAXLINE+1L); /* Free Markstring */
      FreeMem(rootmark,(ULONG) sizeof(struct Markers));
      rootmark= NULL;
    }
#ifdef DEBUG
    for (pmark= rootmark,i= 0; pmark!=NULL; pmark= pmark->NextMark,i++)
      printf("NewMark%02d:%s:%d\n",i,pmark->Mark,pmark->lastpos);
#endif
  } /* end else */

  return rootmark;
}


int
UpdateGroupList(void)
/* copies old GROUPLIST file to GROUPLIST.bak and dumps out new GROUPLIST */
{
  struct Markers *pmark;

  struct GroupInfoList *pGI;

  sprintf(tmp,"%s.bak",grouplistfile);
  DeleteFile(tmp);
  Rename(grouplistfile,tmp);

  if ((fp= fopen(grouplistfile,"w"))==NULL)
  {
    ErrorHandle(GROUPLIST,grouplistfile);
    return 2;
  }

  for (pGI= RootGI; pGI!=NULL; pGI=pGI->NextGI)
  {
    if (pGI->Status & UNSUBSCRIBED)
       fprintf(fp, "%s! ", pGI->Name);
    else
       fprintf(fp, "%s: ", pGI->Name);

    if (pGI->lastArtNr)
    {
      if (GeneralBool & UPDATE_NEWMARKS)
         fprintf(fp,";%lu;",pGI->lastArtNr); /* Update UserlastArtNr!! */
      else
      if (pGI->UserlastArtNr)
         fprintf(fp,";%lu;",pGI->UserlastArtNr); /* Retain UserlastArtNr!! */
    }

    if (pGI->lastArtNr-pGI->firstArtNr>=0 && (pmark=(pGI->Marks))!=NULL)
    {
      while (pmark && pmark->Mark)
      {
        fprintf(fp, " %s\n",pmark->Mark);
#ifdef DEBUG
        printf(":%s:\n",pmark->Mark);
#endif
         pmark=pmark->NextMark;
      }
    }
    else
    fprintf(fp, "\n");
  } /* end for */
  if (fp) fclose(fp);
  return 0;
}


void
CheckLibMarks(char *Markstr, struct GroupInfoList *pGI)
/* Marks articles as read (art_struct list), using ASCII-Markstr */
/* FirstArtNr and NrOfArticles must be set before calling this procedure */
{
  struct art_struct *pArt=RootArt;
  int i;
  ULONG brange= 0UL, erange= 0UL;
  register ULONG x= 0UL;

  /* Parse Markstr */
  /* 0 is a non valid (absolute) article number */

  if (Markstr!=NULL)
     for (i= 0; Markstr[i]!='\0'; i++)
     {
       x= 0UL;

       while (isdigit(Markstr[i]))
       {
         x= x*10UL + (ULONG) (Markstr[i]-'0');
         i++;
       }
#ifdef DEBUG
       printf("x:%lu\n",x);
#endif
       if (Markstr[i]=='-')  /* real range follows */
       {
         brange= erange= x;  /* store beginning of range */
       }
       else                  /* no range OR we just read the erange! */
       {
         if (erange!=0UL) /* we have a beginning of a range, */
         {                /* because erange is set only if we found a '-' before x */
           if (x>=brange) /* seems to be a range */
           {
             if (brange<FirstArtNr) /* brange is not in valid article numbers */
                brange= FirstArtNr; /* set it to first article number */

             if (brange>LastArtNr) brange= erange= x= 0UL; /* brange out of range */

             if (x<FirstArtNr)    /* erange is not in valid range */
             {
               brange= erange= 0UL; /* do not count it */
             }
             else
             {
               if (x>LastArtNr) x= LastArtNr;
               erange= x;        /* real range */
             }
           }
           else  /* x<brange */
              erange= brange;    /* nonsens range */
         }
         else /* erange==0 -> x is single number (we didn't find a '-' before) */
         {
           if (x<FirstArtNr || x>LastArtNr) x= 0UL; /* x out of valid range, ignore it */
           erange= brange= x;
         }

         /* precondition: brange>=FirstArtNr && brange<=erange<=LastArtNr  */
         /* or brange==erange==0UL */

         if (brange!=0UL) /* don't ignore range -> precondition: brange>=FirstArtNr */
         for(x= brange-FirstArtNr ; x<=erange-FirstArtNr; x++)
         { /* make x relative article number for indexing */
           if (x<NrOfArticles) /* x>=0 is redundant because of ULONG */
           {
             pArt= RootArt+x;
             pArt->Status|= READ;
             if (x+FirstArtNr<=pGI->UserlastArtNr || pGI->UserlastArtNr==0)
                pArt->Status&= ~NEW; /* don't mark read articles as new */
#ifdef DEBUG
             printf("%s: Marked %06lu as read\n",newsgroup, pArt->ArticleNr);
#endif
           }
         }
         brange= erange= 0UL;  /* reset range that is to mark */
         if (Markstr[i]=='\0')  /* stop */
           i--; /* The next test will end the loop */
       } /* end if we parsed erange or single number */
     } /* end for all chars in MarkStr */
}


ULONG
CountMarks(char *Markstr) /*** FirstArtNr and LastArtNr must be set!!! ***/
                          /*** UserlastArtNr and NrOfNewArts           ***/
                          /***********************************************/
{
  int i;
  ULONG brange= 0UL, erange= 0UL;
  register ULONG x= 0UL;
  ULONG NrOfReadArts= 0UL;

  /* Parse Markstr */

  if (Markstr!=NULL)
     for (i= 0; Markstr[i]!='\0'; i++)
     {
       x= 0UL;

       while (isdigit(Markstr[i]))
       {
         x= x*10UL + (ULONG) (Markstr[i]-'0');
         i++;
       }
#ifdef DEBUG
       printf("x:%lu\n",x);
#endif
       if (Markstr[i]=='-')  /* real range follows */
       {
         brange= erange= x;
       }
       else                /* end of range or single number */
       {
         if (erange!=0UL)   /* we have brange, now we read erange */
         {
           if (x>=brange)
           {
             if (brange<FirstArtNr)
                brange= FirstArtNr;

             if (brange>LastArtNr) brange= erange= x= 0UL;

             if (x<FirstArtNr)    /* erange is not in valid range */
             {
               brange= erange= 0UL; /* do not count it */
             }
             else
             {
               if (x>LastArtNr) x= LastArtNr;
               erange= x;        /* real range */
             }
           }
           else
              erange= brange;    /* nonsens range */
         }
         else                /* single number */
         {
           if (x<FirstArtNr || x>LastArtNr) x= 0UL;
           erange= brange= x;
         }

         if (brange) /* valid range or number */
         {
            NrOfReadArts+= erange-brange+1UL;
#ifdef DEBUG
            printf("counted:%lu-%lu\n",brange,erange);
#endif
            if (brange>UserlastArtNr)
               NrOfNewArts-= (erange-brange+1UL);
            else
            if (erange>UserlastArtNr)
               NrOfNewArts-= (erange-UserlastArtNr);
         }
         brange= erange= 0UL;  /* reset range that is to mark */
         if (Markstr[i]=='\0')  /* stop */
           i--; /* The next test will end the loop */
       }
     }
#ifdef DEBUG
   printf("==> %s: %lu\n", Markstr, NrOfReadArts);
#endif
   return NrOfReadArts;
}


BOOL
Unsubscribe(struct GroupInfoList *pGI)
/* a pGI == NULL will unsubscribe silently all selected groups */
{
  if (pGI)
  {
    sprintf(tmp,"Really UNSUBSCRIBE %s? [y,N]:",pGI->Name);
    StatusLine();
    DispMsg(tmp);
    *tmp= (*GetKey)();
    ClrLine();
    if (toupper(tmp[0])=='Y')
    {
      pGI->Status|= (UNSUBSCRIBED | G_NEEDS_REFRESH);
      return true;
    }
    return false;
  }
  else
  {
    for(pGI= RootGI; pGI!=NULL; pGI= pGI->NextGI)
       if (pGI->Status & SELECTED_GROUP)
       {
         pGI->Status|= (UNSUBSCRIBED | G_NEEDS_REFRESH);
         pGI->Status&= ~SELECTED_GROUP;
       }
  }
}


void
Resubscribe(struct GroupInfoList *pGI)
{
  if (pGI)
  {
    StatusLine();
    sprintf(tmp,"%s resubscribed.");
    DispMsg(tmp);
    pGI->Status|= G_NEEDS_REFRESH;
    pGI->Status&= ~UNSUBSCRIBED;
  }
  else
  for(pGI= RootGI; pGI!=NULL; pGI= pGI->NextGI)
     if (pGI->Status & SELECTED_GROUP)
     {
       pGI->Status|= G_NEEDS_REFRESH;
       pGI->Status&= ~(SELECTED_GROUP | UNSUBSCRIBED);
     }
}


struct GroupInfoList *
FindGroup(char *s)
/* binary search procedure, _not_ case-sensitive! */
/* returns NULL pointer if not found, uses sortedgrouplist[] */
{
  ULONG lower,upper,mid;
  int result;

  lower= 0;
  upper= NrOfGroups-1;

  while(lower<upper)
  {
    mid= (upper+lower)/2; /* bisect */

    if ((result= stricmp(sortedgrouplist[mid]->Name,s))>0)
       upper= mid==0 ? 0 : mid-1; /* middle element was greater than searchstring */
    else             /*   so we can omit elements mid..end */
    if (result<0)   /* middle element was less than searchstring */
       lower= mid+1;
    else /* hit! */
      return sortedgrouplist[mid];
  }

  if (lower==upper && stricmp(sortedgrouplist[lower]->Name,s)==0)
     return sortedgrouplist[lower];
  else
     return NULL;
}


void
MoveGroups(struct GroupInfoList *pGI)
/* move selected groups to pGI group */
{
  struct GroupInfoList *tmp_pGI, *prevGI, *nextGI;
  ULONG index;

  for (index= 0; index<NrOfGroups; index++)
  {
    tmp_pGI= sortedgrouplist[index];
    if (tmp_pGI->Status & SELECTED_GROUP)
    {
      if (tmp_pGI!=pGI)
      {
        /* let the preceeding node (from the node which we want to move)
           point to the next node where this node to move currently points to:
           or simply: !remove tmp_pGI!
        */
        if (prevGI= tmp_pGI->PrevGI)
           prevGI->NextGI= tmp_pGI->NextGI;
        if (nextGI= tmp_pGI->NextGI)
        {
          nextGI->PrevGI= prevGI;
          if (prevGI==NULL && RootGI==tmp_pGI) /* RootGI was tmp_pGI */
             RootGI= nextGI; /* RootGI must be set to the successor of tmp_pGI */
        }
        /* now insert tmp_pGI _behind_ pGI! */
        tmp_pGI->NextGI= pGI->NextGI;
        pGI->NextGI= tmp_pGI;
        tmp_pGI->PrevGI= pGI;
        if (nextGI= tmp_pGI->NextGI)
           nextGI->PrevGI= tmp_pGI;
        pGI= tmp_pGI; /* next selected group should be inserted behind this one */
      }
      tmp_pGI->Status&= ~SELECTED_GROUP; /* unselect */
    }
  }
}


void
SortGroups(void)
{
  struct GroupInfoList *pGI, *prevGI;
  ULONG index;

  index= 0UL;
  prevGI= NULL;
  for (RootGI= sortedgrouplist[index]; index<NrOfGroups; index++)
  {
    pGI=sortedgrouplist[index];
    pGI->PrevGI= prevGI;
    if (prevGI) prevGI->NextGI= pGI;
    prevGI= pGI;
  }
  pGI->NextGI= NULL;
}


BOOL
ViewGroupMenu(struct GroupInfoList *act_pGI, int *command)
/* returns leave SelectPage, and alters *command */
{
  char Key;
  BOOL leave_SelectGroups= true;

  StatusLine();
  DispMsg("Show U)nsubscribed, R)ead, M)ove (behind cursor), S)ort, Q)uit:");
  Key=GetUpKey();
  ClrLine();
  switch(Key)
  {
    case 'U':*command= SHOW_UNSUBS;
             break;
    case 'R':*command= SHOWALL;
             break;
    case 'M':MoveGroups(act_pGI);
             *command= REBUILD;
             break;
    case 'S':SortGroups();
             *command= REBUILD;
             break;
    default:
             DispMsg(noaction);
             leave_SelectGroups= false;
             *command= NOP;
             break;
  }
  return leave_SelectGroups;
}


BOOL
UnSubscribeMenu(struct GroupInfoList *act_pGI, int *command)
{
  char key1, key2;
  BOOL leave_SelectGroups= false;

  StatusLine();
  DispMsg("U)nsubscribe, R)esubscribe, Q)uit:");
  key1= GetUpKey();
  ClrLine();
  if (key1=='Q')
  {
    *command= NOP;
    DispMsg(noaction);
    return leave_SelectGroups;
  }
  DispMsg("S)elected groups, G)roup at cursor (default):");
  key2= GetUpKey();
  ClrLine();
  if (key2=='S') act_pGI= NULL;

  if (key1=='U')
     Unsubscribe(act_pGI);
  else
     Resubscribe(act_pGI);

  if (act_pGI)
    *command= REFRESHLINE;
  else
    *command= REFRESHLINES;

  return leave_SelectGroups;
}


/* Lists all newsgroups in memory */
void ListGroups(void)
{
  struct GroupInfoList *pGI;
  char c,subscribed, moderated;
  int i;

  for(pGI= RootGI, i= 1; pGI!=NULL; pGI=pGI->NextGI, i++)
  {
    if (i%MaxScrLines==0)
    {
      c= WaitForSpace(NULL);
      c= toupper(c);
    }

    if (c=='Q') break;

    if (pGI->Status & UNSUBSCRIBED)
       subscribed= 'U';
    else
       subscribed= ' ';

    if (pGI->Status & MODERATED)
       moderated= 'M';
    else
       moderated= ' ';

    sprintf(tmp,"%-58s: %c %c %6ld-%-6ld\n", pGI->Name,subscribed,moderated,
                                             pGI->firstArtNr,pGI->lastArtNr);

    (*ConWrite)(tmp);
    if (pGI->NextGI==NULL)
      WaitForSpace(NULL);
  }
}


/* from grouplist.c */
void FreeMarks(struct Markers *);

/***********************/


/* Frees the whole art_structs */
void
FreeNIL(void)
{
  pArt= RootArt;

  if (RootArt && NrOfArticles)
       FreeMem(RootArt,(ULONG) sizeof(struct art_struct)*NrOfArticles);
  RootArt= NULL;
}


void
FreeGIL(void)
{
  struct GroupInfoList *pGI, *plGI;

  pGI= RootGI;

  while (pGI!=NULL)
  {
    plGI= pGI;
    pGI= pGI->NextGI;
    FreeString(plGI->Name);
#ifdef DEBUG
    sprintf(tmp,"Freeing Marks: %s\n", plGI->Name);
    (*ConWrite)(tmp,strlen(tmp));
#endif
    if (plGI->Marks)
      FreeMarks(plGI->Marks);
#ifdef DEBUG
    sprintf(tmp,"Freeing newsgroup: %s\n", plGI->Name);
    (*ConWriteLn)(tmp);
#endif
    FreeMem(plGI,(ULONG) sizeof(struct GroupInfoList));
  }

  RootGI= NULL;

  if (sortedgrouplist)
  {
    FreeMem(sortedgrouplist, (ULONG) (sizeof(struct GroupInfoList *) * NrOfGroups));
    sortedgrouplist= NULL;
  }
}


/* Read in the Grouplist. Lines longer than MAXLINE chars are truncated! */
/* The active-file is already read and grouplist is present */
/* GroupListFP is already open */
/* rc=1 means file not found, 0 ok, other signals error */
int
GetGroups(void)
{
  char *c, *help;
  struct GroupInfoList *pGI,*prevGI, *tmp_pGI, **ppGI;
  struct Markers *pmark;
  BOOL unsubscribed;
  ULONG lines= 0L;
  ULONG index, infolines=0UL;

#ifdef DEBUG
  struct Markers *pmark2;
#endif

  ppGI= &RootGI;
  prevGI= (*ppGI)->PrevGI;

  if ((GroupListFP= fopen(grouplistfile,"r"))!= NULL)
  {
    while (fgets(tmp, SUPERLINE, GroupListFP))
    {
      lines++;
      c=tmp;
      if (*c!=' ' && *c!='\t')  /* if line doesn't start with white space */
      {
        c=NextWhiteSpc(c);
        if (c!=tmp) /* not a blank line */
        {
          pmark=NULL;

          unsubscribed= false;
          if (*(c-1)=='!')       /* unsubscribed newsgroup */
             unsubscribed= true;
          else
          if (*(c-1)!=':')
            *tmp= '\0';     /* produce Error: Non valid entry */

          *(c-1)= '\0';        /* stop it at : or ! */

          if (*tmp=='\0')
          {
            sprintf(tmp,"in line %lu!",lines);
            ErrorHandle(NOVALIDGROUP,tmp);
            return 2;
          }
          else
          {
            pGI= FindGroup(tmp); /* find group in sortedgrouplist */
            if (!pGI)
            {
              ErrorHandle(OBSOLETEGROUP,tmp);
              infolines++;
              if (infolines%MaxScrLines==0)
                 WaitForSpace(NULL);
              continue;
            }
            else
            {
              if (unsubscribed) pGI->Status|= UNSUBSCRIBED;
            }
            pGI->Status&= ~NEW_GROUP; /* this group is already in GROUPLIST */

            if (tmp_pGI= pGI->PrevGI) /* pointers to this struct are invalid */
               tmp_pGI->NextGI= NULL;
            if (tmp_pGI= pGI->NextGI)
               tmp_pGI->PrevGI= NULL;

            pGI->PrevGI= prevGI;
            *ppGI= pGI;           /* NextGI pointer of previous struct points to pGI */
            pGI->NextGI= NULL;
            ppGI= &pGI->NextGI;
            prevGI= pGI;  /* next time, this structure is preceeder */
          }

          c= skpwspc(c);     /* get the marks */
          if (c && *c!='\0')
          {
            if (*c==';')   /* last article mark follows? */
            {
              help= ++c;
              c= strchr(c,';');
              if (c!=NULL)
              {
                *c= '\0';
                if (islong(help))
                {
                   pGI->UserlastArtNr= (ULONG) atol(help);
                   if (pGI->UserlastArtNr<pGI->firstArtNr)
                      pGI->UserlastArtNr= pGI->firstArtNr-1L;
                   if (pGI->UserlastArtNr<pGI->lastArtNr)
                      pGI->Status|= NEWARTS_GROUP;
                   else
                      pGI->UserlastArtNr= pGI->lastArtNr;
                }
                c++;  /* WARNING! *c maybe \0! */
                c= skpwspc(c);
              }
              else
                c= help;
            }

            cuteoln(c);
            pGI->Marks= AllocMem((ULONG) sizeof(struct Markers), MEMTYPE);
            if (pGI->Marks)
            {
              pmark= pGI->Marks;
              pmark->NextMark= NULL;
              if (pmark->Mark= AllocMem((ULONG) MAXLINE+1L, MEMTYPE))
              {
                *(pmark->Mark)= '\0';
                strncpy(pmark->Mark,c,MAXLINE); /* at least MAXLINE bytes */
                *(pmark->Mark+MAXLINE)= '\0';    /* as a precaution */
                pmark->lastpos= strlen(c);
              }
              else
                ErrorHandle(NOMEM,NULL);
            }
            else
            {
              ErrorHandle(NOMEM,NULL);
              pmark= NULL;
            }
          } /* if there are marks */
        } /* it was not a blank line */
        else
        { /* it was a blank line, so skip it and read in the next line */
          continue;
        }
      }
      else /* we've got line continuation */
      {
        c= skpwspc(c);
        cuteoln(c);
        if (*c!='\0') /* Ignore blank-Lines */
        {
          if (!pmark)
           ErrorHandle(STRANGE,"pmark");
          else
          if (isdigit(*c))
          {
            pmark->NextMark= AllocMem((ULONG) sizeof(struct Markers), MEMTYPE);
            if (pmark->NextMark)
            {
              pmark= pmark->NextMark;
              pmark->NextMark= NULL;
              if (pmark->Mark= AllocMem((ULONG) MAXLINE+1L, MEMTYPE))
              {
                strncpy(pmark->Mark,c,MAXLINE); /* at least MAXLINE bytes */
                *(pmark->Mark+MAXLINE)= '\0';    /* as precaution */
                pmark->lastpos= strlen(c);
              }
              else
                ErrorHandle(NOMEM,NULL);
            }
          }
          else
          {
            sprintf(tmp,"in line %lu!",lines);
            ErrorHandle(NOVALIDGROUP,tmp);
          }
        }
      }
#ifdef DEBUG
      printf("Got:%s:", pGI->Name);
      for (pmark2= pGI->Marks; pmark2!=NULL; pmark2=pmark2->NextMark)
        printf("NewMark:%s:%d\n",pmark2->Mark,pmark2->lastpos);
      printf("\n");
#endif
    }
    fclose(GroupListFP);

    /* new groups are inserted at the top of the GROUPLIST */
    /* The newest is the first one */
    ppGI= &RootGI;
    if (*ppGI) /* root node is now in ppGI */
    {
      for (index=0UL; index<NrOfGroups; index++)
          if ((tmp_pGI= sortedgrouplist[index]) && (tmp_pGI->Status & NEW_GROUP))
          { /* tmp_pGI is not NULL, so ppGI!=NULL */
            tmp_pGI->NextGI= *ppGI;          /* new root is linked to old root */
            if ((tmp_pGI->PrevGI= (*ppGI)->PrevGI))  /* previous of new root should always be zero */
               ErrorHandle(STRANGE,"RootGI_prev!=NULL");
            (*ppGI)->PrevGI= tmp_pGI;        /* old root is linked to new root */
            *ppGI= tmp_pGI;                  /* make new root */
          }
    }
  }
  else
   return 1; /* no GROUPLIST found */

  if (GlobalERROR!=OK && infolines%MaxScrLines)
     WaitForSpace(NULL);

  return 0;
}


void
InitLowestHighest(struct GroupInfoList *pGI, ULONG lowest, ULONG highest)
/* Initialize lowest and highest article numbers in newsgroup pGI
   pGI _must_ be not NULL */
{
  pGI->firstArtNr= lowest;
  pGI->lastArtNr= highest;
  /* 0 is a non existing or invalid article number! */
  /* case lowest==highest==0     -> empty group */
  /* case lowest==0 && highest>0 -> set lowest to 1L */
  /* case lowest>highest         -> empty group */

  if (lowest==0UL && highest) pGI->firstArtNr= 1UL;
  else
  if (lowest > highest) pGI->firstArtNr= highest+1; /* empty group! */
}


struct GroupInfoList *
AllocNewGroup(char *newsgroup, ULONG lowest, ULONG highest, char moderated,
              struct GroupInfoList *prevGI, struct GroupInfoList *nextGI)
/* Allocates GroupInfoList structure for a new group named <newsgroup>
   with highest and lowest articles, predecessor <prevGI> and successor <nextGI>
   NULL is returned if allocation failed.
   The name <newsgroup> is copied into a new string */
{
  struct GroupInfoList *pGI;

  pGI= (struct GroupInfoList *)
       AllocMem ((ULONG) sizeof(struct GroupInfoList), MEMTYPE);

  if (pGI)
  {
    pGI->Name= CopyString(newsgroup);  /* copy newsgroup-name */
    pGI->Marks= NULL;              /* Marks are read from GROUPLIST */
    InitLowestHighest(pGI,lowest,highest);
    pGI->OldlastArtNr= pGI->UserlastArtNr= pGI->firstArtNr-1UL;

    pGI->Status= (LASTBIT | NEW_GROUP); /* LASTBIT is always set! */
    if (toupper(moderated)=='M')
       pGI->Status|=MODERATED;
    pGI->RootSubjP= NULL;
    pGI->PrevGI= prevGI;
    pGI->NextGI= nextGI;
  }

  return pGI;
}


/* reads in the active file,
   creates the initial (sorted grouplist),
   and sets MODERATED Flags in pGI-LIST
   as well as firstArtNr, lastArtNr                              */
int
ReadActiveFile(void)
{
  FILE *activefp;
  char moderated, *ngtmp;
  ULONG highest,lowest;
  ULONG index;
  int rc= 1;
  struct GroupInfoList *pGI, **ppGI, *prevGI=NULL, *tmppGI;

  if (!(ngtmp= AllocMem(MAXLINE,MEMTYPE)))
  {
    ErrorHandle(NOMEM,NULL);
    return GlobalERROR;
  }

  NrOfGroups= 0L;

  RootGI= NULL;
  ppGI= &RootGI;  /* first structure is allocated in RootGI */

  activefp= fretryopen(activefile,"r");
  if (activefp)
  {
    setvbuf(activefp,mBuffer,_IOFBF, M_BUFFERSIZE);
    while (fgets(tmp, SUPERLINE, activefp))
    {
      highest= lowest= 0UL;
      sscanf(tmp,"%s %lu %lu %c",ngtmp,&highest,&lowest,&moderated);
      if (*ngtmp) /* not empty */
      { /* prevGI points always to the last element in the list */
        if (prevGI && (rc= stricmp(ngtmp, prevGI->Name))<0)
        { /* new group must be inserted earlier into list */
          for (pGI= RootGI; pGI!=NULL && stricmp(ngtmp, pGI->Name)>0; pGI= pGI->NextGI)
              ;
          if (pGI) /* we've found the place to insert */
          {
            /* pGI now contains a group which must be appended behind ngtmp */
            tmppGI= pGI->PrevGI; /* store predecessor */
            if (tmppGI) /* predecessor exists */
            {
              tmppGI->NextGI= AllocNewGroup(ngtmp, lowest, highest, moderated, tmppGI, pGI);
              if ((pGI->PrevGI= tmppGI->NextGI)==NULL)
              { /* allocation failed */
                ErrorHandle(NOMEM,NULL);
                break;
              }
            }
            else /* pGI was RootGI, so ngtmp must be new RootGI */
            {
              tmppGI= RootGI;
              RootGI= AllocNewGroup(ngtmp, lowest, highest, moderated, NULL, tmppGI);
              tmppGI->PrevGI= RootGI;
            }
            NrOfGroups++;
          }
          else /* we traversed the whole list but didn't find an entry */
          {
            ErrorHandle(STRANGE,"grouplist");
          }
        }
        else /* allocate new group and append it */
        {
          if (rc>0) /* don't allocate double groups twice */
          {
            /* the groupinfolist structure gets allocated here */
            *ppGI= AllocNewGroup(ngtmp, lowest, highest, moderated, prevGI, NULL);

            if (*ppGI==NULL) /* oops! */
            {
              ErrorHandle(NOMEM,NULL);
              break;
            }

            NrOfGroups++;
            prevGI= *ppGI;
            /* set ppGI to next pointer of this structure */
            ppGI= &(*ppGI)->NextGI;
          }
        }
      }
    }

    *ppGI= NULL; /* terminate list */

    fclose(activefp);
  }
  else
  {
    ErrorHandle(NOACTIVE,NULL);
  }

  if (ngtmp) FreeMem(ngtmp,MAXLINE);

  if (NrOfGroups)
  {
    if (sortedgrouplist=
        AllocMem((ULONG) (NrOfGroups*sizeof(struct GroupInfoList *)),MEMTYPE))
    {
      for(pGI= RootGI, index= 0UL; pGI!=NULL; pGI= pGI->NextGI, index++)
      {
         sortedgrouplist[index]= pGI;
      }
    }
    else
    {
      ErrorHandle(NOMEM,NULL);
    }
  }
  else
    ErrorHandle(NOACTIVE,NULL);

  return GlobalERROR; /*all ok*/
}


/* rereads the active file,
   recreates the sorted grouplist, (the order of the grouplist is preserved)
   and updates firstArtNr, lastArtNr, MODERATED Flags in pGI-LIST;
   NrOfGroups is updated! */
int ReReadActiveFile(void)
{
  FILE *activefp;
  char moderated, *ngtmp;
  ULONG highest,lowest;
  ULONG index, tmp_NrOfGroups= 0UL;
  struct GroupInfoList *pGI, **ppGI, *tmp_pGI;
  struct simple_list *roottmp_list= NULL, **pptmp_list, *ptmp_list,
                     *p2tmp_list, *prevtmp_list;


  if (!(ngtmp= AllocMem(MAXLINE,MEMTYPE)))
  {
    ErrorHandle(NOMEM,NULL);
    return GlobalERROR;
  }

  tmp_NrOfGroups= 0UL;
  pptmp_list= &roottmp_list;
  ptmp_list= NULL; /* points to last element of list */

  if (activefp= fretryopen(activefile,"r"))
  {
    while (fgets(tmp, SUPERLINE, activefp))
    {
      highest= lowest= 0UL;

      if (sscanf(tmp,"%s %lu %lu %c",ngtmp,&highest,&lowest,&moderated) && *ngtmp)
      { /* not empty */
        if (pGI= FindGroup(ngtmp))
        { /* we have this group already in our list => update highest,lowest */
          InitLowestHighest(pGI, lowest, highest);
        }
        else
        { /* not found in our previous active-file => newgroup */
          pGI= AllocNewGroup(ngtmp, lowest, highest, moderated, NULL, NULL);
        }

        if (pGI && pGI->UserlastArtNr<pGI->lastArtNr) /* set the 'N' flag */
           pGI->Status|= NEWARTS_GROUP;

        /* allocate new member of temporary list */
        /* this list is used to set up the new sortedgrouplist-array */
        /* ptmp_list points to the last element in the list */
        if (ptmp_list && stricmp(pGI->Name,ptmp_list->pGI->Name)<0)
        { /* New group must be inserted earlier into list */
          prevtmp_list= NULL;
          for (p2tmp_list= roottmp_list; p2tmp_list!=NULL &&
                                        stricmp(pGI->Name,p2tmp_list->pGI->Name)>0;
                                        p2tmp_list= p2tmp_list->nextP)
              prevtmp_list= p2tmp_list;

          if (p2tmp_list) /* we've found a place to insert */
          {
            if (prevtmp_list) /* not to insert before root */
            {
              pptmp_list= &prevtmp_list->nextP;
            }
            else
              pptmp_list= &roottmp_list; /* this new node becomes new root */
          }
          else /* can't be */
            ErrorHandle(STRANGE,"rereadact");
        }
        else
          p2tmp_list= NULL; /* no successor to append */

        if (*pptmp_list= AllocMem((ULONG) sizeof(struct simple_list), MEMTYPE))
        {
          if (p2tmp_list) /* append successor */
             (*pptmp_list)->nextP= p2tmp_list;
          else
             ptmp_list= *pptmp_list; /* set pointer to last node */

          (*pptmp_list)->pGI= pGI;
          pptmp_list= &ptmp_list->nextP; /* point to last node by default */
          tmp_NrOfGroups++;
        }
        else
          ErrorHandle(NOMEM,NULL);
      }
    }

    *pptmp_list= NULL; /* terminate list */

    fclose(activefp);
  }
  else
   ErrorHandle(NOACTIVE,NULL);

  if (ngtmp) FreeMem(ngtmp,MAXLINE);

  /* free old sortedgrouplist */
  if (sortedgrouplist)
     FreeMem(sortedgrouplist, NrOfGroups*sizeof(struct GroupInfoList *));

  /* alloc & fill new sortedgrouplist, free tmp_list */
  NrOfGroups= tmp_NrOfGroups;
  if (sortedgrouplist=
      AllocMem((ULONG) (NrOfGroups*sizeof(struct GroupInfoList *)),MEMTYPE))
  {
    for(ptmp_list= roottmp_list, index= 0UL; ptmp_list!=NULL; index++)
    {
      sortedgrouplist[index]= ptmp_list->pGI;
      p2tmp_list= ptmp_list;
      ptmp_list= ptmp_list->nextP;
      FreeMem(p2tmp_list,(ULONG) sizeof(struct simple_list));
    }
    /* new groups are inserted at the top of the GROUPLIST */
    /* The newest is the first one, groups that had the NEW_GROUP-flag set
       before ReRead are not inserted again! */
    ppGI= &RootGI;
    if (*ppGI) /* root node is now in ppGI */
    {
      for (index=0UL; index<NrOfGroups; index++)
          if ((tmp_pGI= sortedgrouplist[index]) && (tmp_pGI->Status & NEW_GROUP)
              && tmp_pGI->NextGI==NULL && tmp_pGI->PrevGI==NULL)
          { /* tmp_pGI is not NULL, so ppGI!=NULL */
            tmp_pGI->NextGI= *ppGI;          /* new root is linked to old root */
            if ((tmp_pGI->PrevGI= (*ppGI)->PrevGI))  /* previous of new root should always be zero */
               ErrorHandle(STRANGE,"RootGI_prev!=NULL");
            (*ppGI)->PrevGI= tmp_pGI;        /* old root is linked to new root */
            *ppGI= tmp_pGI;                  /* make new root */
          }
    }
  }
  else
  {
    ErrorHandle(NOMEM,NULL);
    return GlobalERROR;
  }

  return 0; /*all ok*/
}



void
ResetGroupList(void)
{
  FILE *newgrouplistFP;
  char *cend;
  BOOL Okay= false;
  ULONG lines= 0UL;

  if ((GroupListFP= fopen(grouplistfile,"r"))!= NULL)
  {
    sprintf(fnametmp,"%s.n",grouplistfile);

    if ( (newgrouplistFP= fopen(fnametmp,"w"))!= NULL)
    {
      while (fgets(tmp, SUPERLINE, GroupListFP))
      {
        if (isalnum(*tmp))
        {
          cend=NextWhiteSpc(tmp);
          if (cend) *cend='\0';
          fprintf(newgrouplistFP, "%s\n", tmp);
          lines++;
        }
      }
      fclose(newgrouplistFP);
      Okay=true;
    }
    fclose(GroupListFP);
    if (Okay)
    {
      DeleteFile(grouplistfile);
      Rename(fnametmp, grouplistfile);
      printf("Processed %lu lines. Ready.\n", lines);
    }
  }
}
