/************************************************************************
 * kill -killfile functions for Arn.                                    *
 * Arn - (C)opyright by Roland Bless, All rights reserved,              *
 * -------- YOU ARE NOT ALLOWED TO DISTRIBUTE A CHANGED SOURCE -------- *
 ************************************************************************/

#ifndef GLOBALVARS_H
#include "globalvars.h"
#endif
#ifndef INCLUDES_SYM
#include "includes.h"
#endif
#ifndef ENUMS_H
#include "enums.h"
#endif


#ifdef LATTICE_50
#include "proto/raw_proto.h"
#endif

/* Prototypes for functions defined in kill.c */
/* proto$start */
char *StripSubject(char *Subject);
int add_killstring(struct kill_field **rootfield,
                   char *field,
                   char *string,
                   short command,
                   UWORD attributes);
BOOL FoundSubjectInKillList(struct kill_field *plist,
                            char *subject);
struct kill_strings *killed(struct kill_field *plist);
ULONG prekill(struct kill_field *plist,
             struct subj_struct *RootSubjP);
void FreeString_List(struct kill_strings *pstrings);
void FreeKill_List(struct kill_field *plist);
int ReadKillFile(struct kill_field **rootp,
                 char *name);
int AppendKillFile(struct kill_field **rootp,
                   char *filename,
                   char *field,
                   char *string,
                   char *mode);
void CheckMeta_regexp(char *dest,
                      char *source);
void CheckMeta(char *dest,
               char *source);
int AddSubjToKill(struct kill_field **plist,
                   char *filename,
                   char *mode,
                   struct art_struct *pArt);
int AddAuthorToKill(struct kill_field **plist,
                     char *filename,
                     char *mode,
                     struct art_struct *pArt);
void AddMsgIDToKill(struct kill_field **plist,
                    char *filename,
                    char *mode);
void EditKill(void);
void KillOnOff(BOOL global);
void ChangeKillFile(struct art_struct *pArt, int depth);
long DoExpireXfile(char *filename);
void ExpireXfiles(void);
/* proto$end */

#include "proto/ArnIO_proto.h"
#include "proto/functions_proto.h"
#include "proto/getconfig_proto.h"
#include "proto/pager_proto.h"
#include "proto/time_proto.h"
#include "proto/strprocs_proto.h"
#include "proto/dirfunc_proto.h"
#include "proto/followup_proto.h"

extern struct kill_field *LocalKill_List;
extern char *pend, *noaction;

extern struct subj_struct *RootSubjP;
/**************************************/

int xlifetime= 30; /* default-lifetime of X-files entries */

/**************************************/



char *
StripSubject(char *Subject)
/* returns a pointer to the beginning of the Subject-String without leading
   Re:'s or Re^x's. Trailing spaces are eliminated, too. */
{
  char *c;

  while (c= firststr(Subject, "Re:"))
  {
    Subject= skpwspc(c);
  }
  while (c= firststr(Subject, "Re^"))
  {
    if (c= strchr(c,':'))
       Subject= skpwspc(c+1);
    else
    {
      c= firststr(Subject, "Re^");
      Subject= skpwspc(NextWhiteSpc(c));
    }
  }

  /* Now cut trailing blanks! */
  for (c= Subject; *c!='\0'; c++);
  if (c>Subject)
  {
    c--;
    while (c>Subject && isspace(*c)) c--;
    *++c='\0';
  }

  return Subject;
}


/* inserts new string into kill-list */
/* if field is NULL, Subject: is default */
int
add_killstring (struct kill_field **rootfield, char *field, char *string,
                short int command, UWORD attributes)
{
  struct kill_field **ppfield;
  struct kill_strings *pstringlist;
  struct kill_strings **ppstringlist;

  if (field==NULL) field= HeaderLines[HL_SUBJECT]; /* default */

  for(ppfield= rootfield; *ppfield!=NULL && strcmp((*ppfield)->headerline,field);
      ppfield=&(*ppfield)->nextfield);

  if ((*ppfield)==NULL) /* field not in list -> add */
  {
    *ppfield=AllocMem((ULONG) sizeof(struct kill_field), MEMTYPE);
    if (*ppfield)
    {
      (*ppfield)->nextfield= NULL; /* has currently no successor */
      /* create new headerline-list */

      (*ppfield)->headerline= CopyString(field);
      if ((*ppfield)->headerline)
      {
        pstringlist= (*ppfield)->stringlist= AllocMem((ULONG) sizeof(struct kill_strings), MEMTYPE);
        if (pstringlist)
        {
          pstringlist->nextstring= NULL;
          pstringlist->regexp= NULL;
          pstringlist->string= NULL;
          pstringlist->table= NULL;
          pstringlist->attributes= attributes;
          if (attributes & REGULAREXPR)  /* we have a regular expression */
          {
            pstringlist->regexp= regcomp(string);
          }
          else
          {
            int string_len= strlen(string);

            pstringlist->table= malloc(string_len*sizeof(short int));
            /* The table has no closing \0 */
            pstringlist->strlen= string_len;
            KMP_ComputeNextTable(string, string_len, pstringlist->table);
            pstringlist->string= malloc(string_len+1);
            strcpy(pstringlist->string, string);
          }
          pstringlist->command= command;
        }
        else
        {
          ErrorHandle(NOMEM,NULL);
          return 1;
        }
      }
      else
      {
        ErrorHandle(NOMEM,NULL);
        return 1;
      }
    }
    else
    {
      ErrorHandle(NOMEM,NULL);
      return 1;
    }
  }
  else /* found the field/stringlist */
  {
    ppstringlist= &((*ppfield)->stringlist);
    if (ppstringlist==NULL)
    {
      ErrorHandle(STRANGE,"ppstring");
      return 1;
    }
    else
    {
      while ((*ppstringlist)!=NULL)
         ppstringlist= &(*ppstringlist)->nextstring;

      if ((*ppstringlist)==NULL) /* string not in list */
      {
        pstringlist= (*ppstringlist)=AllocMem((ULONG) sizeof(struct kill_strings), MEMTYPE);
        if (pstringlist)
        {
          pstringlist->nextstring= NULL;
          pstringlist->regexp= NULL;
          pstringlist->string= NULL;
          pstringlist->table= NULL;
          if (attributes & REGULAREXPR)  /* we have a regular expression */
          {
            pstringlist->regexp= regcomp(string);
          }
          else
          {
            int string_len= strlen(string);

            pstringlist->table= malloc(string_len*sizeof(short int));
            /* The table has no closing \0 */
            pstringlist->strlen= string_len;
            KMP_ComputeNextTable(string, string_len, pstringlist->table);
            pstringlist->string= malloc(string_len+1);
            strcpy(pstringlist->string, string);
          }
          pstringlist->command= command;
        }
        else
        {
          ErrorHandle(NOMEM,NULL);
          return 1;
        }
      }
    }
  }
}


BOOL
FoundSubjectInKillList(struct kill_field *plist, char *subject)
{
  struct kill_strings *pstrings;

  while (plist) /* as long as there are fields */
  {
    if (!strnicmp(plist->headerline,HeaderLines[HL_SUBJECT],7))
    {
      pstrings=plist->stringlist; /* strings to compare */
      while (pstrings!=NULL)
      {
        if (pstrings->regexp==NULL) /* not a regexp */
        {
          if (stricmp(subject,pstrings->string)==0) return true; /* matched! */
        }
        pstrings= pstrings->nextstring;
      }
      break; /* the Subject: field should only occur once in the kill-list!! */
    }
    plist= plist->nextfield;
  }

  return false;
}



struct kill_strings *
killed(struct kill_field *plist)
/* to be called from pager (GetBufField is used)                */
/* returns NULL if not "killed", else a pointer to kill_strings */
/* this is later used to get the command (JUNK)          */
{
  struct kill_strings *pstrings;
  char *t;
  int rc;

  pend=NULL;

  while (plist) /* as long as there are fields */
  {
#ifdef LAZY /* process only kill-entries that are not "prekilled" */
    if (strnicmp(plist->headerline,HeaderLines[HL_FROM],4) && strnicmp(plist->headerline,HeaderLines[HL_SUBJECT],7))
#endif
    if (t= GetBufField(plist->headerline)) /* Look for field in header */
    {
      pstrings= plist->stringlist; /* strings to compare */

      while (pstrings!=NULL)
      {
        if (pstrings->attributes & CASESENSITIVE)
           casedep= 0xFF;
        else
           casedep= 0x5F;

        if (pstrings->regexp)
        {
          if (regexec(pstrings->regexp,t))
          {
            GetBufField(NULL);
#ifdef DEBUG
            printf("Regex matched!\n");
#endif
            return pstrings;
          }
        }
        else
        {
          if ((rc= KMP_Matcher(t,strlen(t),pstrings->string, pstrings->strlen, pstrings->table))
              !=-1)
          {
#ifdef DEBUG
            printf("KMP matched!\n");
#endif
            GetBufField(NULL);

          }
        }
        pstrings=pstrings->nextstring;
      }
    }
    plist= plist->nextfield;
  }
  GetBufField(NULL);  /* restore HeaderBuffer */

  return NULL;
}



ULONG
prekill(struct kill_field *plist, struct subj_struct *RootSubjP)
/* kills articles before showing them in the article selector */
/* or pre-selects them. Returns number of junked articles */
{
  struct kill_strings *fromstrings=NULL, *subjstrings=NULL, *pstrings=NULL;
  register struct art_struct *pArt;
  register struct subj_struct *subjp;
  char *t;
  FILE *article;
  /* register long comparisons=0L; */
  BOOL match= false;
  short int tlen;
  int rc;
  ULONG nr_of_junked= 0UL;

  pend=NULL;

  if (!plist) return nr_of_junked;

  StatusLine();
  DispMsg("Processing X-file...please wait");

  while (plist) /* as long as there are kill-fields */
  {
    if (!strnicmp(plist->headerline,HeaderLines[HL_FROM],4))
      fromstrings= plist->stringlist; /* Is it a From: list? */
    else
    if (!strnicmp(plist->headerline,HeaderLines[HL_SUBJECT],7)) /* Is it a Subject: list? */
      subjstrings= plist->stringlist; /* Is it a From: list? */

    plist= plist->nextfield;
  }

  if (!fromstrings && !subjstrings) return nr_of_junked; /* No kill-lists */

  /* first kill all subjects */
  for (subjp= RootSubjP; subjp!=NULL; subjp=subjp->NextSubject)
  {
    t= subjp->Subject;
    tlen= strlen(t);
    pstrings= subjstrings;

    while (pstrings) /* for all subjects in X-file */
    {
      if (pstrings->attributes & CASESENSITIVE)
         casedep= 0xFF;
      else
         casedep= 0x5F;

      if (pstrings->regexp) /* regexp */
      {
        if (regexec(pstrings->regexp,t))  /* compare - if match do command */
        {
          match= true;
#ifdef DEBUG
          printf("match:%lu %s %x\n",pArt->ArticleNr, pstrings->regexp->startp[1],pArt->Status);
#endif
        }
      }
      else
      {
#ifdef DEBUG
        printf("%s,%s\n",pstrings->string,t);
#endif
        if ((rc= KMP_Matcher(t,tlen,pstrings->string, pstrings->strlen, pstrings->table))!=-1)
        {
          if (pstrings->attributes & MATCHBEGIN)
          {
            if (StripSubject(t)-t==rc)
            {
               match= true;
            }
          }
          else
            match= true;

#ifdef DEBUG
          if (match) printf("matched!%d\n",pstrings->command);
#endif
        }
      }

      if (match)
      {
        /* now apply this command to all articles with that subject */
        for (pArt= subjp->RootArt; pArt!=NULL; pArt= pArt->NextArt)
        {
          if (!(pArt->Status & READ))
          {
            switch (pstrings->command)
            {
              case SELECT: pArt->Status|= SELECTED_ART;
                           break;
              case JUNK: pArt->Status|= READ;
                         nr_of_junked++;
                         sprintf(fnametmp,"%s/%lu",pathtogroup, pArt->ArticleNr);
                         if (pArt->Status & XREF)
                         {
                           if (article=fopen(fnametmp,"r"))
                           {
                             MarkXRefs(article, false);
                             fclose(article);
                           }
                         }
                         break;
            } /* end of switch */
          }
        }
        match= false;
        break; /* no need to examine further kill entries, */
               /* because this article matched already */
      } /* if match */
      pstrings=pstrings->nextstring; /* continue in kill list */
    } /* while pstrings (continue in X-file) */
  } /* end for */

  for (pArt= RootArt; pArt<=LastArt; pArt++)  /* for all articles */
  {
    if (!(pArt->Status & READ)) /* TODO?: search only all unread articles */
    {
      if(t=pArt->From) /* look for From: match */
      {
        tlen= strlen(t);
        pstrings= fromstrings;
        while (pstrings)
        {
          if (pstrings->regexp) /* regexp */
          {
            if (regexec(pstrings->regexp,t))  /* compare - if match do command */
            {
              match= true;

#ifdef DEBUG
            printf("from match:%lu %s\n",pArt->ArticleNr, pstrings->regexp->startp[1]);
#endif
            }
          }
          else
          {
#ifdef DEBUG
            printf("%s,%s\n",pstrings->string,t);
#endif
            if (KMP_Matcher(t,tlen,pstrings->string, pstrings->strlen,
                            pstrings->table)!=-1)
            {
              match= true;
#ifdef DEBUG
              printf("matched!%d\n",pstrings->command);
#endif
            }
          }

          if (match)
          {
            switch (pstrings->command)
            {
              case SELECT: pArt->Status|= SELECTED_ART;
                           break;
              case JUNK: pArt->Status|= READ;
                         nr_of_junked++;
                         sprintf(fnametmp,"%s/%lu",pathtogroup, pArt->ArticleNr);

                         if (pArt->Status & XREF)
                         {
                           if (article= fopen(fnametmp,"r"))
                           {
                             MarkXRefs(article, false);
                             fclose(article);
                           }
                         }
                         break;
            } /* end of switch */
            match= false;
          }

          pstrings= pstrings->nextstring;
        }
      }
    } /* if !READ */
  } /* for all articles */
  ClrLine();

  return nr_of_junked;
}
#undef DEBUG

#ifdef NORMALSTRING
void
printkill_list(struct kill_field *plist)
{
  struct kill_strings *pstrings;

  while (plist)
  {
    printf("Field: %s\n",plist->headerline);
    for(pstrings=plist->stringlist; pstrings!=NULL; pstrings=pstrings->nextstring)
     printf("|%s|\n",pstrings->string);
    plist=plist->nextfield;
  }
}
#endif


void
FreeString_List(struct kill_strings *pstrings)
{
  if (pstrings)
  {
    if (pstrings->nextstring)
    {
      FreeString_List(pstrings->nextstring);
    }
    if (pstrings->regexp)
       free(pstrings->regexp);
    if (pstrings->string)
       free(pstrings->string);
    if (pstrings->table)
       free(pstrings->table);

    FreeMem(pstrings, (ULONG) sizeof(struct kill_strings));
  }
}


void
FreeKill_List(struct kill_field *plist)
{
  if (plist)
  {
    if (plist->nextfield)
    {
      FreeKill_List(plist->nextfield);
    }

    FreeString(plist->headerline);
    FreeString_List(plist->stringlist);
    FreeMem(plist, (ULONG) sizeof(struct kill_field));
  }
}


int
ReadKillFile(struct kill_field **rootp, char *name)
{
  FILE *killfileFP;
  char *s,*t, *field;
  short int command;
  ULONG time= 0UL, currenttime;
  UWORD attributes= 0;

  if (rootp) *rootp=NULL;

  currenttime= GetTime();

  if (name && (killfileFP= fopen(name,"r")))
  {
    while(fgets(tmp, SUPERLINE, killfileFP))
    {
      cuteoln(tmp);
      field= NULL;
      s= skpwspc(tmp);
      attributes= 0; /* default: not case-sensitive, no match begin, no regular expr */
      if (*s=='\0') continue;  /* skip blank lines! */
      time= 0UL;
      s= strchr(tmp,'/');   /* find / */
      if (s!=NULL)
      {                     /* slash found */
        while (s!=NULL && *(s+1)=='/')
        {
          s++; /* skip // */
          delchr(s);
          s=strchr(s,'/');
        }
        if (s && *s=='/')
        {
          *s='\0';
          s++;
          field=tmp;
          t=strchr(s,'~');
        }
        else
          t=strchr(tmp,'~');
      }
      else
       t=strchr(tmp,'~');

      command= JUNK; /* Default Command */
      if(t!=NULL) /* Special command found */
      {
        while (t!=NULL && *(t+1)=='~')
        {
          t++;  /* skip ~~ */
          delchr(t);
          t=strchr(t,'~');
        }

        if (t && *t=='~')
        {
          *t++='\0';
          command= JUNK;
          while (*t!='\0') /* evaluate special command string */
          {                /* TODO */
            switch(*t)
            {
              case 'j': command= JUNK;
                        break;
              case 's': command= SELECT;
                        break;
              case 'r': attributes|= REGULAREXPR;
                        break;
              case 'c': attributes|= CASESENSITIVE;
                        break;
              case '^': attributes|= MATCHBEGIN;
                        break;
            }
            if (isdigit(*t))
            {
              time= EvalDate(t);
              while (isdigit(*t)) t++;
            }
            else
             t++;
          }
        }
        else
        {
          command= JUNK; /* default */
        }
      }
      if (time==0 || time>currenttime)
      {
        if (field)
           add_killstring(rootp, field, s, command, attributes);
        else
           add_killstring(rootp, NULL, tmp, command, attributes);
      }
    }
    fclose(killfileFP);
    return 1;
  }
  else
    return 0;
}


int
AppendKillFile(struct kill_field **rootp, char *filename, char *field, char *string,
               char *mode)
/* returns 0 for action okay */
/* returns 1 if not able to open the kill-file */
/* returns 2 if aborted */
{
  FILE *killfileFP;
  ULONG expiretime;
  UWORD attributes= 0;
  int command=JUNK;
  char *tmp_mode= mode, key, short_tmp[9], *expirestr;


  expiretime= GetTime()+xlifetime*86400;
  short_tmp[0]='\0';
  expirestr= short_tmp;

  DispMsg("Validity period: A)lways, S)pecify, else default, Q)uit (no change):");
  ConClear();
  key= GetUpKey();
  ClrLine();
  if (key=='Q')
  {
    DispMsg(noaction);
    return 2;
  }
  if (key=='S')
  {
    DispMsg("Enter entry-expiration-date (ddmmyyyy):");
    strcpy(short_tmp,ShortDateStr(expiretime));
    Getline(short_tmp,8);
    ClrLine();
  }
  else
  if (key!='A') /* default */
  {
    expirestr= ShortDateStr(expiretime);
  }

  if (rootp) *rootp=NULL;

  if (mode)
  {
    while(*tmp_mode)
    {
      switch (*tmp_mode)
      {
        case 'j': command= JUNK;
                  break;
        case 's': command= SELECT;
                  break;
      }
      tmp_mode++;
    }
  }

  if(killfileFP= fopen(filename,"a"))
  {
    if (mode && strchr(mode,'r')) attributes|= REGULAREXPR; /* should be treated as regexp */
    if (mode && strchr(mode,'^')) attributes|= MATCHBEGIN; /* should be treated as regexp */

    if (field==NULL || stricmp(field,HeaderLines[HL_SUBJECT])==0)
    {
      fprintf(killfileFP, "%s", string);
      add_killstring(rootp, NULL, string, command, attributes);
    }
    else
    {
      fprintf(killfileFP, "%s/%s", field, string);
      add_killstring(rootp, field, string, command, attributes);
    }
    fprintf(killfileFP, "~%s %s\n", mode, expirestr);
    fclose(killfileFP);
    return 0;
  }
  else
    return 1;
}


void
CheckMeta_regexp(char *dest, char *source)
/* check for regular expression meta signs */
{
  char *meta="^$.[]()|?+*\\/~", *c;

  if (source)
  while (*source!='\0')
  {
    if (!isalnum(*source) && (c=strchr(meta,*source)))
    {
      switch(*c)
      {
        case '/': *dest++= '/';
                  break;
        case '~': *dest++= '~';
                  break;
        default: *dest++= '\\';
                 break;
      }
    }
    *dest++= *source++;
  }
  *dest='\0';
}


void
CheckMeta(char *dest, char *source)
/* check for signs we need and double when necessary */
{
  char *meta="/~", *c;

  if (source)
  while (*source!='\0')
  {
    if (!isalnum(*source) && (c=strchr(meta,*source)))
    {
      switch(*c)
      {
        case '/': *dest++='/';
                  break;
        case '~': *dest++='~';
                  break;
      }
    }
    *dest++=*source++;
  }
  *dest='\0';
}



int
AddSubjToKill(struct kill_field **plist,char *filename,
              char *mode,struct art_struct *pArt)
{
  char *subject=NULL, *c;
  int len, rc= 0;

  if (pArt)
  {
    subject= CopyString(pArt->Subject->Subject);
  }
  else
    subject= GetCurrentSubject();

  if (subject)
  {
    len= strlen(subject);
    c= StripSubject(subject);
    CheckMeta(tmp,c);
    tmp[DEFAULT_SUBJECTLEN_FOR_XFILES]='\0';            /* No more than 30 chars */
    rc= AppendKillFile(plist,filename,NULL, tmp, mode);
    FreeMem(subject, (ULONG) len+1UL);
  }

  return rc;
}


int
AddAuthorToKill(struct kill_field **plist,char *filename,
                char *mode, struct art_struct *pArt)
{
  if (pArt)
     return AppendKillFile(plist,filename,"From:", pArt->From, mode);
}


void
AddMsgIDToKill(struct kill_field **plist,char *filename, char *mode)
{
  char *msgid;

  msgid= GetParentMsgID();
  if (!msgid) msgid=GetMsgID();

  if (msgid)
  {
    CheckMeta(tmp,msgid);
    AppendKillFile(plist, filename, "References:", tmp, mode);
    FreeString(msgid);
  }
}


void KillOnOff(BOOL global)
{
  StatusLine();
  ClrLine();
  if (global)
  {
    DispMsg("Global X-file ");
    if (GeneralBool & GLOBALKILLON)
    {
      GeneralBool&= ~GLOBALKILLON;
      DispMsg("off.");
    }
    else
    {
      GeneralBool|= GLOBALKILLON;
      DispMsg("on.");
    }
  }
  else
  {
    DispMsg("Local X-file ");
    if (GeneralBool & LOCALKILLON)
    {
      GeneralBool&= ~LOCALKILLON;
      DispMsg("off.");
    }
    else
    {
      GeneralBool|= LOCALKILLON;
      DispMsg("on.");
    }
  }
}


void
EditKill(void)
{
  BOOL local= true, titleset= false;
  char *filename;
  struct kill_field **plist;

  DispMsg("Edit G)lobal or L)ocal X-file:");
  ConClear();
  *tmp= GetUpKey();
  ClrLine();

  switch (tmp[0])
  {
    case 'G': local= false;
    case 'L':
              break;
    default : return;
              break;
  }

  if (local)
  {
    if (pathtodbase)
    {
      CreateDir_IfNotExists(pathtodbase,pathtodbase+strlen(pathtodbase));
      sprintf(fnametmp,"%s/.X%s",pathtodbase, UserName);
    }
    else
    sprintf(fnametmp,"%s/.X%s",pathtogroup, UserName);

    filename= fnametmp;
    plist= &LocalKill_List;
  }
  else
  {
    filename= globalkillfile;
    plist= &GlobalKill_List;
  }

  if (!(GeneralBool & SCREENTITLE))
  {
    ShowTitle(ArnScreen, TRUE);
    titleset= true;
  }
  DispMsg("Invoking editor...");
  if (GeneralBool & EDSCRSWITCH)
     WBenchToFront();

  if (auxmode && AuxEditor)
     sprintf(tmp,"%s %s", AuxEditor, filename);
  else
     sprintf(tmp,"%s %s", NewsEditor, filename);

  system(tmp);

  if (auxmode) rawcon(1); /* get sure we're in raw mode */

  ScreenToFront(ArnScreen);

  FreeKill_List(*plist);
  DispMsg("rereading X-file...");
  ReadKillFile(plist,filename);
  DispMsg("ready.");
#ifdef NORMALSTRING
  printkill_list(plist);
#endif
  if (titleset) ShowTitle(ArnScreen, FALSE);
}


void ChangeKillFile(struct art_struct *pArt, int depth)
/* pathtogroup must be set! */
{
  char inkey;
  char mode[6];
  char headerfield[51]= "Subject:";
  BOOL local= true;
  char *filename;
  struct kill_field **plist;

  mode[0]='\0';
  DispMsg("E)dit, K)ill, A)utoselect, Q)uit:");
  ConClear();
  inkey=GetUpKey();
  ClrLine();
  switch (inkey)
  {
    case 'A': mode[0]= 's';
              break;
    case 'E': EditKill();
              return;
              break;
    case 'K': break;

    default : ClrLine();
              return;
              break;
  }
  mode[1]='\0';

  DispMsg("S)ubject, A)uthor,");
  if (depth==PAGER)
   DispMsg(" T)hread,");
  DispMsg(" R)egexp, E)xpression, Q)uit:");
  ConClear();
  inkey=GetUpKey();
  ClrLine();
  if (inkey=='Q')
  {
    DispMsg(noaction);
    return;
  }

  DispMsg("Add to G)lobal or L)ocal X-File?:");
  ConClear();
  tmp[0]= GetUpKey();
  ClrLine();
  if(tmp[0]=='G') local= false;

  if (local)
  {
    if (pathtodbase)
    {
      CreateDir_IfNotExists(pathtodbase, pathtodbase+strlen(pathtodbase));
      sprintf(fnametmp,"%s/.X%s",pathtodbase, UserName);
    }
    else
      sprintf(fnametmp,"%s/.X%s",pathtogroup, UserName);

    filename= fnametmp;
    plist= &LocalKill_List;
  }
  else
  {
    filename= globalkillfile;
    plist= &GlobalKill_List;
  }

  switch (inkey)
  {
    case 'A': if (AddAuthorToKill(plist,filename,mode,pArt)) return;
              break;

    case 'S': strcat(mode,"^");
              if (AddSubjToKill(plist,filename,mode,pArt)) return;
              break;

    case 'T': if (pArt==NULL)
                AddMsgIDToKill(plist,filename,mode);
              break;

    case 'R': strcat(mode,"r");
    case 'E': ClrLine();
              DispMsg("Header-Field:");
              Getline(headerfield,50);
              ClrLine();
              DispMsg("Enter an expression:");
              *tmp='\0';
              Getline(tmp,40);
              ClrLine();
              if (strlen(tmp)==0) return;
              if(AppendKillFile(plist, filename, headerfield, tmp, mode)) return;
              break;
    case 'Q': DispMsg(noaction);
              return;
              break;
  } /* end switch */
  DispMsg("added to ");
  if (filename==globalkillfile)
     DispMsg("global");
  else
     DispMsg("local");
  DispMsg("X-file.");
}



long
DoExpireXfile(char *filename) /* returns number of expired lines */
{
  FILE *fp_write, *fp_read;
  char *t;
  ULONG time, currenttime;
  long lines= 0L, preserved_lines= 0L;

  preserved_lines= 0L;

  currenttime= GetTime();

  sprintf(fnametmp, "%s.new",filename);

  if (fp_read=fopen(filename,"r"))
  {
    if (fp_write=fopen(fnametmp,"w"))
    {
      while (fgets(tmp, SUPERLINE, fp_read))
      {
        cuteoln(tmp);
        t= strchr(tmp,'~');
        while (t!=NULL && *(t+1)=='~')
        {
          t+=2;  /* skip ~~ */
          t= strchr(t,'~');
        }
        if (t!=NULL)
        {
          time=0;
          while (*t!='\0')
          if (isdigit(*t))
          {
            time= EvalDate(t);
            break;
          }
          else
            t++;

          if (time==0 || time>currenttime) /* preserve this entry */
          {
            fprintf(fp_write,"%s\n",tmp);
            preserved_lines++;
          }
          else
             lines++;
        }
        else
        {
          fprintf(fp_write,"%s\n",tmp);
          preserved_lines++;
        }
      }
      fclose(fp_write);
      fclose(fp_read);

      if (lines>0) /* lines expired */
      {
        DeleteFile(filename);

        if (preserved_lines>0) /* still lines left */
           Rename(fnametmp,filename);
        else
           DeleteFile(fnametmp); /* No lines left */

      }
      else /* no change */
        DeleteFile(fnametmp);
    }
    else
    {
      printf("ERROR: Couldn't open write-file %s!\n", fnametmp);
      fclose(fp_read);
    }
  }
  return lines;
}


void
ExpireXfiles(void)
{
  FILE *groups;
  char *gtmp, *filename, *c, *xfiletmp;
  long lines;

  printf("X-file expire: StartUp\n");

  if ((gtmp= AllocMem(2*MAXLINE, MEMTYPE))==NULL)
  {
     ErrorHandle(NOMEM,NULL);
     return;
  }
  else
    filename= gtmp+MAXLINE;


  if (GetArnConfig(false)) /* get the full configs */
  {
    ErrorHandle(CONFIGERROR, NULL);
    return;
  }
  else
  {
    if (getconfig())
    {
      ErrorHandle(UUCONFERROR,NULL);
      return;
    }
  }
  strcpy(fnametmp,grouplistfile); /* Get dir where grouplistfile is */
  if(c= PathPart(fnametmp)) *c='\0';
  sprintf(tmp, ".glX%s", UserName);
  AddPart(fnametmp, tmp, MAXLINE);
  if (!(globalkillfile= CopyString(fnametmp))) return;

  lines= DoExpireXfile(globalkillfile);
  if (lines>0)
     printf("%s...%ld line(s) expired.\n",globalkillfile, lines);

  if (activefile)
     groups= fopen(activefile,"r");
  else
     groups= fopen("UULIB:newsgroups","r");

  if (groups)
  {
    sprintf(gtmp,".X%s", UserName);
    xfiletmp= CopyString(gtmp);

    while (fgets(gtmp, MAXLINE, groups))
    {
      c= NextWhiteSpc(gtmp);
      *c= '\0';
      if (GeneralBool & MOREDIRS)
         makeslashes(gtmp);
      strcpy(filename, newsdir);

      if (AddPart(filename, gtmp, MAXLINE))
      {
        if (!AddPart(filename, xfiletmp, MAXLINE))
           ErrorHandle(STRANGE, "AddPart2");
      }
      else
        ErrorHandle(STRANGE, "AddPart1");

      lines= DoExpireXfile(filename);
      if (lines>0)
         printf("%s...%ld line(s) expired.\n", gtmp, lines);
    }
    fclose(groups);
    FreeString(xfiletmp);
  }
  printf("X-file expire: ready.\n");
}
