/***********************************************************************
 * mimedecode.c - MIME decoding funtions for Arn (see RFC1341)         *
 ***********************************************************************
 * Arn - (C)opyright by Roland Bless, All rights reserved,             *
 * -------- YOU ARE NOT ALLOWED TO DISTRIBUTE A CHANGED SOURCE ------- *
 ***********************************************************************/
#include <stdio.h>
#include <ctype.h>

#define true 1
#define false 0

typedef unsigned short BOOL;

/* proto$start */
int
Quoted_Printable_Decode(char *filename_in, char *filename_out);
int
Base64_Decode(char *filename_in, char *filename_out);
/* proto$end */


#define MAXLINE 256

#define hexval(c) (isdigit(c) ? (c-'0') : (isupper(c) ? (c-'A'+0x0A) : (c-'a'+0x0A)))

extern char *tmp;

typedef unsigned char UCHAR;
typedef unsigned long ULONG;


int
Quoted_Printable_Decode(char *filename_in, char *filename_out)
/* decoded text is written into file <filename_out> */
{
  FILE *fp_in, *fp_out;
  UCHAR *c, *d, deststr[MAXLINE], *destend;
  BOOL softlinebreak;

  destend= &deststr[MAXLINE-1];

  if (fp_in= fopen(filename_in,"r"))
  {
    if (fp_out= fopen(filename_out,"w"))
    {
      while (fgets(tmp, MAXLINE, fp_in) && *tmp!='\n') /* skip header */
            fprintf(fp_out,"%s",tmp);
      fputc('\n', fp_out);
      while (fgets(tmp, MAXLINE, fp_in))
      { /* qp-encoded lines are not longer than 76 chars */
        softlinebreak= false;
        for (c= tmp, d= deststr; *c && d<=destend && *c!='\n'; c++,d++)
        {
          if (*c=='=')
          {
            c++;
            if (*c && *c!='\n') /* no soft-line break */
            {
              /* decode */
              if (isxdigit(*c)) /* is hex-digit */
              {
                *d= hexval(*c) << 4; /* multiply with 0x10 */
                if (isxdigit(*(++c)))
                {
                   *d|= hexval(*c);  /* write into lower nibble */
                }
                else /* wrong code at 2nd hex-digit! -> don't change anything */
                {
                  c-= 2; *d= '=';
                }
              }
              else /* wrong code at first hex digit !! */
              {
                c--; *d= '=';
              }
            }
            else /* soft-line break */
            {
              *c--='\0'; /* remove '\n' */
              d--; /* don't increase d */
              softlinebreak= true;
            }
          }
          else
          { /* no coded stuff */
            *d= *c;
          }
        } /* end for */
        *d='\0';
        if (softlinebreak)
           fprintf(fp_out,"%s",deststr);
        else
           fprintf(fp_out,"%s\n",deststr);

      } /* while (fgets()) */
      fclose(fp_out);
    } /* could open write file */
    fclose(fp_in);
  } /* could open read file */
  else
    return 1;

  return 0;
}


ULONG
base64(char c)
/* returns decimal value from 64-base digit */
{
  if (isupper(c))
     return (ULONG) (c-'A');
  else
  if (islower(c))
     return (ULONG) (c-'a'+26UL);
  else
  if (isdigit(c))
     return (ULONG) (c-'0'+52UL);
  else
  if (c=='+')
     return 62UL;
  else
  if (c=='/')
     return 63UL;
}


int
Base64_Decode(char *filename_in, char *filename_out)
/* decodes message encoded in base64 (header is not processed) */
{
  FILE *fp_in, *fp_out;
  ULONG bytes; /* warning! sizeof(bytes)>=4 must be! */
  UCHAR *c, *d, deststr[MAXLINE], *destend;


  destend= &deststr[MAXLINE-1];

  if (fp_in= fopen(filename_in,"r"))
  {
    if (fp_out= fopen(filename_out,"w"))
    {
      while (fgets(tmp, MAXLINE, fp_in) && *tmp!='\n') /* skip header */
            fprintf(fp_out,"%s",tmp);
      fputc('\n', fp_out);
      while (fgets(tmp, MAXLINE, fp_in))
      { /* base64 lines are not longer than 76 chars */
        /* convert 4 6-bit characters into 3 8-bit characters */
        for (c= tmp, d= deststr; *c && d<=destend && *c!='\n'; c++)
        {
          bytes= 0UL;
          /* read group of 4 bytes (with 6-bit filled) */
          bytes|= (base64(*c++)<<18); /* 1st char (leftmost byte) */
          bytes|= (base64(*c++)<<12); /* 2nd char */
          if (*c!='=') /* no pad character */
             bytes|= (base64(*c++)<<6);  /* 3rd char */
          else
            bytes&= 0xFF0000; /* there is only one byte to decode (ignore four bits) */

          if (*c!='=') /* no pad character */
             bytes|= (base64(*c));       /* 4th char */
          else
            bytes&= 0xFFFF00; /* there are only two bytes to decode (ignore two bits) */

          *d++= (UCHAR) (bytes >> 16);

          if (*(c-1)!='=')
             *d++= (UCHAR) ((bytes >> 8) & 0xFF);
          if (*c!='=')
             *d++= (UCHAR) (bytes & 0xFF);
        }
        fwrite(deststr,d-deststr,1,fp_out);
      } /* while (fgets()) */
      fclose(fp_out);
    } /* could open write file */
    fclose(fp_in);
  } /* could open read file */
  else
    return 1;

  return 0;
}
