#ifndef GLOBALVARS_H
 #include "globalvars.h"
#endif
#ifndef ENUMS_H
 #include "enums.h"
#endif

#include "proto/strprocs_proto.h"
#include "proto/pager_proto.h"
#include "proto/followup_proto.h"
#include "proto/mimedecode_proto.h"

/* Prototypes for functions defined in mimesupport.c */
/* proto$start */
void
GetTypeAndSubType(char *source, char **type, char **subtype);
void
GetAttributeAndValue(char *source, char **attribute, char **value);
char *
GetNextParameter(char *parameterpart);
char *IsMIME(char *filename_in, char *filename_out);
/* proto$end */

#define TYPEATTR_CHARSET "charset"
#define KNOWN_CHARSET1 "ISO-8859-1"
#define KNOWN_CHARSET2 "US-ASCII"
#define KNOWN_ENCODING_BASE64 "BASE64"
#define KNOWN_ENCODING_QP "QUOTED-PRINTABLE"

                                   /****************/
char paracstore, *paracstorep;
                                   /****************/


void
GetTypeAndSubType(char *source, char **type, char **subtype)
/* expects non-NULL! source string like "text/plain" and
   allocates two strings for type and subtype which are returned */
{
  char *c;

  *type= *subtype= NULL;

  for (c= source; *c && *c!='/'; c++);
  if (*c=='/')
  {
    *c= '\0';
    *type= CopyString(source);
    *c++='/';
    c= skpwspc(c);
    *subtype= CopyString(c);
  }
}


void
GetAttributeAndValue(char *source, char **attribute, char **value)
/* expects non-NULL! parameter source string like "charset=ISO8859-1" and
   allocates two strings for attribute and value which are returned */
{
  char *c;

  *attribute= *value= NULL;

  for (c= source; *c && *c!='='; c++);
  if (*c=='=')
  {
    *c= '\0';
    *attribute= CopyString(source);
    *c++='=';
    c= skpwspc(c);
    *value= CopyString(c);
  }
}


char *
GetNextParameter(char *parameterpart)
/* returns <parameter> by parsing parameterpart for a separator (;)
   and writes a '\0' to the boundary. Which is restored when
   returning NULL which means that there are no further parameters.
   Note that no new string is allocated, the returned pointer is
   within the string <parameterpart>.
   On every first call of this routine "paracstorep" has to be set to NULL!!
*/
{
  if (parameterpart)
  {
    if (paracstorep) /* restore last byte where we wrote the '\0' */
    {
      *paracstorep= paracstore;
       parameterpart= paracstorep; /* skip field, now on previous '\0' */
    }

    if (*parameterpart!=';')  /* last '\0' was at a space */
    {
      parameterpart= skpwspc(parameterpart); /* skip LWSPC */
      while(*parameterpart && *parameterpart!=';') parameterpart++;
      if (*parameterpart==';') parameterpart++;
    }
    else
      parameterpart++;
       /* now parameterpart is behind the previous ';' */

    if (*parameterpart)
    {
      parameterpart= skpwspc(parameterpart); /* skip leading LWSP */

      if (*parameterpart!='\"') /* quoted string can contain spaces and ';'s */
        for (paracstorep= parameterpart;
             *paracstorep && *paracstorep!=';' && !isspace(*paracstorep);
             paracstorep++);
      else
      { /* find matching closing " */
        for(paracstorep= parameterpart; *paracstorep && *parameterpart!='\"';
            paracstorep++);
        if (*paracstorep) paracstorep++; /* get behind closing " */
      }
      /* paracstorep is now at '\0' first WhiteSpace or ';' */
      paracstore= *paracstorep;

      return parameterpart;
    }
    else /* no further parameters */
      return NULL;
  }
  else
     return NULL;
}


char *
IsMIME(char *filename_in, char *filename_out)
/* looks for MIME-Headerlines and processes article accordingly,
   maybe external metamail is called for help.
   A decoded temporary file is created if there was something to
   process. Its name is returned.
   If no MIME-headers were present or pager cannot/shouldn't display
   output, NULL is returned */
{
  char *tmp_filename_template= "T:mime_arnXXXX.xxxx",
       *tmp_filename= NULL, *headerline, *c, *d, storec,
       *mm_typefield= NULL,
       *mm_parameter_field= NULL,
       *mm_typepara_attribute= NULL,
       *mm_typepara_value= NULL;
  BOOL callmetamail= false, unknowncharset= false, quoted_printable= false,
       base64= false;

  if (headerline= GetHeaderLine(HeaderLines[HL_CONTENT_TYPE]))
  {
#ifdef DEBUG
    printf("%s#\n",headerline);
#endif
    /* get type parameter: */

    c= skpwspc(NextWhiteSpc(headerline)); /* get first header field after : */

    /* search for ; or whitespace as terminator of the type field */
    for (d= c; d && *d && !isspace(*d) && *d!=';'; d++);
    if (d)
    {
      storec= *d;
      *d= '\0';
      mm_typefield= CopyString(c);
#ifdef DEBUG
      printf("%s#\n", mm_typefield);
#endif
      if (stricmp(mm_typefield,"text/plain"))
      {
        callmetamail= true;
      }
      *d= storec;
      c= d;
    }
    /* get type parameter */
    paracstorep= NULL;
    mm_parameter_field= c;
    while (mm_parameter_field= GetNextParameter(mm_parameter_field))
    {
#ifdef DEBUG
      printf("%s ->", mm_parameter_field);
#endif
      GetAttributeAndValue(mm_parameter_field, &mm_typepara_attribute, &mm_typepara_value);
#ifdef DEBUG
printf("attr: %s value: %s\n", mm_typepara_attribute, mm_typepara_value);
#endif
      if (!strnicmp(TYPEATTR_CHARSET, mm_typepara_attribute, sizeof(TYPEATTR_CHARSET)))
      {
        if(strnicmp(KNOWN_CHARSET1, mm_typepara_value, sizeof(KNOWN_CHARSET1))!=0
           &&
           strnicmp(KNOWN_CHARSET2, mm_typepara_value, sizeof(KNOWN_CHARSET2))!=0)
           unknowncharset= true;
      }
      FreeString(mm_typepara_attribute);
      FreeString(mm_typepara_value);
    }
    paracstorep= NULL;
    FreeString(mm_typefield);
  }

  if (headerline= GetHeaderLine(HeaderLines[HL_CONTENT_XFER_ENCODING]))
  {
    if (c= skpwspc(NextWhiteSpc(headerline)))
    {
      if (!stricmp(c,KNOWN_ENCODING_BASE64)) base64= true;
      else
      if (!stricmp(c,KNOWN_ENCODING_QP)) quoted_printable= true;
    }
  }

  if (callmetamail && MetaMail)
  {
    sprintf(tmp,MetaMail,filename_in);
    system(tmp);
  }
  else
  if (quoted_printable)
  {
    tmp_filename= mytmp(tmp_filename_template);
    StatusLine();
    DispMsg("decoding quoted-printable...");
    Quoted_Printable_Decode(filename_in, tmp_filename);
  }
  else
  if (base64)
  {
    tmp_filename= mytmp(tmp_filename_template);
    StatusLine();
    DispMsg("decoding base64...");
    Base64_Decode(filename_in, tmp_filename);
  }


  GetHeaderLine(NULL); /* restore HeaderBuffer */

  if (tmp_filename) strcpy(filename_out, tmp_filename);
#ifdef DEBUG
  printf("%s\n",filename_out);
#endif
  return tmp_filename;
}
