/***********************************************************************
 * pager.c - pager routines for arn "MORE"                             *
 ***********************************************************************
 * Arn - (C)opyright by Roland Bless, All rights reserved,             *
 * -------- YOU ARE NOT ALLOWED TO DISTRIBUTE A CHANGED SOURCE ------- *
 ***********************************************************************/
/* #include <fcntl.h> */
#ifndef ENUMS_H
 #include "enums.h"
#endif

#ifndef GLOBALVARS_H
#include "globalvars.h"
#endif

#ifdef LATTICE_50
 #define size_t long
 #define SEEK_SET 0
 #define SEEK_CURR 1
 #define SEEK_END 2
#endif
#undef DEBUG

#define h_underlined 0x01
#define h_boldfaced  0x02
#define h_inverse    0x04
#define h_slanted    0x08
#define STACK_BLOCKSIZE 64 /* pages/longwords */


struct Stack  /* A stack block for the pager */
{
  long Element[STACK_BLOCKSIZE];
  struct Stack *previous_block;
  short int stack_index;
};

struct kmp_string
{
   char *pattern;
   short int *table;
   short int pattern_len;
};


char *junkedmsg="...junked this article...!\n";
char *noaction ="No action taken.";

char *pend=NULL;
char *actual_pos=NULL, directcommand='\0', *mimefile;
long unread;
int last_direction=NOP;
struct Stack *Top_of_PageStack=NULL;
struct kmp_string s_pattern;

 /****************************************************************/

extern struct kill_field *LocalKill_List; /* from readgroup.c */
extern ULONG LastArtNr;
extern long ToReadSelectedArts;
extern BOOL noselection; /* direct interpretation of artsel-level command */
extern char *waitmsg;    /* from funtcions.c */

/* Prototypes for functions defined in pager.c */
/* proto$start */
long pop(struct Stack **PStackP);
void push(long Element, struct Stack **PStackP);
void clearstack(struct Stack *StackP);
void headerstop(void);
void clrscreenmode(void);
void killsubject(void);
char *ArtStat(struct art_struct *pArt);
BOOL MoreCommands(char key,
                  int *command);
char *hlineend(char *s);
char *GetNextHeaderLine(char **s);
char *GetHeaderLine(char *s);
void PrintHeaderLine(char *line,
                     int headerline,
                     char headerchar,
                     char mode);
int PrintHeader(BOOL verbose);
BOOL KilledArticle(struct kill_field *Kill_List,
                   int *command);
void ProcFF(char *s,
            int Percent,
            long ArtNr,
            char *Status,
            long unread);
BOOL ReadArticle(struct art_struct **ppArt,
                 int *command);
/* proto$end */

#include "proto/strprocs_proto.h"
#include "proto/functions_proto.h"
#include "proto/cmdline_proto.h"
#include "proto/arnhelp_proto.h"
#include "proto/ArnIO_proto.h"
#include "proto/grouplist_proto.h"
#include "proto/followup_proto.h"
#include "proto/reply_proto.h"
#include "proto/kill_proto.h"
#include "proto/arn_proto.h"
#include "proto/time_proto.h"
#include "proto/choosegroups_proto.h"
#include "proto/mimesupport_proto.h"

/*******************/


long
pop(struct Stack **PStackP)
/* a standard stack-function, the stack is organized in linked blocks
   to get rid of memory fragmentation.
   This function simply removes the element at the Top of Stack and returns
   its value */
{
  long Element=0L;
  struct Stack *NewTOS, *StackP;
  short int *index;

  if (StackP= *PStackP) /* if there is a stack block */
  {
    index= &(StackP->stack_index);    /* get TOS */
    Element= StackP->Element[*index]; /* Get element from TOS */
    *index-=1;                                  /* decrease stack_index */
    if (*index<0) /* stack block is now empty, so get the previous block */
    {
      NewTOS= StackP->previous_block;
      free(StackP);
      *PStackP= NewTOS;
    }
  }

  return Element;
}


void
push(long Element, struct Stack **PStackP)
/* See pop for stack-structure.
   This function puts the element given on the Top of Stack, which is
   increased before */
{
  struct Stack *NewTOS, *StackP=*PStackP;
  unsigned short int index;


  if (StackP==NULL || StackP->stack_index==STACK_BLOCKSIZE-1)
  { /* allocate new stack block */
    if (NewTOS= malloc(sizeof(struct Stack)))
    {
      NewTOS->previous_block= StackP;
      NewTOS->stack_index= 0;
      NewTOS->Element[0]= Element;
      *PStackP= NewTOS;
    }
  }
  else  /* push an element onto stack */
  {
    index= ++(StackP->stack_index); /* increase stack index */
    StackP->Element[index]=Element; /* push element */
  }
}


void
clearstack(struct Stack *StackP)
/* This procedure removes all linked blocks and so clears the stack */
/* WARNING: StackP must be set to NULL afterwards !! */
{
  struct Stack *NewTOS;

  while(StackP)
  {
    NewTOS= StackP->previous_block;
    free(StackP);
    StackP= NewTOS;
  }
}


void
headerstop(void)
{
  StatusLine();
  DispMsg("HeaderStop");
  if (GeneralBool & HEADERSTOP)
  {
    GeneralBool&= ~HEADERSTOP; DispMsg(" off.");
  }
  else
  {
    GeneralBool|= HEADERSTOP; DispMsg(" on.");
  }
}


void
clrscreenmode(void)
{
  DispMsg("ClearScreen");
  if (GeneralBool & CLEARSCREEN)
  {
    GeneralBool&= ~CLEARSCREEN; DispMsg(" off.");
  }
  else
  {
    GeneralBool|= CLEARSCREEN; DispMsg(" on.");
  }
}


void
killsubject(void)
{
  sprintf(fnametmp,"%s/.X%s", (pathtodbase ? pathtodbase : pathtogroup), UserName);
  AddSubjToKill(&LocalKill_List,fnametmp,"^",NULL);
  FreeKill_List(LocalKill_List);
  StatusLine();
  DispMsg("rereading X-file...");
  ReadKillFile(&LocalKill_List,fnametmp);
  DispMsg("ready.\n");
}


char *
ArtStat(struct art_struct *pArt)
{
  static char StatTxt[8];
  int i=0;

  StatTxt[i]= '{';

  if (pArt->Status & NEW)
     StatTxt[++i]='N';
  if (pArt->Status & READ)
     StatTxt[++i]='R';
  if (i>0)
  { i++;
    StatTxt[i++]= '}';
  }
  StatTxt[i]= '\0';

  return StatTxt;
}


int
MiscArtMenu(void)
{
  char key;

  ClrLine();
  DispMsg("F)orward, S)pecial quoted follow-up, Q)uit:");
  key= GetUpKey();
  ClrLine();
  switch(key)
  {
    case 'S':return QSP_FOLLOWUP;
             break;
    case 'F':return FORWARD_ART;
             break;
    default :DispMsg(noaction);
             ConCR();
             break;
  }
  return NOP;
}


BOOL
MoreCommands(char key, int *command)
{
  BOOL leave= false;

  switch (key)
  {
    case 'q':*command= QUIT;
             break;
    case 'Q':
    case 'X':*command= TOARTDIR;
             break;
    case ctrl('C'):*command= BREAK;   /* CTRL-C */
             GlobalERROR= QUITARN;
             break;
    case 'P':
    case 'p':*command= PREVART;
             break;
    case 'N':
    case 'n':*command= NEXTART;
             break;
    case 'u':*command= UNREADART;
             break;
    case 'j':*command= JUNK;
             break;
    case 'J':*command= READART;
             break;
    case 'c':*command= CANCELART;
               break;
    case 'C':*command= CATCHUP;
             if (CatchUp(true))
                *command= NOP;
             break;
    case 'e':*command= EXTERN_PRG;
             break;
    case ctrl('T'):ToggleScreenTitle();   /* CTRL-T */
                *command= SAMEP;
                break;
    case ctrl('H'):*command= PREVP;        /* CTRL-H,Backspace */
                GeneralBool&=~ROT13;
                break;
    case ctrl('L'):*command= SAMEP;        /* CTRL-L */
                GeneralBool&=~ROT13;
                break;
    case 'I': *command= STATUS;
              break;
    case 'O':*command= OptionsMenu(PAGER);
             break;
    case ctrl('R'):                        /* CTRL-R */
    case 'T':*command= TOP;
              GeneralBool&=~ROT13;
              break;
    case 'B':*command= BOTTOM;
             DispMsg("jumping to bottom...");
             break;
    case 'd':killsubject();
             *command= NOP;
             break;
    case 'K':*command= EDIT_KILL;
             break;
    case '\t':*command= TAB;
              break;
    case 'k':*command= MARK_THREAD_READ;
             break;
    case 'v':*command= VERBOSE;
             break;
    case 'x':if (GeneralBool & ROT13)
               GeneralBool&=~ROT13;
             else
               GeneralBool|=ROT13;
             *command= SAMEP;
             break;
    case 'f':*command= FOLLOWUP;
             break;
    case 'F':*command= Q_FOLLOWUP;
             break;
    case 'M':*command= MiscArtMenu();
             break;
    case 'r':*command= REPLY;
             break;
    case 'R':*command= Q_REPLY;
             break;
    case 'E':*command= POSTARTICLE;
             break;
    case 'G':*command= SEARCH_PAT;
             break;
    case ' ':*command= defcomm;
             break;
    case '\x0D':*command= NEXTLINE;    /* Return */
                break;
    case 'h':
    case 'H': dohelp(H_MORE);
              *command= NOP;
              break;

    case '\x9b': key= (*GetKey)(); /* Get first char after CSI */
                switch(key)
                {
                  case '?': key=(*GetKey)(); /* HELP KEY */
                            dohelp(H_MORE);
                            break;
                }

                switch(key)
                {
                  /* Function keys */
                  case '0': case '1': case '2': case '3': case '4':
                  case '5': case '6': case '7': case '8': case '9':
                  case ' ':
                            key=(*GetKey)();
                            *command=NOP;
                            break;

                          default:
#ifdef DEBUG
                                   printf("CSI:%c/%d",key,key);
#endif
                                   /* key=(*GetKey)(); */
#ifdef DEBUG
                                   printf(":%c/%d\n",key,key);
#endif
                                   *command=NOP;
                                   break;
#undef DEBUG
                        }

              break;
  }

  last_direction= NOP;

  switch (*command)       /* leave pager ? */
  {
    case       PREVART:
    case       NEXTART:
    case          QUIT:
    case      TOARTDIR:
    case         BREAK:
    case          JUNK:
    case       CATCHUP:
    case      FOLLOWUP:
    case    Q_FOLLOWUP:
    case  QSP_FOLLOWUP:
    case         REPLY:
    case       Q_REPLY:
    case   FORWARD_ART:
    case   POSTARTICLE:
    case     CANCELART:
    case    EXTERN_PRG:
    case     UNREADART:
    case       READART:
    case     EDIT_KILL:
    case MARK_THREAD_READ:
    case           NOP:
                        leave=true;
                        break;
    default: leave= false;
             break;
  }

  directcommand='\0'; /* reset directcommand */

  return leave;
}


char *
hlineend(char *s) /* returns pointer to \n (EOL) or \0 if end of buffer */
{
  while (*s!='\0')
  {
    s=NextCR(s); /* next char after \n */
    if (*s!='\0' && *s!=' ' && *s!='\t') /* no line continuation */
        break;
  }
  if (*s!='\0') /* Not the end of the buffer */
     s--;       /* return pointer to '\n' */
  return s;
}


char *
GetNextHeaderLine(char **s)
/* Gets next HeaderLine out of HeaderBuf */
/* actual_pos is the current line position and should be set to */
/* HeaderBuf if you start from the beginning */
{
  static char mem='\0';

  if (pend)     /* restore last '\n' */
    *pend=mem;

  if (s==NULL)
  {
    pend=NULL;
    return NULL;   /* Just restore the last byte */
  }

  *s=actual_pos;               /* begin of the new line */
  if (*actual_pos=='\0')
     return NULL;

  actual_pos= hlineend(actual_pos); /* Find next end of line */

  if (*actual_pos=='\n' || *actual_pos=='\0')
  {
    pend= actual_pos; /* End is actual_pos */
    if (*actual_pos!='\0') actual_pos++;
    else
      pend--;
    mem=*pend;  /* get \n or \0 */
    *pend='\0'; /* cut off this line */

    return *s;
  }
  else
  {
    ErrorHandle(STRANGE,"getbuffield");
    return NULL;
  }
}


char *
GetHeaderLine(char *s)
/* Gets the line that begins with string s from buffer */
{
  char *c,*r= NULL;
  static char mem= '\0';

  if (pend)      /* restore last byte */
     *pend= mem;

  if (s==NULL)
  {
    pend= NULL;   /* restore Buffer */
    return NULL;
  }

  c= HeaderBuf;              /* search from the beginning of buffer */
  pend=NULL;
  while (*c!='\0')
  {
    if (r= firststr2(c,s))         /* found a matching begin */
    {
      c= hlineend(c);              /* get the eol */
      if (*c=='\n' || *c=='\0')
      {
        if (*c=='\0' && c>r) c--;
        pend= c;
        mem= *c;
        *c='\0';                   /* terminate buffer line */
        break;
      }
      else
      {
        ErrorHandle(STRANGE,"getbuffield");
        break;
      }
    }
    else
    {
      c= hlineend(c);
      if (*c!='\0')
         c++;
    }
  }
  return r;
}



void
PrintHeaderLine(char *line, int headerline, char headerchar, char mode)
{
  char *tail, buf;

  tail= NextWhiteSpc(line);
  if (*tail) tail++;
  buf= *tail;
  *tail= '\0';
  (*ConWrite)(line);
  *tail= buf;

  if (mode & h_underlined)
     (*ConWrite)(tc_UNDERLINE_ON);
  if (mode & h_boldfaced)
     (*ConWrite)(tc_BOLDFACE_ON);
  if (mode & h_inverse)
     InvON();
  if (mode & h_slanted)
     (*ConWrite)(tc_SLANTED_ON);

  switch (headerchar)
  {
    case 'G':
    case 'g': (*ConWrite)(newsgroup);
              break;
    default: (*ConWrite)(tail);
             break;
  }

  if (mode & h_underlined)
     (*ConWrite)(tc_UNDERLINE_OFF);
  if (mode & h_boldfaced)
     (*ConWrite)(tc_BOLDFACE_OFF);
  if (mode & h_inverse)
     InvOFF();
  if (mode & h_slanted)
     (*ConWrite)(tc_SLANTED_OFF);

  ConCR();
}


int
PrintHeader(BOOL verbose)
{
  char *line, *d, *start, *end, c, *t;
  int totlen= 0, lcount= 0, len= 0, newlen= 0;
  BOOL dummy;
  char mode= 0x00;
  short int searchfor;


  actual_pos= HeaderBuf; /* Reset HeaderPointer */
  pend= NULL;

  if (!verbose)
  {
    for (t= headerline; *t!='\0'; t++)
    {
      searchfor= -1;

      switch (*t)
      {
        case '_':mode|= h_underlined;
                 break;
        case '+':mode|= h_inverse;
                 break;
        case '/':mode|= h_slanted;
                 break;
        case '#':mode|= h_boldfaced;
                 break;
        case 'A':searchfor= HL_APPROVED;
                 break;
        case 'C':searchfor= HL_CONTROL;
                 break;
        case 'D':searchfor= HL_DATE;
                 break;
        case 'd':searchfor= HL_DISTRIBUTION;
                 break;
        case 'E':searchfor= HL_EXPIRES;
                 break;
        case 'e':searchfor= HL_SUPERSEDES;
                 break;
        case 'F':searchfor= HL_FROM;
                 break;
        case 'f':searchfor= HL_SENDER;
                 break;
        case 'G':searchfor= HL_NEWSGROUPS;
                 break;
        case 'g':searchfor= HL_NEWSGROUPS;
                 break;
        case 'I':searchfor= HL_MESSAGE_ID;
                 break;
        case 'K':searchfor= HL_KEYWORDS;
                 break;
        case 'L':searchfor= HL_LINES;
                 break;
        case 'N':searchfor= HL_NEWSGROUPS;
                 break;
        case 'n':searchfor= HL_NEWSGROUPS;
                 break;
        case 'O':searchfor= HL_ORGANIZATION;
                 break;
        case 'P':searchfor= HL_PATH;
                 break;
        case 'R':searchfor= HL_REPLY_TO;
                 break;
        case 'r':searchfor= HL_REFERENCES;
                 break;
        case 'S':searchfor= HL_SUBJECT;
                 break;
        case 's':searchfor= HL_SUMMARY;
                 break;
        case 'W':searchfor= HL_FOLLOWUP_TO;
                 break;
        case 'X':searchfor= HL_X_ALL;
                 break;
      } /* end switch */

      if (searchfor>=0 && searchfor<HL_STOP)
      {
        d= GetHeaderLine(HeaderLines[searchfor]);
        if (!d)
        {
          if (searchfor==HL_SUBJECT)
            d= GetHeaderLine(HeaderLines[HL_TITLE]);
          else
          if (searchfor==HL_MESSAGE_ID)
            d= GetHeaderLine(HeaderLines[HL_ARTICLE_ID]);
        }

        if ((*t=='g' || *t=='n') && (strchr(d,',')==NULL)) d=NULL;

        if (d)
        {
          PrintHeaderLine(d,searchfor,*t,mode);
          for (start= d, end= NextCR(d); *start!='\0'; start= end, end= NextCR(end))
          {
            c= *end;
            *end= '\0';
            len= xstrlen(start,&totlen, &dummy);
            if (totlen>MaxChars)
               newlen= totlen/MaxChars+ (totlen%MaxChars ? 1:0);
            else
               newlen= 1;
            lcount+= newlen;
#ifdef DEBUG
                  printf("%s:nw:%d len:%d lc:%d\n",line, newlen, len,lcount);
#endif
                  *end=c;
          }
        }

        mode= 0;
        searchfor= -1;
      } /* if searchfor>=0 */
    } /* for t */
    GetHeaderLine(NULL);
  }
  else /* verbose! */
  {
    while(GetNextHeaderLine(&line))
    {
      (*ConWriteLn)(line);
      for (start= line, end= NextCR(line); *start!='\0'; start= end, end= NextCR(end))
      {
        c= *end;
        *end= '\0';
        len= xstrlen(start,&totlen, &dummy);
        if (totlen>MaxChars)
           newlen= totlen/MaxChars+ (totlen%MaxChars ? 1:0);
        else
           newlen= 1;
        lcount+= newlen;
#ifdef DEBUG
              printf("%s:nw:%d len:%d lc:%d\n",line, newlen, len,lcount);
#endif
              *end= c;
      }
    }
    GetNextHeaderLine(NULL); /* restore last byte */
  }

#ifdef DEBUG
 printf("header:%d\n", lcount);
#endif

  return lcount;
}


BOOL
KilledArticle(struct kill_field *Kill_List, int *command)
{
  struct kill_strings *pstring;

  if (pstring= killed(Kill_List))
  {
    if (pstring->command==SELECT) return false;
    *command= pstring->command;
    return true;
  }

  return false;
}


void
ProcFF(char *s, int Percent, long ArtNr, char *Status, long unread)
/* That's our string to process */
{
  char dummy;

  if (s)
  {
     if (GeneralBool & ROT13)
        return;

     if (*s=='\x0C')
     {
       (*ConWriteLn)("^L");
       s++; /* We already prompted at first ^L */
     }

     while (*s!='\0')
     {
       if (*s!='\x0C')    /* No ^L ? */
         (*ConWriteChar)(*s);
       else
       {
         ConCR();         /* ^L: Prompt MORE in next line */
         dummy= More(Percent, ArtNr, Status, ToReadSelectedArts, unread);
         (*ConWrite)("^L\n");
       }
       s++;
     }
  }
}


/* contains pager-level "MORE" */
BOOL /* leave group? */
ReadArticle(struct art_struct **ppArt,int *command)
{
  struct art_struct *pArt;
  ULONG ArticleNr, unread;
  FILE *Art= NULL;
  char c;
  int lcount= 1, percent= 0, rcommand, len, newlen;
  long size, last, p_last, fppos, lpos= 0L, spos, BufSize;
  BOOL Error= false, killed_it= false;
  BOOL header= false, leave= false, verbose= false, formfeed= false;
  int totlen, tmp_len;

  /* set both global variables */

  pArt= *ppArt;
  if (pArt==NULL) return true; /* Don't read an empty group! */

  defcomm= NEXTP;
  rcommand= *command;

  mimefile= NULL;

  Top_of_PageStack= NULL;
  s_pattern.pattern= NULL;
  s_pattern.table  = NULL;


  unread= CountArticles(READ,UNREAD);

  ArticleNr= pArt->ArticleNr;

  sprintf(fnametmp,"%s/%lu",pathtogroup,ArticleNr);

  Art= fopen(fnametmp, "r");

  if (Art)
  {
    size_t readin;

#ifdef __SASC
    /* SAS-C Mode A specifies to delete CRs at EOLs */
    /* fmode(Art,0); */
#endif
    /* for faster IO! */
    setvbuf(Art, mBuffer, _IOFBF, M_BUFFERSIZE/2);

    tmp[SUPERLINE-1]='\0';
    moretmp[SUPERLINE-1]='\0';

    rcommand= NEXTP;

    if (!fseek(Art, 0L, SEEK_END))
      size= ftell(Art);
    else
    {
      size= 0L;
    }
    if (size==0L)
    {
      actual_pos= NULL;
      fclose (Art);
      pArt->Status|= READ;
      return false;
    }
    rewind(Art);
    p_last= last= 0L;

    /*** GetHeaderBuffer ***/

    readin= 0; /* headerbytes with/out CRs */
    while (fgets(tmp,SUPERLINE-1,Art) && *tmp!='\n' )
           readin+= strlen(tmp);
    /* necessary because fread() kills \r, too [LATTICE-fmode A] */

    /* BufSize= readin+1;  SAS_C kludge, because ftell() doesn't work! */
    BufSize= ftell(Art);

    /* BufSize is file position and is = readin if no CRs are stripped off !*/
#ifdef DEBUG
    printf("BufSize:%ld should be:%ld\n",BufSize,readin);
#endif

    if ((ULONG) BufSize>sizeofHeaderBuf) /* header doesn't fit into Buffer */
    {
      FreeMem(HeaderBuf,sizeofHeaderBuf);
      sizeofHeaderBuf= (ULONG) BufSize;
      HeaderBuf= (char *) AllocMem(sizeofHeaderBuf,MEMTYPE);
    }

    if (HeaderBuf)
    {
      rewind(Art);  /* rewind */
      /* In fmode A fread() kills CRs before \ns! */
      if (fread(HeaderBuf,readin,1,Art)<1) /* Read in whole header */
        Error= true;
      else
        *(HeaderBuf+readin)= '\0';

      actual_pos= HeaderBuf;
    }
    else
     Error= true;

    /**** Header is now in Buffer ****/

    if (Error)
    {
      ErrorHandle(STRANGE,"Header");
      leave= true;
      actual_pos= NULL;
      rcommand= QUIT;
    }
    else /* Header is in buffer */
    {
      if (!(pArt->Status & READ))
      {
        /******* KILL Article *********/
        if (GlobalKill_List && (GeneralBool & GLOBALKILLON))
        {
          if (KilledArticle(GlobalKill_List, &rcommand))
          {
            leave= true;
          }
        }
        if (!leave && LocalKill_List && (GeneralBool & LOCALKILLON))
        {
          if (KilledArticle(LocalKill_List, &rcommand))
          {
            leave= true;
          }
        }
        if (leave)
        {
          StatusLine();
          DispMsg("Xd-Article! ");
          killed_it= true;
        }
      }
    }

    if (IsMIME(fnametmp, tmp))   /* looks for MIME-headers in HeaderBuf and creates a decoded */
    {                            /* tmp-file, whose name is returned in 2nd argument. */
      Art= freopen(tmp, "r", Art); /* reopen new tmp article */
      if (Art)
      {
        mimefile= CopyString(tmp);
        if (!fseek(Art, 0L, SEEK_END))
          size= ftell(Art);
        else
        {
          size= 0L;
        }
        if (size==0L)
        {
          actual_pos= NULL;
          fclose (Art);
          pArt->Status|= READ;
          return false;
        }
        fseek(Art, BufSize-1, SEEK_SET); /* set file pointer behind header */
      }
      else
      {
        ErrorHandle(NOTMPFILE,"mime");
        return false;
      }
    }
    /********************* w h i l e *************************/

    while (!leave && !feof(Art))   /* leave = Quit MORE */
    {
nop:
      if (!header)
      {
        sprintf(tmp,"Article %lu (of %lu) in %s.\n",
                      ArticleNr, LastArtNr, newsgroup);
        ClrScr();
        (*ConWrite)(tmp);

        lcount+= PrintHeader(verbose);

        if (GeneralBool & HEADERSTOP)
        {
          if ((fppos= ftell(Art)-1)>=0L && size!=0L)
          {
             percent= (int) (fppos*100/size);
             c= More(percent, ArticleNr, ArtStat(pArt), ToReadSelectedArts, unread);
          }
          else
             c= More(101, ArticleNr, ArtStat(pArt), ToReadSelectedArts, unread);

          leave= MoreCommands(c, &rcommand);
          if (rcommand==TAB) rcommand=NEXTP;
          if (c!=ctrl('Z')) cmdline[0]= '\0';
          EnterCmdLine(c, PAGER);
          if (cmdline[0]!='\0')
             leave= InterpretCmdLine(cmdline, &rcommand, PAGER);

          if (rcommand==NEXTP)
          {
            header= true;
          }
          *moretmp= '\0';
          goto pos_commands;  /* Sorry for the GOTO, "skip len stuff" */
        }
        else  /* No headerstop */
        {
          header= true;
          rcommand= NEXTP; /* no endless loop! */
          continue; /* read in next lines */
        }
      }
      else  /* Header was already printed! */
      {
        if (rcommand==NEXTP || rcommand==NEXTLINE || rcommand==BOTTOM)
        { /* Don't read further if we have another command */
          if (fgets(moretmp, SUPERLINE-1, Art)==NULL) /* read new line in */
          {
             break;  /* EOF */
          }
          else
          {
            /* calc the length */
            len= xstrlen(moretmp,&totlen, &formfeed);
            if (totlen) totlen--;  /* Don't count the terminating newline! */
            if (totlen>MaxChars)
               newlen= totlen/MaxChars+ (totlen%MaxChars ? 1:0);
            else
               newlen= 1;
#ifdef DEBUG
            printf("%s:nw:%d len:%d lc:%d\n",moretmp, newlen, len,lcount);
#endif
          } /* else fgets */
        }
      } /* end else */


              /* does the newline fit on the screen? - No? -> MORE */
      if (lcount+newlen > (MaxScrLines-2) || *moretmp=='\x0C' || rcommand==NEXTLINE)
      {
        fppos= ftell(Art)-len;

        if (rcommand!=BOTTOM) /* if BOTTOM-command, don't prompt more */
        {
          if (fppos>=0L && size!=0L)
          {
             percent= (int) (fppos*100/size);
             c= More(percent, ArticleNr, ArtStat(pArt), ToReadSelectedArts, unread);
          }
          else
             c= More(101, ArticleNr, ArtStat(pArt), ToReadSelectedArts, unread);

          leave= MoreCommands(c, &rcommand);
          if (rcommand==TAB) rcommand=NEXTP;
          if (c!=ctrl('Z')) cmdline[0]='\0';
          EnterCmdLine(c, PAGER);
          if (cmdline[0]!='\0')
             leave= InterpretCmdLine(cmdline, &rcommand, PAGER);
        }

        if (rcommand==NEXTP || rcommand==BOTTOM)
        {
          lcount= 0;             /* reset line-counter */
          push(last,&Top_of_PageStack);
          last= fppos;  /* store last position */
          verbose= false;
          if (GeneralBool & CLEARSCREEN) ClrScr();/* ClearScreen */
        }
      } /* end: does the newline fit on the screen? */

      /* print line */
      if (rcommand==NEXTP || rcommand==NEXTLINE)
      {
        if (!formfeed)  /* Don't print a ^L (FF) ! */
        {
          if (GeneralBool & ROT13)
             (*ConWrite)(rot(moretmp,rotnumber));       /* PRINT the LINE */
          else
             (*ConWrite)(moretmp);   /* PRINT the LINE */
        }
        else
        {
          ProcFF(moretmp, percent, ArticleNr, ArtStat(pArt), unread);
          formfeed=false;
        }
        lcount+=newlen;
        continue;
      }

pos_commands:  /* Label for display commands */

      if (rcommand==PREVP)
      {
        p_last=pop(&Top_of_PageStack);
        fseek(Art, p_last, SEEK_SET);
        if (p_last==0L)
           rcommand= TOP;
        else
        {
          rcommand= defcomm= NEXTP; lcount= 0;
          last=p_last;
        }
        ClrScr();
      }
      if (rcommand==BOTTOM)
      {
        lcount+=newlen;
        continue;
      }
      if (rcommand==SEARCH_PAT)
      {
        char *c;
        casedep=0x5f; /* CASE INDEPENDENT for regexp() */

        spos= ftell(Art); /* start of search position */

        if (*cmdline!='\0')
        {
          c= NextSpc(cmdline)+1;

          if (s_pattern.table)
          {
            FreeMem(s_pattern.table, (ULONG) (s_pattern.pattern_len*sizeof(short int)));
          }
          FreeString(s_pattern.pattern);

          s_pattern.pattern= CopyString(c);
          s_pattern.pattern_len= strlen(c);
          s_pattern.table= (short int *)
                           AllocMem((ULONG) (s_pattern.pattern_len*sizeof(short int)),
                                     MEMTYPE);
          KMP_ComputeNextTable(c,s_pattern.pattern_len,s_pattern.table);

          rewind(Art);
        }

        if (s_pattern.pattern)
        {
          if (*cmdline=='\0') fseek(Art, lpos, SEEK_SET);

          fppos=ftell(Art);
          while (fgets(tmp, SUPERLINE-1, Art))
          {
            tmp_len= ftell(Art)-fppos;
            fppos= ftell(Art);
            if (KMP_Matcher(tmp, tmp_len, s_pattern.pattern, s_pattern.pattern_len,
                            s_pattern.table)!=-1) break;
          }

          if (feof(Art))        /* pattern not found */
          {
            DispMsg("NOT found!\n");
            if (spos<size-1)
            {
              fseek(Art, spos, SEEK_SET); /* go back to start-search-position */
              rcommand= NOP;
            }
          }
          else
          {
            ClrScr();
            (*ConWrite)(tmp);
            rcommand=NEXTP;
            lcount=1;
            lpos= ftell(Art); /* next start search position */
          }
        } /* if s_pattern.pattern */
        else
        {
          DispMsg("No pattern to search for...!\n");
          rcommand=NOP;
        }
      }
      if (rcommand==SAMEP)
      {
        fseek(Art, last, SEEK_SET);
        if (last==0L)
           rcommand= TOP;
        else
        {
          rcommand= defcomm= NEXTP; lcount= 0;
        }
        ClrScr();
      }
      if (rcommand==TOP || rcommand==VERBOSE)
      {
        fseek(Art, BufSize-1, SEEK_SET); /* Get the \n, too! */
        header= false;
        lcount= 1;
        if (rcommand==VERBOSE) verbose= true;
        else verbose=false;
        rcommand= NEXTP;
        defcomm= NEXTP;
        p_last= last=0L;
        sprintf(tmp,"Article %lu (of %lu) in %s.\n", ArticleNr, LastArtNr, newsgroup);
        ClrScr();
        (*ConWrite)(tmp);

        clearstack(Top_of_PageStack);
        Top_of_PageStack= NULL;
      }
      if (rcommand==STATUS)
      {
        StatusReport(pArt);
        if (!feof(Art)) goto nop;
      }

    } /* end while(!leave && !feof) */

            /********** End of Article, leave==true ! ************/

    if (!leave && feof(Art))        /* when at the end of an article */
    {
      if (rcommand!=TAB)
      {
        defcomm= NEXTP;
      }
      else
      {
        rcommand= NOP;
      }

      if (rcommand==NEXTP || rcommand==NEXTLINE)
      {
        if (*moretmp && moretmp[len-1]!='\n') /* last linefeed missing */
          ConCR();
      }

      if (rcommand!=BOTTOM)
      {
        if (directcommand=='\0')
        {
          if (ToReadSelectedArts>=0)
          sprintf(tmp, "End of article <%lu> in %s [%ld togo]. (SPACE,n,p,q)",
                        ArticleNr,newsgroup,ToReadSelectedArts);
          else
          sprintf(tmp, "End of article <%lu> in %s (%lu unread). (SPACE,n,p,q)",
                        ArticleNr,newsgroup,unread);
          InvON();(*ConWrite)(tmp); InvOFF();

          ConClear();
          c= (*GetKey)();
          ClrLine();
        }
        else
          c=directcommand;

        leave= MoreCommands(c, &rcommand);
        if (c!=ctrl('Z')) cmdline[0]='\0';
        EnterCmdLine(c, PAGER);
        if (cmdline[0]!='\0')
           leave= InterpretCmdLine(cmdline, &rcommand, PAGER);
      }
      else
        rcommand= SAMEP;

      switch (rcommand) /* if one of these cmds ..go back into article */
      {
        case       TOP: case     VERBOSE:
        case     PREVP: case       SAMEP:
        case    STATUS: case  SEARCH_PAT:
                                          goto pos_commands;
                                          break;
                                case TAB:
                             case BOTTOM: ClrLine();
                                          rcommand=NOP;
                                          break;
      }
      if (rcommand==NEXTP || rcommand==NEXTLINE)
      {
        if (pArt->Status & XREF)
           if (!(pArt->Status & READ))
             MarkXRefs(Art, true);
        pArt->Status|= READ; /* Mark Article as READ */

        rcommand= SPACE;
      }
    } /*** end of article ***/


    /******************* commands and actions *******************/

    if (rcommand==SAVEART_WH || rcommand==SAVEART_NH)
    {
      SaveArticle(pArt, cmdline, rcommand);
      rcommand= NOP;
    }
    else
    if (rcommand==FOLLOWUP || rcommand==Q_FOLLOWUP || rcommand==QSP_FOLLOWUP)
    {
      DoFollowUp(pArt,rcommand);
      rcommand= NOP;
    }
    else
    if (rcommand==REPLY || rcommand==Q_REPLY || rcommand==FORWARD_ART)
    {
      short mode;

      switch (rcommand)
      {
        case   REPLY: mode= NOQUOTE;
                      break;
        case Q_REPLY: mode= QUOTE;
                      break;
        case FORWARD_ART: mode= FORWARD;
                      break;
      }
      Reply(pArt->ArticleNr, NULL, mode);
      rcommand=NOP;
    }
    else
    if (rcommand==EDIT_KILL)
    {
      ClrLine();
      ChangeKillFile(pArt, PAGER);
      ConCR();
      rcommand=NOP;
    }
    if (rcommand==POSTARTICLE)
    {
      PostArticle(newsgroup);
      rcommand=NOP;
    }
    else
    if (rcommand == UNREADART)
    {
      if (cmdline[0]=='\0')
         pArt->Status&= (short) ~READ;
      else
         MarkRange(pArt, rcommand);

      rcommand= NOP;
      unread= CountArticles(READ,UNREAD);
    }
    else
    if (rcommand == READART)
    {
      if (pArt->Status & XREF)
        if (!(pArt->Status & READ)) /* don't mark XRefs twice */
          MarkXRefs(Art, true);
      pArt->Status|= READ;
      DispMsg(junkedmsg);

      rcommand= NOP;
      unread= CountArticles(READ,UNREAD);
    }

    /**************************************/

    if (rcommand==NOP)  /* Go back into article, where we left it */
    {
      leave= false;
      if (feof(Art))
        goto pos_commands;
      else
        goto nop;
    }
                   /* quit reading this article !! */

    if (rcommand == JUNK)
    {
      if (cmdline[0]=='\0' || killed_it)
      {
        if (pArt->Status & XREF)
           if (!(pArt->Status & READ)) /* don't mark XRefs twice */
              MarkXRefs(Art, true);
        pArt->Status|= READ;
      }
      else   /* range-command!!! */
      {
        MarkRange(pArt, rcommand);
        rcommand= NOP;
        unread= CountArticles(READ,UNREAD);
      }
    }
  } /* endif Article could be opened */
  else
  {
    ClrScr();
    sprintf(tmp,"skipping unavailable article...%s: %s",pArt->From,
                                                        pArt->Subject->Subject);
    (*ConWrite)(tmp);
    Delay(15L);
    pArt->Status|= READ;
    if (rcommand!=NEXTART && rcommand!=PREVART)
        rcommand= NEXTART;
  }

  if (Art)
  {
    if (rcommand==CANCELART)
    {
      int rc;

      rc= Cancel(newsgroup, Art);   /* Cancel Article and close it automatically
                                       to allow deletion! */
      WaitForSpace(waitmsg);
      if (rc==0) rcommand= NEXTART;   /* TODO: default_ACTION */
      else
      {
        rcommand=NOP; /* Article not cancelled */
        fclose(Art);
      }
    }
    else
      fclose(Art);      /* Close Article */
  }

  if (rcommand==EXTERN_PRG)
  {
    ExternPrg(pArt);
    rcommand=NOP;
  }

  if (s_pattern.pattern)
  {
    if (s_pattern.table)
    {
      FreeMem(s_pattern.table, (ULONG) (s_pattern.pattern_len*sizeof(short int)));
      s_pattern.table= NULL;
    }

    FreeString(s_pattern.pattern);
    s_pattern.pattern= NULL;
  }

  clearstack(Top_of_PageStack);  /* clear stack */
  Top_of_PageStack= NULL;
               /* quit reading this selection of articles? */

  if (mimefile)
  {
    DeleteFile(mimefile);
    FreeString(mimefile);
    mimefile= NULL;
  }

  *command= rcommand;

  switch (*command)
  {
    case CATCHUP:
      case  QUIT:
   case TOARTDIR:
      case BREAK:
                  leave= true;
                        break;

       case JUNK: if (!killed_it)
                     *command= NEXTART;
                  else
                     *command= last_direction; /* TO DO, BUG!! */
                  DispMsg(junkedmsg);
                  leave= false;
                  break;

    case          GOTO:
    case SEARCH_INART_B:
    case SEARCH_INART_F: noselection= true;
                         leave= false;
                         break;

         default: leave= false;
                  break;
  }

  return leave;      /* Quit this group? */
}
