/***********************************************************************
 * readgroup.c - article selector                                      *
 ***********************************************************************
 * Arn - (C)opyright by Roland Bless, All rights reserved,             *
 * -------- YOU ARE NOT ALLOWED TO DISTRIBUTE A CHANGED SOURCE ------- *
 ***********************************************************************/

#ifndef INCLUDES_SYM
 #include "includes.h"
#endif

#ifndef ENUMS_H
 #include "enums.h"
#endif

#ifndef GLOBALVARS_H
#include "globalvars.h"
#endif

#ifndef UCHAR
 #define UCHAR unsigned char
#endif

#undef DEBUG


/* Prototypes for functions defined in readgroup.c */
/* proto$start */
void FreeSubjandFrom(struct GroupInfoList *pGI);
void
InitArtArray(struct art_struct *RootArt, struct art_struct **LastArt);
struct subj_struct *ReadDataBase(struct GroupInfoList *pGI,
                                 struct art_struct *RootField,
                                 ULONG *LastArtNr);
void SetArtCursor(void);
void versioninfo(void);
BOOL SelectArtCommands(char key,
                       int *command);
void GetInfoforCompactView(struct art_struct *firstarticle,
                           ULONG *SizeOfThread,
                           ULONG *last_date,
                           UWORD filter);
void UpdateArtMenu(ULONG firstart,
                   int firstline,
                   int lines,
                   struct art_struct **artarray,
                   UWORD filter,
                   BOOL only_refresh_set);
int MiscMenu(struct art_struct *artarray[]);
int ViewMenu(void);
BOOL SelectArts(ULONG firstart,
                int lines,
                struct art_struct **artarray,
                UWORD filter,
                int *command);
ULONG FilterArtArray(struct subj_struct *RootSubjP,
                    struct art_struct **artarray,
                    UWORD filter, UWORD displaymode);
ULONG MarkThreadRead(struct art_struct *pArt);
int ReadGroup(struct GroupInfoList *pGI,
              int *command);
void
RefreshLines(ULONG firstart,
              int lines,
              struct art_struct *artarray[],
              BOOL only_refresh_set);
void
UnSelectThread(struct art_struct *pArt, UWORD filter),
SelectThread(struct art_struct *pArt, UWORD filter),
SelectCompact(struct art_struct *pArt, UWORD filter),
UnSelectCompact(struct art_struct *pArt, UWORD filter);


/* proto$end */

#include "proto/pager_proto.h"
#include "proto/arnhelp_proto.h"
#include "proto/strprocs_proto.h"
#include "proto/ArnIO_proto.h"
#include "proto/cmdline_proto.h"
#include "proto/functions_proto.h"
#include "proto/kill_proto.h"
#include "proto/grouplist_proto.h"
#include "proto/followup_proto.h"
#include "proto/Arn_proto.h"
#include "proto/choosegroups_proto.h"
#include "proto/time_proto.h"

/* from pager.c */
extern char directcommand, *noaction;

#ifdef DEBUG
extern printkill_list(struct kill_field *rootfield);
#endif

/* from version.c */
extern char *Version, *Last_UpdateD, *Last_UpdateT;
/* from choosegroups.c */
extern char *Menulines[];
extern int grouplines; /* Number of maximum lines that we can display for selection */
                       /* This is the same for groups and articles */

                            /*******************/
                   /* global variables from this module */
struct              /* standard structure to read in */
{
  ULONG ArtNr; /* because of longword alignment, chars must be defined later */
  long Lines;
  ULONG Date;
  UCHAR ArtStat;
  UCHAR FromLen;
} mask;


struct subj_struct *RootSubjP= NULL;
struct kill_field *LocalKill_List= NULL;
ULONG LastDBArtNr,  /* Nr of highest article nr in database */
      FirstArtNr,   /* Article # of the first article */
      LastArtNr,    /*                  last          */
      NrOfActArts,  /* Nr of highest index in artarray/
                      Nr of actual filtered subjects */
      NrOfSubjects;

UWORD displaymode= SUBJ_ALPHABET;

int art_cursory,    /* cursor (line) y-position in article selector */
    max_subjectlen;     /* maximum subject-length */

BOOL noselection= false; /* This skips the article-selection, e.g. a
                           cmd has to be interpreted from lower pager level */

long ToReadSelectedArts; /* must be long! (not ULONG) */
char *ArtMenuTemplate,*ArtMenuTemplate2, *EntryMessage;

                            /*******************/


void
FreeSubjandFrom(struct GroupInfoList *pGI)
/* Frees all Subject_lists allocated by one group and all From-strings, too!
   pGI->RootSubjP must be set! */
{
  struct subj_struct *subjp, *nextsubjp;
  struct art_struct *artp;
  ULONG NrOfArticles;

  if (!pGI || !RootArt) return;

  NrOfArticles= pGI->lastArtNr - pGI->firstArtNr + 1L;

  if (NrOfArticles<=0) return;

  for (subjp= pGI->RootSubjP; subjp!=NULL; subjp= nextsubjp)
  {
    FreeString(subjp->Subject);
    nextsubjp= subjp->NextSubject;
    FreeMem(subjp, (ULONG) sizeof(struct subj_struct));
  }
  /* ALL Subject-strings AND STRUCTURES are now free */
  /* Now free all FROM'S */
  for (artp= RootArt; artp<=LastArt; artp++)
  {
    FreeString(artp->From);
    artp->From= NULL;
  }

  pGI->RootSubjP= NULL;
  RootSubjP= NULL;
}


void
InitArtArray(struct art_struct *RootArt, struct art_struct **LastArt)
/* FirstArtNr and NrOfArticles must be set!! */
{
  register struct art_struct *ActArtP;
  register ULONG ArtNr;

  if (RootArt)
  {
    for (ArtNr= 0UL, ActArtP= RootArt; ArtNr<NrOfArticles; ArtNr++, ActArtP++)
    {
      ActArtP->Status= (NONEXISTENT | READ);
      ActArtP->NextArt= ActArtP->PrevArt= NULL;
      ActArtP->From= NULL;
      ActArtP->Subject= NULL;
      ActArtP->NrofFollowUps=
      ActArtP->Date= 0UL;
      ActArtP->ArticleNr= ArtNr+FirstArtNr;
      ActArtP->ExtData= NULL;
    }
    *LastArt= --ActArtP;
  }
}


                         /* READ DATABASE see arnmaster.c */
/* returns: pointer to the subject threads, LastArtNr
   needs:   pointer to group and base pointer from the array of pointers to art_structs
            pathtogroup must be set */
struct subj_struct *
ReadDataBase(struct GroupInfoList *pGI, struct art_struct *RootField, ULONG *LastArtNr)
{
  FILE *dbasefp;                        /* file pointer */
  ULONG readnr;                         /* nr of read datasets */
  struct art_struct *ActArtP,           /* actual article struct */
                    **PrevNextArtPP, *PrevArtP;
  ULONG FollowUpCount= 0UL;              /* FollowUps following */
  struct subj_struct *SubjectStructP= NULL, **NextSubjPP= NULL;
  char *Subject= NULL, *FromStr= NULL;
  BOOL VALID;                           /* article in active range? */
  UCHAR SubjLen= 0;                      /* length of the subj */
  int error;

  NrOfSubjects= 0UL;

  ActArtP= RootField;      /* TODO */
  *LastArtNr= FirstArtNr= pGI->firstArtNr;

  if (pGI->lastArtNr < pGI->firstArtNr) return NULL; /* empty group! */
  PrevNextArtPP= NULL;
  PrevArtP= NULL;

  RootSubjP= NULL;

  /* We suppose that pathtogroup was set before!!!! */
  sprintf(fnametmp,"%s/.newsdb",(pathtodbase!=NULL ? pathtodbase : pathtogroup));

  if (dbasefp= fretryopen(fnametmp,"r"))  /* Open the DataBase */
  {
    error= setvbuf(dbasefp, mBuffer, _IOFBF, M_BUFFERSIZE); /* for faster IO! */
    if (error)
       ErrorHandle(STRANGE,"setvbuf");

    if (!feof(dbasefp))
    do
    {
      /* read in one standard mask */
#ifdef DEBUG
      printf("ftell: %ld ",ftell(dbasefp));
#endif
      readnr= fread((char *) &mask,sizeof(mask),1 ,dbasefp);
#ifdef DEBUG
      printf("readnr: %x feof:%d ftell:%ld\n",readnr,feof(dbasefp),ftell(dbasefp));
#endif
      if (readnr)  /* we got it */
      {
        if (mask.ArtNr > pGI->lastArtNr || mask.ArtNr < pGI->firstArtNr)
        {
           VALID= false; /* article not in the active-range (expired) */
                         /* This must be ignored, but skipped! */
        }
        else
        {
          VALID= true;      /* Valid article-number */
          ActArtP= &RootField[mask.ArtNr-FirstArtNr]; /* Get struct in direct-access */
          ActArtP->Status= mask.ArtStat; /* copy status */
          ActArtP->Lines = mask.Lines;   /* copy lines */
          ActArtP->Date  = mask.Date;
          if (mask.ArtNr>*LastArtNr) *LastArtNr= mask.ArtNr;
        }
#ifdef DEBUG
        printf("#%ld Stat:%x=%x %ld FromLen:%d\n",mask.ArtNr,mask.ArtStat,ActArtP->Status,mask.Lines,mask.FromLen);
#endif
        /* From is a necessary entry */
        if (mask.FromLen>0)  /* should always be >0 */
        {
#ifdef DEBUG
          if (mask.FromLen>16) printf("From: > 16 chars!\n");
#endif
          FromStr= AllocMem((ULONG) mask.FromLen+1UL, MEMTYPE);

          if (FromStr)
          {
            readnr= fread(FromStr,mask.FromLen,1,dbasefp); /* Read Author */
            FromStr[mask.FromLen]= '\0';
#ifdef DEBUG
            printf("From: %s\n",FromStr);
#endif
          }
          else
            ErrorHandle(NOMEM,NULL);

          if (VALID)
             ActArtP->From= FromStr; /* link FROM into struct */
          else
             if (FromStr) FreeMem(FromStr, (ULONG) mask.FromLen+1UL);
        }
        else
        {
          sprintf(tmp,"%s/%lu",pGI->Name,mask.ArtNr);
          StatusLine();
          ErrorHandle(NOFROM,tmp); /* DBASE CORRUPT */
          break; /* exit do ..while */
        }

        if (FollowUpCount==0) /* Subject follows */
        {
          if (SubjectStructP && SubjectStructP->RootArt==NULL)
          { /* No valid article-nr followed yet, so free the reserved mem */
            if (SubjectStructP)
               FreeMem(SubjectStructP, (ULONG) sizeof(struct subj_struct));
            if (Subject)
               FreeMem(Subject, (ULONG) SubjLen+1);

            Subject= NULL;
          }

          if (SubjectStructP= AllocMem((ULONG) sizeof(struct subj_struct),MEMTYPE))
          {
            if (VALID)
            {
              if (RootSubjP==NULL) RootSubjP= SubjectStructP;
              ActArtP->Subject= SubjectStructP;   /* link SUBJECT-STRUCT */

              PrevNextArtPP= &ActArtP->NextArt;    /* For FollowUps */
              PrevArtP= ActArtP;                  /* For FollowUps */

              /* link NextSubject from last Subj to this one */
              if (NextSubjPP) *NextSubjPP= SubjectStructP;
              /* reference back to this first article containing subject */
              SubjectStructP->RootArt= ActArtP;
              /* preserve address of NextSubj link */
              NextSubjPP= &SubjectStructP->NextSubject;
              NrOfSubjects++;
            }
            else
            { /* preserve Subject if VALID FollowUp follows */
              SubjectStructP->RootArt= NULL;
              SubjectStructP->NextSubject= NULL;
              PrevArtP= NULL;
            }
            readnr= fread(&SubjLen,sizeof(UCHAR), 1, dbasefp);
#ifdef DEBUG
            printf("SubjLen: %d ",SubjLen);
#endif
            if (SubjLen>0) /* should always be >0 */
            {
              Subject= SubjectStructP->Subject= AllocMem((ULONG) SubjLen+1UL,MEMTYPE);
              if (Subject)
              {
                readnr= fread(Subject,SubjLen,1,dbasefp); /* get the SUBJECT */
                Subject[SubjLen]= '\0';
#ifdef DEBUG
                printf("Subject: %s\n",Subject);
#endif
              }
              else
                ErrorHandle(NOMEM,NULL);
            }
            else
            {
              StatusLine();
              sprintf(tmp,"%s/%lu",pGI->Name,mask.ArtNr);
              ErrorHandle(NOSUBJ,tmp);
              break;
            }
          }
          else
           ErrorHandle(NOMEM,NULL);

          if (mask.ArtStat & (PA_WITHF | FUP_WITHF))
          {     /* readin FollowUp count */
            readnr= fread((char *) &FollowUpCount, sizeof(ULONG), 1, dbasefp);
#ifdef DEBUG
            printf("FUP_count: %lu\n",FollowUpCount);
#endif
          }
        } /* endif FollowUpcount==0 */
        else
        {           /**** FOLLWUP-handling ****/
          if (VALID) /* link followup into art_struct list */
          {
            ActArtP->Subject= SubjectStructP; /* Fups have the same subject */
            ActArtP->PrevArt= PrevArtP;
            if (SubjectStructP->RootArt) /* VALID preceeded subject */
            {
              SubjectStructP->RootArt->NrofFollowUps++;
#ifdef DEBUG
            printf("NrOfFupS: %lu\n",SubjectStructP->RootArt->NrofFollowUps);
#endif
              *PrevNextArtPP= ActArtP; /* Link last NextArt to this struct */
            }
            else /** preceeding article containing subject was INVALID **/
            {
              /* This is maybe our first article now */
              if (RootSubjP==NULL) RootSubjP= SubjectStructP;
              /* link NextSubject from last Subj to this one */
              if (NextSubjPP) *NextSubjPP= SubjectStructP;
              /* reference back to this first article containing subject */
              SubjectStructP->RootArt= ActArtP;
              /* preserve address of NextSubj link */
              NextSubjPP= &SubjectStructP->NextSubject;
              NrOfSubjects++;
            }

            /* preserve addresses of of this struct for next links */
            PrevNextArtPP= &ActArtP->NextArt;
            PrevArtP= ActArtP;
          }
          FollowUpCount--;
        }
      }
    }
    while(!feof(dbasefp));

    if (PrevNextArtPP) *PrevNextArtPP= NULL;
    if (NextSubjPP) *NextSubjPP= NULL;

    fclose (dbasefp);
  } /* opened database */

  if (GlobalERROR)
  {
    pGI->RootSubjP= RootSubjP;
    FreeSubjandFrom(pGI);       /* DBASE CORRUPT */
    DeleteFile(fnametmp); /* Give arnmaster a chance */
  }

  return RootSubjP;
}


void SetArtCursor(void)
{
  GotoXY(1,art_cursory+2);
  (*ConWrite)(">\r");
  /* (*ConWriteLn)(">\x9b\x44"); */
}


void
versioninfo(void)
{
  sprintf(tmp,"Version: This is Arn %s (%s/%s) (C) by R. Bless",
                                    Version, Last_UpdateD, Last_UpdateT);
  StatusLine(); DispMsg(tmp); SetArtCursor();
}


BOOL SelectArtCommands(char key, int *command)
{
   BOOL leave;
   short int cu_num;

   if (cu_num= CursorKeys(&key))
   {
     switch (cu_num)
     {
       case CUNUM_UP: *command= LINEUP;
                 return false;
                 break;
       case CUNUM_DOWN: *command= LINEDOWN;
                 return false;
                 break;
       case CUNUM_RIGHT: *command= PAGEFORWARD;
                 return true;
                 break;
       case CUNUM_LEFT: *command= PAGEBACK;
                 return true;
                 break;
     }
   }


   switch(key)
   {
     case '-': *command= SELECTRANGE;
               break;
     case '<': *command= PAGEBACK;
               break;
     case '>': *command= PAGEFORWARD;
               break;
     case ' ': *command= SPACE;
               break;
     case '^': *command= FIRSTPAGE;
               break;
     case '$': *command= LASTPAGE;
               break;
     case '*': *command= SELECT_THREAD;
               break;
     case '+': *command= SELECT_NEW;
               break;
     case '@': *command= SELECT_INVERSE;
               break;
     case '#': *command= SELECT_ALL;
               break;
     case '%': *command= UNSELECT_ALL;
               break;

     case '\x0d': *command= NEXTP; /* Enter - READ this article */
                  break;
     case 'N': *command= NEXTGROUP;
               break;
     case 'P': *command= PREVGROUP;
               break;
     case 'C': *command= CATCHUP;
               break;
     case 'A': *command= SHOWALL;
               break;
     case 'E': *command= POSTARTICLE;
               break;
     case 'J': *command= JUNK;
               break;
     case 'U': *command= UNREADART;
               break;
     case 'M': *command= MISCMENU;
               break;
     case 'V': *command= VIEWMENU;
               break;
     case 'O': *command= OptionsMenu(ARTSEL);
               break;
     case 'K': *command= EDIT_KILL;
               break;
     case ctrl('T'):*command= TOGGLE_SCREENTITLE;        /* CTRL-T */
                 break;
     case ctrl('L'): *command= REFRESHSCREEN;    /* CTRL-L */
                  break;
     case 'B': versioninfo();
               break;
     case 'H': dohelp(H_ARTSEL);
               *command= REFRESHSCREEN;
               break;
     case 'Z': *command= READNOW;
               break;
     case '\t': *command= READNOW_ANDCATCHUP;
                break;
     case 'X': *command= TOGROUPDIR;
               break;
     case 'Q': *command= QUIT;
               break;
     case ctrl('C'): *command= BREAK;  /* CTRL-Q Quick exit */
               GlobalERROR= QUITARN;
               break;
     default: *command= NOP;
              break;
   }


   switch (*command) /* which command aborts THIS selection-page? */
   {                 /* those which are not handled by selectarticle */
     case READNOW:   /* or needs a REFRESH */
     case READNOW_ANDCATCHUP:
     case NEXTP:
     case SPACE:
     case PAGEFORWARD:
     case PAGEBACK:
     case FIRSTPAGE:
     case LASTPAGE:
     case REFRESHSCREEN:
     case QUIT:
     case BREAK:
     case NEXTGROUP:
     case PREVGROUP:
     case SHOWALL:
     case TOGROUPDIR:
     case CATCHUP:
     case POSTARTICLE:
     case EDIT_KILL:
     case JUNK:
     case UNREADART:
     case TOGGLE_SCREENTITLE:
          leave= true;
          break;
     default: leave= false;
              break;

   }

   return leave;
}


#ifdef DEBUG
/* diagnostic routine */
void
PrintGroup(struct GroupInfoList *pGI)
{
  ULONG NrOfArticles;
  ULONG ArtNr;
  struct subj_struct *subjP;
  struct art_struct *ActArtP;

  if (!pGI) return; /* Error! */

  NrOfArticles= pGI->lastArtNr - pGI->firstArtNr + 1L;
  printf("%lu Articles in %s:\n", NrOfArticles, pGI->Name);

  for (subjP= pGI->RootSubjP; subjP!=NULL; subjP= subjP->NextSubject)
  {
    strncpy(tmp,subjP->Subject,45);
    tmp[45]= '\0';
    for (ActArtP= subjP->RootArt; ActArtP!=NULL; ActArtP= ActArtP->NextArt)
    {
      ArtNr= GetRelArtNr(ActArtP)+(ULONG) FirstArtNr;
      printf("%6ld: %-45s %-16s %6lu\n",ArtNr, tmp, ActArtP->From, ActArtP->Lines);
    }
  }
}
#endif

int
MiscMenu(struct art_struct *artarray[])
/* Unsubscribe, Mark-actions on selected articles */
{
  char key1,key2;
  ULONG index;

  StatusLine();
  DispMsg("J)unk, U)nread, unS)ubscribe, Q)uit:");
  key1= GetUpKey();
  ClrLine();
  if (key1=='Q') return NOP;
  if (key1=='S') return UNSUB;
  if (key1!='J' && key1!='U')
  {
    DispMsg(noaction);
    return NOP;
  }
  DispMsg("S)elected or at C)ursor (default):");
  key2= GetUpKey();
  ClrLine();
  if (key2!='S')
  {
    switch(key1)
    {
      case 'J': return JUNK;
                break;
      case 'U': return UNREADART;
                break;
       default: DispMsg(noaction);
                return NOP;
                break;
    }
  }
  else
  for (index= 0L; index<NrOfActArts; index++)
  {
    if (artarray[index]->Status & SELECTED_ART)
    {
      artarray[index]->Status|= NEEDS_REFRESH;
      switch(key1)
      {
        case 'J': artarray[index]->Status|= READ;
                  break;
        case 'U': artarray[index]->Status&= ~(READ);
                  break;
         default: return NOP;
      }
    }
  }

  return REFRESHLINES;
}


int
ViewMenu(void)
{
  char key;
  UWORD display_choice;

  StatusLine();
  DispMsg("1)Subject alphabetical, 2)Subj. date, 3)Article date, 4)Compacted, Q)uit:");
  key= GetUpKey();
  ClrLine();
  switch(key)
  {
    case '1': display_choice= SUBJ_ALPHABET;
              break;
    case '2': display_choice= SUBJ_DATE;
              break;
    case '3': display_choice= ART_DATE;
              break;
    case '4': if (displaymode==SUBJ_ALPHABET)
                 display_choice= COMPACT_ALPHABET;
              else
                 display_choice= COMPACT_DATE;
              break;
    default:  DispMsg(noaction);
              return NOP;
              break;
  }

  if (display_choice!=displaymode) /* real change -> do a refresh */
  {
    displaymode= display_choice;
    return REBUILD_DISPLAY;
  }
  else
     return NOP;
}



void GetInfoforCompactView(struct art_struct *firstarticle, ULONG *SizeOfThread, ULONG *last_date, UWORD filter)
/* Get information that is displayed for compacted view */
/* return values in both ULONG variables */
{
  struct art_struct *part;

  *SizeOfThread= 1UL;
  *last_date= firstarticle->Date;
  for (part= firstarticle->NextArt; part!=NULL; part= part->NextArt)
  {
    if ((part->Status & filter)==0)
    {
      *last_date= part->Date;
      *SizeOfThread+=1;
    }
  }
}


void /* destroys GROUPMENU!!!! */
UpdateArtMenu(ULONG firstart, int firstline, int lines,
              struct art_struct *artarray[], UWORD filter, BOOL only_refresh_set)
/* Update ArtArray: firstart is the index of the first article to update
   (this is viewed as the first article on page {Menulines[0]})
   firstline is the "offset" to firstart (page), lines: nr of articles
   to update from offset on. artarray: array containing all pointers to
   valid articles. The RESULT is found in MENULINES[].
   only_refresh_set is true if only those lines should be "refreshed"
   which need refresh (indicated by NEEDS_REFRESH-flag)                  */
{
  char key, StatusFlags[5];
  ULONG i, lastline= firstline+lines, SizeOfThread, last_date;
  struct art_struct *pArt;
  struct subj_struct *prevSubj= NULL;
  char linesstr[30], date1[7], date2[7];

  if (firstline>0)
     prevSubj= artarray[firstart+firstline-1]->Subject;

  for (i= firstline; i<lastline; i++) /* begin with line firstline, then refresh */
                                     /* number of lines entries */
  {
    if (firstart+i<NrOfArticles)
    {
      pArt= artarray[firstart+i];
      if (!only_refresh_set || (pArt->Status & NEEDS_REFRESH))
      {
        strcpy(StatusFlags,"    "); /* reset/init Statusflags */
        if (pArt->Status & NEW)
           StatusFlags[0]= 'N';
        if (pArt->Status & READ)
           StatusFlags[1]= 'R';

        if (pArt->Status & (FOLLOWUPART | FUP_WITHF))
        {
          if (pArt->Subject!=prevSubj || i==0)
          {
            strcpy(tmp,"Re: ");
            strncpy(tmp+4,pArt->Subject->Subject,max_subjectlen-4);
          }
          else
            strcpy(tmp,">"); /* means normal followup */
        }
        else
        {
          strncpy(tmp,pArt->Subject->Subject,max_subjectlen);
        }

        tmp[max_subjectlen]= '\0';
        if (pArt->Lines<0)
           strcpy(linesstr,"-");
        else
           ltostr(pArt->Lines,linesstr,29);

        if (i<MENU_OFFSET_1)
           key= i+'a';
        else
        if (i<=MENU_OFFSET_1+'z'-'a')
           key= i+'a'-MENU_OFFSET_1;
        else
           key= i+'A'-(MENU_OFFSET_1+'z'-'a'+1);

        if (displaymode!=COMPACT_ALPHABET && displaymode!=COMPACT_DATE)
        {
           sprintf(Menulines[i],ArtMenuTemplate,
                    key, pArt->From, tmp, linesstr, StatusFlags, pArt->ArticleNr);
        }
        else
        {
          tmp[max_subjectlen-14]= '\0';
          GetInfoforCompactView(pArt, &SizeOfThread, &last_date, filter);

          strcpy(date1,SuperShortDateStr(pArt->Date));
          strcpy(date2,SuperShortDateStr(last_date));
          sprintf(Menulines[i],ArtMenuTemplate2,
                   key, pArt->From, tmp, SizeOfThread, StatusFlags,
                   linesstr, date1, date2);
          /* key, pArt->From, Subject, NrOfFollowUps, StatusFlags, linestr, first_date, last_date */
        }
        Menulines[i][MaxChars]= '\0';

      } /* endif !only_refresh_set || NEEDS_REFRESH */
      prevSubj= pArt->Subject;
    } /* endif valid article */
    else
       Menulines[i][0]= '\0';
  } /* end for */
}


void
RefreshLines(ULONG firstart, int lines,
             struct art_struct *artarray[], BOOL only_refresh_set)
/* only printfs those lines from Menulines with NEEDS_REFRESH set */
/* NOTE: This function doesn't alter MenuLines[]! See UpdateArtMenu! */
{
  int i;
  BOOL tmpbool;
  UWORD stat;
  ULONG actart;

  if (!only_refresh_set) /* start with headerline */
  {
    /* headerline */
    sprintf(tmp,"%s - page %3lu of %3lu - %lu of %lu", newsgroup,
                      longceil(firstart+1,grouplines), longceil(NrOfActArts,grouplines),
                      NrOfActArts, NrOfArticles);

    /* fill to end of line with spaces, so we get a black bar */
    for(i= strlen(tmp); i<MaxChars-1-17; i++)
       tmp[i]= ' ';
    strcpy(tmp+i,"lines       artnr");
    tmp[MaxChars]= '\0';

    /* print headerline */
    ClrScr();
    InvON();
    (*ConWriteLn)(tmp);
    InvOFF();
  }

  /* display page of article subjects */
  for(i= 0,actart= firstart; i<lines; i++,actart++)
  {
    stat= artarray[actart]->Status;

    if (!only_refresh_set)
    {
      (*ConWriteChar)(i<MENU_OFFSET_1 ? ' ' : OFFSET_1_CHAR);
    }
    else
    {
      if (stat & NEEDS_REFRESH)
      {
        artarray[actart]->Status&= ~NEEDS_REFRESH; /* delete flag */
        GotoXY(2,i+2); /* position cursor */
      }
      else
        continue; /* don't print the line */
    }

    /* Print the line inverse video or normal */
    if (tmpbool= (stat & SELECTED_ART))
       InvON();

    (*ConWrite)(Menulines[i]);

    if (tmpbool) InvOFF();
  }

}


void
SelectCompact(struct art_struct *pArt, UWORD filter)
{
  if (pArt==NULL || (displaymode!=COMPACT_ALPHABET && displaymode!=COMPACT_DATE))
     return;
  else
  {
    pArt->Status|= NEEDS_REFRESH;

    for (pArt= pArt->Subject->RootArt; pArt!=NULL; pArt= pArt->NextArt)
    {
      if (!(pArt->Status & filter))
        pArt->Status|= SELECTED_ART;
    }
  }
}


void
UnSelectCompact(struct art_struct *pArt, UWORD filter)
{
  if (pArt==NULL || (displaymode!=COMPACT_ALPHABET && displaymode!=COMPACT_DATE))
     return;
  else
  {
    pArt->Status|= NEEDS_REFRESH;

    for (pArt= pArt->Subject->RootArt; pArt!=NULL; pArt= pArt->NextArt)
    {
      if (!(pArt->Status & filter))
        pArt->Status&= ~SELECTED_ART;
    }
  }
}



void
SelectThread(struct art_struct *pArt, UWORD filter)
{
  if (pArt)
  for (pArt= pArt->Subject->RootArt; pArt!=NULL; pArt= pArt->NextArt)
      if (!(pArt->Status & filter)) pArt->Status|= (SELECTED_ART | NEEDS_REFRESH);
}


void
UnSelectThread(struct art_struct *pArt, UWORD filter)
{
  if (pArt)
  for (pArt= pArt->Subject->RootArt; pArt!=NULL; pArt= pArt->NextArt)
  {
    if (!(pArt->Status & filter))
    {
      pArt->Status&= ~SELECTED_ART;
      pArt->Status|= NEEDS_REFRESH;
    }
  }
}


                     /* ARTICLE SELECTION LEVEL */
BOOL
SelectArts(ULONG firstart, int lines, struct art_struct *artarray[],
             UWORD filter, int *command) /* uses tmp !!!! */
/* Input-procedure to select articles. */
{
  int i, begin, end, selected, seltmp;   /* range and selection indices */
  char key;                      /* input chars */
  BOOL leave= false;             /* leave this selection level */
  ULONG actart;
  struct art_struct *pArt;
  FILE *article;

  *command= NOP;

  /* Built Screen */

  UpdateArtMenu(firstart, 0, lines, artarray, filter, false);
  RefreshLines(firstart,lines,artarray,false);
  if (EntryMessage)
  {
    StatusLine();
    DispMsg(EntryMessage);
    FreeString(EntryMessage);
    EntryMessage= NULL;
  }

  leave= false;
  begin= selected= end= -1; /* begin & selected:0..25 ! */
  SetArtCursor(); /* Set cursor to begin */

  do  /* main loop */
  {
    *command= NOP;
    key=(*GetKey)();  /* get the command */

    /* for a range last <selected> must be retained, so don't call GetLineNrFromKey() */
    if ((seltmp= GetLineNrFromKey(key, lines))>=0)
    {
      selected= seltmp;
      if (begin>=0)  /* if begin of range was given then end of range is now given */
      {
        end= selected;
        GetKeyFromLine(end);
        ClrLine();
        SetArtCursor();
        if (begin>end) /* swap range */
        {
          int help= begin;

          begin= end-1;
          end= help-1;
        }

        /* select range */
        for (i= begin+1,actart= firstart+i; i<=end; i++,actart++)
        {
          if (artarray[actart]->Status & SELECTED_ART)
          {
            artarray[actart]->Status&= ~SELECTED_ART;
            DeSelectLine(i, &art_cursory);
            UnSelectCompact(artarray[actart],filter);
          }
          else
          {
            artarray[actart]->Status|= SELECTED_ART;
            SelectLine(i, &art_cursory);
            SelectCompact(artarray[actart],filter);
          }
        }
        end= begin= -1;
      }
      else
      { /* normal SINGLE selection */
        actart= firstart+selected;
        /* select or deselect depending on previous state */
        if (artarray[actart]->Status & SELECTED_ART)
        {
          artarray[actart]->Status&= ~SELECTED_ART;
          DeSelectLine(selected, &art_cursory);
          UnSelectCompact(artarray[actart],filter);
        }
        else
        {
          artarray[actart]->Status|= SELECTED_ART;
          SelectLine(selected, &art_cursory);
          SelectCompact(artarray[actart],filter);
        }
      }
      SetArtCursor();
    }
    else /* check for command */
    {
      if (begin>=0 && *command==NOP)/* no real range */
      {
        begin= selected= -1;
        ClrLine();
        *command= NOP;
      }
      else
      {
        leave= SelectArtCommands(key, command);
        if (key!=ctrl('Z')) cmdline[0]= '\0';       /* check for command-line */
        EnterCmdLine(key, ARTSEL);            /* This MUST hold a disjoint set
                                                  of keys with GroupSelComm */
        if (*cmdline!='\0')
        {
          leave= InterpretCmdLine(cmdline, command, ARTSEL);
        }
      }
    }

    if (*command==MISCMENU)
       *command= MiscMenu(artarray);
    else
    if (*command==VIEWMENU)
       *command= ViewMenu();

    switch(*command) /* Which action can we examine? */
    {
      case LINEUP  : if (art_cursory>0)
                     {
                       (*ConWriteChar)(art_cursory<MENU_OFFSET_1 ? ' ' : OFFSET_1_CHAR);
                       art_cursory--;
                       SetArtCursor();
                     }
                     break;

      case LINEDOWN: if (art_cursory<lines-1)
                     {
                       (*ConWriteChar)(art_cursory<MENU_OFFSET_1 ? ' ' : OFFSET_1_CHAR);
                       art_cursory++;
                       SetArtCursor();
                     }
                     break;

      case SELECTRANGE: if (selected>=0)
                        {
                          begin= selected;
                          (*ConWriteChar)(art_cursory<MENU_OFFSET_1 ? ' '
                                                                    : OFFSET_1_CHAR);
                          StatusLine();
                          (*ConWrite)(">Select range: ");
                          GetKeyFromLine(selected);
                          (*ConWriteChar)('-');
                        }
                        break;

      case SELECT_NEW: for (actart= 0L; actart<NrOfActArts; actart++)
                         if ((artarray[actart]->Status & (NEW | READ))==NEW)
                         {
                           artarray[actart]->Status|= (SELECTED_ART | NEEDS_REFRESH);
                         }
                       *command= REFRESHLINES;
                       break;

      case SELECT_THREAD: if (selected>=0)
                          { /* select all articles with same subject, but only unread articles! */
                            SelectThread(artarray[firstart+selected], filter);
                          }
                          *command= REFRESHLINES;
                          break;

      case SELECT_ALL: for (actart= 0L; actart<NrOfActArts; actart++)
                           if (!(artarray[actart]->Status & SELECTED_ART))
                             artarray[actart]->Status|= (SELECTED_ART | NEEDS_REFRESH);
                       *command= REFRESHLINES;
                       break;

      case UNSELECT_ALL: for (actart= 0L; actart<NrOfActArts; actart++)
                             if (artarray[actart]->Status & SELECTED_ART)
                             {
                               artarray[actart]->Status&= ~SELECTED_ART;
                               artarray[actart]->Status|= NEEDS_REFRESH;
                             }
                         *command= REFRESHLINES;
                         break;

      case SELECT_INVERSE: for (actart= firstart; actart<firstart+lines; actart++)
                             if (artarray[actart]->Status & SELECTED_ART)
                                 artarray[actart]->Status&= ~SELECTED_ART;
                              else
                                 artarray[actart]->Status|= SELECTED_ART;
                           *command= REFRESHSCREEN; /* complete refresh is a must */
                           leave= true;
                           break;

      case POSTARTICLE: ClrScr();
                        PostArticle(newsgroup);
                        *command= REFRESHSCREEN;
                        break;

      case EDIT_KILL: StatusLine();
                      ChangeKillFile(artarray[firstart+art_cursory], ARTSEL);
                      if (GlobalERROR) WaitForSpace(NULL);
                      *command= NOP;
                      break;

      case CATCHUP: if (!CatchUp(true))
                       *command= NEXTGROUP;
                    else
                      *command= NOP;
                    break;

      case JUNK: pArt= artarray[firstart+art_cursory];
                 if (*cmdline=='\0')
                 {
                   pArt->Status|= (READ| NEEDS_REFRESH);
                   sprintf(fnametmp, "%s/%lu",pathtogroup, pArt->ArticleNr);
                   if (pArt->Status & XREF)
                   {
                     if (article= fopen(fnametmp,"r"))
                     {
                       MarkXRefs(article, false);
                       fclose(article);
                     }
                   }
                   *command= REFRESHLINE;
                 }
                 else
                 {
                   MarkRange(pArt, *command);
                   *command= REFRESHSCREEN;
                 }
                 break;

     case SAVEART_WH:
     case SAVEART_NH: SaveArticle(artarray[firstart+art_cursory], cmdline, *command);
                      *command= REFRESHSCREEN;
                      break;

        case UNREADART: pArt= artarray[firstart+art_cursory];
                        if (cmdline[0]=='\0')
                        {
                          pArt->Status&= ~READ;
                          pArt->Status|= NEEDS_REFRESH;
                          *command= REFRESHLINE;
                        }
                        else
                        {
                          MarkRange(pArt, *command);
                          *command= REFRESHSCREEN;
                        }
                        break;
        case SELECT:   /* only returned by interpretcmdline! */
        case UNSELECT: MarkRange(artarray[firstart+art_cursory], *command);
                       *command= REFRESHSCREEN;
                       break;
        case REBUILD_DISPLAY:
        case UNSUB: leave= true;
                    break;
       case PAGEFORWARD: if (firstart+lines>=NrOfActArts) *command= NOP;
                         break;
       case PAGEBACK:    if (firstart<grouplines) *command= NOP;
                         break;
    } /* end of switch */

    if (*command==REFRESHLINE) /* refresh menuline in cursor-line */
    {
      UpdateArtMenu(firstart, art_cursory, 1, artarray, filter, false);
      GotoXY(2,art_cursory+2);
      if (artarray[firstart+art_cursory]->Status & SELECTED_ART)
         InvON();
      (*ConWrite)(Menulines[art_cursory]);
      InvOFF();
      *command= NOP; /* SetArtCursor */
    }
    else
    if (*command==REFRESHLINES)
    {
      UpdateArtMenu(firstart, 0, lines, artarray, filter, true);
      RefreshLines(firstart,lines,artarray,true);
      *command= NOP;
    }

    if (*command==NOP)
    {
      /* (*ConWriteChar)(art_cursory<MENU_OFFSET_1 ? ' ' : OFFSET_1_CHAR); */
      SetArtCursor();
      leave= false;
    }
  }
  while (!leave);

  switch (*command) /* which commands terminate ARTICLESELECTION? */
  {
    case PREVGROUP:
    case NEXTGROUP:
    case TOGROUPDIR:
    case CATCHUP:
    case QUIT:
    case BREAK:
         StatusLine();
         leave= true;
         break;
    default: leave= false;
             break;
  }

  return leave;
}


int
subj_datecmp(struct subj_struct **firstsubjP, struct subj_struct **secondsubjP)
{
  ULONG date1, date2;

  date1= (*firstsubjP)->RootArt->Date;
  date2= (*secondsubjP)->RootArt->Date;

  if (date1 < date2)
     return -1;
  else
  if (date1 == date2)
     return 0;
  else
     return 1;
}


int
art_datecmp(struct art_struct **firstartP, struct art_struct **secondartP)
{
  ULONG date1, date2;

  date1= (*firstartP)->Date;
  date2= (*secondartP)->Date;

  if (date1 < date2)
     return -1;
  else
  if (date1 == date2)
     return 0;
  else
     return 1;
}


ULONG
FilterArtArray(struct subj_struct *RootSubjP, struct art_struct **artarray ,
               UWORD filter, UWORD displaymode)
/* writes only those articles into artarray if article status has not a bit
   of filter-mask set */
{
  ULONG count, NrOfFiltered, index, SizeOfThread;
  struct subj_struct *subjP, **subjParray= NULL;
  struct art_struct *ActArtP;

  switch(displaymode)
  {
    case SUBJ_DATE:
    case COMPACT_DATE:
                      if ((subjParray= AllocMem(NrOfSubjects *
                                                sizeof(struct subj_struct *), MEMTYPE))==NULL)
                      {
                        ErrorHandle(NOMEM, NULL);
                        return 0UL;
                      }
                      else
                      { /* sort subjects to date order */
                        /* fill array with pointers */
                        for (index= 0UL, subjP= RootSubjP; subjP!=NULL; subjP= subjP->NextSubject, index++)
                        {
                          subjParray[index]= subjP;
                        }
                        /* sort them */
                        StatusLine();
                        DispMsg("Sorting...");
                        qsort(subjParray, NrOfSubjects,
                              sizeof(struct subj_struct *), subj_datecmp);
                        DispMsg("Ready!");
                        ClrLine();
                      }
                      break;
    default: break;
  }

  count= 0UL;
  switch(displaymode)
  {
    case SUBJ_DATE:
             for (index= 0UL; index<NrOfSubjects; index++)
             {
               for (ActArtP= subjParray[index]->RootArt; ActArtP!=NULL; ActArtP= ActArtP->NextArt)
                   if (count<NrOfArticles && (ActArtP->Status & filter)==0)
                      artarray[count++]= ActArtP;
             };
             break;
    case COMPACT_DATE:
    case COMPACT_ALPHABET:
             for (index= 0UL; index<NrOfSubjects; index++)
             {
               SizeOfThread= 0UL;
               for (ActArtP= subjParray[index]->RootArt; ActArtP!=NULL; ActArtP= ActArtP->NextArt)
               {
                 if (count<NrOfArticles && (ActArtP->Status & filter)==0)
                 {
                   artarray[count++]= ActArtP;
                   break; /* store only first subject in artarray */
                 }
               }
             };
             break;
    case SUBJ_ALPHABET:
    case ART_DATE:
    default:
             for (subjP= RootSubjP; subjP!=NULL; subjP= subjP->NextSubject)
             {
               for (ActArtP= subjP->RootArt; ActArtP!=NULL; ActArtP= ActArtP->NextArt)
                   if (count<NrOfArticles && (ActArtP->Status & filter)==0)
                      artarray[count++]= ActArtP;
             };
             break;
  }

  NrOfFiltered= count; /* We now have this # of articles filtered */

  if (displaymode==ART_DATE)
  {
    StatusLine();
    DispMsg("Sorting...");
    qsort(artarray, NrOfFiltered,
          sizeof(struct art_struct *), art_datecmp);
    DispMsg("Ready!");
    ClrLine();
  }
  /* set the rest of the artarray to NULL */
  while (count<NrOfArticles) artarray[count++]= NULL;

  if (subjParray)
  {
    FreeMem(subjParray, NrOfSubjects*sizeof(struct subj_struct *));
    ClrLine();
  }
  return NrOfFiltered;
}


ULONG MarkThreadRead(struct art_struct *pArt)
/* returns number of marked articles */
{
  struct art_struct *tmp_pArt;
  ULONG counter= 0UL;

  for (tmp_pArt= pArt->Subject->RootArt; tmp_pArt; tmp_pArt= tmp_pArt->NextArt)
  {
    tmp_pArt->Status&= ~SELECTED_ART;
    tmp_pArt->Status|= READ;
    counter++;
  }
  return counter;
}



int ReadGroup(struct GroupInfoList *pGI, int *command)
/* main routine which shows article list */
{
  /* Remarks:
     <artarray> is an array of pointers to articles, big enough for all articles,
     i.e. sizeof(artarray)==NrOfArticles. It is filled up only upto NrOfActArts,
     which is done by FilterArtarray().
  */
  struct art_struct *pArt,
         *pgotoArt= NULL; /* pointer to article to jump to */
  ULONG count;
  int lastcommand;
  struct subj_struct *RootSubjP;
  ULONG ArtNr;
  struct art_struct *ActArtP, **artarray;   /*** artarray[] ****/
  struct Markers *pmark;
  int lines, oldMaxScrLines;
  UWORD filter= READ; /* We don't want to see READ articles */
  ULONG i, firstart, beforegoto= 0L, nr_of_junked= 0UL;
  BOOL leave= false, showall= false, gotomode= false, catchupgroup= false;

  /******************/

  ToReadSelectedArts= -1L;

  /* Less lines than MAX_SELECT_LINES ? */
  EntryMessage= NULL;
  lines= grouplines;
  max_subjectlen= MaxChars-40;
  firstart= 0L;
  art_cursory= 0;

  /* Init Template "%c %-16s %-40s %6s %4s %6ld\n" */
  sprintf(tmp, "%%c %%-16s %%-%ds %%6s %%4s %%6ld\n", max_subjectlen);
               /* key, pArt->From, Subject, linesstr, StatusFlags, pArt->ArticleNr */
               /* 40 */
  ArtMenuTemplate= CopyString(tmp); /* is checked for NULL later! */
  sprintf(tmp, "%%c %%-16s %%-%ds [%%4ld] %%4s %%6s %%6s/%%6s\n", max_subjectlen-14);
   /* key, pArt->From, Subject, NrOfFollowUps, StatusFlags, linestr, first_date, last_date */
   /* 52 */
  ArtMenuTemplate2= CopyString(tmp); /* is checked for NULL later! */

  lastcommand= *command;
  strcpy(newsgroup,pGI->Name);

  if (!(switches & SW_DIRONLY))
  {
    MakeDirName(tmp, newsdir, newsgroup);
    if (dbasedir)
    {
      char stmp[MAXLINE+1];

      MakeDirName(stmp, dbasedir, newsgroup);
      pathtodbase= CopyString(stmp);
    }
  }
  else
    strcpy(tmp,pGI->Name);

  strncpy(pathtogroup,tmp,MAXLINE); /* In pathtogroup is now the path
                                       to the group! */
  pathtogroup[MAXLINE]= '\0';

  GlobalERROR= 0;

  if (pGI->Status & MODERATED)
     GeneralBool|= G_MODERATED;

  if (!pGI) return false; /* Error! */
  LastArtNr= pGI->lastArtNr;
  FirstArtNr= pGI->firstArtNr;
  NrOfArticles= LastArtNr - FirstArtNr + 1L;
  oldMaxScrLines= MaxScrLines; /* oldMaxScrLines is used for recognizing
                                  a ToggleScreenTitle() in the pager level */

  if (NrOfArticles<=0L)
  {
    StatusLine(); DispMsg("Empty group!");
    Delay(50L);
    return false;
  }

  /* Reserve array of article-structures */
  RootArt= AllocMem((ULONG) sizeof(struct art_struct)*NrOfArticles, MEMTYPE);
  artarray= AllocMem((ULONG) sizeof(struct art_struct *)*NrOfArticles, MEMTYPE);

  if (!RootArt || !artarray || !ArtMenuTemplate || !ArtMenuTemplate2)
  {
    ErrorHandle(NOMEM, NULL);
    return false;
  }

  /* Initialize Array */
  InitArtArray(RootArt, &LastArt);

  /* Read in the DATABASE */
  for (count= 1UL; count<=2; count++)
  {
    StatusLine();
    DispMsg("reading database...");
    RootSubjP= ReadDataBase(pGI, RootArt, &LastDBArtNr);
    /* RootSubjP is set to NULL, if no dbase exists */
    pGI->RootSubjP= RootSubjP;
    ClrLine();
    if (!RootSubjP || LastDBArtNr<LastArtNr)
    {
      StatusLine();

      if (RootSubjP)
      {
        if (count==1) /* update obsolete dbase only once! */
        {
          DispMsg("Obsolete Database, updating...");
          FreeSubjandFrom(pGI); /* Free all stuff we have allocated so far */
          /* reset complete array */
          InitArtArray(RootArt,&LastArt);
        }
      }
      else
      {
        if (!RootSubjP && count>1)
        {
          DispMsg("Empty newsgroup...");
          Delay(50L);
          break;
        }
        else
          DispMsg("Creating Database...");
      }
      DispMsg("please wait a moment...");
      /* TODO, Signal arnmaster to collect data */
      sprintf(tmp, "arnmaster -g %s", newsgroup);
      Execute(tmp, (BPTR) NullFH, (BPTR) NullFH);
      DispMsg("ready.");
      ClrLine();
    }
    else /* Leave loop */
      break;
  }

  if (RootSubjP) /* we have articles */
  {
    /* mark all new articles as new */

    if (pGI->UserlastArtNr)
       ArtNr= pGI->UserlastArtNr+1;
    else
       ArtNr= FirstArtNr; /* no UserlastArtNr existed before, so all articles are new */

    if (ArtNr>=FirstArtNr && ArtNr<=LastArtNr)
       for (ArtNr-= FirstArtNr,ActArtP= &RootArt[ArtNr]; ArtNr<NrOfArticles; ArtNr++, ActArtP= &RootArt[ArtNr])
       {
         ActArtP->Status|= NEW;
       }

    /* mark read articles as read */
    for (pmark= pGI->Marks; pmark!=NULL; pmark= pmark->NextMark)
      if (pmark->Mark)
        CheckLibMarks(pmark->Mark,pGI);        /* Get the Marks */
      else
        ErrorHandle(STRANGE,"nomarkstr"); /* There must be a markstring */

    /* local X-file processing */
    if (GeneralBool & LOCALKILLON)
    {
      FreeKill_List(LocalKill_List);
      LocalKill_List= NULL;
      sprintf(fnametmp,"%s/.X%s",(pathtodbase ? pathtodbase : pathtogroup),UserName);
      ReadKillFile(&LocalKill_List, fnametmp);
#ifdef DEBUG
            printf("Local:\n");
            /* printkill_list(LocalKill_List); */
#endif

      /* kill all Subjects: and From: before we show the articles */
      nr_of_junked+= prekill(LocalKill_List, RootSubjP);
    }

    if (GeneralBool & GLOBALKILLON)
       nr_of_junked+= prekill(GlobalKill_List, RootSubjP);

    *tmp='\0';
    /* Fill array with pointers */
    NrOfActArts= FilterArtArray(RootSubjP, artarray, filter, displaymode);
    if (NrOfActArts<=0L) /* none with that filter? */
    {
      filter= NONEXISTENT; /* Display all articles! */
      NrOfActArts= FilterArtArray(RootSubjP, artarray, filter, displaymode);
      if (NrOfActArts<=0L) leave= true; /* We still have nothing, return */
      else
         strcpy(tmp,"No unread articles. ");
    }

    if (nr_of_junked)
    {
      sprintf(fnametmp, "Junked %lu articles by X-file!", nr_of_junked);
      strcat(tmp, fnametmp);
    }
    if (*tmp) EntryMessage= CopyString(tmp);

    /* main loop */
    while (*command!=QUIT && *command!=BREAK && !leave)
    {
      lines= NrOfActArts-firstart;
      /* firstart is the NR of the first article that was not already displayed */
      if (lines>grouplines) lines= grouplines;
      else
      if (art_cursory>=lines) art_cursory= lines/2; /* cursor is behind lines, reset */

      switch(*command) /* paging set the cursor into middle of each page */
      {
        case PAGEFORWARD:
        case PAGEBACK:
        case SPACE:
        case LASTPAGE:
        case FIRSTPAGE: art_cursory= lines/2;
                        break;
      }

      /* action goes here */
      if (!noselection)
         leave= SelectArts(firstart, lines, artarray, filter, command);
      else
        leave= noselection= false; /* skip selection only once */

               /******************************************************/
      /* leave means to quit Article-SELECTION */
      if (!leave)
      {
        if (*command==SHOWALL || *command==REBUILD_DISPLAY)
        {
          int tmpfilter= 0;

          if (*command==SHOWALL)
          {
            if (showall)  /* toggle */
            {
              tmpfilter= filter;
              showall= false;
            }
            else
              showall= true;
          }
          else
            tmpfilter= filter;

          NrOfActArts= FilterArtArray(RootSubjP, artarray, tmpfilter, displaymode); /* show all */
          if (NrOfActArts<=0L)
          {
            filter= NONEXISTENT; /* Display all articles! */
            NrOfActArts= FilterArtArray(RootSubjP, artarray, filter, displaymode);
          }
          if (firstart>NrOfActArts)
             *command= FIRSTPAGE;
          else
          {
            *command= REFRESHSCREEN;
            if (art_cursory+firstart>NrOfActArts)
               art_cursory= (NrOfActArts-firstart)/2;
          }
        }

        if (*command==PAGEBACK)
        {
          if (firstart>=grouplines) firstart-= grouplines;
          else
          {
            firstart= 0L;
          }
        }
        else
        if (*command==PAGEFORWARD || *command==SPACE)
        {
          if (firstart+lines<NrOfActArts) firstart+= lines;
          else
          {
            if (*command==SPACE)
            {
              if (CountArticles(SELECTED_ART,SELECTED_ART))
                *command= READNOW;
              else
              {
                *command= NEXTGROUP;
                StatusLine();
                leave= true;
              }
            }
          }
        }
        else
        if (*command==LASTPAGE)
        {
          firstart= ((NrOfActArts-1)/grouplines)*grouplines;
        }
        else
        if (*command==FIRSTPAGE)
        {
          firstart= 0L;
        }
        else
        if (*command==UNSUB)
        {
          Unsubscribe(pGI);
          if (pGI->Status & UNSUBSCRIBED)
          {
            *command= NEXTGROUP;
            leave= true;
          }
          else
            *command= NOP;
        }
        else
        if (*command==SEARCH_INART_F || *command==SEARCH_INART_B)
        {
          ULONG index= firstart+art_cursory, found;
          /* index is the current cursor position */

          StatusLine();
          if(index==(found= SearchInArt(artarray, NrOfActArts, index, *command)))
          {
            *command= NOP; /* not found */
          }
          else
          {
            pgotoArt= artarray[found];
            firstart= (found/grouplines)*grouplines;
            art_cursory= found-firstart;
            *command= NEXTP;
          }
        }

        if (*command==GOTO)  /* GOTO ARTICLE */
        {
          int i= 0;
          ULONG artnr;

          pgotoArt= NULL;

          while(isdigit(cmdline[i])) i++;
          cmdline[i]= '\0';
          if (cmdline[0]!='\0')
          {
            artnr= (ULONG) atol(cmdline);
            if (artnr>=FirstArtNr && artnr<=LastArtNr)
            {
              pgotoArt= &RootArt[artnr-FirstArtNr]; /* valid goto pointer */
              /* set new cursor position */
              *command= NEXTP; /* read this single article */
            }
            else
            {
              StatusLine();
              sprintf(tmp, "Couldn't find that article %lu!",artnr);
              DispMsg(tmp);
              Delay(50L);
              if (gotomode) *command= READNOW; /* return to selection */
              else
                *command=NOP;
            }
          }
          else
          {
            if (gotomode) *command= READNOW; /* return to selection */
            else
              *command= NOP;
          }
        } /* if GOTO */

        if (*command==NEXTP) /* read one single article */
        {
          if (!pgotoArt)
             pArt= artarray[firstart+art_cursory];
          else
            pArt= pgotoArt;

          oldMaxScrLines= MaxScrLines;
          leave= ReadArticle(&pArt, command);

          if (*command==MARK_THREAD_READ)
          {
            MarkThreadRead(pArt);
            *command= SPACE;
          }

          if (*command==SPACE)
             *command= REFRESHSCREEN;

          if (*command==QUIT || *command==TOARTDIR)
          {
            leave= false;
            *command= REFRESHSCREEN;
          }

          if (pgotoArt)
          {
            pgotoArt= NULL;
            if (gotomode) *command= READNOW; /* return to selections */
          }
        } /* endif NEXTP */

        if (*command==READNOW || *command==READNOW_ANDCATCHUP)
        {
          ULONG NrOfSelectedArts= 0L;

          i= 0L;


          if (gotomode) /* returning from goto to article */
          {
            i= beforegoto;
            gotomode= false;
          }
          else
          {
            catchupgroup= (*command==READNOW_ANDCATCHUP);
            if ((NrOfSelectedArts= CountArticles(SELECTED_ART,SELECTED_ART))==0L) /* nothing selected */
            {
              catchupgroup= false;
              if (!CatchUp(true))  /* ask before catchup */
              {
                *command= NEXTGROUP;
                leave= true;
              }
              else
                 *command= NOP;
            }
          }

          ToReadSelectedArts= NrOfSelectedArts-1;

          /************* while *************/
          while (i<NrOfActArts && !leave)
          {
            if (artarray[i]->Status & SELECTED_ART)
            {
              if (*command==NEXTART && i<NrOfArticles) ToReadSelectedArts--;
              else
              if (*command==PREVART) ToReadSelectedArts++;

              oldMaxScrLines= MaxScrLines;
              pArt= artarray[i];
              leave= ReadArticle(&pArt, command);

              if (*command==MARK_THREAD_READ)
              {
                ToReadSelectedArts-= MarkThreadRead(pArt);
                *command= NEXTART;
              }

              if (*command==QUIT || *command==SPACE)
              {
                leave= false;
                *command= NEXTART; /* return to article-selection */
              } /* QUIT returned by READARTICLE doesn't
                   mean to quit ArticleSelection!*/

              if (*command==TOARTDIR) /* exit while */
              {
                *command= REFRESHSCREEN;
                catchupgroup= false;  /* abort catchup */
                leave= false;

                i= NrOfActArts; /* leave this loop, but return to art-selection */
              }
              if (*command==TOGROUPDIR) /* exit while */
              {
                i= NrOfActArts;
                leave= true; /* leave this article selection, too */
              }
              else
              if (*command==GOTO)
              {
                gotomode= true;
                beforegoto= i; /* preserve article-# for returning */
                i= NrOfActArts; /* leave while */
              }
            } /* end if SELECTED_ART */


            if (*command==PREVART)
            {
               if (i>0L)
                 i--;
               else
               {
                 *command= NEXTART;
                 ToReadSelectedArts++;
               }
            }
            else
            if (*command==NEXTART || !leave) i++;
          } /************* while *************/
          ToReadSelectedArts= -1L;

          if (*command!=GOTO)
          {
            /* unselect all selected read articles */
            for (i= 0; i<NrOfActArts; i++)
                if ((artarray[i]->Status & (READ | SELECTED_ART)) == (READ | SELECTED_ART))
                   artarray[i]->Status&= ~SELECTED_ART;

            if (catchupgroup)
            {
              CatchUp(false); /* don't ask, just do it */
              *command= NEXTGROUP;
              leave= true;
            }
          }

        } /* if READNOW */

        /* command returned by readarticle doesn't mean to quit
           article-selection, except BREAK!!! */
        if (*command==NOP)
        {
          SetArtCursor();
        }
      } /* if (!leave) */
      if (*command==TOGGLE_SCREENTITLE || oldMaxScrLines!=MaxScrLines)
      {
        int index;
        /* we must be careful, because if at the last group we must
             go to the next page and reset the cursor */
        index= art_cursory+firstart; /* store old cursor position */
        if (*command==TOGGLE_SCREENTITLE) ToggleScreenTitle();
        grouplines= MaxScrLines-2;
        oldMaxScrLines= MaxScrLines;
        firstart= (index/grouplines)*grouplines; /* calculate new page */
        art_cursory= index-firstart;           /* set new cursor position */
        *command= REFRESHSCREEN;
      }
    } /* while !QUIT && !leave */
  }

  StatusLine();
  DispMsg("leaving newsgroup...");
  /* Cleanup */

  FreeMarks(pGI->Marks);  /* if it's NULL FreeMarks() can handle it */

  pGI->Marks= NewMark(NULL);  /* Set new Marks */

  if (artarray) FreeMem (artarray, (ULONG) NrOfArticles*sizeof(struct art_struct *));

  FreeSubjandFrom(pGI); /* Free all Froms and Subjects */

  FreeMem(RootArt, (ULONG) sizeof(struct art_struct)*(ULONG) NrOfArticles);
  RootArt= NULL;

  FreeKill_List(LocalKill_List);
  LocalKill_List= NULL;

  FreeString(ArtMenuTemplate);
  FreeString(ArtMenuTemplate2);
  FreeString(EntryMessage);
  FreeString(pathtodbase);

  GeneralBool&= ~G_MODERATED; /* reset flag */

  /* if leave==true then return to group-selection level! */
  switch (*command)
  {
    case BREAK: leave= true;
                break;
    default   : leave= false;
                break;
  }

  return leave;
}
