/***********************************************************************
 * strprocs.c -- string procedures for various use                     *
 *                                                                     *
 * Arn - (C)opyright by Roland Bless, All rights reserved,             *
 * -------- YOU ARE NOT ALLOWED TO DISTRIBUTE A CHANGED SOURCE ------- *
 ***********************************************************************/

#define STRPROCS_C

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#ifndef GLOBALVARS_H
#include "globalvars.h"
#endif

#ifndef ENUMS_H
#include "enums.h"
#endif

#ifndef true
#define true 1
#define false 0
#endif

#define MAXBASE 64
#define FIELDSIZE 3  /* characters left for filling base range */

/* Prototypes for functions defined in strprocs.c */
/* proto$start */
short islong(char *s);
void cuteoln(char *s);
char *skpwspc(char *s);
int xstrlen(register  char *s,
            int *len,
            BOOL *formfeed);
char *NextCR(char *s);
char *NextSpc(char *s);
char *NextWhiteSpc(register  char *s);
char *firststr(register  char *s,
               register  char *t);
char *firststr2(register  char *r,
                register  char *t);
#ifdef HAS_NO_STRSTR
char *strstr(char *s,
             char *t);
#endif
void delchr(char *s);
int getline(FILE *fp,
            char *s,
            int lim);
void mirrorstr(char *string,
               int n);
void ltostr(long number,
            char *string,
            int n);
void makeslashes(char *s);
char *strmetachr(char *s,
                 char meta);
#ifdef SIMPLE_WILDMATCHER
char *Match(register  char *s,
            register  char *fpat);
#endif
char *rot(char *s,
          int n);
char Numbers(int number);
void ToBase(unsigned long number,
            char *result,
            int base);
int KMP_Matcher(char *string,
                short length_string,
                char *pattern,
                short length_pattern,
                short *next_table);
void KMP_ComputeNextTable(char *pattern,
                          int pattern_length,
                          short *next_table);
char *CopyString(char *source);
void FreeString(char *string);
/* proto$end */
                      /**********************/

    /****************************************************************/

short int
islong(char *s)    /* returns true if s doesn't contain any non-numbers */
{
  char *ps= s;

  while (*s!='\0' && (isdigit(*s))) s++;
  if (*s=='\0' && s!=ps)
     return true;
  else
     return false;
}


void cuteoln(char *s)   /* Cuts off the '\n' from s */
{
  if (s)
  {
    while(*s!='\0' && *s!='\n')
         s++;
    *s='\0';
  }
}


char *skpwspc(s) /* returns pointer to first non white-space character */
char *s;
{
  if (s)
     while (*s!='\0' && (*s==' ' || *s=='\t' || *s=='\n')) s++;

  return s;
  /* skips spaces and tabs & LFs, returns first occurance of non white space */
}


/* returns length of string (termin. by \0) and counts tabs as 8 spaces */
int
xstrlen(s,len,formfeed)
register char *s;
int *len;
BOOL *formfeed;
{
  register int i= 0;

  *len=0;
  while (*s!='\0')
  {
    if (*s=='\t') *len=(*len/8+1)*8;
    else *len+=1;
    if (*s=='\x0C') *formfeed=true;
                               /* Set flag if we detected a ^L (FF) */
    i++;
    s++;
  }
  return i;
}


char *NextCR(s) /* returns pointer to next char after a newline '\n' */
char *s;
{
  while (*s!='\0' && *s!='\n') s++;
  if (*s!='\0')
     s++;
  return s;
}


char *
NextSpc(s)  /* returns pointer to next space in s (or p to end of s) */
char *s;
{
  if (s)
     while (*s!='\0' && *s!=' ')
           s++;
  return (char *) s;
}


char *
NextWhiteSpc(s)
register char *s;
{
  if (s)
    while (*s!='\0' && *s!=' ' && *s!='\t' && *s!='\n') s++;

  return s;
}


char *firststr(s, t)  /* returns pointer to first byte following t in s */
register char *s,*t;
{
  if (s==NULL || *s=='\0' || toupper(*s)!=toupper(*t))
     return NULL;
  else
  { s++;t++; }

  while (*t!='\0' && toupper(*s)==toupper(*t))
  {
    t++; s++;
  }
  if (*t != '\0')
     return NULL;
  else
     return s;
}


char *firststr2(register char *r, register char *t)  /* returns r if r begins with t */
                                                     /* otherwise NULL */
{
  register char *s=r;

  if (s==NULL || *s=='\0' || toupper(*s)!=toupper(*t))
     return NULL;
  else
  { s++;t++; }

  while (*t!='\0' && toupper(*s)==toupper(*t))
  {
    t++; s++;
  }
  if (*t != '\0')
     return NULL;
  else
     return r;
}


#ifdef HAS_NO_STRSTR
char *strstr (s, t)   /* returns pointer to first occurence of t in s */
char *s,*t;           /* This is a kludge, straightforward algorithm */

{
  register int i;
  short notfound = 1;

  while (notfound)
  {
    while (*s != *t && *s != '\0')
          s++;
    if (*s == '\0')
        return NULL;
    else
    {
      for (i=1; (notfound = *(t+i) != '\0') && *(s+i) == *(t+i);i++);
      if (notfound)
         s++;
    }
  }
  if (notfound == 0)
     return s;
  else
     return NULL;
}
#endif


/* moves all characters one position to the left */

void
delchr(char *s)
{
  if (s && *s!='\0')
  {
    do
    {
      s++;
      *(s-1)=*s;
    }
    while (*s!='\0');
  }
}



int getline(fp, s, lim)
FILE *fp;
char s[];
int lim;
{
  int c, i;

  i = 0;
  while (--lim > 0 && (c=fgetc(fp)) != EOF && c != '\n')
        s [i++] = c;
  if (c == '\n')
     s[i++] = c;
  s[i] = '\0';
  return i;
}


void
mirrorstr(char *string, int n)
/* mirrors string s[0]..s[n], returns nothing */
{
  char c;
  register int i;

  if (n>0) /* only mirror more than one char */
  for (i= n; i>n-i; i--)
  {
    c= string[i];
    string[i]= string[n-i];
    string[n-i]= c;
  }
}


void
ltostr(number, string, n)    /* n == Max stringlength */
long number;
char *string;
int n;
{
  register int i= 0;

  n-=1;

  if (n>0)
  {
    if (number==0)
      string[i++]='0';
    else
    for (i= 0; i<n && number>0; i++)
    {
      string[i]= (char) ('0'+ (char) (number % 10));
      number/=10;
    }
  }
  string[i]= '\0';
  mirrorstr(string,i-1);
}


void
makeslashes(char *s)
{
  if (s)
  while (*s!='\0')
  {
    if (*s=='.') *s='/';
    s++;
  }
}


char *
strmetachr(char *s, char meta) /* returns pointer to first occurrence of
                                  meta in s, where metameta is changed to
                                  meta itself */
{
  s=strchr(s,meta);

  while (s!=NULL && *(s+1)==meta)
  {
    s++;  /* skip metameta */
    delchr(s);
    s=strchr(s,meta);
  }

  return s;
}


#ifdef SIMPLE_WILDMATCHER
char *
Match(s, fpat)   /* returns NULL if pattern doesn't match */
register char *s,*fpat;  /* This is a kludge, a straightforward algorithm */
{
  while (*fpat!='\0')
  {
    if (*fpat == '*')
    {
      while (*fpat=='*' || *fpat=='?')
      {
        fpat++;         /* skip following wildcards, until letter or \0 */
      }
      if (*fpat!='\0')        /* letter/pattern detected! */
      {
        while (*s!='\0' && (*s&casedep) != (*fpat&casedep))
              s++;
        if (*s=='\0') return NULL;
      }
      else
         while (*s!='\0')    /* ingore all chars till end */
               s++;
    }
    else
    if (*fpat == '?')
    {
      fpat++;                    /* Just ignore that char */
      if (*s!='\0') s++;
    }
    else
    if ((*s&casedep) == (*fpat&casedep))
    {
      register char *stmp=s,*ptmp=fpat;

      /* test whether sub-string matches or not */
      while (*stmp!='\0' && ((*stmp&casedep) == (*ptmp&casedep)))
      {
         stmp++;ptmp++;
      }
      if (*stmp=='\0' || *ptmp=='*' || *ptmp=='?' || *ptmp=='\0')
      {
        fpat=ptmp;s=stmp;     /* it does! next wildcard or so... */
      }
      else
      {
        s++;
        while (*s!='\0' && (*s&casedep) != (*fpat&casedep)) s++;
      }
    }
    else
        return false;
  }
  return (*s=='\0' ? s : NULL);
}
#endif



char *rot(char *s,int n)
/* "rotates" letters A..Z,a..z of string s with offset n (ROT13) */
{
  char *t=s;

  while(*s!='\0')
  {
    if (*s>='A' && *s<='Z')
       *s='\x41'+(*s-'\x41'+n)%26;
    else
    if (*s>='a' && *s<='z')
       *s='\x61'+(*s-'\x61'+n)%26;
    s++;
  }
  return t;
}


char
Numbers(int number)
/* returns single character as digit to base 64 */
{
  static char field[FIELDSIZE+1]="+-*";
  /* field contains all digits above 62 (including) */

  if (number>=0 && number<=9)
    return (char) ('0'+number);
  else
  if (number>=10 && number<=35)
    return (char) ('a'+number-10);
  else
  if (number>=36 && number<=61)
    return (char) ('A'+number-36);
  else
  if (number>=62)
    return field[(number-62) % FIELDSIZE];
}


void
ToBase(unsigned long number, char *result, int base)
/* returns "number" converted to base in string result */
{
  int k= 0;

  if (number>0)
   for (k=0; number>0; k++)
   {
     result[k]=Numbers((int) (number % base));
     number/=base;
   }
  else
    result[k++]='0';

  result[k]='\0';
}


/* Knuth Morris Pratt Algorithm */

int
KMP_Matcher(char *string,  short int length_string,
            char *pattern, short int length_pattern,
            short int *next_table)
/* returns -1 if not found, else position of first match */
{
  short int pat_pos, string_pos, Start;

  for(pat_pos=string_pos= 0, Start= -1; Start==-1 && string_pos<length_string;)
  {
    if ((pattern[pat_pos]&casedep) == (string[string_pos]&casedep))
    {
      pat_pos++;string_pos++;
    }
    else
    {
      pat_pos= next_table[pat_pos]+1;
      if (pat_pos==-1)
      {
        pat_pos= 0;
        string_pos++;
      }
    }
    if (pat_pos>=length_pattern) Start= string_pos-length_pattern;
  }

  return Start;
}


void
KMP_ComputeNextTable(char *pattern, int pattern_length, short int *next_table)
/* calculates the next_table which must be allocated before */
{
  short int i,j;

  next_table[0]=-2;
  next_table[1]=-1;
  for(i=2; i<pattern_length; i++)
  {
    j= next_table[i-1]+1;
    while ((pattern[i-1]!=pattern[j]) && j>=0)
    {
      j= next_table[j]+1;
    }
    next_table[i]= j;
  }
}


char *
CopyString(char *source)
/* Allocates String, copies it and returns pointer to it */
{
  char *dest;

  if (source==NULL) return NULL;

  dest=AllocMem((ULONG) strlen(source)+1UL,NULL);
  if (dest)
     strcpy(dest,source);
  else
    ErrorHandle(NOMEM,NULL);

  return dest;
}



void
FreeString(char *string)
/* Frees string which was allocated with AllocMem() */
{
  if (string)
     FreeMem(string, (ULONG) strlen(string)+1UL);
}
