/***********************************************************************
 * time.c - time routines for Arn.c                                    *
 * Arn - (C)opyright by Roland Bless, All rights reserved,             *
 * -------- YOU ARE NOT ALLOWED TO DISTRIBUTE A CHANGED SOURCE ------- *
 ***********************************************************************/
#include <string.h>
#include <time.h>
#ifndef EXEC_LISTS_H
 #include   <exec/lists.h>
#endif
#ifdef MANUALLY
#ifndef DEVICES_TIMER.H
 #include   <devices/timer.h>
#endif
#else
#ifndef LIBRARIES_DOS_H
#include <libraries/dos.h>
#endif
#endif
#ifndef ENUMS_H
 #include "enums.h"
#endif

#ifdef TIMEZONES
#include "timezones.h"
#endif

#ifndef PROTO_DOS_H
#include <proto/dos.h>
#endif

#ifndef YEAR0
#define YEAR0 1978
#endif

extern long TZ_Offset;
extern char timez[];

/* Prototypes for functions defined in time.c */
/* proto$start */
ULONG GetTime(void);
ULONG GetClock(void);
int leapyear(int year);
struct tm *mytime(struct tm *ptm,
                  register  ULONG clock);
char *TimeStr(int mode);
ULONG EvalDate(char *s);
char *ShortDateStr(unsigned long time);
char *ShortTimeStr(unsigned long time);
char *SuperShortDateStr(ULONG time);
/* proto$end */


/************/

static short int DaysofMonth[]=
  { 31, 28, 31, 30,
    31, 30, 31, 31,
    30, 31, 30, 31 };


static char *Weekday[]=
{
  "Sun",
  "Mon",
  "Tue",
  "Wed",
  "Thu",
  "Fri",
  "Sat"
};

static char *Month[]=
{
  "Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"
};



#ifdef MANUALLY
ULONG
GetTime(void)
{
  struct MsgPort *CreatePort();
  unsigned long time;
  struct timerequest trq;

  if (OpenDevice(TIMERNAME, UNIT_VBLANK, &trq, 0L))
  {
      ErrorHandle(NOTIMER,NULL);
      return 0L;
  }

  trq.tr_node.io_Message.mn_ReplyPort = CreatePort(0L, 0L);
  trq.tr_node.io_Command = TR_GETSYSTIME;
  DoIO(&trq);
  time = trq.tr_time.tv_secs + (trq.tr_time.tv_micro + 500000)/1000000;
  CloseDevice(&trq);
  DeletePort(trq.tr_node.io_Message.mn_ReplyPort);

  return time;
}
#else
ULONG GetTime(void)
{
  struct DateStamp MyDSt;

  DateStamp(&MyDSt);
  return (ULONG) MyDSt.ds_Days*86400+MyDSt.ds_Minute*60+MyDSt.ds_Tick/TICKS_PER_SECOND;
}
#endif


ULONG GetClock(void)
{
  ULONG clock;

  clock= GetTime();
  clock-=(ULONG) (((TZ_Offset/100)*60+TZ_Offset%100)*60); /* subtract TZ-Offset */

  return clock;
}

int leapyear (int year)
{
  return (year % 4 == 0 && year % 100 != 0 || year % 400 == 0) ? 366 : 365;
}

struct tm *
mytime(struct tm *ptm, register ULONG clock)
{
  register int year=YEAR0, month=0;

  ptm->tm_sec= clock % 60;
  clock/= 60;
  ptm->tm_min= clock % 60;
  clock/= 60;
  ptm->tm_hour= clock % 24;
  clock/= 24;
  ptm->tm_wday= clock % 7;

  year=YEAR0;
  while (clock >= 366)
  {
    clock-= leapyear(year);
    year++;
  }
  if (clock == 365L && leapyear(year)!= 366)
  {
    clock-=365;
    year++;
  }

  DaysofMonth[1]= leapyear(year)==366? 29 : 28;

  ptm->tm_yday= clock;

  while (clock >= DaysofMonth[month])
  {
    clock-=DaysofMonth[month];
    month++;
  }

  clock++;
  month++;

  ptm->tm_mday= (short) clock;
  ptm->tm_mon=  month;
  ptm->tm_year= year;

  return ptm;
}



char*        /* returns valid ARPANET-Format string of current time */
TimeStr(int mode) /* mode: 0=ARPANET Format, 1=short log format, 2=Only dd-mmm */
{
  struct tm Time;
  static char timestr[40];


  if (mytime(&Time, GetClock()))
  {
#ifdef DEBUG
    printf("Time: %02d-%02d-%02d %02d:%02d:%02d\n",
           Time.tm_mday,Time.tm_mon,Time.tm_year,
           Time.tm_hour,Time.tm_min,Time.tm_sec);
#endif
    switch (mode)
    {
      case 0:
             sprintf(timestr, "%s, %02d %s %04d %02d:%02d:%02d %s",
                      Weekday[Time.tm_wday],
                      Time.tm_mday,
                      Month[Time.tm_mon-1],
                      Time.tm_year,
                      Time.tm_hour,
                      Time.tm_min,
                      Time.tm_sec,
                      timez);
             break;
      case 1:
             sprintf(timestr, "%02d %s %02d %02d:%02d:%02d",
                     Time.tm_mday,
                     Month[Time.tm_mon-1],
                     Time.tm_year-1900,
                     Time.tm_hour,
                     Time.tm_min,
                     Time.tm_sec);
             break;
      case 2:
             sprintf(timestr, "%02d-%s",
                      Time.tm_mday,
                      Month[Time.tm_mon-1]
                     );
             break;
    }

    return timestr;
  }
  else
    return NULL;
}



/* returns system-time from format ddmmyyyy */
ULONG
EvalDate(char *s)
{
  char dtmp[5];
  short day, month, year, i;
  ULONG time=0L;

  if (strlen(s)!=8) return 0L;
  else
  {
    strncpy(dtmp, s, 2);
    dtmp[2]='\0';
    s+=2;
    day=atol(dtmp);
    strncpy(dtmp, s, 2);
    dtmp[2]='\0';
    month=atol(dtmp);
    s+=2;
    strncpy(dtmp, s, 4);
    dtmp[4]='\0';
    year=atol(dtmp);

    if (--day>0) time+=day*86400;
    DaysofMonth[1]= leapyear(year)==366? 29 : 28;
    for (i= month-1; i>0; i--)
        time+=DaysofMonth[i-1]*86400;
    for (i= year-1; i>=YEAR0; i--)
        time+= leapyear(i)*86400;

    return time;
  }
}


char*        /* returns ddmmyyyy-Format string of time */
ShortDateStr(ULONG time)
{
  struct tm Time;
  static char timestr[9];

  if (mytime(&Time,time))
  {
    sprintf(timestr,"%02d%02d%04d",
           Time.tm_mday,Time.tm_mon,Time.tm_year);

    return timestr;
  }
  else
    return NULL;
}


char*        /* returns dd-mmm-(AmigaDOS)Format string of time */
SuperShortDateStr(ULONG time)
{
  struct tm Time;
  static char timestr[7];

  if (mytime(&Time,time))
  {
    sprintf(timestr,"%02d-%s",
           Time.tm_mday,Month[Time.tm_mon-1]);

    return timestr;
  }
  else
    return NULL;
}



char*        /* returns ddmmyyyy-Format string of time */
ShortTimeStr(ULONG time)
{
  struct tm Time;
  static char timestr[25];

  if (mytime(&Time,time))
  {
    sprintf(timestr,"%02d %3s %04d %02d:%02d:%02d",
           Time.tm_mday,Month[Time.tm_mon-1],Time.tm_year,
           Time.tm_hour,Time.tm_min,Time.tm_sec);

    return timestr;
  }
  else
    return NULL;
}


#ifdef TIMEZONES
int
GetTimeOffSet(char *s)
{
  ULONG hashval= 0UL;
/*
  ADT    -03     Atlantic Daylight Time
  AHST   -10     Alaska-Hawaii Standard Time
  AST    -04     Atlantic Standard Time
  AT     -02     Azores Time
  BST    +01     British Summer Time
* BST    -03     Brazil Standard Time
* BST    -11     Bering Standard Time
* BT     -11     "Bering Time"
  BT     +03     Baghdad Time, USSR Zone 2
  CADT   +10.5   Central Australian Daylight Time
  CAST   +09.5   Central Australian Standard Time
  CAT    -10     Central Alaska Time
  CCT    +08     China Coast Time, USSR Zone 7
  CDT    -05     Central Daylight Time
  CEDT   +02     "Central European Daylight Time"
  CES    +02     "Central European Summer time"
  CEST   +02     Central European Summer Time
  CET    +01     Central European Time
  CST    -06     Central Standard Time
  EADT   +11     East Australian Daylight Time
  EAST   +10     East Australian Standard Time
  ECT    +01     "European Central Time"? (Norway)
  EDT    -04     Eastern Daylight Time
  EET    +02     Eastern European Time, USSR Zone 1
  EST    -05     Eastern Standard Time
  FST    +02     French Summer Time
* FST    +02     "Finish Standard Time" ?
  FWT    +01     French Winter Time
  GMT    +00     Greenwich Mean Time
  GST    -03     Greenland Standard Time
* GST    +10     Guam Std Time, USSR Zone 9
  GVA    +01     ? (Switzerland)
  HFE    +02     Heure Francais d'Ete
  HFH    +01     Heure Francais d'Hiver
  IDLE   +12     International Date Line, East
  IDLW   -12     INternational Date Line, West
  IST    +02     Israelian Standard Time
* IST    +05.5   Indian Standard Time
  IT     +03.5   Iran Time
  JST    +09     Japan Std Time, USSR Zone 8
  JT     +07.5   Java Time
  MDT    -06     Mountain Daylight Time
  MEST   +02     Middle European Summer Time
  MESZ   +02     "Mittel Europaeische Sommer Zeit" (Germany)
  MET    +01     Middle European Time
  MEWT   +01     Middle European Winter Time
  MEZ    +01     "Mittel Europaeische Zeit" (Germany)
  MST    -07     Mountain Standard Time
  MT     +08.5   Moluccas Time
  NDT    -02.5   Newfoundland Daylight Time
  NST    -03.5   Newfoundland Standard Time
* NST    +06.5   North Sumatra Time
  NT     -11     Nome Time
  NZDT   +13     New Zealand Daylight Time
  NZST   +12     New Zealand Standard Time
  NZT    +12     "New Zealand Time"
  PDT    -07     Pacific Daylight Time
  PST    -08     Pacific Standard Time
  SADT   +10.5   South Australian Daylight Time
  SAST   +09.5   South Australian Standard Time
  SET    +01     "Standard European Time"
  SST    +02     "Swedish Summer Time"
* SST    +08     Singapore Standard Time
* SST    +07     South Sumatra Time, USSR Zone 6
  SWT    +01     "Swedish Winter Time"
  UT     +00     Universal Time
  UTC    +00     Universal Time, coordinated
  WADT   +08     West Australian Daylight Time
  WAST   +07     West Australian Standard Time
  WAT    -01     West Africa Time
  WET    +00     Western European Time
  WZ     +00     "Weltzeit"
  YST    -09     Yukon Standard Time
  Z      +00     Universal Time
  ZP4    +04     GMT+4 hours.
  ZP5    +05     GMT+5 hours.
  ZP6    +06     GMT+6 hours.
*/
 if (*s=='+' || *s=='-' || isdigit(*s)) /* local differential (hhmm) */
   return atol(s);
 else
 {
   while(*s)
   {
     hashval|=(((toupper(*s)-'A')+1) & 0x001F);
     hashval<<=5;
     s++;
   }
   switch (hashval)
   {
     case TZ_ADT : /* -03 Atlantic Daylight Time */
     case TZ_GST : /* -03     Greenland Standard Time */
     case TZ_C   :
                   return -300;
                   break;
     case TZ_AHST: /* -10 Alaska-Hawaii Standard Time */
     case TZ_CAT : /* -10 Central Alaska Time */
     case TZ_K   :
                   return -1000;
                   break;
     case TZ_AST : /* -04 Atlantic Standard Time */
     case TZ_EDT : /* -04 Eastern Daylight Time */
     case TZ_D   :
                   return -400;
                   break;
     case TZ_AT  : /* -02 Azores Time */
     case TZ_B   :
                   return -200;
                   break;
     case TZ_BST : /* +01 British Summer Time */
     case TZ_CET : /* +01 Central European Time */
     case TZ_ECT : /* +01 "European Central Time"? (Norway) */
     case TZ_FWT : /* +01 French Winter Time */
     case TZ_GVA : /* +01 ? (Switzerland) */
     case TZ_HFH : /* +01 Heure Francais d'Hiver */
     case TZ_MET : /* +01 Middle European Time */
     case TZ_MEWT: /* +01 Middle European Winter Time */
     case TZ_MEZ : /* +01 "Mittel Europaeische Zeit" (Germany) */
     case TZ_N   :
     case TZ_SET : /* +01 "Standard European Time" */
     case TZ_SWT : /* +01 "Swedish Winter Time" */
                   return 100;
                   break;
     case TZ_BT  : /* -11 "Bering Time" */
     case TZ_NT  : /* -11 Nome Time */
     case TZ_L   :
                   return -1100;
                   break;
     case TZ_CADT: /* +10.5 Central Australian Daylight Time */
     case TZ_SADT: /* +10.5   South Australian Daylight Time */
                   return 1030;
                   break;
     case TZ_CAST: /* +09.5 Central Australian Standard Time */
     case TZ_SAST: /* +09.5   South Australian Standard Time */
                   return 930;
                   break;
     case TZ_CCT : /* +08 China Coast Time, CIS Zone 7 */
     case TZ_WADT: /* +08 West Australian Daylight Time */
     case TZ_U   :
                   return 800;
                   break;
     case TZ_CDT : /* -05 Central Daylight Time */
     case TZ_EST : /* -05 Eastern Standard Time */
     case TZ_E   :
                   return -500;
                   break;
     case TZ_CEDT: /* +02 "Central European Daylight Time" */
     case TZ_CES : /* +02 "Central European Summer time" */
     case TZ_CEST: /* +02 Central European Summer Time */
     case TZ_FST : /* +02 French Summer Time */
     case TZ_EET : /* +02 Eastern European Time, CIS Zone 1 */
     case TZ_HFE : /* +02 Heure Francais d'Ete */
     case TZ_IST : /* +02 Israelian Standard Time */
     case TZ_MEST: /* +02 Middle European Summer Time */
     case TZ_MESZ: /* +02 "Mittel Europaeische Sommer Zeit" (G */
     case TZ_O   :
     case TZ_SST : /* +02 "Swedish Summer Time" */
                   return 200;
                   break;
     case TZ_P   :
                   return 300;
                   break;
     case TZ_Q   :
                   return 400;
                   break;
     case TZ_R   :
                   return 500;
                   break;
     case TZ_S   :
                   return 600;
                   break;
     case TZ_CST : /* -06 Central Standard Time */
     case TZ_MDT : /* -06 Mountain Daylight Time */
     case TZ_F   :
                   return -600;
                   break;
     case TZ_EADT: /* +11 East Australian Daylight Time */
     case TZ_X   :
                   return 1100;
                   break;
     case TZ_EAST: /* +10 East Australian Standard Time */
     case TZ_W   :
                   return 1000;
                   break;
     case TZ_GMT : /* +00 Greenwich Mean Time */
     case TZ_UT  : /* +00 Universal Time */
     case TZ_UTC : /* +00 Universal Time, coordinated */
     case TZ_WET : /* +00 Western European Time */
     case TZ_WZ  : /* +00 "Weltzeit" */
     case TZ_Z   : /* +00     Universal Time */
                   return 0;
                   break;
     case TZ_IDLE: /* +12 International Date Line, East */
     case TZ_NZST: /* +12 New Zealand Standard Time */
     case TZ_NZT : /* +12 "New Zealand Time" */
     case TZ_Y   :
                   return +1200;
                   break;
     case TZ_IDLW: /* -12     INternational Date Line, West */
     case TZ_M   :
                   return -1200;
                   break;
     case TZ_IT  : /* +03.5   Iran Time */
                   return 330;
                   break;
     case TZ_JST : /* +09     Japan Std Time, CIS Zone 8 */
     case TZ_V   :
                   return 900;
                   break;
     case TZ_JT  : /* +07.5   Java Time */
                   return 730;
                   break;
     case TZ_MST : /* -07 Mountain Standard Time */
     case TZ_PDT : /* -07 Pacific Daylight Time */
     case TZ_G   :
                   return -700;
                   break;
     case TZ_MT  : /* +08.5   Moluccas Time */
                   return 830;
                   break;
     case TZ_NDT : /* -02.5   Newfoundland Daylight Time */
                   return -230;
                   break;
     case TZ_NST : /* -03.5   Newfoundland Standard Time */
                   return -330;
                   break;
     case TZ_NZDT: /* +13     New Zealand Daylight Time */
                   return +1300;
                   break;
     case TZ_PST : /* -08 Pacific Standard Time */
     case TZ_H   :
                   return -800;
                   break;
     case TZ_WAST: /* +07     West Australian Standard Time */
     case TZ_T   :
                   return 700;
                   break;
     case TZ_A   : /* -01 Military */
     case TZ_WAT : /* -01 West Africa Time */
                   return -100;
                   break;
     case TZ_YST : /* -09     Yukon Standard Time */
     case TZ_I   :
                   return -900;
                   break;
     default: return 0;
              break;

   }
 }
}



int
GetMonth(char *s)
/* returns number of month from MMM string */
{
  switch(toupper(*s))
  {
    case 'J':s++;
             if (toupper(*s)=='U')
             {
               s++;
               if (toupper(*s)=='N') /* Jun */
                 return 6;
               else /* Jul */
                 return 7;
             }
             else /* Jan */
               return 1;
             break;
    case 'F': return 2;
              break;
    case 'M': s+=2;
              if (toupper(*s)=='R') /* Mar */
                return 3;
              else
                return 5; /* May */
              break;
    case 'A': s++;
              if (toupper(*s)=='P') /* Apr */
                return 4;
              else
                return 8;
              break;
    case 'S': return 9;
              break;
    case 'O': return 10;
              break;
    case 'N': return 11;
              break;
    case 'D': return 12;
              break;
     default: return 12;
              break;
  }
}



ULONG
ParseDate(char *datestring)
{
  char month[4],timezone[10];
  struct tm tmstruct;
  ULONG dateval=0UL;
  short int i, offset;

  if (isalpha(*datestring))
     while(*datestring && (!isspace(*datestring))) datestring++; /* skip leading weekday */

  while(isspace(*datestring)) datestring++;

  if(isalpha(*datestring)) /* ctime format Wdy Mon DD HH:MM:SS YYYY */
  {
    sscanf(datestring,"%3s %d %d:%d:%d %d %9s",
           &month,
           &tmstruct.tm_mday,
           &tmstruct.tm_hour,
           &tmstruct.tm_min,
           &tmstruct.tm_sec,
           &tmstruct.tm_year,
           &timezone
           );

  }
  else          /* Wdy, DD Mon YY HH:MM:SS TIMEZONE */
  {
    sscanf(datestring,"%d %3s %d %d:%d:%d %9s",
           &tmstruct.tm_mday,
           &month,
           &tmstruct.tm_year,
           &tmstruct.tm_hour,
           &tmstruct.tm_min,
           &tmstruct.tm_sec,
           &timezone
           );
  }
  tmstruct.tm_mon= GetMonth(month);
  offset= GetTimeOffSet(timezone);
  if (tmstruct.tm_year<(YEAR0-1900)) /* Year 2000? :-) */
     tmstruct.tm_year+=2000;
  else
  if (tmstruct.tm_year<=100) /* between 78..99 -> mapped to 1978-1999 */
  {
    tmstruct.tm_year+=1900;
  }
  if (--tmstruct.tm_mday>0) dateval+=tmstruct.tm_mday*86400;
  DaysofMonth[1]= leapyear(tmstruct.tm_year)==366? 29 : 28;
  for (i= tmstruct.tm_mon-1; i>0; i--)
      dateval+=DaysofMonth[i-1]*86400;
  for (i= tmstruct.tm_year-1; i>=YEAR0; i--)
      dateval+= leapyear(i)*86400;

  dateval+= (LONG) ((tmstruct.tm_hour-offset/100)*60
                    +(tmstruct.tm_min-offset%100)   )*60
            +tmstruct.tm_sec;

  return dateval;
}
#endif
