
/*
 *  BMSENCODE.C
 *
 *  Encode/Decode a BMS coded file
 */

#include "defs.h"

#ifdef TEST

List	RegList;
char	TmpBuf1[512];
char	TmpBuf2[512];

#include "subs.c"


#endif

Prototype void	 MailDataBase(RegNode *, char *);
Prototype bool_t BMSDecodeFile(FILE *, FILE *);
Prototype long	 BMSEncodeFile(FILE *, FILE *);
Prototype long BMSEncodeSize(long);

char	EncodeTable[64] = {
    "abcdefghijklmnop"
    "qrstuvwxyzABCDEF"
    "GHIJKLMNOPQRSTUV"
    "WXYZ0123456789+-"
};

char	DecodeTable[256] = {
    -1,-1,-1,-1,-1,-1,-1,-1,	-1,-1,-1,-1,-1,-1,-1,-1,    /*	00  */
    -1,-1,-1,-1,-1,-1,-1,-1,	-1,-1,-1,-1,-1,-1,-1,-1,    /*	10  */
    -1,-1,-1,-1,-1,-1,-1,-1,	-1,-1,-1,62,-1,63,-1,-1,    /*	20  */
    52,53,54,55,56,57,58,59,	60,61,-1,-1,-1,-1,-1,-1,    /*	30  */
    -1,26,27,28,29,30,31,32,	33,34,35,36,37,38,39,40,    /*	40  */
    41,42,43,44,45,46,47,48,	49,50,51,-1,-1,-1,-1,-1,    /*	50  */
    -1, 0, 1, 2, 3, 4, 5, 6,	 7, 8, 9,10,11,12,13,14,    /*	60  */
    15,16,17,18,19,20,21,22,	23,24,25,-1,-1,-1,-1,-1,    /*	70  */

    -1,-1,-1,-1,-1,-1,-1,-1,	-1,-1,-1,-1,-1,-1,-1,-1,    /*	80  */
    -1,-1,-1,-1,-1,-1,-1,-1,	-1,-1,-1,-1,-1,-1,-1,-1,    /*	90  */
    -1,-1,-1,-1,-1,-1,-1,-1,	-1,-1,-1,-1,-1,-1,-1,-1,    /*	A0  */
    -1,-1,-1,-1,-1,-1,-1,-1,	-1,-1,-1,-1,-1,-1,-1,-1,    /*	B0  */
    -1,-1,-1,-1,-1,-1,-1,-1,	-1,-1,-1,-1,-1,-1,-1,-1,    /*	C0  */
    -1,-1,-1,-1,-1,-1,-1,-1,	-1,-1,-1,-1,-1,-1,-1,-1,    /*	D0  */
    -1,-1,-1,-1,-1,-1,-1,-1,	-1,-1,-1,-1,-1,-1,-1,-1,    /*	E0  */
    -1,-1,-1,-1,-1,-1,-1,-1,	-1,-1,-1,-1,-1,-1,-1,-1,    /*	F0  */
};

#ifndef TEST

/*
 *  Mail a BMS database
 */

void
MailDataBase(rn, dbFileName)
RegNode *rn;
char *dbFileName;
{
    FILE *fi;
    FILE *fo;
    char *tmpFileName = AllocTemporaryDiskFile();
    long r;

    if (fi = fopen(dbFileName, "r")) {
	if (fo = fopen(tmpFileName, "w")) {
	    long size;

	    fprintf(fo, "To: %s!bms_in\n", rn->rn_Node.ln_Name);
	    if (rn->rn_RetPath && rn->rn_RetPath[0])
		fprintf(fo, "Reply-To: %s!bms_err\n", rn->rn_RetPath);
	    fprintf(fo, "Subject:\n");
	    fprintf(fo, "X-Bms:\n");
	    fprintf(fo, "\n");
	    size = BMSEncodeFile(fi, fo);
	    fprintf(fo, "\n");
	    fclose(fo);
	    sprintf(TmpBuf1, "%s <%s -f bms_err -R BMS", SendMailName(), tmpFileName);
	    r = system(TmpBuf1);
	    if (r != 0)
		bmslog("BMSSend", "sendmail failed to %s!bms_in exit %d\n", rn->rn_Node.ln_Name, r);
	    remove(tmpFileName);

	    AddHistogram(rn, size);
	}
	fclose(fi);
    }
}

#endif

/*
 *  BMSDecodeFile(fi, fo)
 *
 *  Decode a BMS file, return 0 on success, -1 on failure
 */


bool_t
BMSDecodeFile(fi, fo)
FILE *fi;
FILE *fo;
{
    char *ptr;
    long size;
    long crc;
    bool_t r = 0;

    /*
     *	Obtain header
     */

    while (ptr = fgets(TmpBuf1, sizeof(TmpBuf1), fi)) {
	if (strnicmp(TmpBuf1, "BMSBegin", 8) == 0)
	    break;
    }
    if (ptr == NULL)
	return(-1);

    /*
     *	Parse header.  Header contains length of file
     */

    if (sscanf(TmpBuf1 + 8, "%li", &size) != 1) {
	bmslog("BMSDecode", "Bad BMSBegin line: %s", TmpBuf1);
	return(-2);
    }

    /*
     *	Parse BMS encoded file, lines begin with
     */

    CrcInit(&crc);

    {
	short so;
	short lo;
	short c;
	short nl = 1;	/*  new line mark	*/
	long i; 	/*  output file index	*/

	for (i = so = 0; i < size && (c = getc(fi)) != EOF; ) {
	    if (nl == 1) {
		if (c == 'B' || c == 'b')   /*  BMSEnd  */
		    break;
		if (c != '+')               /*  look for + initiator    */
		    continue;
		nl = 0;
		c = getc(fi);               /*  skip it */
	    }
	    if (c == '\n' || c == '\r') {
		nl = 1;
		continue;
	    }
	    if (DecodeTable[c] < 0)
		break;
	    lo = (lo << 6) | DecodeTable[c];
	    so += 6;
	    if (so >= 8) {
		short c = lo >> (so - 8);

		putc(c, fo);
		CrcBump(&crc, c);

		++i;
		so -= 8;
	    }
	}

	/*
	 *  if nl == 0 then throw away remainder of line
	 */

	if (nl == 0) {
	    while ((c = getc(fi)) != EOF && c != '\n')
		;
	    c = getc(fi);   /*  first character of next line    */
	}

	/*
	 *  Obtain terminator line
	 */

	TmpBuf1[0] = c;
	fgets(TmpBuf1 + 1, sizeof(TmpBuf1) - 1, fi);

	if (strnicmp(TmpBuf1, "BMSEnd", 6) != 0 || i != size) {
	    bmslog("BMSDecode", "Corrupt File/EndMark %d/%d bytes: %s",
		i, size, TmpBuf1
	    );
	    r = -2;
	}
    }

    /*
     *	Compare CRC against BMSEnd crc
     */

    if (r == 0) {
	long file_crc = 0;

	if (sscanf(TmpBuf1 + 6, "%i", &file_crc) == 0 || file_crc != crc) {
	    bmslog("BMSDecode", "Corrupt File CRC failed: 0x%08lx expected 0x%08lx",
		file_crc, crc
	    );
	    r = -2;
	}
    }
    return(r);
}

/*
 *  BMS encode (part) of a file
 */

long
BMSEncodeFile(fi, fo)
FILE *fi;
FILE *fo;
{
    long size;
    long crc;
    long r = 0;

    fseek(fi, 0L, 2);
    size = ftell(fi);
    fseek(fi, 0L, 0);

    fprintf(fo, "BMSBegin %d", size);

    CrcInit(&crc);

    {
	short so;
	short lo;
	long i;
	long j;

	for (i = j = so = 0; i < size; ++i) {
	    short c;

	    if ((c = getc(fi)) == EOF)
		break;
	    CrcBump(&crc, c);
	    lo = (lo << 8) | c;
	    so += 8;
	    while (so >= 6) {
		short co;

		if ((j & 63) == 0)
		    fprintf(fo, "\n+");

		co = EncodeTable[(lo >> (so - 6)) & 0x3F];
		putc(co, fo);
		++j;
		so -= 6;
	    }
	}
	if (so) {
	    short co;

	    if ((j & 63) == 0)
		fprintf(fo, "\n+");
	    lo <<= 6 - so;
	    co = EncodeTable[lo & 0x3F];
	    fputc(co, fo);
	}
	r = j;
    }
    fprintf(fo, "\nBMSEnd 0x%lx\n", crc);
    return(r);
}

/*
 *  Return the approximate size of a BMS encoding, in bytes
 */

long
BMSEncodeSize(bytes)
long bytes;
{
    return(bytes * 8 / 6 + 160);
}

#ifdef TEST

/*
 *  TESTCODE
 *
 *  test e infile	encode
 *  test d infile	decode
 *
 */

main(ac, av)
char *av[];
{
    FILE *fi;
    long r = 0;

    if (ac != 3) {
	puts("test e infile     -encode to stdout");
	puts("test d infile     -decode to stdout");
	return(0);
    }
    if (av[1][0] == 'e') {
	if (fi = fopen(av[2], "r"))
	    r = BMSEncodeFile(fi, stdout);
	else
	    fprintf(stderr, "Unable to open %s\n", av[2]);
    } else if (av[1][0] == 'd') {
	if (fi = fopen(av[2], "r"))
	    r = BMSDecodeFile(fi, stdout);
	else
	    fprintf(stderr, "Unable to open %s\n", av[2]);
    } else {
	fprintf(stderr, "bad option\n");
    }
    return(r);
}


#endif
