
/*
 *  DEQUEUE.C
 */

#include "defs.h"

Prototype void InitDequeue(void);
Prototype void DequeueBMS(char *, char *);
Prototype void EnterDequeueState(RegNode *, CtlNode *, Comm *, char *, short);

Prototype void SMRxDequeue(DBHan *, record_t, RegNode *, CtlNode *);
Prototype void SMTxDequeue(DBHan *, record_t, RegNode *, CtlNode *);
Prototype void SMToDequeue(DBHan *, record_t, RegNode *, CtlNode *);

void
InitDequeue(void)
{
    SetStateMachine(SMACH_DEQUEUE, SMRxDequeue, SMTxDequeue, SMToDequeue);
}

/*
 *  Dequeue a transfer	(either local or remote request)
 *
 *  - search for the ID
 *  - dequeue the comm
 */

void
DequeueBMS(id, msg)
char *id;
char *msg;
{
    RegNode *rn;
    CtlNode *cn;
    long idno = strtol(id, NULL, 0);

    if (idno == 0) {
	bmslog("Dequeue", "Non-Zero ID required, use SID field of item to dequeue");
	return;
    }
    if (msg == NULL)
	msg = "Manual Dequeue";

    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
	LoadBMSRegTree(rn);
	if (rn->rn_Comm.com_SID == idno) {
	    EnterDequeueState(rn, NULL, &rn->rn_Comm, msg, 1);
	    break;
	}
	for (cn = GetHead(&rn->rn_CtlList); cn; cn = GetSucc(&cn->cn_Node)) {
	    if (cn->cn_Comm.com_SID == idno) {
		EnterDequeueState(rn, cn, &cn->cn_Comm, msg, 1);
		break;
	    }
	}
	if (cn)
	    break;
	UnLoadBMSRegTree(rn);
    }
    if (rn) {
	SaveBMSDataBase();
	UnLoadBMSRegTree(rn);
    }

    if (rn == NULL)
	bmslog("Dequeue", "Unable to find ID %d\n", idno);
}


/*
 *  Dequeue a COMM.  Sets the COMM node to a dequeue state and, if txMsg
 *  is non-zero, send a dequeue message to the remote.
 *
 *  msg is mailed to the user locally
 */

void
EnterDequeueState(rn, cn, comm, msg, txMsg)
RegNode *rn;
CtlNode *cn;
Comm *comm;
char *msg;
short txMsg;
{
    if (msg == NULL)
	msg = "";
    bmslog("Dequeue", "Dequeueing connection %s id %d (%s)\n", rn->rn_Node.ln_Name, comm->com_SID, msg);

    /*
     *	Send mail
     */

    {
	char *tmpFileName = AllocTemporaryDiskFile();
	FILE *fo;
	long r;

	if (fo = fopen(tmpFileName, "w")) {
	    fprintf(fo, "Subject: BMS DEQUEUE NOTIFICATION (%s)\n", msg);
	    fprintf(fo, "\n");
	    fprintf(fo, "REGISTRATION %-15s ALIAS=%-8s STAT=%s\n",
		(rn->rn_Node.ln_Name ? rn->rn_Node.ln_Name : "<defaults>"),
		(rn->rn_Alias ? rn->rn_Alias : "<none>"),
		StatusOfComm(&rn->rn_Comm)
	    );
	    if (cn == NULL) {
		PrintParamStatus(fo, &rn->rn_Param, 4);
		PrintCommStatus(fo, &rn->rn_Comm, 4);
		PrintStatistics(fo, &rn->rn_Stat, 4);
		PrintFileAccList(fo, &rn->rn_AccList, 4);
	    } else {
		PrintFileAccList(fo, &rn->rn_AccList, 4);
		fprintf(fo, "    CONTROL: %s STAT=%s\n",
		    cn->cn_Node.ln_Name ? cn->cn_Node.ln_Name : "",
		    StatusOfComm(&cn->cn_Comm)
		);
		PrintCommStatus(fo, &cn->cn_Comm, 8);
		PrintStatistics(fo, &cn->cn_Stat, 8);
		PrintFileList(fo, &cn->cn_FileList, 8);
	    }
	    fclose(fo);

	    sprintf(TmpBuf1, "%s <%s -f bms_err -t bms-manager", SendMailName(), tmpFileName);
	    r = system(TmpBuf1);
	    if (r != 0)
		bmslog("Dequeue", "Unable to sendmail bms-manager");
	    remove(tmpFileName);
	}
	zfree(tmpFileName);
    }

    /*
     *	DeQ, comm state machine
     */

    InitCommChangeStateMachine(SMACH_DEQUEUE, rn, comm);
    SetCommTimeout(comm, ONEMONTH);
    SetCommMessage(comm, msg);
    comm->com_TOMisc = gettime();
    if (txMsg)
	comm->com_State = 1;

    /*
     *	Set GF_NODISP flag
     */

    if (cn)
	cn->cn_Flags |= GF_NODISP | GF_MODIFIED;
    else
	rn->rn_Flags |= GF_NODISP | GF_MODIFIED;
}


/************************************************************************
 *									*
 *			  DEQUEUE STATE MACHINE 			*
 *									*
 ************************************************************************
 *
 *  state 0	Initiate dequeue with no send
 *  state 1	Initiate dequeue with send
 *  state 2	Dequeue timed out (after 1 month), destroy connection
 */

/*
 *  If we receive any data (other then a remote dequeue or error), and
 *  no dequeue has been sent for two days, send a dequeue
 */

void
SMRxDequeue(db, rid, rn, cn)
DBHan *db;
record_t rid;
RegNode *rn;
CtlNode *cn;
{
    Comm *comm = (cn) ? &cn->cn_Comm : &rn->rn_Comm;
    long n = gettime() - comm->com_TOMisc;
    long size;
    long type;
    short sendDeq = 0;

    for (rid = ScanSubRecord(db, rid, &size, &type); rid; rid = ScanNextRecord(db, rid, &size, &type)) {
	switch(type) {
	case DBTYPE_BMS_ERROR:
	case DBTYPE_BMS_DEQUEUE:
	    break;
	defeault:
	    sendDeq = 1;
	    break;
	}
    }
    if (sendDeq && (n < 0 || n > ONEDAY * 2)) {
	comm->com_TOMisc = gettime();
	comm->com_Flags |= GF_ALERT;
	comm->com_State = 1;
    }
}

/*
 *  Dequeue initiate
 */

void
SMTxDequeue(db, rid, rn, cn)
DBHan *db;
record_t rid;
RegNode *rn;
CtlNode *cn;
{
    Comm *comm = (cn) ? &cn->cn_Comm : &rn->rn_Comm;
    char buf[64];

    switch(comm->com_State) {
    case 0:
	SetCommTimeout(comm, ONEMONTH);
	comm->com_State = 2;
	break;
    case 1:
	SetCommTimeout(comm, ONEWEEK * 2);
	comm->com_State = 2;

	rid = WriteBMSChannelRecord(db, 0, rid, comm);
	WriteRecord(db, rid, 0, DBTYPE_BMS_DEQUEUE, comm->com_Msg, comm->com_Msg ? strlen(comm->com_Msg) : 0);
	bmslog("TxDequeue", "Dequeue sent to %s for sid=%d did=%d (%s)",
	    rn->rn_Node.ln_Name,
	    comm->com_SID,
	    comm->com_DID,
	    comm->com_Msg ? comm->com_Msg : ""
	);
	break;
    case 2:
	break;
    }
}

/*
 *  Timeout
 */

void
SMToDequeue(db, rid, rn, cn)
DBHan *db;
record_t rid;
RegNode *rn;
CtlNode *cn;
{
    if (cn)
	TerminateComm(rn, cn, &cn->cn_Comm, "Dequeue period complete");
    else
	TerminateComm(rn, cn, &rn->rn_Comm, "Dequeue period complete");
}

