
/*
 *  HISTOGRAM.C
 */

#include "defs.h"

Prototype void LoadHistogram(RegNode *, record_t);
Prototype void AddHistogram(RegNode *, long);
Prototype void SaveHistogram(RegNode *);
Prototype void UnLoadHistogram(RegNode *);

Prototype void CalculateHistogramSent(RegNode *, long *, long *);
Prototype void CalculateHistogramAvail(RegNode *, long *, long *);

/*
 *  Database operations
 */

void
LoadHistogram(rn, rid)
RegNode *rn;
record_t rid;
{
    HistNode *hn;
    long size;
    long type;

    rn->rn_TxHId = rid;
    for (rid = ScanSubRecord(Db, rid, &size, &type); rid; rid = ScanNextRecord(Db, rid, &size, &type)) {
	if (type == DBTYPE_TXHIST_NODE) {
	    hn = LoadExtRecord(Db, rid, NULL, markersize(HistNode), sizeof(HistNode));
	    AddTail(&rn->rn_HistList, &hn->hn_Node);
	}
    }
}

void
AddHistogram(rn, bytes)
RegNode *rn;
long bytes;
{
    HistNode *hn = zalloc(sizeof(HistNode));

    hn->hn_TimeStamp = gettime();
    hn->hn_Bytes = bytes;
    hn->hn_Flags = GF_MODIFIED;
    AddTail(&rn->rn_HistList, &hn->hn_Node);
}

void
SaveHistogram(rn)
RegNode *rn;
{
    HistNode *hn;

    for (hn = GetHead(&rn->rn_HistList); hn; hn = GetSucc(&hn->hn_Node)) {
	if (rn->rn_TxHId == 0)
	    rn->rn_TxHId = WriteRecord(Db, rn->rn_RId, 0, DBTYPE_TXHIST, NULL, 0);
	if (hn->hn_Flags & GF_DELETED) {
	    if (hn->hn_RId)
		DeleteRecord(Db, hn->hn_RId);
	} else if (hn->hn_Flags & GF_MODIFIED) {
	    hn->hn_Flags &= ~GF_MODIFIED;
	    hn->hn_RId = WriteRecord(Db, rn->rn_TxHId, hn->hn_RId, DBTYPE_TXHIST_NODE, hn, markersize(HistNode));
	}
    }
}

void
UnLoadHistogram(rn)
RegNode *rn;
{
    HistNode *hn;

    while (hn = RemHead(&rn->rn_HistList))
	zfree(hn);
}


/*
 *  Calculate the bytes sent over the last week and last day.
 */

void
CalculateHistogramSent(rn, pdaily_used, pweekly_used)
RegNode *rn;
long *pdaily_used;
long *pweekly_used;
{
    HistNode *hn;
    time_t t = gettime();

    *pdaily_used = 0;
    *pweekly_used = 0;

    for (hn = GetHead(&rn->rn_HistList); hn; hn = GetSucc(&hn->hn_Node)) {
	long n = t - hn->hn_TimeStamp;

	if (n < 0 || n > 60 * 60 * 24 * 7) {
	    hn->hn_Flags |= GF_DELETED;
	} else {
	    *pweekly_used += hn->hn_Bytes;
	    if (n < 60 * 60 * 24)
		*pdaily_used += hn->hn_Bytes;
	}
    }
}

/*
 *  Calculate space remaining over the last 24 hours and the last week.
 *  That is, calculate the maximum daily/weekly bytes and subtract the
 *  bytes that have already been sent.
 */
#ifdef NOTDEF

void
CalculateHistogramAvail(rn, pdaily_left, pweekly_left)
RegNode *rn;
long *pdaily_left;
long *pweekly_left;
{
    long daily;
    long weekly;
    long n;

    CalculateHistogramSent(rn, &daily, &weekly);

    if ((n = rn->rn_Param.pm_MaxTxBpd) == 0)
	n = DefRN->rn_Param.pm_MaxTxBpd;
    n = n - daily;
    if (n < 0)
	n = 0;
    *pdaily_left = n;

    if ((n = rn->rn_Param.pm_MaxTxBpw) == 0)
	n = DefRN->rn_Param.pm_MaxTxBpw;
    n = n - weekly;
    if (n < 0)
	n = 0;
    *pweekly_left = n;
}

#endif
