
/*
 *  MISC.C
 */

#include "defs.h"

Prototype void RenameBMS(char *, char *);
Prototype void AliasBMS(char *);
Prototype void DestroyBMS(char *);
Prototype void RetryBMS(char *);
Prototype void AddFileBMS(char *);
Prototype void RemFileBMS(char *);
Prototype void ResFileBMS(char *);

void
RenameBMS(sname, dname)
char *sname;
char *dname;
{
    RegNode *rn;

    if (dname == NULL) {
	bmslog("Rename", "No TO field specified");
	return;
    }
    if (rn = FindRegNode(sname, 1)) {
	if (rn->rn_Node.ln_Name)
	    zfree(rn->rn_Node.ln_Name);
	rn->rn_Node.ln_Name = zstrdup(dname);
	rn->rn_Flags |= GF_MODIFIED;
    } else {
	bmslog("Rename", "Unable to find node: %s", sname);
    }
}

void
AliasBMS(alias)
char *alias;
{
    RegNode *rn;

    if (alias == NULL) {
	bmslog("Alias", "no alias!");
	return;
    }
    if (rn = FindFirstMarked()) {
	if (rn->rn_Alias)
	    zfree(rn->rn_Alias);
	rn->rn_Alias = zstrdup(alias);
	rn->rn_Flags |= GF_MODIFIED;
    } else {
	bmslog("Alias", "No specified destinations");
    }
}

/*
 *  FILE ACCESS CONTROL (also see FILES.C)
 *
 *  Add     Add a new file to the access list, allowed file
 *  Rem     Remove a file from the access list
 *  Res     Add a new file to the access list, restricted file
 */

void
AddFileBMS(fileName)
char *fileName;
{
    RegNode *rn;
    FileAccess *fa;

    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
	if (rn->rn_Flags & GF_MARKED) {
	    LoadBMSRegTree(rn);
	    for (fa = GetHead(&rn->rn_AccList); fa; fa = GetSucc(&fa->fa_Node)) {
		if (stricmp(fa->fa_Node.ln_Name, fileName) == 0)
		    break;
	    }
	    if (fa == NULL)
		fa = AllocFileAccess(rn, fileName);
	    fa->fa_Flags |= GF_MODIFIED;
	    fa->fa_Flags &= ~FAF_RESTRICT;
	    SaveBMSDataBase();
	    UnLoadBMSRegTree(rn);
	}
    }
}

void
RemFileBMS(fileName)
char *fileName;
{
    RegNode *rn;
    FileAccess *fa;

    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
	if (rn->rn_Flags & GF_MARKED) {
	    LoadBMSRegTree(rn);
	    for (fa = GetHead(&rn->rn_AccList); fa; fa = GetSucc(&fa->fa_Node)) {
		if (stricmp(fa->fa_Node.ln_Name, fileName) == 0)
		    break;
	    }
	    if (fa) {
		fa->fa_Flags |= GF_DELETED;
	    } else {
		bmslog("Remfile", "Node not found");
	    }
	    SaveBMSDataBase();
	    UnLoadBMSRegTree(rn);
	}
    }
}

void
ResFileBMS(fileName)
char *fileName;
{
    RegNode *rn;
    FileAccess *fa;

    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
	if (rn->rn_Flags & GF_MARKED) {
	    LoadBMSRegTree(rn);

	    for (fa = GetHead(&rn->rn_AccList); fa; fa = GetSucc(&fa->fa_Node)) {
		if (stricmp(fa->fa_Node.ln_Name, fileName) == 0)
		    break;
	    }
	    if (fa == NULL)
		fa = AllocFileAccess(rn, fileName);
	    fa->fa_Flags |= GF_MODIFIED;
	    fa->fa_Flags |= FAF_RESTRICT;

	    SaveBMSDataBase();
	    UnLoadBMSRegTree(rn);
	}
    }
}

void
DestroyBMS(id)
char *id;
{
    RegNode *rn;
    CtlNode *cn;
    long idno = strtol(id, NULL, 0);


    if (idno == 0) {
	bmslog("Destroy", "Non-zero ID required, use SID field of COMM element to destroy");
	return;
    }

    printf("Destroy should only be used as a last resort\n");
    printf("Use Dequeue whenever possible\n");

    if (getyn("Continue with destroy?", NO) == NO)
	return;

    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
	LoadBMSRegTree(rn);
	if (rn->rn_Comm.com_SID == idno) {
	    rn->rn_Flags |= GF_DELETED;
	    break;
	}
	for (cn = GetHead(&rn->rn_CtlList); cn; cn = GetSucc(&cn->cn_Node)) {
	    if (cn->cn_Comm.com_SID == idno) {
		cn->cn_Flags |= GF_DELETED;
		break;
	    }
	}
	if (cn)
	    break;
	UnLoadBMSRegTree(rn);
    }
    if (rn) {
	SaveBMSDataBase();
	UnLoadBMSRegTree(rn);
    } else {
	bmslog("Destroy", "Unable to find ID %d\n", idno);
    }
}

void
RetryBMS(id)
char *id;
{
    RegNode *rn;
    CtlNode *cn;
    long idno = strtol(id, NULL, 0);

    if (idno == 0) {
	bmslog("Retry", "Non-zero ID required, use SID field of COMM element to retry");
	return;
    }

    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
	LoadBMSRegTree(rn);
	if (rn->rn_Comm.com_SID == idno) {
	    SetCommTimeout(&rn->rn_Comm, 1);
	    rn->rn_Comm.com_Retries = 0;
	    break;
	}
	for (cn = GetHead(&rn->rn_CtlList); cn; cn = GetSucc(&cn->cn_Node)) {
	    if (cn->cn_Comm.com_SID == idno) {
		SetCommTimeout(&cn->cn_Comm, 1);
		cn->cn_Comm.com_Retries = 0;
		break;
	    }
	}
	if (cn)
	    break;
	UnLoadBMSRegTree(rn);
    }
    if (rn) {
	SaveBMSDataBase();
	UnLoadBMSRegTree(rn);
    } else {
	bmslog("Retry", "Unable to find ID %d\n", idno);
    }
}


