
/*
 *  PROTOCOL.H
 *
 *  The BMS email protocol is quite simple.  Control and Data is combined
 *  together into a DATABASE (DATABASE.C style) which is then BMS encoded
 *  and transfered via email to some remote host, which BMS decodes it and
 *  then scans the resulting database.
 *
 *  A BMS email file may contain any number of separately encoded BMS
 *  chunks.
 *
 *  All database data is ascii except for the contents of the data files
 *  themselves.
 */

#define ONEHOUR     (60*60)
#define ONEDAY	    (60*60*24)
#define ONEWEEK     (ONEDAY * 7)
#define ONEMONTH    (ONEDAY * 30)
#define ONEYEAR     (ONEDAY * 365)

#define DBTYPE_BMS_REGISTER	    (DBTYPE_CONTROL|0x0000)
#define DBTYPE_BMS_CHANNEL	    (DBTYPE_CONTROL|0x0001)

/*
 *  Except for the registration packet all data is 'under' a
 *  channel of some sort.  channel-opens, for example, run under
 *  a registration channel.
 */

#define DBTYPE_BMS_ACK_REGISTER     (DBTYPE_CONTROL|0x4000)
#define DBTYPE_BMS_NAK_REGISTER     (DBTYPE_CONTROL|0x4001)

#define DBTYPE_BMS_OPEN_GET	    (DBTYPE_CONTROL|0x0010)
#define DBTYPE_BMS_OPEN_PUT	    (DBTYPE_CONTROL|0x0011)
#define DBTYPE_BMS_OPEN_MIRROR	    (DBTYPE_CONTROL|0x0012)
#define DBTYPE_BMS_DEQUEUE_CHANNEL  (DBTYPE_CONTROL|0x001F)

#define DBTYPE_BMS_ACK_OPEN_GET     (DBTYPE_CONTROL|0x4010)
#define DBTYPE_BMS_ACK_OPEN_PUT     (DBTYPE_CONTROL|0x4011)
#define DBTYPE_BMS_ACK_OPEN_MIRROR  (DBTYPE_CONTROL|0x4012)

#define DBTYPE_BMS_FILE 	    (DBTYPE_CONTROL|0x0020)
#define DBTYPE_BMS_DIR		    (DBTYPE_CONTROL|0x0021)
#define DBTYPE_BMS_MIRROR_ALERT     (DBTYPE_CONTROL|0x0022)
#define DBTYPE_BMS_DEQUEUE	    (DBTYPE_CONTROL|0x0023)
#define DBTYPE_BMS_RETRY	    (DBTYPE_CONTROL|0x0024)
#define DBTYPE_BMS_DONE 	    (DBTYPE_CONTROL|0x0025)
#define DBTYPE_BMS_NODE 	    (DBTYPE_CONTROL|0x0026)

#define DBTYPE_BMS_DATA 	    (DBTYPE_CONTROL|0x0030)
#define DBTYPE_BMS_AMI_PROTECTION   (DBTYPE_CONTROL|0x0031)
#define DBTYPE_BMS_AMI_TIMESTAMP    (DBTYPE_CONTROL|0x0032)
#define DBTYPE_BMS_AMI_COMMENT	    (DBTYPE_CONTROL|0x0033)
#define DBTYPE_BMS_SIZE 	    (DBTYPE_CONTROL|0x0034)
#define DBTYPE_BMS_OFFSET	    (DBTYPE_CONTROL|0x0035)
#define DBTYPE_BMS_FILENO	    (DBTYPE_CONTROL|0x0036)
#define DBTYPE_BMS_TTLFILES	    (DBTYPE_CONTROL|0x0037)
#define DBTYPE_BMS_NOFILE	    (DBTYPE_CONTROL|0x0038)
#define DBTYPE_BMS_NOACCESS	    (DBTYPE_CONTROL|0x0039)

#define DBTYPE_BMS_RETRY_CHUNK	    (DBTYPE_CONTROL|0x003A)

/*
 *  Common node types
 */

#define DBTYPE_BMS_LOID 	    (DBTYPE_CONTROL|0x0080)
#define DBTYPE_BMS_RMID 	    (DBTYPE_CONTROL|0x0081)
#define DBTYPE_BMS_ERROR	    (DBTYPE_CONTROL|0x0082)
#define DBTYPE_BMS_MESSAGE	    (DBTYPE_CONTROL|0x0083)
#define DBTYPE_BMS_DEQUEUE	    (DBTYPE_CONTROL|0x0084)
#define DBTYPE_BMS_RETPATH	    (DBTYPE_CONTROL|0x0085)
#define DBTYPE_BMS_KEEPALIVE	    (DBTYPE_CONTROL|0x0086)
#define DBTYPE_BMS_STOP 	    (DBTYPE_CONTROL|0x0087)
#define DBTYPE_BMS_START	    (DBTYPE_CONTROL|0x0088)
#define DBTYPE_BMS_FWDPATH	    (DBTYPE_CONTROL|0x0089)

/*
 *  State machines
 */

#define MAXSTATES   16

#define SMACH_REGISTRATION  1
#define SMACH_GET	    2
#define SMACH_PUT	    3
#define SMACH_LIST	    4
#define SMACH_DEQUEUE	    5

