
/*
 *  STATUS.C
 */

#include "defs.h"

Prototype void StatusBMS(void);
Prototype void PrintCtlList(FILE *, List *, long);
Prototype void PrintFileList(FILE *, List *, long);
Prototype void PrintFileAccList(FILE *, List *, long);
Prototype void PrintChunkList(FILE *, FileNode *, long);
Prototype void PrintCommStatus(FILE *, Comm *, long);
Prototype void PrintParamStatus(FILE *, Param *, long);
Prototype void PrintStatistics(FILE *, Stat *, long);
Prototype void PrintParamVariable(FILE *, short, const  char *, long, long);
Prototype char *HowCompleteChunkList(FileNode *);
Prototype char *StatusOfComm(Comm *);

/*
 *  Display STATUS for specified nodes or default node
 */

void
StatusBMS()
{
    RegNode *rn;

    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
	if (GetHead(&DestList) && (rn->rn_Flags & GF_MARKED) == 0)
	    continue;
	if ((rn->rn_Flags & GF_NODISP) && AllOpt == 0)
	    continue;

	LoadBMSRegTree(rn);

	printf("REGISTRATION %-15s ALIAS=%-8s STAT=%s\n",
	    (rn->rn_Node.ln_Name ? rn->rn_Node.ln_Name : "<defaults>"),
	    (rn->rn_Alias ? rn->rn_Alias : "<none>"),
	    StatusOfComm(&rn->rn_Comm)
	);
	if (VerboseOpt)
	    printf("    RETPATH %s\n", rn->rn_RetPath ? rn->rn_RetPath : "<unknown>");
	if (VarsOpt)
	    PrintParamStatus(stdout, &rn->rn_Param, 4);
	if (DebugOpt)
	    PrintCommStatus(stdout, &rn->rn_Comm, 4);
	if (StatsOpt)
	    PrintStatistics(stdout, &rn->rn_Stat, 4);
	PrintCtlList(stdout, &rn->rn_CtlList, 4);
	if (FilesOpt)
	    PrintFileAccList(stdout, &rn->rn_AccList, 4);
	UnLoadBMSRegTree(rn);
    }
}

void
PrintCtlList(fo, list, tab)
FILE *fo;
List *list;
long tab;
{
    CtlNode *cn;

    for (cn = GetHead(list); cn; cn = GetSucc(&cn->cn_Node)) {
	if ((cn->cn_Flags & GF_NODISP) && AllOpt == 0)
	    continue;
	tbfprintf(fo, tab, "CONTROL: %s  STAT=%s\n",
	    cn->cn_Node.ln_Name ? cn->cn_Node.ln_Name : "",
	    StatusOfComm(&cn->cn_Comm)
	);
	if (DebugOpt)
	    PrintCommStatus(fo, &cn->cn_Comm, tab + 4);
	if (StatsOpt)
	    PrintStatistics(fo, &cn->cn_Stat, tab + 4);
	if (ActiveOpt)
	    PrintFileList(fo, &cn->cn_FileList, tab + 4);
    }
}

void
PrintFileList(fo, list, tab)
FILE *fo;
List *list;
long tab;
{
    FileNode *fn;
    long numFiles = 0;

    for (fn = GetHead(list); fn; fn = GetSucc(&fn->fn_Node)) {
	if (fn->fn_Files)
	    numFiles = fn->fn_Files;
    }

    for (fn = GetHead(list); fn; fn = GetSucc(&fn->fn_Node)) {
	tbfprintf(fo, tab, "FILE %d/%d: %s%s %s%s size=%d flags=%s%s%s%s %d%%\n",
	    fn->fn_FileNo,
	    numFiles,
	    (fn->fn_RemFileName) ? "rem=" : "",
	    fn->fn_RemFileName,
	    (fn->fn_RemFileName) ? "loc=" : "",
	    fn->fn_LocFileName,
	    fn->fn_Size,
	    (fn->fn_Flags & FNF_DIR) ? " DIR" : "",
	    (fn->fn_Flags & FNF_ILLEGAL) ? " NOACCESS" : "",
	    (fn->fn_Flags & FNF_BAD) ? " NOTFND" : "",
	    (fn->fn_Flags & FNF_DONE) ? " DN" : "",
	    HowCompleteChunkList(fn)
	);
	if (ActiveOpt > 1)
	    PrintChunkList(fo, fn, tab + 4);
    }
}

void
PrintChunkList(fo, fn, tab)
FILE *fo;
FileNode *fn;
long tab;
{
    FileChunk *fc;

    for (fc = GetHead(&fn->fn_ChunkList); fc; fc = GetSucc(&fc->fc_Node)) {
	tbfprintf(fo, tab, "CHUNK: off=%-6d len=%-6d\n", fc->fc_Offset, fc->fc_Bytes);
    }
}

void
PrintCommStatus(fo, comm, tab)
FILE *fo;
Comm *comm;
long tab;
{
    tbfprintf(fo, tab, "COMM: ctl=%d SID=%08lx CID=%08lx DID=%08lx SeqNo=%d State=%d\n",
	comm->com_Control,
	comm->com_SID,
	comm->com_CID,
	comm->com_DID,
	comm->com_SeqNo,
	comm->com_State
    );
    tbfprintf(fo, tab, "      retry#%d TxReq=%d TxOk=%d TmpTx=%d RemTxBpd=%d RemTxBpw=%d\n",
	comm->com_Retries,
	comm->com_TxReq,
	comm->com_TxOk,
	comm->com_TmpTx,
	comm->com_RemTxBpd,
	comm->com_RemTxBpw
    );
    tbfprintf(fo, tab, "      flags=0x%04x TOTime=%08lx TOStamp=%08lx msg=%s\n",
	comm->com_Flags,
	comm->com_TOTime,
	comm->com_TOStamp,
	comm->com_Msg ? comm->com_Msg : ""
    );
}

void
PrintParamStatus(fo, param, tab)
FILE *fo;
Param *param;
long tab;
{
    tab += 4;

    tbfprintf(fo, tab - 4, "PARAMS:\n");
    PrintParamVariable(fo, tab, "MaxRxBpd",     param->pm_MaxRxBpd, DefRN->rn_Param.pm_MaxRxBpd);
    PrintParamVariable(fo, tab, "MaxTxBpd",     param->pm_MaxTxBpd, DefRN->rn_Param.pm_MaxTxBpd);
    PrintParamVariable(fo, tab, "MaxRxBpw",     param->pm_MaxRxBpw, DefRN->rn_Param.pm_MaxRxBpw);
    PrintParamVariable(fo, tab, "MaxTxBpw",     param->pm_MaxTxBpw, DefRN->rn_Param.pm_MaxTxBpw);
    PrintParamVariable(fo, tab, "MaxRxMail",    param->pm_MaxRxMail,DefRN->rn_Param.pm_MaxRxMail);
    PrintParamVariable(fo, tab, "MaxTxMail",    param->pm_MaxTxMail,DefRN->rn_Param.pm_MaxTxMail);
    PrintParamVariable(fo, tab, "MinMail",      param->pm_MinMail,  DefRN->rn_Param.pm_MinMail);
    PrintParamVariable(fo, tab, "MirRxBpd",     param->pm_MirRxBpd, DefRN->rn_Param.pm_MirRxBpd);
    PrintParamVariable(fo, tab, "MirTxBpd",     param->pm_MirTxBpd, DefRN->rn_Param.pm_MirTxBpd);
    PrintParamVariable(fo, tab, "CacBwPerc",    param->pm_CacBwPerc,DefRN->rn_Param.pm_CacBwPerc);
    PrintParamVariable(fo, tab, "ReqSwamp",     param->pm_ReqSwamp, DefRN->rn_Param.pm_ReqSwamp);
    PrintParamVariable(fo, tab, "MirSwamp",     param->pm_MirSwamp, DefRN->rn_Param.pm_MirSwamp);
}

void
PrintParamVariable(fo, tab, name, node_value, def_value)
FILE *fo;
short tab;
const char *name;
long node_value;
long def_value;
{
    if (node_value)
	tbfprintf(fo, tab, "%-9s %d\n", name, node_value);
    else
	tbfprintf(fo, tab, "%-9s %d (default)\n", name, def_value);
}

void
PrintStatistics(fo, stat, tab)
FILE *fo;
Stat *stat;
long tab;
{
}

void
PrintFileAccList(fo, list, tab)
FILE *fo;
List *list;
long tab;
{
    FileAccess *fa;

    for (fa = GetHead(list); fa; fa = GetSucc(&fa->fa_Node)) {
	if (fa->fa_Flags & GF_DELETED)
	    continue;
	tbfprintf(fo, tab, "ACCESS: %c%s\n", ((fa->fa_Flags & FAF_RESTRICT) ? '-' : ' '), fa->fa_Node.ln_Name);
    }
}

char *
HowCompleteChunkList(fn)
FileNode *fn;
{
    long n;

    n = FileBytesRemaining(fn);
    if (n == -1)
	return(0);
    if (n == 0) {
	if (fn->fn_Flags & FNF_DONE)
	    return(100);
	if (fn->fn_Flags & FNF_DIR)
	    return(0);
    }

    /*
     *	n / fn->fn_Size  percent remaining
     */

    n = 100 - n * 100 / fn->fn_Size;	/*  percent complete	*/
    return(n);
}

char *
StatusOfComm(com)
Comm *com;
{
    static char Buf[128];
    long n;

    n = sprintf(Buf, "%d:%s to=", com->com_SID, com->com_Msg ? com->com_Msg : "<unknown>");
    if (com->com_TOTime == -1) {
	n +=sprintf(Buf + n, "off");
    } else {
	long delta = com->com_TOStamp - gettime();
	if (delta < 0)
	    n += sprintf(Buf + n, "now");
	else if (delta > ONEDAY * 3)
	    n += sprintf(Buf + n, "%dd", delta / ONEDAY);
	else
	    n += sprintf(Buf + n, "%02d:%02d", delta / ONEHOUR, delta / 60 % 60);
    }
    return(Buf);
}


