
/*
 *  SUBS.C
 */

#include "defs.h"
#include <config.h>

/*
 *  DEBUG_TIMEMPX is the time multiplication factor, used for debugging so
 *  one doesn't have to wait days...
 *
 */

/*#define DEBUG_TIMEMPX   1440*/

#define CRCSEED 0xA15F423B

Prototype long GenCrcBuf(long *, unsigned char *, long);
Prototype void CrcInit(long *);
Prototype void CrcBump(long *, unsigned char);
Prototype char *zstrdup(const char *);
Prototype void *zalloc(long);
Prototype void zfree(void *);
Prototype Node *MakeNode(const char *);
Prototype void bmslog(const char *, const char *, ...);
Prototype bool_t getyn(const char *, char);
Prototype short tbprintf(short, const char *, ...);
Prototype short tbfprintf(FILE *, short, const char *, ...);
Prototype short tbvfprintf(FILE *, short, const char *, va_list);
Prototype RegNode *FindRegNode(const  char *, short);
Prototype CtlNode *FindCtlNode(RegNode *, long, long);
Prototype RegNode *FindFirstMarked(void);
Prototype RegNode *FindNextMarked(RegNode *);
Prototype char *AllocTemporaryDiskFile(void);
Prototype time_t gettime(void);
Prototype char *SendMailName(void);

long
GenCrcBuf(pcrc, buf, bytes)
long *pcrc;
unsigned char *buf;
long bytes;
{
    long crc = *pcrc;

    while (bytes) {
	if (crc & 0x80000000)
	    crc = (crc << 1) ^ *buf ^ 1;
	else
	    crc = (crc << 1) ^ *buf;
	++buf;
	--bytes;
    }
    *pcrc = crc;
    return(crc);
}

void
CrcInit(pcrc)
long *pcrc;
{
    *pcrc = CRCSEED;
}

void
CrcBump(pcrc, c)
long *pcrc;
unsigned char c;
{
    long crc = *pcrc;

    if (crc & 0x80000000)
	*pcrc = (crc << 1) ^ c ^ 1;
    else
	*pcrc = (crc << 1) ^ c;
}

void *
zalloc(bytes)
long bytes;
{
    void *ptr;

    if (ptr = malloc(bytes)) {
	clrmem(ptr, bytes);
	return(ptr);
    }
    bmslog("Memory", "Allocate %d bytes failed", bytes);
    exit(1);
}

void
zfree(ptr)
void *ptr;
{
    free(ptr);
}

char *
zstrdup(str)
const char *str;
{
    return(strdup(str));
}


Node *
MakeNode(str)
const char *str;
{
    Node *node = zalloc(sizeof(Node) + strlen(str) + 1);
    node->ln_Name = (char *)(node + 1);
    strcpy(node->ln_Name, str);
    return(node);
}

void
bmslog(where, ctl, ...)
const char *where;
const char *ctl;
{
    va_list va;
    static char Buf[512];
    long n;
    FILE *fi;
    time_t t = time(NULL);
    struct tm *tm = localtime(&t);

    va_start(va, ctl);
    n = strftime(Buf, sizeof(Buf), "%d-%b-%y %H:%M ", tm);
    n += sprintf(Buf + n, "%s:", where);
    n += vsprintf(Buf + n, ctl, va);
    if (n && Buf[n-1] == '\n')
	--n;
    printf("%.*s\n", n, Buf);
    va_end(va);

    if (fi = fopen("BMS:BMS.Log", "a")) {
	va_start(va, ctl);
	fprintf(fi, "%.*s\n", n, Buf);
	va_end(va);
	fclose(fi);
    }
}

bool_t
getyn(msg, def)
const char *msg;
char def;
{
    char buf[256];
    bool_t r = def;

    printf("%s (%c): ", msg, def);
    fflush(stdout);

    while (gets(buf)) {
	if (buf[0] == 0)
	    break;
	if (buf[0] == NO) {
	    r = NO;
	    break;
	}
	if (buf[0] == YES) {
	    r = YES;
	    break;
	}
    }
    return(r);
}

short
tbprintf(tab, ctl, ...)
short tab;
const char *ctl;
{
    va_list va;
    short n;

    va_start(va, ctl);
    n = tbvfprintf(stdout, tab, ctl, va);
    va_end(va);
    return(n);
}

short
tbfprintf(fo, tab, ctl, ...)
FILE *fo;
short tab;
const char *ctl;
{
    va_list va;
    short n;

    va_start(va, ctl);
    n = tbvfprintf(fo, tab, ctl, va);
    va_end(va);
    return(n);
}

short
tbvfprintf(fo, tab, ctl, va)
FILE *fo;
short tab;
const char *ctl;
va_list va;
{
    short n;

    n = fprintf(fo, "%*.*s", tab, tab, "");
    n += vfprintf(fo, ctl, va);
    return(n);
}


RegNode *
FindRegNode(name, allNodes)
const char *name;
short allNodes;
{
    RegNode *rn;

    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
	if ((rn->rn_Flags & GF_NODISP) && allNodes == 0)
	    continue;
	if (rn->rn_Node.ln_Name && stricmp(name, rn->rn_Node.ln_Name) == 0)
	    break;
	if (rn->rn_Alias && stricmp(name, rn->rn_Alias) == 0)
	    break;
    }
    return(rn);
}

CtlNode *
FindCtlNode(rn, sid, did)
RegNode *rn;
long sid;
long did;
{
    CtlNode *cn;

    for (cn = GetHead(&rn->rn_CtlList); cn; cn = GetSucc(&cn->cn_Node)) {
	if (sid && cn->cn_Comm.com_SID == sid)
	    break;
	if (did && cn->cn_Comm.com_DID == did)
	    break;
    }
    return(cn);
}

RegNode *
FindFirstMarked(void)
{
    RegNode *rn;

    for (rn = GetHead(&RegList); rn; rn = GetSucc(&rn->rn_Node)) {
	if (rn->rn_Flags & GF_MARKED)
	    break;
    }
    return(rn);
}

RegNode *
FindNextMarked(rn)
RegNode *rn;
{
    if (rn) {
	for (rn = GetSucc(&rn->rn_Node); rn; rn = GetSucc(&rn->rn_Node)) {
	    if (rn->rn_Flags & GF_MARKED)
		break;
	}
    }
    return(rn);
}

char *
AllocTemporaryDiskFile()
{
    static long id;
    char buf[64];

    sprintf(buf, "T:BMS%06lx%d.TMP", FindTask(NULL), ++id);
    return(zstrdup(buf));
}

/*
 *  Return the current time.  For debugging purposes we multiply it
 *  by DEBUG_TIMEMPX (1 when debugging is off) to 'speed up' the retry
 *  rate.
 */

time_t
gettime(void)
{
    time_t t = time(NULL);

    return(t * DEBUG_TIMEMPX);
}

char *
SendMailName()
{
    return(GetConfig(SENDMAIL, "sendmail"));
}

