
/*
 *  TRANSFER.C
 *
 *  High Level Communications
 */

#include "defs.h"

Prototype void InitCommLocalSide(short, RegNode *, Comm *);
Prototype void InitCommRemoteSide(short, RegNode *, Comm *, long);
Prototype void InitCommChangeStateMachine(short, RegNode *, Comm *);
Prototype void SetCommTimeout(Comm *, long);
Prototype void SetCommMessage(Comm *, char *);
Prototype void TerminateComm(RegNode *, CtlNode *, Comm *, char *);
Prototype void SendMailResults(RegNode *, CtlNode *, const char *);

Prototype char *AllocSemiPermanentFileName(Comm *);
Prototype void	SetStateMachine(short, void (*)(DBHan *, record_t, RegNode *, CtlNode *), void (*)(DBHan *, record_t, RegNode *, CtlNode *), void (*)(DBHan *, record_t, RegNode *, CtlNode *));

Prototype void DispatchCommRx(DBHan *, record_t, RegNode *, CtlNode *, Comm *);
Prototype void DispatchCommTx(DBHan *, record_t, RegNode *, CtlNode *, Comm *);
Prototype void DispatchCommTo(DBHan *, record_t, RegNode *, CtlNode *, Comm *);

typedef struct StateDis {
    void (*sd_RxFunc)(DBHan *, record_t, RegNode *, CtlNode *);
    void (*sd_TxFunc)(DBHan *, record_t, RegNode *, CtlNode *);
    void (*sd_ToFunc)(DBHan *, record_t, RegNode *, CtlNode *);
} StateDis;

StateDis	StateDisAry[MAXSTATES];

/*
 *  InitCommLocalSide()
 *
 *  Initialize communications for a new state machine.
 */

void
InitCommLocalSide(smId, rn, comm)
short smId;
RegNode *rn;
Comm *comm;
{
    DefRN->rn_Flags |= GF_MODIFIED;
    rn->rn_Flags |= GF_ALERT | GF_MODIFIED;
    if (++DefRN->rn_CommId == 0)    /*  XXX conflicts w/ existing ID's? */
	DefRN->rn_CommId = 1;
    comm->com_Flags |= GF_MODIFIED | GF_ALERT;
    comm->com_Control = smId;
    comm->com_SID = DefRN->rn_CommId;
    comm->com_CID = 0;
    comm->com_DID = 0;
    comm->com_SeqNo = 0;
    comm->com_State = 0;
    comm->com_Retries = 0;
    comm->com_TmpFileNo = 0;
    comm->com_RemTxBpd = 100000000;	/*  arbitrarily large number	*/
    comm->com_RemTxBpw = 100000000;	/*  arbitrarily large number	*/
    comm->com_TOTime = -1;		/*  disable timeout	*/
}

/*
 *  InitCommRemoteSide()
 *
 *  Initialize communications for a new state machine.
 */

void
InitCommRemoteSide(smId, rn, comm, did)
short smId;
RegNode *rn;
Comm *comm;
long did;
{
    DefRN->rn_Flags |= GF_MODIFIED;
    rn->rn_Flags |= GF_ALERT | GF_MODIFIED;
    if (++DefRN->rn_CommId == 0)    /*  XXX conflicts w/ existing ID's? */
	DefRN->rn_CommId = 1;
    comm->com_Flags |= GF_MODIFIED | GF_ALERT;
    comm->com_Control = smId;
    comm->com_SID = DefRN->rn_CommId;
    comm->com_CID = 0;
    comm->com_DID = did;
    comm->com_SeqNo = 0;
    comm->com_State = 0;
    comm->com_Retries = 0;
    comm->com_TmpFileNo = 0;
    comm->com_RemTxBpd = 100000000;	/*  arbitrarily large number	*/
    comm->com_RemTxBpw = 100000000;	/*  arbitrarily large number	*/
    comm->com_TOTime = -1;		/*  disable timeout	*/
}

void
InitCommChangeStateMachine(smId, rn, comm)
short smId;
RegNode *rn;
Comm *comm;
{
    rn->rn_Flags |= GF_ALERT | GF_MODIFIED;
    comm->com_Flags |= GF_MODIFIED | GF_ALERT;
    comm->com_Control = smId;
    comm->com_SeqNo = 0;
    comm->com_State = 0;
    comm->com_Retries = 0;
    comm->com_TmpFileNo = 0;
    comm->com_RemTxBpd = 100000000;	/*  arbitrarily large number	*/
    comm->com_RemTxBpw = 100000000;	/*  arbitrarily large number	*/
    comm->com_TOTime = -1;		/*  disable timeout	*/
}


/*
 *  Set COMM timeout
 */

void
SetCommTimeout(comm, to)
Comm *comm;
long to;
{
    if (to < 0) {
	if (comm->com_TOTime != -1) {
	    comm->com_TOTime = -1;
	    comm->com_Flags |= GF_MODIFIED;
	}
    } else {
	comm->com_TOTime = to;
	comm->com_TOStamp = gettime() + to;
	comm->com_Flags |= GF_MODIFIED;
    }
}

/*
 *  SetCommMessage()    - Set message field for Comm port
 */

void
SetCommMessage(comm, msg)
Comm *comm;
char *msg;
{
    if (comm->com_Msg) {
	zfree(comm->com_Msg);
	comm->com_Msg = NULL;
    }
    if (msg) {
	comm->com_Msg = zstrdup(msg);
    }
    comm->com_Flags |= GF_MODIFIED;
}

/*
 *  Termination COMM with a message.  COMM node is not deleted, but
 *  is set to a 'done' state for a month.  This is not the same as
 *  a dequeue, this call is meant for protocol completion only.
 *
 *  msg is mailed to the user locally
 */

void
TerminateComm(rn, cn, comm, msg)
RegNode *rn;
CtlNode *cn;	/*  may be NULL */
Comm *comm;
char *msg;
{
    comm->com_Flags |= GF_DELETED | GF_MODIFIED;
    comm->com_Control = 0;
    if (cn)
	cn->cn_Flags |= GF_DELETED | GF_MODIFIED;
    else
	rn->rn_Flags |= GF_DELETED | GF_MODIFIED;
}

/*
 *  Allocate a filename for a semi-permanent file.. that is, a
 *  temporary file that will be filled with file chunks as they
 *  come in over a period of days.
 *
 *  Use sequence numbering for default node
 *
 *  BMS:Tmp/
 */

char *
AllocSemiPermanentFileName(comm)
Comm *comm;
{
    char buf[64];

    sprintf(buf, "BMS:Tmp/Comm%d.%d.TMP", comm->com_SID, comm->com_TmpFileNo++);
    comm->com_Flags |= GF_MODIFIED;
    return(zstrdup(buf));
}

void
SetStateMachine(smId, smRxFunc, smTxFunc, smToFunc)
short smId;
void (*smRxFunc)(DBHan *, record_t, RegNode *, CtlNode *);
void (*smTxFunc)(DBHan *, record_t, RegNode *, CtlNode *);
void (*smToFunc)(DBHan *, record_t, RegNode *, CtlNode *);
{
    StateDis *sd = StateDisAry + smId;

    sd->sd_RxFunc = smRxFunc;
    sd->sd_TxFunc = smTxFunc;
    sd->sd_ToFunc = smToFunc;
}

void
DispatchCommRx(db, rid, rn, cn, comm)
DBHan *db;
record_t rid;
RegNode *rn;
CtlNode *cn;	/*  can be NULL */
Comm *comm;
{
    StateDis *sd = StateDisAry + comm->com_Control;

    if (sd->sd_RxFunc)
	sd->sd_RxFunc(db, rid, rn, cn);
}

void
DispatchCommTx(db, rid, rn, cn, comm)
DBHan *db;
record_t rid;
RegNode *rn;
CtlNode *cn;	/*  can be NULL */
Comm *comm;
{
    StateDis *sd = StateDisAry + comm->com_Control;

    if (sd->sd_TxFunc)
	sd->sd_TxFunc(db, rid, rn, cn);
}


void
DispatchCommTo(db, rid, rn, cn, comm)
DBHan *db;
record_t rid;
RegNode *rn;
CtlNode *cn;	/*  can be NULL */
Comm *comm;
{
    StateDis *sd = StateDisAry + comm->com_Control;

    if (sd->sd_ToFunc)
	sd->sd_ToFunc(db, rid, rn, cn);
}

void
SendMailResults(rn, cn, subject)
RegNode *rn;
CtlNode *cn;
const char *subject;
{
    char *tmpFileName = AllocTemporaryDiskFile();
    FILE *fo;
    long r;

    if (fo = fopen(tmpFileName, "w")) {
	fprintf(fo, "Subject: %s\n", subject);
	fprintf(fo, "\n");
	fprintf(fo, "REGISTRATION %-15s ALIAS=%-8s STAT=%s\n",
	    (rn->rn_Node.ln_Name ? rn->rn_Node.ln_Name : "<defaults>"),
	    (rn->rn_Alias ? rn->rn_Alias : "<none>"),
	    StatusOfComm(&rn->rn_Comm)
	);
	if (cn == NULL) {
	    PrintCommStatus(fo, &rn->rn_Comm, 4);
	    PrintStatistics(fo, &rn->rn_Stat, 4);
	} else {
	    fprintf(fo, "    CONTROL: %s STAT=%s\n",
		cn->cn_Node.ln_Name ? cn->cn_Node.ln_Name : "",
		StatusOfComm(&cn->cn_Comm)
	    );
	    PrintCommStatus(fo, &cn->cn_Comm, 8);
	    PrintStatistics(fo, &cn->cn_Stat, 8);
	    PrintFileList(fo, &cn->cn_FileList, 8);
	}
	fclose(fo);

	sprintf(TmpBuf1, "%s <%s -f bms_err -R BMS -t bms-manager", SendMailName(), tmpFileName);
	r = system(TmpBuf1);
	if (r != 0)
	    bmslog("Dequeue", "Unable to sendmail bms-manager");
	remove(tmpFileName);
    }
    zfree(tmpFileName);
}

