
/*
 *  LOCKFILE.C
 *
 *  (C) Copyright 1989-1990 by Matthew Dillon,  All Rights Reserved.
 *
 *  Lock and unlock a file.  Under AmigaDOS, openning a file mode 1006
 *  (accomplished with fopen(,"w"), locks the file exclusively.  That
 *  is, further fopen()s will fail.  Thus, we need only keep a live
 *  file descriptor to 'lock' the file.
 *
 *  This is advantagious because if the program exits without removing
 *  the lock file we are still ok... the file is unlocked by virtue of
 *  the file descriptor getting closed.
 *
 */

#include <exec/types.h>
#include <exec/lists.h>
#include <libraries/dos.h>
#ifdef LATTICE
#include <proto/exec.h>
#include <proto/dos.h>
#endif
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#define Prototype extern
#define Local static
#define LOCKDIR		"T:"

typedef struct List LIST;
typedef struct Node NODE;

typedef struct {
    NODE    Node;
    FILE    *Fi;
    short   Refs;
} LNode;

Prototype void LockFile(char *);
Prototype void UnLockFile(char *);
Prototype void UnLockFiles(void);
Prototype int FileIsLocked(char *);
Local void FreeLockNode(LNode *);



LIST LockList = { (NODE *)&LockList.lh_Tail, NULL, (NODE *)&LockList.lh_Head };



void
LockFile(char *file)
{
    char *ptr;
    LNode *node;
    LNode *n;

    for (ptr = file + strlen(file); ptr >= file && *ptr != '/' && *ptr != ':'; --ptr);
    ++ptr;

    if (node = malloc(sizeof(LNode) + strlen(LOCKDIR) + 16 + strlen(ptr))) {
	node->Node.ln_Name = (char *)(node + 1);

	sprintf(node->Node.ln_Name,"%s%s.LOCK",LOCKDIR,ptr);

	for (n = (LNode *)LockList.lh_Head; n != (LNode *)&LockList.lh_Tail; n = (LNode *)n->Node.ln_Succ) {
	    if (stricmp(node->Node.ln_Name, n->Node.ln_Name) == 0) {
		++n->Refs;
		free(node);
		return;
	    }
	}

	while ((node->Fi = fopen(node->Node.ln_Name, "w")) == NULL) {
	    Delay(100L);
	    /*chkabort();*/
	}
	node->Refs = 1;
	AddTail(&LockList, &node->Node);
    }
}



/*
 *  Check to see whether a file is locked.  We could try to fopen the
 *  file for 'w', but this causes unnecesary filesystem activity
 */

int
FileIsLocked(char *file)
{
    char *ptr;
    char buf[128];
    long lock;

    for (ptr = file + strlen(file); ptr >= file && *ptr != '/' && *ptr != ':'; --ptr);
    ++ptr;

    sprintf(buf,"%s%s.LOCK",LOCKDIR,ptr);

    if (lock=Lock(buf,EXCLUSIVE_LOCK)) {
	UnLock(lock);
	return(0);
    }
    if (IoErr()==ERROR_OBJECT_IN_USE)
	return(1);
    return(0);
}



void
UnLockFile(char *file)
{
    LNode *node;
    char buf[128];
    char *ptr;

    for (ptr = file + strlen(file); ptr >= file && *ptr != '/' && *ptr != ':'; --ptr);
    ++ptr;

    sprintf(buf,"%s%s.LOCK",LOCKDIR,ptr);

    for (node = (LNode *)LockList.lh_Head; node != (LNode *)&LockList.lh_Tail; node = (LNode *)node->Node.ln_Succ) {
	if (stricmp(buf, node->Node.ln_Name) == 0) {
	    if (--node->Refs == 0)
		FreeLockNode(node);
	    break;
	}
    }
}



void
UnLockFiles()
{
    LNode *node;

    while ((node = (LNode *)LockList.lh_Head) != (LNode *)&LockList.lh_Tail)
	FreeLockNode(node);
}



static void
FreeLockNode(LNode *node)
{
    Remove((struct Node *)node);
    fclose(node->Fi);
    remove(node->Node.ln_Name);
    free(node);
}

