#include "prefs.h"

#include <exec/types.h>
#include <exec/io.h>
#ifdef LATTICE
#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/intuition.h>
#endif
#include <graphics/text.h>
#include <intuition/screens.h>
#include <intuition/intuition.h>
#include <utility/tagitem.h>
#include <graphics/view.h>
#include <devices/console.h>
#include <devices/inputevent.h>
#include <exec/devices.h>
#include <string.h>



struct Library  *IntuitionBase = NULL;
struct Library  *GfxBase = NULL;
struct Screen   *screen = NULL;
struct Window   *window = NULL;
struct IOStdReq *writereq = NULL;
struct MsgPort  *writeport = NULL;
struct IOStdReq *readreq = NULL;
struct MsgPort  *readport = NULL;
struct TextAttr textAttr;
BOOL consoleOpen = FALSE;
BOOL showtitle   = FALSE;
int Win_Height = 0;
int Win_Width  = 0;
int Screen_Height = 200;
int Screen_Width  = 640;
int Screen_Interlace = 0;



Prototype void close_intuition(void);
Prototype BOOL open_intuition(void);
Prototype BOOL intui_interface(void);
Prototype void SetScreenTitle(char *);
Prototype char getch(void);
Prototype long getst(char *,long);
Prototype void putch(char);
Prototype void putst(char *);
Prototype void putstl(char *,long);
Prototype void ElmToFront(void);
Prototype void WorkbenchToFront(void);



void close_intuition(void)
{
  if (consoleOpen) CloseDevice(writereq);		consoleOpen = FALSE;
  if (readreq) DeleteExtIO(readreq);			readreq = NULL;
  if (readport) DeletePort(readport);			readport = NULL;
  if (writereq) DeleteExtIO(writereq);			writereq = NULL;
  if (writeport) DeletePort(writeport);			writeport = NULL;
  if (window) CloseWindow(window);			window = NULL;
  if (screen) CloseScreen(screen);			screen = NULL;
  if (IntuitionBase) CloseLibrary(IntuitionBase);	IntuitionBase = NULL;
  if (GfxBase) CloseLibrary(GfxBase);			GfxBase = NULL;
}



BOOL open_intuition(void)
{
  struct NewScreen newscreen;
  struct NewWindow newwindow;
  UWORD pen_array = ~0;
  struct TagItem screen_tags[] = {SA_Pens,NULL,TAG_END,NULL,TAG_END,NULL};  /* NULL: dummy */
  char str[80];
  int pos=0;
  int i;
  char *scanpnt;
  long color;

  /* initialize */
  screen_tags[0].ti_Data = (ULONG)&pen_array;

  if (!(IntuitionBase=OpenLibrary("intuition.library",33L))) {
    fprintf(stderr,"elm: cannot open intuition.library\n");
    close_intuition();
    return(FALSE);
  }

  if (!(GfxBase=OpenLibrary("graphics.library",33L))) {
    fprintf(stderr,"elm: cannot open graphics.library\n");
    close_intuition();
    return(FALSE);
  }

  textAttr.ta_Name  = "topaz";
  textAttr.ta_YSize = 8;
  textAttr.ta_Style = FS_NORMAL;
  textAttr.ta_Flags = FPF_DESIGNED;

  newscreen.LeftEdge     = 0;
  newscreen.TopEdge      = 0;
  newscreen.Width        = Screen_Width;
  newscreen.Height       = Screen_Height;
  newscreen.Depth        = 2;
  newscreen.DetailPen    = AUTOFRONTPEN;
  newscreen.BlockPen     = AUTOBACKPEN;
  if (Screen_Interlace)
    newscreen.ViewModes  = HIRES|LACE;
  else
    newscreen.ViewModes  = HIRES;
  newscreen.Type         = CUSTOMSCREEN;
  newscreen.Font         = &textAttr;
  newscreen.DefaultTitle = "Amiga Elm  1991 Andreas M. Kirchwitz (original Heiko W. Rupp)";
  newscreen.Gadgets      = NULL;
  newscreen.CustomBitMap = NULL;

  if (IntuitionBase->lib_Version<36) {
    struct Screen WBScreen;
    if (GetScreenData(&WBScreen,sizeof(struct Screen),WBENCHSCREEN,NULL)) {
      newscreen.ViewModes = WBScreen.ViewPort.Modes;
    }

    if (!(screen=OpenScreen(&newscreen))) {
      fprintf(stderr,"elm: cannot open screen\n");
      close_intuition();
      return(FALSE);
    }
  }
  else {
    long WBModeID;
    struct Screen *WBScreen;
    if (WBScreen=LockPubScreen("Workbench")) {
      WBModeID=GetVPModeID(&WBScreen->ViewPort);
      UnlockPubScreen(NULL,WBScreen);
      screen_tags[1].ti_Tag  = SA_DisplayID;
      screen_tags[1].ti_Data = WBModeID;
    }

    if (!(screen=OpenScreenTagList(&newscreen,screen_tags))) {
      fprintf(stderr,"elm: cannot open screen\n");
      close_intuition();
      return(FALSE);
    }
  }

  if (scanpnt=ColorData) {
    for(i=0;i<(1L<<newscreen.Depth);i++) {
      color = strtol(scanpnt,&scanpnt,0);
      SetRGB4(&(screen->ViewPort),i,(color >> 8) & 0xf, (color >> 4) & 0xf, color & 0xf);
    }
    /*RemakeDisplay();*/
  }

/* test ausgabe fuer OS 1.3 -vs- OS 2.0 */

/*
  printf("BarHeight: %d, BarVBorder: %d, BarHBorder: %d\n",
		screen->BarHeight,screen->BarVBorder,screen->BarHBorder);
  printf("MenuVBorder: %d, MenuHBorder: %d\n",
		screen->MenuVBorder,screen->MenuHBorder);
  printf("WBorTop: %d, WBorLeft: %d, WBorRight: %d, WBorBottom: %d\n",
		screen->WBorTop,screen->WBorLeft,screen->WBorRight,screen->WBorBottom);
*/

/* ende der test ausgabe */

  newwindow.LeftEdge     = 0;
  newwindow.TopEdge      = 0;
  newwindow.Width        = newscreen.Width;
  newwindow.Height       = newscreen.Height;
  newwindow.DetailPen    = AUTOFRONTPEN;
  newwindow.BlockPen     = AUTOBACKPEN;
  newwindow.IDCMPFlags   = 0L;
  newwindow.Flags        = BACKDROP|BORDERLESS|RMBTRAP|NOCAREREFRESH|ACTIVATE;
  newwindow.FirstGadget  = NULL;
  newwindow.CheckMark    = NULL;
  newwindow.Title        = NULL;
  newwindow.Screen       = screen;
  newwindow.BitMap       = NULL;
  newwindow.MinWidth     = newscreen.Width;
  newwindow.MinHeight    = newscreen.Height;
  newwindow.MaxWidth     = newscreen.Width;
  newwindow.MaxHeight    = newscreen.Height;
  newwindow.Type         = CUSTOMSCREEN;

  if (!(window=OpenWindow(&newwindow))) {
    fprintf(stderr,"elm: cannot open window\n");
    close_intuition();
    return(FALSE);
  }

  /* ShowTitle(screen,FALSE); */

  if (!(writeport=CreatePort("Elm-Console.write",0L))) {
    fprintf(stderr,"elm: cannot open console write port\n");
    close_intuition();
    return(FALSE);
  }

  if (!(readport=CreatePort("Elm-Console.read",0L))) {
    fprintf(stderr,"elm: cannot open console read port\n");
    close_intuition();
    return(FALSE);
  }

  if (!(writereq=CreateExtIO(writeport,(long)sizeof(struct IOStdReq)))) {
    fprintf(stderr,"elm: cannot open console write request\n");
    close_intuition();
    return(FALSE);
  }

  if (!(readreq=CreateExtIO(readport,(long)sizeof(struct IOStdReq)))) {
    fprintf(stderr,"elm: cannot open console read request\n");
    close_intuition();
    return(FALSE);
  }

  writereq->io_Data = (APTR)window;
  writereq->io_Length = sizeof(struct Window);

  if (OpenDevice("console.device",0L,writereq,0L)) {
    fprintf(stderr,"elm: cannot open console.device\n");
    close_intuition();
    return(FALSE);
  }

  consoleOpen = TRUE;

  readreq->io_Device = writereq->io_Device;
  readreq->io_Unit   = writereq->io_Unit;

  if (IntuitionBase->lib_Version<36)
    sprintf(TBuffer,"%c%d%c",0x9b,screen->BarHeight+3,0x79);
  else
    sprintf(TBuffer,"%c%d%c",0x9b,screen->BarHeight+4,0x79);
  outline();

  /* get window height and width */

  sprintf(TBuffer,"%c%c%c%c",0x9b,0x30,0x20,0x71);
  outline();
  do {
    str[pos++]=getch();
  }
  while (str[pos-1]!='r');
  pos--;  /* 'r' */
  pos--;  /* ' ' */
  str[pos]='\0';

#define isnum(x) ((x)>='0' && (x)<='9')

  do {pos--;} while (isnum(str[pos]));
  pos++;

  Win_Width = atoi(&str[pos]);

  pos--;
  str[pos]='\0';
  do {pos--;} while (isnum(str[pos]));
  pos++;

  Win_Height = atoi(&str[pos]);

/*  printf("Dimensions: %dx%d chars\n",Win_Width,Win_Height);*/

  if (Win_Width<80 || Win_Height<20) {
    fprintf(stderr,"elm: minimum display is 80x20 characters (you have %dx%d)\n",Win_Width,Win_Height);
    close_intuition();
    return(FALSE);
  }

  return(TRUE);
}



BOOL intui_interface(void)
{
  return(window!=NULL);
}



void SetScreenTitle(char *text)
{
  SetWindowTitles(window,(char *)-1L,text);
  ShowTitle(screen,TRUE);
}



void ElmToFront(void)
{
  ScreenToFront(screen);
}



void WorkbenchToFront(void)
{
  WBenchToFront();
}



char getch(void)
{
  char ch = '\0';
  readreq->io_Command = CMD_READ;
  readreq->io_Data    = (APTR)&ch;
  readreq->io_Length  = 1;
  DoIO(readreq);
  return(ch);
}



long getst(char *buffer,long limit)
{
  register long len = 0L;
  char ch;

  while ((ch=getch())!='\r') {
    if (ch=='\b') {
      if (len>0) {
        putst("\b \b");
        len--;
      }
    }
    else if (ch==0x1b || ch==0x9b) {
    }
    else {
      if (len<limit)
        putch( buffer[len++] = ch );
    }
  }

  buffer[len] = '\0';

  return(len);
}



void putch(char ch)
{
  writereq->io_Command = CMD_WRITE;
  writereq->io_Data    = (APTR)&ch;
  writereq->io_Length  = 1;
  DoIO(writereq);
}



void putst(char *buffer)
{
  writereq->io_Command = CMD_WRITE;
  writereq->io_Data    = (APTR)buffer;
  writereq->io_Length  = strlen(buffer);
  DoIO(writereq);
}



void putstl(char *buffer,long length)
{
  writereq->io_Command = CMD_WRITE;
  writereq->io_Data    = (APTR)buffer;
  writereq->io_Length  = length;
  DoIO(writereq);
}
