
/*
 *  DISPLAY.C
 *
 *  This module handles all display windows.
 *
 */

#include <exec/types.h>
#include <exec/memory.h>
#include "defs.h"
#include <intuition/intuition.h>
#include <clib/exec_protos.h>
#include <clib/intuition_protos.h>

Prototype int getyn(char *);
Prototype void *OpenBrowseDisplay(char *, char *, short);
Prototype void SetBrowseTitle(void *, char *);
Prototype void CloseBrowseDisplay(void *);
Prototype void ChangeBrowseDisplay(void *, char *, char *);
Prototype int WaitBrowseEvent(void **, int *, int *);
Prototype void	LoadDisplayConfig(void);
Prototype void	SaveDisplayConfig(void);
Prototype void	RefreshBrowseDisplay(void *, short);

Local void TabExpand(char *, short);



#define MAXCONF 3

typedef struct Window	    Window;
typedef struct NewWindow    NewWindow;
typedef struct IntuiMessage IMsg;

typedef struct Display {
    struct Display *Next;
    Window  *Win;	    /*	associated window	*/
    char    *Name;	    /*	name of file		*/
    char    *Title;	    /*	window title		*/
    FILE    *Fi;	    /*	file pointer		*/
    long    WMask;	    /*	wait mask for window	*/
    short   Id; 	    /*	config id		*/
    short   FlagRefresh;
    short   ColStart;
    long    FirstNonHdrLine;/*	first non-header line	*/

    long    TLineNo;	    /*	line # for top line	*/
    long    TPos;	    /*	position (top line)     */
    long    BPos;	    /*	position (bottom line)  */
    long    XPos;	    /*	end of file (or scan)   */
    long    WLines;	    /*	lines displayed in win	*/
    long    WMaxLines;	    /*	maximum #lines in window*/
    long    FLines;	    /*	lines in file if known	*/
    struct  Gadget *PropGad;/*	prop gadget for window	*/
    long    LastTLine;
} Display;

typedef struct Config {
    short   TopEdge;
    short   LeftEdge;
    short   Width;
    short   Height;
} Config;

static long    WMask;	   /*  Window mask	   */
static Display *CachDisp;  /*  last display that had an event		   */
static Display *DispBase;  /*  list of displays    */
static short   FlagRefresh;

static Config  ConfAry[MAXCONF];

Local int   HandleEvent(Display *, int *, int *);
Local void  SaveWindowConfig(Display *);
Local long  BackupFile(Display *, long, long);
Local long  ForwardFile(Display *, long, long);
Local void  SetRefresh(Display *, int);

void
FixPropGadget(_disp, all)
void *_disp;
short all;
{
    Display *disp=_disp;
    UWORD VertPot,VertBody;
    short hidden;
    short TopLine;
    short WLines;

    if (disp->PropGad==0)
	return;

    WLines=(disp->Win->Height-disp->Win->BorderTop)-(disp->Win->BorderBottom+1);
    WLines/=disp->Win->RPort->TxHeight;

    if (WLines<disp->FLines) {
	hidden=disp->FLines-WLines;

	if (disp->TLineNo>hidden)
	    TopLine=hidden;
	else
	    TopLine=disp->TLineNo;

	VertBody=((WLines-1)*MAXBODY)/(disp->FLines-1);
	VertPot=(TopLine*MAXPOT)/hidden;
    } else {
	VertPot=MAXPOT;
	VertBody=MAXBODY;
    };
    if (all)
	NewModifyProp(disp->PropGad, disp->Win, 0, AUTOKNOB|PROPBORDERLESS|PROPNEWLOOK|FREEVERT, -1, VertPot,-1,VertBody,1);
}


void *
OpenBrowseDisplay(fileName, title, id)
char *fileName;
char *title;
short id;
{
    static NewWindow Nw = {
	0, 0, 320, 200, -1, -1,
	CLOSEWINDOW|NEWSIZE|MOUSEBUTTONS|VANILLAKEY|REFRESHWINDOW|RAWKEY|
	GADGETUP|GADGETDOWN|MOUSEMOVE,
	WINDOWSIZING|WINDOWDRAG|WINDOWDEPTH|WINDOWCLOSE|
	SIMPLE_REFRESH|ACTIVATE,
	NULL, NULL, NULL, NULL, NULL,
	32, 16, -1, -1, WBENCHSCREEN
    };
    static struct PropInfo PropInfoTemplate = {
	PROPBORDERLESS|FREEVERT|AUTOKNOB|PROPNEWLOOK,	/* Flags */
	-1,				    /* HorizPot     */
	-1,				    /* VertPot	    */
	-1,				    /* HorizBody    */
	-1,				    /* VertBody     */
    };
    static struct Gadget WindowGad = {
	0,				    /* Next	    */
	0,0,0,0,			    /* Position     */
	GFLG_RELHEIGHT|GFLG_RELRIGHT,	    /* Flags	    */
	GACT_RIGHTBORDER|		    /* Activation   */
	GACT_FOLLOWMOUSE|GACT_IMMEDIATE,
	GTYP_PROPGADGET,		    /* Type	    */
	NULL, NULL,			    /* Render	    */
	NULL,				    /* GadgetText   */
	NULL,				    /* MX	    */
	NULL,				    /* SpecialInfo  */
	1,				    /* Gadget ID    */
	NULL,				    /* UserData     */
    };

    Display *disp = malloc(sizeof(Display));

    if (title == NULL)
	title = fileName;

    setmem(disp, sizeof(Display), 0);

    if (id < 0 || id >= MAXCONF)
	id = MAXCONF - 1;

    {
	Config *conf = ConfAry + id;
	Nw.LeftEdge = conf->LeftEdge;
	Nw.TopEdge  = conf->TopEdge;
	Nw.Width    = conf->Width;
	Nw.Height   = conf->Height;
	if (Nw.Width < 32) {
	    Nw.LeftEdge = 40;
	    Nw.Width = 600;
	}
	if (Nw.Height < 16)
	    Nw.Height = 150;
    }

    disp->Win = OpenWindow(&Nw);
    if (disp->Win == NULL) {
	Nw.LeftEdge = 0;
	Nw.TopEdge  = 0;
	Nw.Width    = 320;
	Nw.Height   = 100;
	disp->Win = OpenWindow(&Nw);
    }

    if (disp->Win) {
	struct Gadget *Gad;
	struct PropInfo *Prop;

	disp->Name = strdup(fileName);
	disp->Title= strdup(title);
	disp->WMask = 1 << disp->Win->UserPort->mp_SigBit;
	disp->Id = id;
	disp->LastTLine = -20;

	disp->Win->UserData = 0;
	Gad = AllocRemember(&disp->Win->UserData, sizeof(*Gad), MEMF_CLEAR);
	Prop = AllocRemember(&disp->Win->UserData, sizeof(*Prop), MEMF_CLEAR);

	if (Gad && Prop) {
	    struct Window *Window=disp->Win;

	    *Gad=WindowGad;
	    *Prop=PropInfoTemplate;

	    Gad->LeftEdge=-(Window->BorderRight-5);
	    Gad->TopEdge=Window->BorderTop+1;
	    Gad->Width=Window->BorderRight-8;
	    Gad->Height=-(Window->BorderBottom+Window->BorderTop+11);

	    Gad->SpecialInfo=(APTR)Prop;
	    Gad->GadgetRender=AllocRemember(&disp->Win->UserData, sizeof(struct Image), MEMF_CLEAR);
	    if (Gad->GadgetRender) {
		disp->PropGad=Gad;
		AddGadget(Window, Gad, -1);
		RefreshGadgets(Gad, Window, NULL);
	    }
	}

	SetAPen(disp->Win->RPort, 1);

	SetWindowTitles(disp->Win, disp->Title, (char *)-1L);

	disp->Next = DispBase;
	DispBase = disp;

	WMask |= disp->WMask;

	SetRefresh(disp, 0);
	disp->FLines = 0;
	if (disp->Fi = fopen(disp->Name, "r")) {
	    disp->FirstNonHdrLine = 0;
	    while (fgets(TmpBuf, sizeof(TmpBuf), disp->Fi) && TmpBuf[0] != '\n') {
		++disp->FirstNonHdrLine;
		++disp->FLines;
	    }
	    while (fgets(TmpBuf, sizeof(TmpBuf), disp->Fi))
		++disp->FLines;
	    disp->XPos = ftell(disp->Fi);
	}
    } else {
	free(disp);
	disp = NULL;
    }
    return((void *)disp);
}

void
SetBrowseTitle(_disp, title)
void *_disp;
char *title;
{
    Display *disp = _disp;
    char *old = disp->Title;

    disp->Title = strdup(title);
    SetWindowTitles(disp->Win, disp->Title, (char *)-1L);
    free(old);
}

void
CloseBrowseDisplay(_disp)
void *_disp;
{
    Display *disp = _disp;
    Display *d;

    if (disp == CachDisp)
	CachDisp = NULL;

    if (DispBase == disp) {                     /*  unlink display  */
	DispBase = disp->Next;
    } else {
	for (d = DispBase; d; d = d->Next) {
	    if (d->Next == disp) {
		d->Next = disp->Next;
		break;
	    }
	}
    }

    SetWindowTitles(disp->Win, NULL, NULL);

    if (disp->Win) {
	SaveWindowConfig(disp);
	CloseWindow(disp->Win);
	FreeRemember(&disp->Win->UserData, TRUE);
    }

    WMask &= ~disp->WMask;

    free(disp->Title);
    free(disp->Name);
    if (disp->Fi)
	fclose(disp->Fi);
    free(disp);
}

/*
 *  Change file being displayed
 */

void
ChangeBrowseDisplay(_disp, fileName, title)
void *_disp;
char *fileName;
char *title;
{
    Display *disp = _disp;
    Window *win = disp->Win;

    if (title == NULL)
	title = fileName;

    {
	char *old = disp->Title;
	disp->Title = strdup(title);
	SetWindowTitles(disp->Win, disp->Title, (char *)-1L);
	free(old);
    }

    free(disp->Name);
    disp->Name = strdup(fileName);
    if (disp->Fi) {
	fclose(disp->Fi);
	disp->Fi = NULL;
    }

    disp->TLineNo = 0;
    disp->LastTLine = -20;
    disp->TPos = 0;
    disp->XPos = 0;
    disp->WLines = 0;
    disp->WMaxLines = 0;
    disp->FLines = 0;

    SetRefresh(disp, 0);
}

/*
 *  Wait for an event.	While waiting finish up scanning files for which we
 *  do not know the line number extent.  Returns 0 if there are no windows
 *  left.
 */

int
WaitBrowseEvent(_pdisp, prow, pcol)
void **_pdisp;
int *prow;
int *pcol;
{
    long mask;
    Display **pdisp = _pdisp;

    if (WMask == 0)
	return(0);

    for (;;) {
	Display *disp;

	if (FlagRefresh && (SetSignal(0,0) & WMask) == 0) {
	    for (disp = DispBase; disp; disp = disp->Next) {
		if (disp->FlagRefresh)
		    RefreshBrowseDisplay(disp, 0);
	    }
	    FlagRefresh = 0;
	}
	mask = Wait(WMask);
	if ((disp = CachDisp) && (disp->WMask & mask)) {
	    mask &= ~disp->WMask;
	    if (HandleEvent(disp, prow, pcol))
		goto event_handled;
	}
	if (mask) {
	    for (disp = DispBase; disp; disp = disp->Next) {
		if (mask & disp->WMask) {
		    mask &= ~disp->WMask;
		    if (HandleEvent(disp, prow, pcol)) {
event_handled:
			*pdisp = disp;
			if (GetHead(&disp->Win->UserPort->mp_MsgList))
			    mask |= disp->WMask;
			if (mask)
			    SetSignal(mask, mask);
			return(1);
		    }
		}
	    }
	}
    }
}

static int
HandleEvent(disp, prow, pcol)
Display *disp;
int *prow;
int *pcol;
{
    int ret = 0;
    IMsg *im;
    Window *win = disp->Win;
    static short PropGadgetDown;

    while ((ret == 0) && (im = GetMsg(win->UserPort))) {
	switch(im->Class) {
	case GADGETDOWN:
	    PropGadgetDown = 1;
	    break;
	case GADGETUP:
	    PropGadgetDown = -1;
	case MOUSEMOVE:
	    if (PropGadgetDown) {
		UWORD hidden,WLines;
		long lineNo;

		if (disp->FLines>disp->WLines) {
		    struct PropInfo *PI=(struct SpecialInfo *)disp->PropGad->SpecialInfo;

		    WLines=(disp->Win->Height-disp->Win->BorderTop)-(disp->Win->BorderBottom+1);
		    WLines/=disp->Win->RPort->TxHeight;
		    hidden=disp->FLines-WLines;
		    lineNo = ((hidden*PI->VertPot)+(MAXPOT/2))/MAXPOT;
		    if (lineNo != disp->TLineNo) {
			disp->TLineNo = lineNo;
			disp->TPos = GotoLine(disp, lineNo);
			SetRefresh(disp,0);
		    }
		}
	    }
	    if (PropGadgetDown < 0)
		PropGadgetDown = 0;
	    break;
	case CLOSEWINDOW:
	    *prow = -1;
	    *pcol = -1;
	    ret = 1;
	    break;
	case VANILLAKEY:
	    switch(im->Code) {
	    case 27:
		*prow = -1;
		*pcol = -1;
		ret = 1;
		break;
	    case 13:
		if (disp->BPos != disp->XPos) {
		    disp->TPos = ForwardFile(disp, disp->TPos, 2);
		    ++disp->TLineNo;
		    SetRefresh(disp, 0);
		}
		break;
	    case '2':       /*  page down   */
	    case ' ':
		if (disp->BPos != disp->XPos) {
		    long move;

		    for (move = disp->WMaxLines; disp->TLineNo + move > disp->FLines - disp->WMaxLines; --move)
			;
		    if (move < 0)
			move = 0;

		    disp->TPos = ForwardFile(disp, disp->TPos, move);
		    disp->TLineNo += move;
		    SetRefresh(disp, 0);
		}
		break;
	    case '8':       /*  page up   this article  */
		if (disp->TPos != 0) {
		    disp->TPos = BackupFile(disp, disp->TPos, disp->WMaxLines);
		    disp->TLineNo -= disp->WMaxLines;
		    if (disp->TLineNo < 0)
			disp->TLineNo = 0;
		    SetRefresh(disp, 0);
		}
		break;
	    case '3':
		if (disp->BPos != disp->XPos) {
		    disp->TPos = BackupFile(disp, disp->XPos, disp->WLines);
		    disp->TLineNo = disp->FLines - disp->WLines;    /* XXX */
		    if (disp->TLineNo < 0)
			disp->TLineNo = 0;
		    SetRefresh(disp, 0);
		}
		break;
	    case '9':
		if (disp->TPos != 0) {
		    disp->TPos = 0;
		    disp->TLineNo = 0;
		    SetRefresh(disp, 0);
		}
		break;
	    case '6':
		SetRefresh(disp, disp->ColStart + 40);
		break;
	    case '4':
		SetRefresh(disp, disp->ColStart - 40);
		break;
	    default:
		*prow = -1;
		*pcol = im->Code;
		ret = 1;
	    }
	    break;
	case RAWKEY:
	    switch (im->Code) {
	    case 0x4c:
		if (im->Qualifier & (IEQUALIFIER_LSHIFT|IEQUALIFIER_RSHIFT)) {
		    if (disp->TPos != 0) {
			disp->TPos = BackupFile(disp, disp->TPos, disp->WMaxLines);
			disp->TLineNo -= disp->WMaxLines;
			if (disp->TLineNo < 0)
			    disp->TLineNo = 0;
			SetRefresh(disp, 0);
		    }
		} else {
		    if (disp->TPos != 0) {
			disp->TPos = BackupFile(disp, disp->TPos, 1);
			disp->TLineNo--;
			if (disp->TLineNo < 0)
			    disp->TLineNo = 0;
			SetRefresh(disp, 0);
		    }
		}
		break;
	    case 0x4d:
		if (im->Qualifier&(IEQUALIFIER_LSHIFT|IEQUALIFIER_RSHIFT)) {
		    if (disp->BPos != disp->XPos) {
			int move;

			move=disp->WMaxLines;
			while ((disp->TLineNo+move) >= (disp->FLines-disp->WMaxLines))
			    move--;

			if (move < 0)
			    move = 0;

			disp->TPos = ForwardFile(disp, disp->TPos, move);
			disp->TLineNo += move;
			SetRefresh(disp, 0);
		    }
		} else {
		    if (disp->BPos != disp->XPos && (disp->TLineNo+disp->WMaxLines < disp->FLines)) {
			disp->TPos = ForwardFile(disp, disp->TPos, 1);
			++disp->TLineNo;
			SetRefresh(disp, 0);
		    }
		}
		break;
	    case 0x5f:
		*prow=-1;
		*pcol='h';
		ret=1;
		break;
	    default:
		break;
	    }
	    break;

	case MOUSEBUTTONS:
	    if (im->MouseX < win->Width - win->BorderLeft && im->MouseY > win->BorderTop) {
		if (im->Code == SELECTDOWN) {
		    *pcol = (im->MouseX - win->BorderLeft) / win->RPort->TxWidth;
		    *prow = ((im->MouseY - win->BorderTop) / win->RPort->TxHeight) + disp->TLineNo;
		    if (*prow >= 0)
			ret = 1;
		}
	    }
	    break;
	case NEWSIZE:
	    SetRefresh(disp, 0);
	    break;
	case REFRESHWINDOW:
	    BeginRefresh(win);
	    EndRefresh(win, 1);
	    RefreshBrowseDisplay(disp, 1);
	    break;
	default:
	    break;
	}
	ReplyMsg(im);
    }
    return(ret);
}

void
RefreshBrowseDisplay(vdisp, all)
void *vdisp;
short all;
{
    Display *disp = (Display *)vdisp;
    long x, y;
    int maxx;
    int dispLines = 0;
    int lineNo;
    short skipLineBeg = 0;
    short skipLineEnd = 0;
    Window *win = disp->Win;
    short b;
    short isHdr;
    short rvsMode;

    {
	long diff = disp->LastTLine-disp->TLineNo;

	if (all == 0 && diff > 0 && diff < disp->WLines/2) {
	    skipLineBeg = diff;
	    skipLineEnd = disp->WLines;

	    ScrollRaster(
		disp->Win->RPort,
		0,
		-(disp->Win->RPort->TxHeight * diff),
		disp->Win->BorderLeft, disp->Win->BorderTop,
		disp->Win->Width - (disp->Win->BorderRight + 1),
		disp->Win->Height - (disp->Win->BorderBottom + 1)
	    );
	} else if (all == 0 && diff < 0 && diff > -disp->WLines/2) {
	    skipLineBeg = 0;
	    skipLineEnd = disp->WLines + diff;

	    ScrollRaster(
		disp->Win->RPort,
		0,
		-(diff * disp->Win->RPort->TxHeight),
		disp->Win->BorderLeft, disp->Win->BorderTop,
		disp->Win->Width - (disp->Win->BorderRight + 1),
		disp->Win->Height - (disp->Win->BorderBottom + 1)
	    );
	} else {
	    SetAPen(win->RPort, 0);
	    RectFill(win->RPort, win->BorderLeft, win->BorderTop, win->Width - (win->BorderRight+1), win->Height - (win->BorderBottom+1));
	    SetAPen(win->RPort, 1);
	    all = 1;
	}
    }
    disp->LastTLine=disp->TLineNo;
    disp->FlagRefresh = 0;

    if (disp->Fi == NULL) {
	disp->Fi = fopen(disp->Name, "r");
	if (disp->Fi == NULL)       /*  can't refresh if can't open file */
	    return;
	for (disp->FLines = 0; fgets(TmpBuf, sizeof(TmpBuf), disp->Fi); ++disp->FLines)
	    ;
	fseek(disp->Fi, 0L, 2);
	disp->XPos = ftell(disp->Fi);
    }
    if (ftell(disp->Fi) != disp->TPos)
	fseek(disp->Fi, disp->TPos, 0);
    lineNo = disp->TLineNo - 1;
    x = win->BorderLeft;
    y = win->BorderTop + win->RPort->TxBaseline;
    maxx = (win->Width - win->BorderLeft - win->BorderRight) / win->RPort->TxWidth;

    b = 0;
    TmpBuf[0] = 0;

    disp->WMaxLines = (win->Height - win->BorderBottom - win->BorderTop - 2) / win->RPort->TxHeight;

    while (y < win->Height - win->BorderBottom - 2) {
	int len;

	if (TmpBuf[b] == 0) {
	    if (fgets(TmpBuf, sizeof(TmpBuf), disp->Fi) == NULL)
		break;
	    rvsMode = 0;
	    if (++lineNo < disp->FirstNonHdrLine) {
		isHdr = 1;
		if (strncmp(TmpBuf, "Subject:", 8) == 0)
		    rvsMode = 1;
	    } else {
		isHdr = 0;
	    }
	    TabExpand(TmpBuf, sizeof(TmpBuf));
	    b = 0;
	}
	len = strlen(TmpBuf + b) - disp->ColStart;
	if (len > maxx)
	    len = maxx;
	if (len > 0) {
	    if (rvsMode) {
		SetAPen(win->RPort, 0);
		SetBPen(win->RPort, 1);
	    }
	    if (skipLineBeg > 0 || skipLineEnd <= 0) {
		Move(win->RPort, x, y);
		Text(win->RPort, TmpBuf + b + disp->ColStart, len);
	    }

	    if (rvsMode) {
		SetAPen(win->RPort, 1);
		SetBPen(win->RPort, 0);
	    }
	}
	--skipLineBeg;
	--skipLineEnd;
	++dispLines;
	y += win->RPort->TxHeight;
	if (isHdr || disp->ColStart || len <= 0)
	    b = strlen(TmpBuf);
	else
	    b += len;
    }
    disp->BPos = ftell(disp->Fi);
    disp->WLines = dispLines;
    FixPropGadget(disp, all);
}

void
LoadDisplayConfig()
{
    FILE *fi;

    if (fi = fopen("s:dnews.config", "r")) {
	fread(ConfAry, 1, sizeof(ConfAry), fi);
	fclose(fi);
    }
}

void
SaveDisplayConfig()
{
    FILE *fi;

    if (fi = fopen("s:dnews.config", "w")) {
	fwrite(ConfAry, 1, sizeof(ConfAry), fi);
	fclose(fi);
    }
}

static void
SaveWindowConfig(disp)
Display *disp;
{
    Config *conf = ConfAry + disp->Id;
    Window *win = disp->Win;

    conf->TopEdge   = win->TopEdge;
    conf->LeftEdge  = win->LeftEdge;
    conf->Width     = win->Width;
    conf->Height    = win->Height;
}

/*
 *  Count lines forwards
 */

static long
ForwardFile(disp, pos, lines)
Display *disp;
long pos;
long lines;
{
    long n;
    long i;

    if (disp->Fi == NULL) {
	disp->Fi = fopen(disp->Name, "r");
	if (disp->Fi == NULL)
	    return(0);
    }
    fseek(disp->Fi, pos, 0);
    while (lines) {
	if (fgets(TmpBuf, sizeof(TmpBuf), disp->Fi) == NULL)
	    break;
	--lines;
    }
    return(ftell(disp->Fi));
}


/*
 *  Find a given line
 */

long
GotoLine(disp, lines)
Display *disp;
long lines;
{
    if (disp->Fi == NULL) {
	disp->Fi = fopen(disp->Name, "r");
	if (disp->Fi == NULL)
	    return(0);
    }
    fseek(disp->Fi, 0, 0);
    while (lines) {
	if (fgets(TmpBuf, sizeof(TmpBuf), disp->Fi) == NULL)
	    break;
	--lines;
    }
    return(ftell(disp->Fi));
}

/*
 *  Count lines backwards
 */

static long
BackupFile(disp, pos, rlines)
Display *disp;
long pos;
long rlines;
{
    long n;
    long i;

    if (disp->Fi == NULL) {
	disp->Fi = fopen(disp->Name, "r");
	if (disp->Fi == NULL)
	    return(0);
    }
    ++rlines;	/*  think of the case where rlines is 1 in the below code   */

    while (rlines && pos) {
	n = sizeof(TmpBuf);
	if (pos < n)
	    n = pos;
	fseek(disp->Fi, pos - n, 0);
	fread(TmpBuf, 1, n, disp->Fi);
	for (i = n - 1; i >= 0; --i) {
	    if (TmpBuf[i] == '\n') {
		if (--rlines == 0)
		    break;
	    }
	}
	++i;
	pos = pos - n + i;
    }
    return(pos);
}

/*
 *  Flag that the display should be refreshed without doing so immediately.
 *  This allows the user to skip around articles without having to wait
 *  for the display to catch up.
 */

static void
SetRefresh(disp, lrc)
Display *disp;
int lrc;
{
    FlagRefresh = 1;

    if (lrc < 0)
	lrc = 0;

    if (lrc > TMPBSIZE - 1)
	lrc = TMPBSIZE - 1;

    disp->FlagRefresh = 1;
    disp->ColStart = lrc;
}


/*
 *			    YES OR NO
 */

typedef struct IntuiText    ITEXT;
typedef unsigned char	ubyte;

int
getyn(text)
char *text;
{
    int result;
    ITEXT *body, *pos, *neg;

    if (DispBase == NULL)
	return(0);

    body = (ITEXT *)AllocMem(sizeof(ITEXT), 0);
    pos  = (ITEXT *)AllocMem(sizeof(ITEXT), 0);
    neg  = (ITEXT *)AllocMem(sizeof(ITEXT), 0);
    setmem(body, sizeof(ITEXT), 0);
    setmem(pos , sizeof(ITEXT), 0);
    setmem(neg , sizeof(ITEXT), 0);
    body->BackPen = pos->BackPen = neg->BackPen = 1;
    body->DrawMode= pos->DrawMode= neg->DrawMode= AUTODRAWMODE;
    body->LeftEdge = 10;
    body->TopEdge  = 12;
    body->IText    = (ubyte *)text;
    pos->LeftEdge = AUTOLEFTEDGE;
    pos->TopEdge = AUTOTOPEDGE;
    pos->IText = (ubyte *)"OK";
    neg->LeftEdge = AUTOLEFTEDGE;
    neg->TopEdge = AUTOTOPEDGE;
    neg->IText = (ubyte *)"CANCEL";
    result = AutoRequest(DispBase->Win, body, pos, neg, 0, 0, 320, 58);
    FreeMem(body, sizeof(ITEXT));
    FreeMem(pos , sizeof(ITEXT));
    FreeMem(neg , sizeof(ITEXT));
    return(result);
}

void
TabExpand(buf, max)
char *buf;
short max;
{
    short i;
    short len = strlen(buf);
    char *ptr;

    max -= 9;
    for (i = 0, ptr = buf; *ptr; ++ptr, ++i) {
	if (*ptr == '\n') {
	    *ptr = 0;
	    break;
	}
	if (*ptr == 9 && len < max) {
	    int n = ~i & 7;
	    movmem(ptr, ptr + n, len - i + 1);
	    setmem(ptr, n + 1, ' ');
	    len += n;
	}
    }
}

