
/*
 *  RNEWS.C
 *
 *  RNEWS [file]    (if not specified, stdin is used)
 *
 *  Copyright 1988 by William Loftus.  All rights reserved.
 *  Extensive Changes Copyright 1990 by Matthew Dillon, All Rights Reserved
 *
 *  This is rnews for compressed news.	News 2.11 will uux
 *  a file to this system that will be in compressed format.
 *  This program will strip off the "#! cunbatch", uncompress
 *  the news, and call unbatch.  If the news is not in compressed
 *  format it will just pass it to unbatch.
 *
 *  Each newsgroup directory has a .next file which indicates the next
 *  article number to write.  This number rolls over at 32767.
 */

#include <exec/types.h>
#include <exec/ports.h>
#include <dos/dosextens.h>
#include <clib/exec_protos.h>
#include <clib/dos_protos.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>

#include "news.h"

#include <log.h>
#include <ForwardNews.h>    /* MRR, 09-23-90. */
#include <owndevunit.h>

typedef struct MsgPort	MsgPort;

typedef struct group {
    struct group *next;
    long actual;	/*  actual bytes	*/
    long bytes; 	/*  uncompressed bytes	*/
    long articles;	/*  articles		*/
    char name[1];
} group;

static group groups;


void	myexit(void);
long	dofile(FILE *, int);
long	unbatch(FILE *, long);
long	CopyToGroup(char *, long, char *, char **);
long	LinkToGroup(char *, char *);
void	initgroups(void);
group	*ValidGroup(const char *);
void	HandleSequenceNumber(char *, long *);


IDENT(".14");

static char TmpBuf[2048];	/*  at least 1K */
char HomeDir[256];

char *NewsDir;
short DDebug;
short UHaveNews;
short DummyOpt;
long  LinksOk = 0;
long  OnlyOne = 0;

struct Library *OwnDevUnitBase;

extern struct DosLibrary *DOSBase;

#define PROG_NAME "RNews"

int _bufsiz = 8192;	/*  for stdio	*/

/*
 *  MAIN()
 *
 *  The main loop.  Note that OwnDevUnit must be valid for ulog() to work!
 */

int
main(ac, av)
int ac;
char *av[];
{
    int ret = 0;
    short i;
    short delFile = 0;
    FILE *fi;
    char *newsFile = NULL;

    LogProgram = PROG_NAME;

    atexit(myexit);

    /*
     *	File Locking (ulog, LockFile, etc...)
     */

    if ((OwnDevUnitBase = OpenLibrary(ODU_NAME, 0)) == NULL) {
	printf("Unable to open %s\n", ODU_NAME);
	exit(20);
    }

    /*
     *	Hard Links are possible for 2.04 or later.  NOTE!  Do not use
     *	Hard Links for anything less then V37 !!
     */

    if (DOSBase->dl_lib.lib_Version >= 37)
	LinksOk = 1;

    getcwd(HomeDir, sizeof(HomeDir));
    NewsDir = GetConfigDir(UUNEWS);

    if (ac == 0) {
	ulog(-1, "cannot call RNews from workbench!");
	exit(10);
    }

    for (i = 1; i < ac; ++i) {
	char *ptr = av[i];
	if (*ptr != '-') {
	    if (newsFile)
		ulog(-1, "Extranious news file %s ignored", ptr);
	    else
		newsFile = ptr;
	    continue;
	}

	switch(ptr[1]) {
	case 'n':
	    DummyOpt = 1;
	    break;
	case 'd':
	    DDebug = (ptr[2]) ? atoi(ptr + 2) : 1;
	    break;
	case 'x':
	    LogLevel = atoi(ptr + 2);
	    break;
	case 'h':
	    LinksOk = 0;
	    break;
	case '1':
	    OnlyOne = 1;
	    break;
	default:
	    printf("unknown flag '%s' ignored\n", ptr);
	    break;
	}
    }

    initgroups();
    if (GetForwardingInfo() != 0) {     /* MRR, 09-23-90. */
	ulog(-1,"Failed to get news forwarding information - abort.");
	exit(1);
    }

    /*
     *	open file, ensure it is seekable.
     */

    if (newsFile) {     /*  news file       */
	if ((fi = fopen(newsFile, "r")) == NULL) {
	    ulog(-1, "Unable to open %s", newsFile);
	    exit(10);
	}
    } else {		/*  else on stdin   */
	if (fseek(stdin, 0L, 1) >= 0) {
	    fi = stdin;
	} else {
	    newsFile = strdup(TmpFileName("news-work"));
	    if (newsFile == NULL) {
		ulog(-1, "strdup() : out of memory");
		exit(20);
	    }
	    if (fi = fopen(newsFile, "w+")) {
		long n;
		while ((n = fread(TmpBuf, 1, sizeof(TmpBuf), stdin)) > 0) {
		    long rslt = fwrite(TmpBuf, 1, n, fi);
		    if (rslt != n) {
			ulog(-1, "Write failed file %s", newsFile);
			/* XXX */
			break;
		    }
		}
		fseek(fi, 0L, 0);
		delFile = 1;
	    } else {
		ulog(-1, "Unable to create scratch file %s", newsFile);
		exit(20);
	    }
	}
    }

    if (chdir(NewsDir) != 0) {
	ulog(0, "NewsDir %s doesn't exist; quitting", NewsDir);
	exit(1);
    }

    ret = dofile(fi, 1);
    fclose(fi);

    if (delFile && ret == 0)
	remove(newsFile);

    if (UHaveNews) {
	MsgPort *port;
	Forbid();
	if ((port = FindPort("T:NewsRdy")) && port->mp_SigTask) {
	    Signal(port->mp_SigTask, 1 << port->mp_SigBit);
	    UHaveNews = 0;
	}
	Permit();
    }
    if (UHaveNews) {
	int fd = open("T:NewsRdy", O_RDONLY);
	if (fd < 0) {
	    char *cmd;

	    if (cmd = FindConfig(NEWSREADYCMD)) {
		fd = open("T:NewsRdy", O_CREAT|O_TRUNC, 0666);
		if (fd >= 0)
		    close(fd);
		sprintf(TmpBuf, "Run %s -x T:NewsRdy", cmd);
		Execute(TmpBuf, (BPTR)0, (BPTR)0);
	    }
	} else {
	    close(fd);
	}
    }
    exit(ret);
}

/*
 *  handle a news file
 */

long
dofile(fi, isReal)
FILE *fi;
int isReal;
{
    char buf[64];
    long pos = ftell(fi);
    long siz;
    long res = 0;
    long r;

    fseek(fi, 0L, 2);
    siz = ftell(fi);
    fseek(fi, pos, 0);

    if (isReal)
	groups.actual += siz;

    for (;;) {
	pos = ftell(fi);

	if (fgets(buf, sizeof(buf), fi) == NULL)
	    break;

	if (strncmp(buf, "#! cunbatch", 11) == 0) {
	    /*
	     *	Compressed Batch, uncompress first.  If corrupted do not
	     *	attempt to parse it.
	     */

	    char *tmpname = strdup(TmpFileName("news-work"));
	    FILE *f2;

	    r = uncompress_to_file(fi, tmpname);

	    if (r < 0) {
		ulog(-1, "Newsfile Corrupt (%s), file not processed or deleted", tmpname);
		res = 30;
		break;
	    } else if (f2 = fopen(tmpname, "r")) {
		r = dofile(f2, 0);
		fclose(f2);

		if (res < r)
		    res = r;
		if (DDebug < 10)
		    remove(tmpname);
	    } else {
		ulog(-1, "Unable to decompress news (%ld)", IoErr());
		if (DDebug < 10)
		    remove (tmpname);
		if (res < 20)
		    res = 20;
	    }
	    free(tmpname);
	} else if (strncmp(buf, "#! rnews ", 9) == 0) {
	    /*
	     *	uncompressed batch, new style
	     */

	    r = unbatch(fi, atol(buf + 9));
	    if (res < r)
		res = r;
	} else if (strncmp(buf, "! rnews ", 8) == 0) {
	    /*
	     *	uncompressed batch, old style
	     */

	    r = unbatch(fi, atol(buf + 8));
	    if (res < r)
		res = r;
	} else {
	    /*
	     *	treat unknown as an unbatched file
	     */

	    fseek(fi, pos, 0);
	    r = unbatch(fi, siz - pos);
	    if (res < r)
		res = r;
	}
    }
    if (res)
	ulog(-1, "dofile() error %d", res);
    return(res);
}

/*
 *  UNBATCH()
 *
 *  unbatches a news article starting at the current seek position and
 *  running for the specified number of bytes.	If #!rnews is encountered,
 *  the file is seeked backed after the current batch is done so dofile()
 *  catches it.
 *
 *  The article is stored into the appropriate newsgroup(s) as specified
 *  in LIB/NewsGroups, or in the 'junk' directory.  Individual article
 *  sequence numbers are kept in a file in each newsgroup directory.
 *
 *  note that realFileName is actually a dot-formatted newsgroup/artno,
 *  not a directory hierarchy.
 *
 */

long
unbatch(fi, bytes)
FILE *fi;
long bytes;
{
    char *base = malloc(bytes + 1);
    char *realFileName = NULL;
    short junkFlag;

    if (base == NULL) {
	ulog(-1, "Unable to malloc %d bytes", bytes);
	return(20);
    }
    bytes = fread(base, 1, bytes, fi);
    if (bytes <= 0) {
	ulog(-1, "unbatch: read failed %d", bytes);
	free(base);
	return(20);
    }
    base[bytes] = 0;	/*  terminate buffer	*/

    /*
     *	check for #! rnews line within buffer, seek back accordingly
     */

    {
	char *ptr;
	if (ptr = strstr(base, "#! rnews ")) {
	    fseek(fi, -(bytes - (ptr - base)), 1);
	    bytes = ptr - base;
	    *ptr = 0;
	}
    }

    /*
     *	Global statistics
     */

    groups.bytes += bytes;
    ++groups.articles;

    /*
     *	scan for news groups.  If any news group in the list exists the
     *	article is not stored to junk.	If no news groups in the list
     *	exists the article is stored to junk.
     *
     *	The JunkSave configuration entry can be used to disable the
     *	saving of articles to junk.
     */

    {
	char *ptr = GetConfig(JUNKSAVE, "Y");

	if (ptr[0] == 'y' || ptr[0] == 'Y')
	    junkFlag = 1;
	else
	    junkFlag = -1;
    }

    {
	char *grp;

	ScanHeader(base, "Newsgroups:");
	while (grp = ScanNext()) {
	    if (ValidGroup(grp))
		junkFlag = 0;
	}
    }

    if (junkFlag) {
	if (junkFlag > 0) {
	    CopyToGroup(base, bytes, "junk", &realFileName);
	    ++UHaveNews;
	}
    } else {
	char *grp;		/*  current newsgroup name  */
	short didFwd = 0;	/*  at least one copy has been written to disk	*/

	ScanHeader(base, "Newsgroups:");
	while (grp = ScanNext()) {
	    group *gp;
	    if (gp = ValidGroup(grp)) {
		++gp->articles;
		gp->bytes += bytes;

		if (didFwd == 0 || LinkToGroup(grp, realFileName) < 0) {
		    if (CopyToGroup(base, bytes, grp, &realFileName) < 0)
			ulog(-1, "Unable to copy article to group %s", grp);
		    else
			didFwd = 1;
		}
		++UHaveNews;
	    }
	}
    }

    /*
     *	forward article to downstream sites if it was stored.
     */

    if (realFileName) {
	if (DummyOpt == 0)
	    ForwardArticle(base, bytes, realFileName);
	free(realFileName);
    }
    free(base);
    return(0);
}

/*
 *  LINKTOGROUP()
 *
 *  If we are allowed to hard link and we have already written out
 *  one file (realFileName != NULL), then create a hard link instead
 *  of copying the file.
 *
 *  note that realFileName is actually a newsgroup/artno (dot format)
 *
 *  If this call fails RNews will attempt to make a copy of the file
 *  instead.
 */

long
LinkToGroup(grp, realFileName)
char *grp;
char *realFileName;
{
    long r = -1;
    BPTR lock;
    char *real;
    char *path;

    if (LinksOk && realFileName) {
	/*
	 *  ok to link, already made at least one copy of the file
	 *  (need something to link to!)
	 */

	if (real = HandleHeirarchy(GetConfigDir(UUNEWS), realFileName, 0)) {
	    if (lock = Lock(real, SHARED_LOCK)) {
		/*
		 *  got a handle on the original file, convert group
		 *  name to path and
		 */

		if (path = HandleHeirarchy(GetConfigDir(UUNEWS), grp, 0)) {
		    /*
		     *	attempt to link.
		     */

		    long seqNo;

		    HandleSequenceNumber(path, &seqNo);
		    sprintf(TmpBuf, "%s/%ld", path, seqNo);

		    if (MakeLink(TmpBuf, (long)lock, 0L) == 0) {
			ulog(-1, "Couldn't create hardlink to %s (%ld), copying instead", TmpBuf, IoErr());
		    } else {
			r = 0;
		    }
		}
		UnLock(lock);
	    }
	    free(real);
	}
    }
    return(r);
}

/*
 *  COPYTOGROUP()
 *
 *  In the event that a hardlink is disallowed or not possible (due to
 *  this being the first instance of a news file), write out the file.
 */

long
CopyToGroup(buf, bytes, grp, rfn)
char *buf;
long bytes;
char *grp;
char **rfn;
{
    long seqNo;
    char *path;
    FILE *fo;
    long r = 0;

    if (DummyOpt)
	return(0);

    if ((path = HandleHeirarchy(GetConfigDir(UUNEWS), grp, 0)) == NULL)
	return(-1);

    /*
     *	Obtain a sequence number (and bump .next file)
     */

    HandleSequenceNumber(path, &seqNo);

    /*
     *	write the file.  When we write a real file, we set *rfn to point
     *	to the actual file name for ForwardArticle().  This allows us
     *	to implement hard links for multiple references
     */

    if (*rfn == NULL) {
	sprintf(TmpBuf, "%s/%d", grp, seqNo);
	*rfn = strdup(TmpBuf);
    }
    sprintf(TmpBuf, "%s/%d", path, seqNo);

    if (fo = fopen(TmpBuf, "w")) {
	long n = fwrite(buf, 1, bytes, fo);
	fclose(fo);
	if (n != bytes)
	    r = -1;
    }
    return(r);
}

void
HandleSequenceNumber(path, pseq)
char *path;
long *pseq;
{
    FILE *fi;

    sprintf(TmpBuf, "%s/.next", path);

    if (fi = fopen(TmpBuf, "r+")) {
	if (fscanf(fi, "%ld", pseq) < 1)
	    *pseq = 1;
	fseek(fi, 0L, 0);
    } else {
	*pseq = 1;
	fi = fopen(TmpBuf, "w");
    }

    if (fi) {
	fprintf(fi, "%-6ld", (*pseq + 1) & 0x7FFF);
	fclose(fi);
    }
}

void
myexit(void)
{
    chdir(HomeDir);
    {
	group *gp;

	OpenLog();
	for (gp = groups.next; gp; gp = gp->next) {
	    if (gp->articles) {
		ulog(-1, "%-24s %4ld ARTS %7ld UNCOMP",
		    gp->name,
		    gp->articles,
		    gp->bytes
		);
	    }
	}

	ulog(-1, "%-24s %4ld ARTS %7ld/%-7ld BYTES/UNCOMPRESSED",
	    "TOTAL",
	    groups.articles,
	    groups.actual,
	    groups.bytes
	);
	CloseLog();
    }

    UnLockFiles();

    if (OwnDevUnitBase) {
	CloseLibrary(OwnDevUnitBase);
	OwnDevUnitBase = NULL;
    }
}

/*
 *  Newsgroups
 */

void
initgroups(void)
{
    struct group *gp;
    long len;
    FILE *fp;
    char buf[200];

    if (groups.next != NULL)
	return;

    if ((fp = openlib("newsgroups")) == NULL) {
	ulog(-1, "Couldn't open LIB/newsgroups file");
	exit(2);
    }

    gp = &groups;
    while (fgets(buf, sizeof(buf), fp)) {
	short c;
	for (len = 0; (c = buf[len]) != ' ' && c != 9 && c != '\n'; ++len);
	buf[len] = 0;
	if ((gp->next = malloc(sizeof(group)+len+1)) == NULL) {
	    ulog(-1, "Malloc failed!");
	    exit(4);
	}
	gp = gp->next;
	clrmem(gp, sizeof(group));
	strcpy(gp->name, buf);
    }
    fclose(fp);
}

group *
ValidGroup(grp)
const char *grp;
{
    struct group *gp;

    for (gp = groups.next; gp; gp = gp->next) {
	if (stricmp(gp->name, grp) == 0)
	    break;
    }
    return(gp);
}

