/***************************************************************************

                                 TrimSpool
                      Copyright 1994 by Gregg Giles

***************************************************************************/


/***************************************************************************
GET COMMAND LINE ARGUMENTS
***************************************************************************/

parse arg arguments



/***************************************************************************
DEFINITIONS
Warning: be sure to check the scope!
***************************************************************************/

/* Progam definitions */
PROG_AUTHOR    = 'Gregg Giles'      /* author */
PROG_COPYRIGHT = '1994'             /* Copyright */
PROG_DATE      = '04 June 1994'     /* release date */
PROG_NAME      = 'TrimSpool'        /* program name */
PROG_VERSION   = '1.0'              /* program version */

/* global definitions */
OFF       = 0                       /* test condition */
ON        = 1                       /* test condition */
NO        = 0                       /* test condition */
YES       = 1                       /* test condition */

/* counters */
ctr_expiredage    = 0               /* average age of expired files (days) */
ctr_expiredbytes  = 0               /* cumulative size of expired files (bytes) */
ctr_expiredfiles  = 0               /* expired file counter */
ctr_rembytes      = 0               /* cumulative size of remaining files (bytes) */
ctr_remfiles      = 0               /* remaining file counter */
ctr_totalbytes    = 0               /* cumulative size of all files (bytes) */
ctr_totaldays     = 0               /* cumulative age of files (days) */

/* defaults */
QUIET     = OFF                     /* controls console output */

/* Totals */



/***************************************************************************
 Open rexxsupport.library if not already open
***************************************************************************/

/* Open rexxsupport.library if not open */
if ~show('L', 'rexxsupport.library') then do
   if addlib('rexxsupport.library', 0, -30, 0) then do
      if (DEBUG >= 6) then do
         say 'Added rexxsupport.library'
      end
   end
   else do
      say 'Error: cannot open rexxsupport.library'
      exit 10
   end
end


/***************************************************************************
CONSOLE OUTPUT
If told to do so, turn off console I/O completely
***************************************************************************/

if (find(upper(arguments), 'QUIET')>0) then do
   QUIET = ON
end



/***************************************************************************
DISPLAY PROGRAM INFORMATION
***************************************************************************/

if (QUIET = OFF) then do
   say PROG_NAME 'Version' PROG_VERSION '('PROG_DATE')'
   say 'Copyright 'PROG_COPYRIGHT 'by' PROG_AUTHOR
   say ''
end


/***************************************************************************
DISPLAY USAGE TEMPLATE IF NO ARGUMENTS ARE GIVEN
***************************************************************************/

if (words(arguments)=0) then do
	say 'Usage: TrimSpool SOURCE <spool_dir> DAYS <days> [DEBUG=<level>] [QUIET]'
   exit
end



/***************************************************************************
DEBUGGING LEVEL
Higher levels increase verbosity
***************************************************************************/

DEBUG = 0                           /* default debugging value */
LOC_DEBUG = pos('DEBUG=', upper(arguments))

if (LOC_DEBUG > 0) then do
   DEBUG = substr(arguments, LOC_DEBUG+length('DEBUG='), 1)
   if ((DEBUG<0)|(DEBUG>9)) then do
      say 'Error: invalid DEBUG value - not displaying debug information'
   end
end



/***************************************************************************
GET NAME OF DIRECTORY FROM WHICH PROGRAM WAS LAUNCHED
This information is used later in finding the complete pathnames used to 
specify the source and destination directories
***************************************************************************/

LAUNCH_DIR = pragma(directory)
if (DEBUG>=5) then say '*** LAUNCH_DIR='LAUNCH_DIR



/***************************************************************************
GET SOURCE DIRECTORY FROM PARAMETERS
***************************************************************************/

/* If no source directory was supplied, show error message, then exit */
POS_SOURCE = find(upper(arguments), 'SOURCE')
if (POS_SOURCE=0) then do
   say 'Error: no source directory specified'
   exit                          /* exit program */
end

SOURCE_DIR = word(arguments, POS_SOURCE+1)   /* get source from arguments */

if (DEBUG>=5) then say '*** Specified source directory:' SOURCE_DIR 

/* If source directory doesn't exist, issue error message and exit */
if (exists(SOURCE_DIR)=0) then do
   say 'Error: source directory doesn''t exist ("'SOURCE_DIR'")'
   exit                                      /* exit program*/
end

/* Get complete pathname of source directory */
pragma(directory, SOURCE_DIR)    /* change directory to specified source */
SOURCE_DIR = pragma(directory)   /* get complete pathname of source dir */   
pragma(directory, LAUNCH_DIR)    /* change directory back to original */

/* Ensure source dir has colon or slash on end, otherwise program will fail */
TMP_CHAR = right(SOURCE_DIR, 1)
if ((TMP_CHAR ~= '/')&(TMP_CHAR ~= ':')) then do
   if (DEBUG >= 5) then do
      say '*** Forcing slash onto end of source directory path'
   end
   SOURCE_DIR = SOURCE_DIR'/'             /* force slash onto end */
end

if (DEBUG >= 5) then say '*** Modified source directory:' SOURCE_DIR



/***************************************************************************
GET DAYS VALUE FROM PARAMATERS
**************************************************************************/

POS_DAYS = find(upper(arguments), 'DAYS')

if (POS_DAYS > 0) then do
	ARG_DAYS = word(arguments, POS_DAYS+1)
	if (DEBUG >= 5) then do
		say '*** ARG_DAYS='ARG_DAYS
	end
end

/* If days value doesn't exist, issue error message and exit */
if (ARG_DAYS <= 0) then do
   say 'Error: invalid DAYS value supplied'
   exit                                      /* exit program*/
end



/***************************************************************************
GET SOURCE_DIR INFORMATION
***************************************************************************/

if (QUIET = OFF) then do
   say 'Scanning spool directory...'
   say ''
end

/* get filenames from directory and stuff them into a string */
if (DEBUG >= 1) then do
   say 'Processing:' SOURCE_DIR
end
DIR_CONTENTS = showdir(SOURCE_DIR, 'f')
/*say 'DIR_CONTENTS='DIR_CONTENTS*/

/* get the total number of files in specified directory */
NUM_FILES = words(DIR_CONTENTS)
if (DEBUG >= 5) then do
   say '*** NUM_FILES='NUM_FILES
end


/**************************************************************************
DELETE EXPIRED FILES
**************************************************************************/

do ctr_file = 1 to NUM_FILES
   /* get file next file to be examined */
   if (DEBUG >= 5) then do
      say ''
      say '*** ctr_file='ctr_file
   end
   CURRENT_FILE = word(DIR_CONTENTS, ctr_file)  /* get name from string */
   CURRENT_FILEPATH = SOURCE_DIR''CURRENT_FILE  /* add path to filename */

   /* get file information */
   FILE_INFO = statef(CURRENT_FILEPATH)      /* get file information */
   if (DEBUG >= 5) then do
      say '*** CURRENT_FILE='CURRENT_FILE
      say '*** FILE_INFO='FILE_INFO
   end

   /* Verify that file information was returned - information can't be
      obtained on mail packets being received when being examined. No
      information was returned if the file info string is blank. */

   if (FILE_INFO ~= '') then do
      /* extract all information from string into variables */
      FILE_TYPE = word(FILE_INFO, 1)            /* file type (file, dir) */
      FILE_BYTES = word(FILE_INFO, 2)           /* file size (bytes) */
      FILE_BLOCKS = word(FILE_INFO, 3)          /* file size (blocks) */
      FILE_PROTECTION = word(FILE_INFO, 4)      /* file protection */
      FILE_DAYS = word(FILE_INFO, 5)            /* file age (days since 01/01/01) */
      FILE_MINUTES = word(FILE_INFO, 6)         /* file age (minutes) */
      FILE_TICKS = word(FILE_INFO, 7)           /* file age (ticks) */
      FILE_COMMENT = word(FILE_INFO, 8)         /* file comment */

      /* determine age of file */
      BASEDATE_FILE = date(basedate, FILE_DAYS) /* file age (since 01/01/01) */
      BASEDATE_CURRENT = date(basedate)         /* current day (since 01/01/01) */
      BASEDATE_DIFF = BASEDATE_CURRENT - BASEDATE_FILE   /* age of file (days since 01/01/01) */
      if (DEBUG >= 5) then
         say '*** file='BASEDATE_FILE'  current='BASEDATE_CURRENT'  diff='BASEDATE_DIFF'  date='date(normal, FILE_DAYS)         

      /* if file age is equal to or exceeds trim age, delete it */
      if (BASEDATE_DIFF >= ARG_DAYS) then do
         /* delete expired file */
            if (QUIET = OFF) then say 'Deleting "'CURRENT_FILE'" ('FILE_BYTES' bytes) ('BASEDATE_DIFF' days old)'
         /* delete file unless debugging is turned on */
         if (DEBUG ~> 0) then
            address command 'delete 'CURRENT_FILEPATH' quiet'  /* delete file */      
         else say 'Debug override - "'CURRENT_FILE'" not deleted.'

         /* bump appropriate counters */
         ctr_expiredfiles = ctr_expiredfiles + 1         /* bump expired file counter */
         ctr_expiredbytes = ctr_expiredbytes + FILE_BYTES    /* total expired bytes */
         ctr_totaldays = ctr_totaldays + BASEDATE_DIFF   /* total days */
      end
      ctr_totalbytes = ctr_totalbytes + FILE_BYTES
   end
   else say 'Error: couldn''t get file information on' CURRENT_FILE
end

/* Display total information */
if (QUIET = OFF) then do
   /* determine "remaining" values */
   if (ctr_expiredfiles > 0) then do
      ctr_expiredage = (ctr_totaldays/ctr_expiredfiles)   /* avg. expired file age (days) */   
   end
   ctr_rembytes   = (ctr_totalbytes - ctr_expiredbytes)  /* bytes */
   ctr_remfiles   = (NUM_FILES-ctr_expiredfiles)         /* files */

   say ''
   say 'Initial  :' NUM_FILES 'files ('ctr_totalbytes' bytes)'
   say 'Expired  :' ctr_expiredfiles 'files ('ctr_expiredbytes' bytes) (average 'ctr_expiredage 'days old)'
   say 'Remaining:' ctr_remfiles 'files ('ctr_rembytes' bytes)'
   if (DEBUG >= 1) then
      say 'DEBUGGING: Debugging enabled - no files were deleted'
end
