/*
 *
 *  AM --- AmigaMail
 *  (C) 1991, 1992 by Christian Riede
 *
 *  AM is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY.  No author or distributor accepts responsibility to anyone
 *  for the consequences of using it or for whether it serves any
 *  particular purpose or works at all, unless he says so in writing.
 *  Refer to the GNU General Public License, Version 1, for full details.
 *  
 *  Everyone is granted permission to copy, modify and redistribute AM,
 *  but only under the conditions described in the GNU General Public
 *  License, Version 1.  A copy of this license is supposed to have been 
 *  given to you along with AM so you can know your rights and responsi-
 *  bilities.  It should be in a file named COPYING.  Among other things,
 *  the copyright notice and this notice must be preserved on all copies.
 *
 *  
 *
 */


#include "am.h"

/* perform action on listview-item with given number */

void ActionRead(ULONG num, struct List *Selected)
{
	char Command[80];
	char Mailname[80];
	struct Mail *Mail;
	BPTR MailLock;
	
	/* does mail with this number exist? */
	if (!(Mail = GetMail(num,Selected)))
	return;

	/* sender asked for Return-Receipt-To: ? */
	if ((Mail->read & (MAIL_RRT|MAIL_READ))==MAIL_RRT)  /* show message only once */
		SimpleRequest(Window,"Sender requested Return-Receipt-To");

	/* build filename */
	sprintf(Mailname,"uumail:%s.mail/%ld",Username,Mail->Number);

	/* lock file */
	if (!(MailLock = Lock((UBYTE *)Mailname,ACCESS_READ)))
	{
		SimpleRequest(Window,"Can't lock file");
		return;
	}

	/* mark as read */
	Mail->read |= MAIL_READ;
	MarkRead(Mail->Number);

	/* show file */
	/* run text filter */
#if 0
	sprintf(Command,
		"run <nil: >nil: uucp:c/metamail -r uumail:%s.mail/%ld",Username,Mail->Number);

	System((UBYTE *)Command,AMSystemTags);
#endif

	sprintf(Command,
		"uumail:%s.mail/%ld",Username,Mail->Number);

	CreateNewProcTags(
		NP_Entry,ReadMail,
		NP_StackSize,32768,
		NP_Arguments,Command,
		NP_Name,"AM_ReadMail",
		TAG_END,0);

	/* release lock */
	UnLock(MailLock);
}

void ActionSave(ULONG num, struct List *Selected)
{
	struct Mail *Mail;
	char Command[160];
	char Filename[160];

	/* does mail with this number exist? */
	if (!(Mail = GetMail(num,Selected)))
		return;

	*FReq->rf_File=0;
	*FReq->rf_Dir=0;
	if (AslRequestTags(FReq,ASL_Window,Window,TAG_END,0)) 
	{
		strcpy(Filename,(char *)FReq->rf_Dir);
		AddPart((UBYTE *)Filename,(UBYTE *)FReq->rf_File,160);
		sprintf(Command,"run c:copy uumail:%s.mail/%ld %s",
			Username,Mail->Number,Filename);
		Execute((UBYTE *)Command,AMSystemTags[0].ti_Data,AMSystemTags[1].ti_Data);
	}
}

void ActionPrint(ULONG num, struct List *Selected)
{
	struct Mail *Mail;
	char Command[160];

	/* does mail with this number exist? */
	if (!(Mail = GetMail(num,Selected)))
		return;

	sprintf(Command,"run c:type uumail:%s.mail/%ld to prt:",Username,Mail->Number);
	Execute((UBYTE *)Command,AMSystemTags[0].ti_Data,AMSystemTags[1].ti_Data);
}

void ActionReply(ULONG num, struct List *Selected,ULONG mode)
{
	struct Mail *Mail;
	char SendParms[160];

	if (!(Mail = GetMail(num,Selected)))
		return;

	sprintf(SendParms,"%ld %lx %s",
		SFLAGS_REPLY|mode,
		(ULONG)Mail,
		"laber");

	CreateNewProcTags(
		NP_Entry,SendMail,
		NP_Arguments,SendParms,
		NP_StackSize,32768,
		NP_Name,"AM_ReplyMail",
		TAG_END,0);
}

void ActionForward(ULONG num, struct List *Selected,ULONG mode)
{
	struct Mail *Mail;
	char SendParms[160];

	if (!(Mail = GetMail(num,Selected)))
		return;

	sprintf(SendParms,"%ld %lx %s",
		SFLAGS_FORWARD|mode,
		(ULONG)Mail,
		"laber");

	CreateNewProcTags(
		NP_Entry,SendMail,
		NP_Arguments,SendParms,
		NP_StackSize,32768,
		NP_Name,"AM_ForwardMail",
		TAG_END,0);
}

void ActionDelete(ULONG num, struct List *Selected)
{
	struct Mail *Mail;

	if (!(Mail = GetMail(num,Selected)))
		return;

	DeleteMail(Mail);
}

void ActionAddAddress(ULONG num, struct List *Selected)
{
	struct Mail *Mail;
	char Realname[160],Username[160];
	struct List Addressbook;

	if (!(Mail = GetMail(num,Selected)))
		return;

	ReadAddressBook(&Addressbook);
	AddAddress(&Addressbook,GetRealname(Mail->From,Realname),GetUsername(Mail->From,Username),-1,"",2);
	WriteAddressBook(&Addressbook);

	/* refresh listview gadget */
	RefreshDisplay(0,0,0,&Mailbox,
		Selected,LastSelected,(void *)SearchPattern);

}

