/* windows.c
 *
 * Copyright (c) 1993,1994 by Ezra Story.  All rights reserved.  Permission to
 * copy this program is given provided that the copy is not sold and that
 * this copyright notice is included.
 *
 * Window handling for UW.  Could be called Channel handling, but we use
 * pty's so I guess the analogy still holds. :-)
 */
static char rcsid[] = "$Id: windows.c 1.3 1994/06/07 02:57:56 Ezra_Story Exp $";

#include "config.h"
#include <sys/types.h>
#include <sys/file.h>
#include <stdlib.h>

#ifdef SYSSELECT
#include <sys/select.h>
#endif

#include "lists.h"
#include "mydefs.h"

proto WIN *wins[MAXWINDOWS];
proto LIST winlist;
proto OpenWindow();
proto WindowPri();
proto CheckWindows();
proto CloseWindow();
proto BuftoPty();
proto InitWindows();

WIN     *wins[MAXWINDOWS];

LIST    winlist;

/* initial setup of window array and priority list */
InitWindows()
{
    int x;
    for (x=0;x<MAXWINDOWS;x++)
        wins[x] = 0;

    NewList(&winlist);
}

/* Sets up a window.  If i is zero, it chooses a free window
 * number, otherwise it attempts (and perhaps fails) to open
 * the specific window number given. return window number or
 * -1 on failure.
 */
OpenWindow(i,pri)
int i;
char pri;
{
    WIN *w;

    /* find us a window number */
    if (i == 0)
        for (i=1;(i<MAXWINDOWS)&&(wins[i]);i++);
    if (i>= MAXWINDOWS) return(-1);

    /* alloc space for window struct */
    w = (WIN *)malloc(sizeof(WIN));
    if (w == 0) return(-1);
    w->id = i;
    w->ib = 0;

    /* open up a pty for this window */
    w->fd = forkpty(w);
    if (w->fd == -1)
        {
        free(w);
        return(-1);
        }

    /* add it to the priority list */
    w->n.Pri = pri;
    Enqueue(&winlist,&w->n);
    wins[i] = w;
    return(i);
}

/* gives window a new priority */
WindowPri(w,pri)
WIN *w;
char pri;
{
    Remove(&w->n);
    w->n.Pri = pri;
    Enqueue(&winlist,&w->n);
}

/* deallocates window */
CloseWindow(w)
WIN *w;
{
    wins[w->id] = 0;
    Remove(&w->n);
    closepty(w);
    free(w);
}

/* Dumps window input buffers to their pty's in priority order */
BufstoPty()
{
    WIN *w;
    int x;

    for (w = (WIN *)winlist.Head; w->n.Succ ; w = (WIN *)w->n.Succ)
        while (w->ib)
            {
            x = write(w->fd,w->ibuf,w->ib);
            if(x>0) w->ib -= x;
            }
}

/* Checks window pty's for output data in priority order */
CheckWindows()
{
    WIN *w;

    for (w = (WIN *)winlist.Head; w->n.Succ ; w = (WIN *)w->n.Succ)
        if (FD_ISSET(w->fd,&selmask[1].sm_rd)) p_SendData(w);
}

